#!/bin/ksh
#
#	(c) Copyright 1993 IBM Corp
#
##############################################################################
##############################################################################
#
#	Quickie script to ship files to the ${SHIP} directory and to
#	update stanza based inventory file and apply list
#	to enforce owner, group, and permissions of shipped code
#	This script is presented as the macro ${CS} in the Makefiles
#
#	ARGUMENTS:
#		${1}	The full pathname of ${SHIP}
#		${2}	The full pathname of ${ROOT}
#		${3}	The name of the file to be shipped ("NULL" for dirs)
#		${4}	The absolute pathname of the file as it would have been
#			on a AIX 3.1 machine ("NULL" defaults to same as 3.2)
#		${5}	Absolute pathname of the file on the target 3.2 machine
#		${6}	The name of the lpp_option to which this file belongs
#		${7}	The type of the file:
#			( FILE DIRECTORY FIFO BLK_DEV CHAR_DEV MPX_DEV )
#		${8}	The name of the owner on the target machine
#		${9}	The name of the group on the target machine
#		${10}	The (optional) mode (TCB TP SVTX SUID SGID), and comma
#			then the (required) access permissions
#			for this file on the target machine in octal
#			(eg. tcb,777 or 755)
#			It should be noted that certain mode type files (tcb)
#			will also be entered into the ${LPP_OPT}.tcb file.
#		${11}	The checksum type of either G or V
#			N - no checksum (directories only)
#			G - checksum is determined by this script
#			V - checksum is volatile
#			(file will be dynamically changed on target machine)
#
##############################################################################
#
#	The following must all be included if any are.
#	If it is desired that any not appear in the inventory file,
#	their positional parameter should be filled in with the string NULL
#
#	OPTIOAL ARGUMENTS:
#		${12}	One or more links to the file delimited by commas
#		${13}	One or more symbolic links delimited by commas
#		${14}	The verify program
#		${15}	The original source location
#
##############################################################################
#
#	The size of the file is determined by this shell script.
#
##############################################################################

# echo arg0, ${0}
# echo arg1, ${1}
# echo arg2, ${2}
# echo arg3, ${3}
# echo arg4, ${4}
# echo arg5, ${5}
# echo arg6, ${6}
# echo arg7, ${7}
# echo arg8, ${8}
# echo arg9, ${9}
# echo arg10, ${10}
# echo arg11, ${11}

# echo arg12, ${12}
# echo arg13, ${13}
# echo arg14, ${14}
# echo arg15, ${15}

if [ $# -ne 11 ] ;then
	if [ $# -ne 15 ] ;then
		print\
		'usage: code_ship:\
		<shipdir>\
		<rootdir>\
		[<source file> | "NULL"(directory)]\
		[< / relative name 3.1 destination> | "NULL"(same as 3.2)]\
		< / relative name 3.2 destination>\
		<option_name>\
		<filetype>
		<owner>\
		<group>\
		<checksum type>\
		[mode]<permissions>(octal)
	[<links>, <symlinks>, <verify program>, <source location>]'
		exit 1
	fi
	if [ ${12} != "NULL" ] ;then LINKS=${12} ;fi
	if [ ${13} != "NULL" ] ;then SLINKS=${13} ;fi
	if [ ${14} != "NULL" ] ;then VERIFY=${14} ;fi
	if [ ${15} != "NULL" ] ;then ORIG=${15} ;fi
fi

SHIP=${1}
ROOT=${2}
SRC=${3}
ODST=${4}
DST=${5}
LPP_OPT=${6}
TYP=${7}
OWN=${8}
GRP=${9}
MOD=${10}
CKS=${11}

APPLY=${LPP_OPT}.al
INVENTORY=${LPP_OPT}.inventory
TCBF=${LPP_OPT}.tcb

#	append filename onto apply list
#
echo .${DST} >> ${APPLY}


#	append stanza onto inventory file
#
if [ ${TYP} = "DIRECTORY" ]
then
	CKS=N
	mkdir ${ROOT}${DST} || ( echo ${0}: mkdir failed; exit 1 )

else
	if [ ${CKS} != "G" ]
	then
		if [ ${CKS} != "V" ]
		then
			echo ${0}: invalid checksum paremeter
			exit 1
		fi
	fi
	cp ${SRC} ${ROOT}${DST} || ( echo ${0}: copy failed ; exit 1 )
fi

echo ${DST}: >> ${INVENTORY}
echo "	owner = ${OWN}" >> ${INVENTORY}
echo "	group = ${GRP}" >> ${INVENTORY}
echo "	mode = ${MOD}" >> ${INVENTORY}
echo "	type = ${TYP}" >> ${INVENTORY}

if [ ${TYP} != "DIRECTORY" ]
then
	echo "	size = `wc -c ${ROOT}${DST} | cut -f1 -d"/"" >> ${INVENTORY}
	if [ ${CKS} = "G" ]
	then
		echo "	checksum = \"`sum ${ROOT}${DST} | cut -f1 -d"/"`\"" >> ${INVENTORY}
	else
		echo "	checksum = VOLATILE" >> ${INVENTORY}
	fi
fi

if [ ${ODST} = "NULL" ]
then
	echo "	oldpath = ${DST}" >> ${INVENTORY}
else
	echo "	oldpath = ${ODST}" >> ${INVENTORY}
fi

echo "	class = apply,inventory,${LPP_OPT}" >> ${INVENTORY}

if [ ${LINKS} ] ;then echo "	links = ${ODST}" >> ${INVENTORY} ;fi
if [ ${SLINKS} ] ;then echo "	symlinks = ${ODST}" >> ${INVENTORY} ;fi
if [ ${VERIFY} ] ;then echo "	program = ${ODST}" >> ${INVENTORY} ;fi
if [ ${ORIG} ] ;then echo "	source = ${ODST}" >> ${INVENTORY} ;fi

echo >> ${INVENTORY}


##############################################################################
#
#	The following does whatever is specific to individual file modes
#
##############################################################################

FMODE=`echo ${MOD} | cut -f1 -d","`


################################
#	Trusted compute base
################################

if [ ${FMODE} = "TCB" ]
then
	echo ${DST} will be shipped as part of the Trusted Compute Base
	echo ${DST}: >> ${TCBF}
	echo "	owner = ${OWN}" >> ${TCBF}
	echo "	group = ${GRP}" >> ${TCBF}
	echo "	mode = ${MOD}" >> ${TCBF}
	echo "	type = ${TYP}" >> ${TCBF}

	if [ ${TYP} != "DIRECTORY" ]
	then
		echo "	size = `wc -c ${ROOT}${DST} | cut -f1 -d"/"`" >> ${TCBF}
		if [ ${CKS} != "G" ]
		then
			echo "	checksum = \"`sum ${ROOT}${DST} | cut -f1 -d"/"`\"" >> ${TCBF}
		else
			echo "	checksum = VOLATILE" >> ${TCBF}
		fi
	fi

	if [ ${ODST} = "NULL" ]
	then
		echo "	oldpath = ${DST}" >> ${TCBF}
	else
		echo "	oldpath = ${ODST}" >> ${TCBF}
	fi

	echo "	class = apply,inventory,${LPP_OPT}" >> ${TCBF}

	if [ ${LINKS} ] ;then echo "	links = ${ODST}" >> ${TCBF} ;fi
	if [ ${SLINKS} ] ;then echo "	symlinks = ${ODST}" >> ${TCBF} ;fi
	if [ ${VERIFY} ] ;then echo "	program = ${ODST}" >> ${TCBF} ;fi
	if [ ${ORIG} ] ;then echo "	source = ${ODST}" >> ${TCBF} ;fi

	echo >> ${TCBF}
fi



################################
#	Trusted process
################################

if [ ${FMODE} = "TP" ]
then
	echo ${DST} will be shipped as a Trusted Process
fi


################################
#	Save Text on Swap File
################################

if [ ${FMODE} = "SVTP" ]
then
	echo ${DST} will be shipped as a Text Saved on Swap file
fi


################################
#	Set User ID set
################################

if [ ${FMODE} = "SUID" ]
then
	echo ${DST} will be shipped with the Set User ID bit set
fi


################################
#	Set Group ID set
################################

if [ ${FMODE} = "SGID" ]
then
	echo ${DST} will be shipped with the Set Group ID bit set
fi




