#ifndef _gm_arch_h_
#define _gm_arch_h_

/* NOTE: The include of gm_arch_debug.h MUST be first! */
#include "gm_arch_debug.h"

#include "gm_internal.h"
#include "gm_impl.h"
#include "gm_arch_types.h"
#include "gm_lanai.h"

#include <stdio.h>

#define GM_ARCH_MAX_MINOR 64
#define GM_ARCH_MAX_INSTANCE 4

#define GM_ARCH_DMA_READ	0x10
#define GM_ARCH_DMA_WRITE	0x20
#define GM_ARCH_DMA_RDWR	(GM_ARCH_DMA_READ | GM_ARCH_DMA_WRITE)
#define GM_ARCH_DMA_CONSISTENT  0x40
#define GM_ARCH_DMA_STREAMING   0x80
#define GM_ARCH_DMA_CONTIGUOUS  0x100

#define GM_ARCH_INTR_CLAIMED         1
#define GM_ARCH_INTR_UNCLAIMED       2
#define GM_ARCH_SYNC_FOR_DEVICE      3
#define GM_ARCH_SYNC_FOR_CPU         4

#define GM_SLEEP_WOKE 0
#define GM_SLEEP_INTERRUPTED 1
#define GM_SLEEP_TIMED_OUT           16
  
#define GM_EACCES		EACCES
#define GM_EBUSY		EBUSY
#define GM_EFAULT       	EFAULT
#define GM_EINTR		EINTR
#define GM_EINVAL		EINVAL
#define GM_ENOMEM		ENOMEM
#define GM_ENOTTY		ENOTTY
#define GM_EPERM		EPERM
#define GM_EPROTO		EPROTO
#define GM_EUNATCH		EUNATCH

/* GM_PRINT etc. support */

#define gm_aix_print(x) do { \
printf x; \
} \
while (0)

#define gm_aix_panic(x) do { \
printf x ; \
/* fflush(stdout); don't know how to do this in aix */ \
printf("\n"); \
gm_arch_spin(NULL, 10000); \
} \
while (0)

#define GM_ARCH_PRINT(args) gm_aix_print(args)
#define GM_ARCH_INFO(args) gm_aix_print(args)
#define GM_ARCH_NOTE(args) gm_aix_print(args)
#define GM_ARCH_WARN(args) gm_aix_print(args)
#define GM_ARCH_PANIC(args) gm_aix_panic(args)


/************************************************************************
 *
 * Locking macros
 *
 * NOTE WELL: These macros are for use in the process environment only.
 *            Do _not_ use them in the interrupt environment!
 *
 *  * GM_AIX_LOCK_ENTER  Seizes the lock.  If the lock is already busy,
 *                       blocks until it is available.
 *
 *  * GM_AIX_LOCK_HELD   Tests whether we already hold this lock.
 *                       Returns TRUE if we do, otherwise FALSE.
 *
 *  * GM_AIX_LOCK_EXIT   Releases the lock.  Note: AIX's man page for
 *                       simple_unlock() says, "The lock must be held by
 *                       the thread which calls the simple_unlock kernel
 *                       service," but it doesn't specify what happens
 *                       when this condition is not met.
 *
 *  TODO: Not all lock manipulations use these macros.  Standardize and
 *        explain the use of GM_* macros for lock manipulations.
 *
 ***********************************************************************/
#if GM_AIX_DEBUG_LOCKS && GM_DEBUG
#define GM_AIX_LOCK_ENTER(l_p) do { \
printf("Entering lock at 0x%x\n", l_p); \
 simple_lock(l_p); } \
 while (0)

inline boolean_t
GM_AIX_LOCK_HELD(l_p) {
  boolean_t function_status;

  function_status = lock_mine((void *)l_p);
  printf("Lock at 0x%x %s mine\n",
	 l_p, (function_status ? "IS" : "is NOT"));
  return (function_status);
}

#define GM_AIX_LOCK_EXIT(l_p) do { \
   simple_unlock(l_p); \
 printf("Exiting lock at 0x%x\n", l_p);} \
 while (0)
#else
#define GM_AIX_LOCK_ENTER(l_p) simple_lock(l_p)
#define GM_AIX_LOCK_HELD(l_p)  lock_mine(l_p)
#define GM_AIX_LOCK_EXIT(l_p)  simple_unlock(l_p)
#endif


/************
 * OS interface function prototypes
 ************/

int
gm_aix_open (
	dev_t devno,
	int devflag,
	int chan,
	int ext
);

int
gm_aix_close (
	dev_t devno,
	int chan,
	int ext
);

int
gm_aix_ioctl (
	dev_t devno,
	int op,
	int arg,
	ulong dev_flag,
	int chan,
	int ext
);

int
gm_aix_cfg (
	dev_t devno,
	int op,
	struct uio *uiop
);

int
gm_aix_mpx (
	dev_t devno,
	chan_t *chanp,
	char *channame
);

extern	int	nodev ();

/* No declaration for this in gm_impl.h? */
gm_status_t
gm_arch_page_len (unsigned long *r);

#define	PCI_VENDOR_ID	0x00
#define	PCI_DEVICE_ID	0x02
#define	PCI_COMMAND	0x04
#define	PCI_BASE_0	0x10
#define	PCI_INTR_LINE	0x3C

#define	PCI_CMD_IO_SPACE	0x01
#define	PCI_CMD_MEM_SPACE	0x02
#define	PCI_CMD_BUS_MASTER	0x04
#define	PCI_CMD_PARITY_ERROR	0x040
#define	PCI_CMD_SYSTEM_ERROR	0x100

#define	PCI_CFG_READ	0
#define	PCI_CFG_WRITE	1

#endif /*_gm_arch_h_*/
