/******************************************************************-*-c-*-
 * Myricom GM networking software and documentation			 			 *
 * Copyright (c) 2002 by Myricom, Inc.					 *
 * All rights reserved.	 See the file `COPYING' for copyright notice.	 *
 *************************************************************************/

/* author: gallatin@myri.com */

/* NOTE: The include of gm_arch_debug.h MUST be first! */
#include "gm_arch_debug.h"
#include "gm_arch_types.h"
#include "gm_internal.h"
#include "gm_instance.h"
#include "gm_impl.h"
#include "gm_ether.h"
#include "gmip.h"
#include "gm_debug.h"

#include <sys/dump.h>

#ifdef  SUPPORT_64BIT_OS
#define DO_MAP_PAGE 1
#else
#define DO_MAP_PAGE 0
#endif  /* SUPPORT_64BIT_OS */

int gm_aix_ether_xmit_copies = 0;
int silent_drops = 0;

static struct mbuf * gm_aix_ether_arch_rxalloc (int how, int max_segs);
static void gm_aix_ether_arch_unencap (struct gm_aix_ether_sc *sc, struct mbuf *m_head,
				       struct gm_aix_ether_arch_dma_info *dma);

static int
gm_aix_ether_arch_encap (struct gm_aix_ether_sc *sc, struct mbuf *m_head,
			 struct gm_aix_ether_arch_dma_info *dma);

/*****************************************************************************/
/*
 * NAME:     cdt_func
 *
 * FUNCTION: return the address of the component dump table 
 *
 * EXECUTION ENVIRONMENT: 
 *
 * NOTES:
 *
 * RETURNS:  	the address of the component dump table
 */
/*****************************************************************************/
struct cdt *cdt_func(void)
{
	return((struct cdt *)&dd_ctrl.cdt.head);
}

int
gm_aix_ip_close (ndd_t	*the_ndd)
{
  struct gm_aix_ether_sc *sc;
  dev_ctrl_t	*p_dev;
  struct gm_aix_ether_pkt *pkt;
  struct mbuf *m;
  ulong		mapped_addr;
  int i;
  int ipri;			/* Current priority */
  
  gm_aix_lock_driver();		/* Protect driver data from other threads */

  sc = (struct gm_aix_ether_sc *) the_ndd->ndd_correlator;
  p_dev = sc->p_dev;
  
  ipri = disable_lock(DDS.intr_priority, &p_dev->wrk.dev_lock);
  
  mapped_addr = (ulong) iomem_att (IO_MAP);


  NDD.ndd_flags &= ~(NDD_RUNNING | NDD_BROADCAST | NDD_UP);
  WRK.adap_state = CLOSED_STATE;

  gm_aix_ip_clear_offlevel (p_dev);

  /* Free packets queued for reception */
  
  while (WRK.rx_head) 
    {
      m = WRK.rx_head;
      WRK.rx_head = m->m_nextpkt;
      m->m_nextpkt = NULL;
      m_freem (m);
    }

  /* Free packets owned by the lanai */
  for (i = 0; i < GM_RECV_RING_CNT; i++)
    {
      pkt = sc->recv_ring + i;
      if (pkt->m)
	{
	  gm_aix_ether_arch_unencap (sc, pkt->m, &pkt->dma);
	  m_freem (pkt->m);
	  pkt->m = NULL;
	}
    }

  /* Free packets queued for transmit but owned by lanai */
  for (i = 0; i < GM_SEND_RING_CNT; i++)
    {
      pkt = sc->send_ring + i;
      if (pkt->m)
	{
	  gm_aix_ether_arch_unencap (sc, pkt->m, &pkt->dma);
	  m_freem (pkt->m);
	  pkt->m = NULL;
	}
    }

  /* Free packets queued for transmit but not given to lanai */
  for (i = 0; i < GM_SW_RING_CNT; i++)
    {
      pkt = sc->sw_ring + i;
      if (pkt->m)
	{
	  gm_aix_ether_arch_unencap (sc,  pkt->m, &pkt->dma);
	  m_freem (pkt->m);
	  pkt->m = NULL;
	}
    }

  /* enable interrupts and close the port */
  i_enable(ipri);
  gm_disable_ethernet (sc->port);
  gm_close (p_dev->port);
  p_dev->port = NULL;

  GM_PRINT(GM_AIX_DEBUG_IP, ("gm_aix_ether_xmit_copies = %d\n",
			     gm_aix_ether_xmit_copies));
  GM_PRINT(GM_AIX_DEBUG_IP, ("silent drops %d\n", silent_drops));

  iomem_det ((void *) mapped_addr);
  simple_unlock(&p_dev->wrk.dev_lock);

  gm_aix_unlock_driver();

  simple_lock(&dd_ctrl.i_lock);	/* Protect the dd_ctrl struct */
  dd_ctrl.num_ip_opens--;
  if (!dd_ctrl.num_ip_opens) 
    {
      dmp_del (cdt_func);
    }
  simple_unlock(&dd_ctrl.i_lock);

  GM_PRINT(GM_AIX_DEBUG_IP, ("gm_aix_ip_close: done\n"));
  return (0);
}


int
gm_aix_ip_init (dev_ctrl_t *p_dev)
{
  
  ndd_t *dev_ndd;
  gm_instance_state_t *is;
  struct gm_aix_ether_sc *sc;
  struct gm_aix_ether_pkt *pkt;
  int i, status;
  struct mbuf *m;

  is = p_dev->is;
  gm_always_assert (is != NULL);

  for (i = 0; i < 6; i++)
    {
      p_dev->wrk.net_addr[i] = is->lanai.eeprom.lanai_board_id[i];
    }	

  sc = gm_malloc (sizeof (*sc));
  if (sc == NULL)
    return GM_OUT_OF_MEMORY;
  bzero (sc, sizeof (*sc));


  sc->d_handle = d_map_init (BID_ALTREG (DDS.bus_id, PCI_BUSMEM),
				DMA_MASTER | DMA_MAXMIN_16M |       \
				DMA_ADDRESS_64_BIT|DMA_ENABLE_64,
				0, 0);
  if ((sc->d_handle == (d_handle_t) DMA_FAIL) ||
      (sc->d_handle == (d_handle_t) NULL)) {
    GM_WARN(("IP failed to get dhandle\n"));
    goto abort_with_sc;
  }
  sc->bogus_dma.dio.total_iovecs = GM_MAX_ETHERNET_SCATTER_CNT;
  sc->bogus_dma.dio.dvec = sc->bogus_dma.iovec;

  sc->send_dma.dio.total_iovecs = GM_MAX_ETHERNET_GATHER_CNT;
  sc->send_dma.dio.dvec = sc->send_dma.iovec;

  sc->bogus_m = gm_aix_ether_arch_rxalloc (M_DONTWAIT, GM_MAX_ETHERNET_SCATTER_CNT);
  if (sc->bogus_m == NULL)
    {
      goto abort_with_d_handle;
    }
  if (!gm_aix_ether_arch_encap (sc, sc->bogus_m, &sc->bogus_dma))
    {
      goto abort_with_bogus_m;
    }
  

  sc->p_dev = p_dev;
  sc->is = p_dev->is;

  dev_ndd = &p_dev->ndd;
  dev_ndd->ndd_name  = (char *)(p_dev->dds.lname);
  dev_ndd->ndd_alias = (char *)(p_dev->dds.alias);
  dev_ndd->ndd_correlator = (caddr_t)sc;  
  dev_ndd->ndd_open      = gm_aix_ip_open;
  dev_ndd->ndd_close     = gm_aix_ip_close;
  dev_ndd->ndd_output    = gm_aix_ip_output;
  dev_ndd->ndd_ctl       = gm_aix_ip_ioctl;
  dev_ndd->ndd_addrlen   = ENT_NADR_LENGTH;
  dev_ndd->ndd_hdrlen    = ENT_HDR_LEN;
  dev_ndd->ndd_mtu       = GM_IP_MTU + ENT_HDR_LEN;
  dev_ndd->ndd_mintu     = ENT_MIN_MTU;
  dev_ndd->ndd_type      = NDD_ISO88023;	/* Includes std ethernet */
  dev_ndd->ndd_specstats = (caddr_t)&(ENTSTATS);
  dev_ndd->ndd_speclen   = sizeof(ENTSTATS);
  dev_ndd->ndd_physaddr  = (caddr_t)(p_dev->wrk.net_addr);
  
#ifdef	SUPPORT_64BIT_OS
  dev_ndd->ndd_flags |= NDD_64BIT;
#endif	/* SUPPORT_64BIT_OS */ 

  dev_ndd->ndd_flags |= NDD_PSEG;

  WRK.adap_state = CLOSED_STATE;

  status = ns_attach (dev_ndd);
  if (status)
    {
      GM_WARN (("ns_attach () failed is status %d\n", status));
    }
  return status;

 abort_with_bogus_m:
  m_freem (sc->bogus_m);

 abort_with_d_handle:
  D_MAP_CLEAR (sc->d_handle);  

 abort_with_sc:
  gm_free (sc);

  return GM_OUT_OF_MEMORY;

}

int
gm_aix_ip_open (ndd_t *the_ndd)
{
  struct gm_aix_ether_sc *sc;
  dev_ctrl_t	*p_dev;
  struct gm_aix_ether_pkt *pkt;
  struct mbuf *m;
  ndd_statblk_t stat_blk;		/* status block */
  ulong		mapped_addr;
  int status, ipri;
  gm_u8_t i;

  gm_aix_lock_driver();		/* Protect driver data from other threads */

  sc = (struct gm_aix_ether_sc *) the_ndd->ndd_correlator;
  p_dev = sc->p_dev;

  GM_PRINT(GM_AIX_DEBUG_IP, ("gm_aix_ip_open: p_dev 0x%x\n", p_dev));

  /* If the device is not in a closed state, error */
  if (WRK.adap_state != CLOSED_STATE) 
    {
      GM_WARN (("gm_aix_ip_open: not closed error\n"));
      gm_aix_unlock_driver();
      return(EBUSY);
  }

  simple_lock (&p_dev->wrk.dev_lock);
  mapped_addr = (ulong) iomem_att (IO_MAP);

  status =  gm_open (&p_dev->port, p_dev->unit, GM_ETHERNET_PORT_ID,
		     "ethernet", (enum gm_api_version)GM_API_VERSION);

  sc->port = p_dev->port;
  if (status != GM_SUCCESS)
   {
     GM_WARN (("error opening ethernet port, status = %d\n", status));
     goto abort_with_dev_lock;
   }

  status = gm_enable_ethernet (sc->port, GM_NUM_ETHERNET_RECV_TOKENS,
                               GM_NUM_SEND_TOKENS - 2);
  if (status != GM_SUCCESS)
    {
      GM_NOTE (("error enabling ethernet,status=%d\n", status));
      status = EIO;
      goto abort_with_open_port;
    }

  
  sc->sreq = 0;
  sc->sdone = 0;
  sc->sw_req = 0;
  sc->sw_done = 0;
  sc->rdone = 0;
  sc->sdone_index = 0;
  sc->sreq_index = 0;

  gm_always_assert (GM_NUM_SEND_TOKENS >= GM_SEND_RING_CNT);
  gm_always_assert (GM_NUM_ETHERNET_RECV_TOKENS >= GM_RECV_RING_CNT);

  for (i = 0; i < GM_RECV_RING_CNT; i++)
        {
      pkt = sc->recv_ring + i;
      gm_always_assert (pkt->m == 0 && pkt->dma.dio.used_iovecs == 0);
      pkt->dma.dio.total_iovecs = GM_MAX_ETHERNET_SCATTER_CNT;
      pkt->dma.dio.dvec =  pkt->dma.iovec; 
      m = gm_aix_ether_arch_rxalloc (M_DONTWAIT, GM_MAX_ETHERNET_SCATTER_CNT);
      if (m == NULL)
        {
          status = ENOMEM;
          goto abort_with_open_port;
        }
      pkt->m = m;
      if (!gm_aix_ether_arch_encap (sc, m, &pkt->dma))
        {
          status = ENOMEM;
          goto abort_with_open_port;
        }
    }

  for (i = 0; i < GM_SEND_RING_CNT; i++)
    {
      pkt = sc->send_ring + i;
      pkt->dma.dio.total_iovecs = GM_MAX_ETHERNET_GATHER_CNT;
      pkt->dma.dio.dvec =  pkt->dma.iovec; 
    }

  for (i = 0; i < GM_SW_RING_CNT; i++)
    {
      pkt = sc->send_ring + i;
      pkt->dma.dio.total_iovecs = GM_MAX_ETHERNET_GATHER_CNT;
      pkt->dma.dio.dvec =  pkt->dma.iovec; 
    }


  gm_pause_lanai (sc->is);
  gm_ethernet_set_recv_intr_callback (sc->port, gm_aix_ether_rxeof, sc);
  gm_ethernet_set_sent_intr_callback (sc->port, gm_aix_ether_txeof, sc);
  for (i = 0; i < GM_RECV_RING_CNT; i++)
    {
      pkt = sc->recv_ring + i;
      gm_ethernet_provide_scatter_list_with_tag (sc->port, 
						pkt->dma.dio.used_iovecs,
						pkt->dma.desc,
						i);
    }
  gm_unpause_lanai (sc->is);
  sc->open = 1;

  NDD.ndd_flags |= NDD_UP;
  /* If first open, register dump table */
  simple_lock(&dd_ctrl.i_lock);	/* Protect the dd_ctrl struct */
  if ((++dd_ctrl.num_ip_opens) == 1) 
    {
      dmp_add(cdt_func);
    }
  simple_unlock(&dd_ctrl.i_lock);

  /* Update the device state and send a status block to demuxer */
  NDD.ndd_flags |= NDD_RUNNING | NDD_BROADCAST | NDD_SIMPLEX;
  WRK.adap_state = OPEN_STATE;
  
  /* Clear all stats */
  bzero(&GENSTATS, sizeof GENSTATS);
  bzero (&ENTSTATS, sizeof ENTSTATS);
  bzero (&ADAPSTATS, sizeof ADAPSTATS);
  
  WRK.enable_promiscuous = 0;
  WRK.send_bad_rx_pkts   = 0;
  WRK.enable_multicast   = 0;
  WRK.multi_addr_count   = 0;
  
  /* save start times for statistics */
  WRK.ndd_stime = WRK.dev_stime = lbolt;
  
  gm_aix_ip_init_offlevel (p_dev);

  bzero(&WRK.multi_table, sizeof(ent_multi_t));
  
  bzero(&stat_blk, sizeof(ndd_statblk_t));
  stat_blk.code = NDD_CONNECTED;
  NDD.nd_status(&NDD, &stat_blk);

  iomem_det ((void *) mapped_addr);
  simple_unlock(&p_dev->wrk.dev_lock);
  gm_aix_unlock_driver();

  GM_PRINT(GM_AIX_DEBUG_IP, ("gm_aix_ip_open returning GM_SUCCESS\n"));
  return GM_SUCCESS;

 abort_with_open_port:
  gm_disable_ethernet (sc->port);
  gm_close (sc->port);
  for (i = 0; i < GM_RECV_RING_CNT; i++)
    {
      pkt = sc->recv_ring + i;
      if (pkt->dma.dio.used_iovecs)
        {
          gm_aix_ether_arch_unencap (sc, pkt->m,  &pkt->dma);
        }
      if (pkt->m)
        {
          m_freem (pkt->m);
          pkt->m = NULL;
        }
    }

 abort_with_dev_lock:
  iomem_det ((void *) mapped_addr);
  simple_unlock(&p_dev->wrk.dev_lock);
  gm_aix_unlock_driver();

  GM_PRINT(GM_AIX_DEBUG_IP, ("gm_aix_ip_open returning 0x%x\n", status));
  return status;
}

int
gm_aix_ip_term (dev_ctrl_t *p_dev)
{
  int status;
  struct gm_aix_ether_sc *sc;

  GM_PRINT(GM_AIX_DEBUG_IP, ("gm_aix_ip_term: p_dev 0x%x\n", p_dev));

  sc = (struct gm_aix_ether_sc *)p_dev->ndd.ndd_correlator;
  if (WRK.adap_state != CLOSED_STATE) 
    {
      GM_WARN (("gm_aix_ip_term: adapter not closed\n"));
      return (-1);
    }

  if (status = ns_detach (&NDD)) 
    {
      GM_WARN (("gm_aix_ip_term: ns_detach failed 0x%x\n",status));
      return (-1);
    }

  gm_aix_ether_arch_unencap (sc, sc->bogus_m, &sc->bogus_dma);
  D_MAP_CLEAR (sc->d_handle);  
  gm_always_assert(sc->bogus_m);
  m_freem (sc->bogus_m);
  gm_free (sc);
  p_dev->ndd.ndd_correlator = NULL;

  GM_PRINT(GM_AIX_DEBUG_IP, ("gm_aix_ip_term returning GM_SUCCESS\n"));
  return (GM_SUCCESS);
}


static struct mbuf *
gm_aix_ether_arch_rxalloc (int how, int max_segs)
{
  struct mbuf *m_head, *mp, *m_tail;
  int len;
  unsigned long *ref_count;

  len = GM_ETHERNET_MTU;

  /* get the first header mbuf */
  mp = m_gethdr (how, MT_DATA);
  if (mp == NULL)
    {
      return NULL;
    }

  /* make it into a cluster mbuf */
  m_clgetm (mp, how, MCLBYTES);
  if ((mp->m_flags & M_EXT) == 0)
    {
      m_free (mp);
      return NULL;
    }
  /* misalign 14 byte ethernet header so that IP header ends up on a 4
     byte boundary */
  mp->m_data += 2;
  mp->m_len = MCLBYTES - 2;

  /* start building the chain of PAGE_SIZE mbufs */
  m_head = m_tail = mp;
  len -= MCLBYTES - 2;

  while (len > 0)
    {
      /* get the next header mbuf */
      mp = m_get (how, MT_DATA);
      if (mp == NULL)
        {
          goto abort_with_cluster;
        }
      /* attach to chain */
      m_tail->m_next = mp;
      m_tail = mp;
      /* use a "cluster" */
      m_clgetm (mp, how, MCLBYTES);
      if ((mp->m_flags & M_EXT) == 0)
	{
	  goto abort_with_cluster;
	}
      mp->m_len = MCLBYTES;
      len -= MCLBYTES;
    }

  return m_head;

 abort_with_cluster:
  m_freem (m_head);
  return (NULL);
}

static void
gm_aix_ether_arch_unencap (struct gm_aix_ether_sc *sc, struct mbuf *m,
			   struct gm_aix_ether_arch_dma_info *dma)
{
  D_UNMAP_LIST(sc->d_handle, &dma->dio);
  M_DO_PSEG_UNMAP(m);
  dma->dio.used_iovecs = 0;
}


int
gm_aix_ether_arch_encap (struct gm_aix_ether_sc *sc, struct mbuf *m_head,
			 struct gm_aix_ether_arch_dma_info *dma)
{
  struct mbuf *m;
  struct dio_64 vdio;
  struct d_iovec_64 iovec[GM_MAX_ETHERNET_SG_CNT];
  int i, status;

  M_DO_PSEG_MAP (m_head);
  i = 0;
  for (m = m_head; m != NULL; m = m->m_next)
    {
      if (m->m_len == 0)
	continue;	/* skip zero-length mbufs */

      if (i == GM_MAX_ETHERNET_SG_CNT)
	return 0;	/* check for a chain which is too long */
      
      iovec[i].iov_len = m->m_len;
      iovec[i].iov_base = (unsigned long long)m->m_data;
      iovec[i].xmp = M_XMEMD(m);
      i++;
    }
  
  vdio.total_iovecs = vdio.used_iovecs = i;
  vdio.dvec = iovec;

  dma->dio.used_iovecs = 0;

  status = D_MAP_LIST (sc->d_handle, DMA_READ | DMA_BYPASS, 1, &vdio, 
		       &dma->dio);
  if (status != DMA_SUCC)
    {
      if (status == DMA_NORES || status == DMA_DIOFULL) 
	{  
	  D_UNMAP_LIST(sc->d_handle, &dma->dio);
	  dma->dio.used_iovecs = 0;
	}
      goto abort_with_pseg;
    }

  for (i = 0; i < dma->dio.used_iovecs; i++)
    {
      dma->desc[i].ptr = dma->iovec[i].iov_base;
      dma->desc[i].len = dma->iovec[i].iov_len;
    }
  return dma->dio.used_iovecs;

 abort_with_pseg:
  M_DO_PSEG_UNMAP(m_head);
  return 0;


}

void
gm_aix_ip_xmit_mbuf (struct gm_aix_ether_sc *sc, struct mbuf *m_head, 
		     struct gm_aix_ether_arch_dma_info *dma)
{

  struct gm_aix_ether_pkt *pkt;
  struct mbuf *m;
  int broadcast, index, must_copy, nsegs;
  unsigned char *address;

  pkt = sc->send_ring + sc->sreq_index;

  address = mtod (m_head, unsigned char *);
  
  broadcast = address[0] & 0x1;
  bcopy (dma,  &pkt->dma, sizeof (pkt->dma));
  pkt->dma.dio.dvec = pkt->dma.iovec; 

  pkt->m = m_head;

  /* update the send_ring prior to actually sending so the 
   * interrupt handler is sure to have up-to-date send_ring
   * info 
   */
  sc->sreq += 1;
  sc->sreq_index += 1;
  if (sc->sreq_index > GM_SEND_RING_MAX_INDEX)
    {
      sc->sreq_index = 0;
    }

  if (broadcast)
    {
      gm_ethernet_broadcast (sc->port, pkt->dma.dio.used_iovecs, 
			     pkt->dma.desc);
    }		
  else	
    {
      gm_ethernet_send (sc->port, pkt->dma.dio.used_iovecs, 
			pkt->dma.desc, address);
    }
}

int
gm_aix_ip_output (ndd_t *the_ndd, struct mbuf *mbuf_head)
{
  dev_ctrl_t      *p_dev;
  struct gm_aix_ether_sc *sc;
  struct gm_aix_ether_pkt *pkt;
  ulong mapped_addr;
  struct mbuf *m, *mnext;
  int s, status, must_return_zero, nsegs, index;

  status = 0;
  must_return_zero = 0;

  sc = (struct gm_aix_ether_sc *) the_ndd->ndd_correlator;
  p_dev = sc->p_dev;

  mapped_addr = (ulong) iomem_att (IO_MAP);

  s = disable_lock(DDS.intr_priority, &p_dev->wrk.dev_lock); 

  for (m = mbuf_head; m; m = mnext)
    {
      /* check to see if there is space to queue the packet */
      if ((GM_SEND_RING_CNT == (sc->sreq - sc->sdone)) &&
	  (GM_SW_RING_CNT == (sc->sw_req - sc->sw_done)))
	{
	  status = EAGAIN;
	  goto abort_with_status;
	  
	}
      /* we've accepted a packet.  From here on out, we own it and we
      are responsible for freeing it.  We must return 0 to indicate
      that we own the packet and have freed  or will free it */
      must_return_zero = 1;

      mnext = m->m_nextpkt;
      m->m_nextpkt = NULL;

      nsegs = gm_aix_ether_arch_encap (sc, m, &sc->send_dma);

      /* if nsegs was zero, something went wrong and we need to
	 collapse the packet so we can DMA it */
      if (nsegs == 0)
	{
	  struct mbuf *mnew;

	  mnew = m_collapse (m, 1);
	  gm_aix_ether_xmit_copies++;
	  /* m_collapse will always free the chain even if it fails.
	     Unless we return 0, the upper layers will assume they own
	     the packet & will free it (again).  So we must always be
	     sure to return 0 if we've called m_collapse.
	     Furthermore, the docs are wrong -- it does not return a
	     chain with N page sized mbuf clusters.  Rather it returns
	     N clusters, with some of them possibly crossing page
	     boundaries and which are physically discountigous, making
	     it nearly useless for consoladating an mbuf chain for
	     DMA.  So we just give a limit of 1 mbuf and for 9K, it
	     can never cross more than 2 page boundaries.  Ugh.
	  */

	  if (mnew == NULL)
	    {
	      silent_drops++;
	      continue;
	    }
	  m = mnew;
	  gm_always_assert (m->m_next == NULL);

	  nsegs = gm_aix_ether_arch_encap (sc, m, &sc->send_dma);
	  if (nsegs == 0)
	    {
	      silent_drops++;
	      m_freem (m);
	      continue;
	    }
	}
	 
      if (sc->sreq - sc->sdone == GM_SEND_RING_CNT)
	{
	  /* no send tokens left, queue the packet */
	  if (sc->sw_req - sc->sw_done != GM_SW_RING_CNT)
	    {
	      index = sc->sw_req & GM_SW_RING_MAX_INDEX;
	      pkt = sc->sw_ring + index;
	      bcopy (&sc->send_dma,  &pkt->dma, sizeof (pkt->dma));
	      pkt->dma.dio.dvec = pkt->dma.iovec;
	      pkt->m = m;
	      sc->sw_req += 1;
	    }
	  else
	    {
	      /* no space in SW queue: this cannot happen */
	      gm_always_assert (0);
	    }
	  continue;
	}
      if (the_ndd->ndd_trace)
	{
	  (*(the_ndd->ndd_trace)) (the_ndd, m, m->m_data, the_ndd->ndd_trace_arg);
	}
      gm_aix_ip_xmit_mbuf (sc, m, &sc->send_dma);
    }

 abort_with_status:
  iomem_det ((void *) mapped_addr);
  unlock_enable(s, &p_dev->wrk.dev_lock);

  if (status && must_return_zero)
    {
      /* we need to free the remainder of the chain */
      while (m)
	{
	  silent_drops++;
	  mnext = m->m_nextpkt;
	  m_freem (m);
	  m = mnext;
	}
      return 0;
    }
  return status;
}

/* transmit complete interrupt handler.  This is entered via 
 * gm_handle_sent_tokens() from the __gm_ethernet_wake_callback().
 * The caller takes care of locking and unlocking the dev_lock.
 */

void
gm_aix_ether_txeof (void *ptr)
{
  struct gm_aix_ether_sc *sc = (struct gm_aix_ether_sc *) ptr;
  struct gm_aix_ether_pkt *pkt;
  struct mbuf *m;
  dev_ctrl_t *p_dev;
  int index, status, s;

  p_dev = sc->p_dev;

  if (sc->sreq - sc->sdone == 0)
    {
      GM_WARN (("Internal error: SENT_EVENT but no mbuf "
                "in send_ring!??,sreq=%u,sdone=%u\n", sc->sreq,
                sc->sdone));
      return;
    }
  pkt = sc->send_ring + sc->sdone_index;
  INCREMENT_DINT(GENSTATS.ndd_opackets_lsw, GENSTATS.ndd_opackets_msw);
  ADD_DINT(GENSTATS.ndd_obytes_lsw,  GENSTATS.ndd_obytes_msw, 
	   pkt->m->m_pkthdr.len);
  
  if (pkt->m->m_flags & M_BCAST)
    {
      INCREMENT_DINT(GENSTATS.ndd_ifOutBcastPkts_lsw, 
		     GENSTATS.ndd_ifOutBcastPkts_msw);
      ENTSTATS.bcast_tx_ok++;
    }
  else if (pkt->m->m_flags & M_MCAST)
    {
      INCREMENT_DINT(GENSTATS.ndd_ifOutMcastPkts_lsw, 
		     GENSTATS.ndd_ifOutMcastPkts_msw);
      ENTSTATS.mcast_tx_ok++;
    }
  else
{
      INCREMENT_DINT(GENSTATS.ndd_ifOutUcastPkts_lsw, 
		     GENSTATS.ndd_ifOutUcastPkts_msw);
    }

  gm_aix_ether_arch_unencap (sc, pkt->m, &pkt->dma);

  m_freem (pkt->m);
  pkt->m = NULL;
  sc->sdone += 1;
  sc->sdone_index += 1;
  if (sc->sdone_index > GM_SEND_RING_MAX_INDEX)
    {
      sc->sdone_index = 0;
    }

  /* restart xmit if it is waiting for space in the send ring */
  if (sc->sw_req != sc->sw_done)
    {
      index = sc->sw_done & GM_SW_RING_MAX_INDEX;
      sc->sw_done += 1;
      pkt = sc->sw_ring + index;

      if (NDD.ndd_trace)
	{
	  (*(NDD.ndd_trace)) (&NDD, pkt->m, pkt->m->m_data, 
			       NDD.ndd_trace_arg);
	}      

      gm_aix_ip_xmit_mbuf (sc, pkt->m, &pkt->dma);
      /* mbuf is now owned by the hardware transmit
	 ring and will be freed from there.  Zero it out
	 here to prevent a double free in gm_aix_ip_close */
      pkt->m = NULL; 
			
    }
}

#define IS_BCAST(addr)                                   	\
        ((*(uint32_t *)(&(addr)[0]) == 0xffffffff) &&           \
                (*(uint32_t *)(&(addr)[2]) == 0xffffffff))      \
                        ? (1) : (0)

void
gm_aix_ether_rxeof (void *ptr, gm_ethernet_recv_t *e) 
{
  struct gm_aix_ether_sc *sc;
  dev_ctrl_t *p_dev;
  struct mbuf *m, *m_head, *m_new;
  struct gm_aix_ether_pkt *pkt;
  int len, s;
  gm_u16_t myritype;
  unsigned int ulen = gm_ntoh_u32 (e->length);
  gm_u16_t csum = gm_ntoh_u16 (e->ip_checksum);
  gm_u8_t index = gm_ntoh_u8 (e->tag);
  gm_u8_t flags = gm_ntoh_u8 (e->flags);

  sc = (struct gm_aix_ether_sc *) ptr;
  p_dev = sc->p_dev;

  s = disable_lock(DDS.intr_priority, &p_dev->wrk.dev_lock);

  ENTSTATS.start_rx++;
  pkt = sc->recv_ring + index;
  m_head = pkt->m;

  if (m_head == sc->bogus_m)
    {
      /* 
       * used the "bogus" mbuf chain previous, data may be bad. Hope
       * we have enough resources & try to alloc a new chain this time
       */
      goto abort_need_realloc;
    }
  len = ulen;
  if (len < (sizeof (struct ether_header)))
    {
      GM_WARN (("received invalid packet: len=%d\n", len));
      goto abort_need_restock;
    }
  if (len < MHLEN - 2)
    {
      /* copy packet rather than waste space, this also avoids
       * calling dma_map_{un,}load again & improves small packet
       * latency */
      m_new = m_gethdr (M_DONTWAIT, MT_DATA);
      if (m_new == NULL)
        {
          goto abort_need_restock;
        }
      else
        {
	  m_new->m_data += 2;
          bcopy (mtod (m_head, char *), mtod (m_new, char *), len);

          m_new->m_len = len;
          m_head = m_new;
          /* recycle m_head (pkt->m) */
          gm_provide_receive_buffer_with_tag (sc->port,
					      sc->port->ethernet->scatter.list[index].ptr,
					     GM_ETHERNET_TAG_SIZE, GM_LOW_PRIORITY,
					     (gm_u32_t)index);
        }
    }
  else  /* don't copy, but pass the data up the stack */
    {  
      /* alloc new rx buffer */
      m_new = gm_aix_ether_arch_rxalloc (M_DONTWAIT, GM_MAX_ETHERNET_SCATTER_CNT);      
      if (m_new == NULL)
        {
          goto abort_need_restock;
        }
      gm_aix_ether_arch_unencap (sc, pkt->m, &pkt->dma);
      if (gm_aix_ether_arch_encap (sc, m_new, &pkt->dma))
        {
          pkt->m = m_new;
          gm_ethernet_provide_scatter_list_with_tag (sc->port, 
						     pkt->dma.dio.used_iovecs,
						     pkt->dma.desc,
						     index);
        }
      else
        {
          /* if the encapsulation fails, we have already unloaded the
           * old DMA map, so we have no choice but to attach a known
           * "bogus" mbuf chain so the lanai has buffers to dma into
           */
          m_freem (m_new);
          pkt->m = sc->bogus_m;
          gm_ethernet_provide_scatter_list_with_tag (sc->port, 
						      sc->bogus_dma.dio.used_iovecs,
                                                      sc->bogus_dma.desc,
						      index);
        }
#if (0)
      /* grab the HW checksum to save the IP stack from doing
       * software checksums. Note that we do this only when
       * we're not already touching the data by doing a copy 
       */
      if (flags & GM_IPV4_PARTIAL_CHECKSUM)
	{
	  gm_mbuf_rx_checksum (m_head, ulen, csum);
	}
#endif
    }

  /* let the upper layers know the len of the packet */
  m_head->m_pkthdr.len = len;

  /* update the interface counters */

  INCREMENT_DINT(GENSTATS.ndd_ipackets_lsw,
		 GENSTATS.ndd_ipackets_msw);
  ADD_DINT(GENSTATS.ndd_ibytes_lsw, GENSTATS.ndd_ibytes_msw, len);

    /* AIX wants flags set in the mbuf for mcast an bcast pkts */
  if (IS_BCAST (mtod(m_head, caddr_t)))
    {
      ENTSTATS.bcast_rx_ok++;
      m_head ->m_flags |= M_BCAST;
    }
  else if (mtod(m_head, gm_u8_t) & 0x01)
    {
      ENTSTATS.mcast_rx_ok++;
      m_head ->m_flags |= M_MCAST;
    }

  /* chop off excess elements of mbuf chain */
  for (m = m_head; (len > 0) && (m != NULL); m = m->m_next)
    {
      if (m->m_len >= len)
        {
          /* adjust length & free rest of chain */
          m->m_len = len;
          if (m->m_next != NULL)
            {
              m_freem (m->m_next);
              m->m_next = NULL;
            }
        }
      len -= m->m_len;
    }


  /* pass the packet up to the upper layers */

  /* make sure to put this mbuf at the tail of the queue
     so as to prevent TCP from seeing out-of-order packets */

  m_head->m_nextpkt = NULL;
  
  if (WRK.rx_tail == NULL)
    {
      WRK.rx_head = m_head;
    }
  else
    {
      WRK.rx_tail->m_nextpkt = m_head;
    }
  WRK.rx_tail = m_head;

  /* wake up thread to pass message to upper layers */
  
  gm_aix_ip_sched_offlevel (p_dev);


  unlock_enable(s, &p_dev->wrk.dev_lock);
  return;                       /* done with this packet, return */

  /* error handling code, expects old mbuf in m_head & valid dma
   * for m_head in pkt->dma. */

 abort_need_realloc:            /* get here if bogus_m was used previously */
  m_head = gm_aix_ether_arch_rxalloc (M_DONTWAIT, GM_MAX_ETHERNET_SCATTER_CNT);
  if (m_head == NULL)
    {
      m_head = sc->bogus_m;
   }
  else
    {
      if (!gm_aix_ether_arch_encap (sc, m_head, &pkt->dma))
        {
          m_freem (m_head);
          m_head = sc->bogus_m;
        }
    }
 abort_need_restock:
  /* restock the lanai, re-using mbuf chain in m_head
     or the bogus dma pages in bogus_dma */
  if (m_head == sc->bogus_m)
    {
      gm_ethernet_provide_scatter_list_with_tag (sc->port, 
						 sc->bogus_dma.dio.used_iovecs,
                                         	 sc->bogus_dma.desc,
						 index);
    }
  else
    {
      gm_ethernet_provide_scatter_list_with_tag (sc->port, 
						 pkt->dma.dio.used_iovecs,
                                         	 pkt->dma.desc,
						 index);
    }

  GENSTATS.ndd_ipackets_drop++;
  unlock_enable(s, &p_dev->wrk.dev_lock);
}




/*
  This file uses GM standard indentation.

  Local Variables:
  c-file-style:"gnu"
  tab-width:8
  End:
*/

