/*
  gmip_out.c
  architecture specific code for AIX 4 GM CDLI driver
*/

/* NOTE: The include of gm_arch_debug.h MUST be first! */
#include "gm_arch_debug.h"
#include "gmip.h"
#include "gm_debug.h"

extern dd_ctrl_t dd_ctrl;

int
gm_aix_ip_start_offlevel (struct intr *p_ihs)
{
	int ipri;			/* Current priority */
	dev_ctrl_t	*p_dev;
	struct mbuf	*mbufp;

	p_dev = (dev_ctrl_t *) (((ulong) p_ihs) - (ulong) sizeof (struct intr));

	GM_PRINT(GM_AIX_DEBUG_IP,
		 ("gm_aix_ip_start_offlevel: p_dev 0x%x\n", p_dev));

	/* Handover all mbufs to the demuxer */
	ipri = disable_lock(DDS.intr_priority, &p_dev->wrk.dev_lock);
	mbufp = WRK.rx_head;
	WRK.rx_head = WRK.rx_tail = NULL;
	unlock_enable(ipri, &p_dev->wrk.dev_lock);

	if (mbufp)
		NDD.nd_receive(&NDD, mbufp);

	return (INTR_SUCC);
}

void
gm_aix_ip_init_offlevel (dev_ctrl_t *p_dev)
{
	GM_PRINT(GM_AIX_DEBUG_IP,
		 ("gm_aix_ip_init_offlevel: Doing init OFFLVL3\n"));

	(p_dev)->dpc_ihs.next = (struct intr *) NULL;
	(p_dev)->dpc_ihs.handler = gm_aix_ip_start_offlevel;
	(p_dev)->dpc_ihs.bus_type = BUS_NONE;
	(p_dev)->dpc_ihs.flags = 0;
	(p_dev)->dpc_ihs.level = INT_OFFLVL;
	(p_dev)->dpc_ihs.priority = INTOFFL2;;
	(p_dev)->dpc_ihs.bid = DEV_BID;
}


void
gm_aix_ip_sched_offlevel (dev_ctrl_t *p_dev)
{
	GM_PRINT(GM_AIX_DEBUG_IP, ("gm_aix_ip_sched_offlevel: Calling i_sched\n"));
	i_sched (&(p_dev->dpc_ihs));
}

void
gm_aix_ip_clear_offlevel (dev_ctrl_t *p_dev)
{
	if (p_dev->dpc_ihs.flags & I_SCHED) {
		GM_PRINT(GM_AIX_DEBUG_IP,
			 ("gm_aix_ip_clear_offlevel: Calling i_clear\n"));
		i_clear (&(p_dev->dpc_ihs));
	}
	else {
		GM_PRINT(GM_AIX_DEBUG_IP,
			 ("gm_aix_ip_clear_offlevel: Not Calling i_clear\n"));
	}
}
