/*
  gmip_ioctl.c
  architecture specific code for AIX 4 GM CDLI driver
*/

/* NOTE: The include of gm_arch_debug.h MUST be first! */
#include "gm_arch_debug.h"
#include "gmip.h"
#include "gm_debug.h"

/*
 * CONTENTS:
 *      "Forward" declarations
 *      gm_aix_ip_cmdname [conditional]
 *	gm_aix_ip_ioctl
 *	get_mib
 *	update_stats
 *	clear_stats
 *	get_rx_addrs
 *	add_filter
 *	del_filter
 *	add_multi_addr
 *	del_multi_addr
 */



/***********************************************************************
 * "Forward" function declarations
 ***********************************************************************/

int update_stats(dev_ctrl_t *p_dev);
int clear_stats(dev_ctrl_t *p_dev);
int add_filter(dev_ctrl_t *p_dev, ns_8022_t *filter);
int del_filter(dev_ctrl_t *p_dev, ns_8022_t *filter);
int get_mib(dev_ctrl_t *p_dev);
int get_rx_addrs(dev_ctrl_t *p_dev, char *pbuf, uint bsize);



#if GM_AIX_DEBUG_IP
char *
gm_aix_ip_cmdname(int the_cmd)
{
  char invalid_cmd_text[19];

  char *ioctl_cmd_names[] = {
    "*** Invalid (0) code",	/*  0 */
    "NDD_GET_STATS",		/*  1 */
    "NDD_ENABLE_MULTICAST",	/*  2 */
    "NDD_DISABLE_MULTICAST",	/*  3 */
    "NDD_PROMISCUOUS_ON",	/*  4 */
    "NDD_PROMISCUOUS_OFF",	/*  5 */
    "NDD_ADD_FILTER",		/*  6 */
    "NDD_DEL_FILTER",		/*  7 */
    "NDD_MIB_QUERY",		/*  8 */
    "NDD_MIB_GET",		/*  9 */
    "NDD_MIB_SET",		/*  A */
    "NDD_ADD_STATUS",		/*  B */
    "NDD_DEL_STATUS",		/*  C */
    "NDD_ENABLE_ADDRESS",	/*  D */
    "NDD_DISABLE_ADDRESS",	/*  E */
    "NDD_MIB_ADDR",		/*  F */
    "NDD_CLEAR_STATS"		/* 10 */
  } ;
#define MAX_CMD_NAME_IX NDD_CLEAR_STATS


  if ((the_cmd < 0) || (the_cmd > MAX_CMD_NAME_IX)) {
    sprintf(invalid_cmd_text, "0x%x", the_cmd);
    return (invalid_cmd_text);
  }
  else {
    return (ioctl_cmd_names[the_cmd]);
  }
}
#endif /* GM_AIX_DEBUG_IP */

uchar ent_broad_adr[ENT_NADR_LENGTH] = {0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF};

/* This table identifies the MIB statistics supported by this driver.  It is 
 * returned on an NDD_MIB_QUERY ioctl cmd.  It is constant information.
 */
const static ethernet_all_mib_t ent_mib_status = {
	/* Generic Interface Extension Table */
        MIB_READ_ONLY,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, /* ifExtnsChipSet        */
        MIB_READ_ONLY,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, /* ifExtnsRevWare        */
        MIB_READ_ONLY,		/* ifExtnsMulticastsTransmittedOks */
        MIB_READ_ONLY,		/* ifExtnsBroadcastsTransmittedOks */
        MIB_READ_ONLY,		/* ifExtnsMulticastsReceivedOks    */
        MIB_READ_ONLY,		/* ifExtnsBroadcastsReceivedOks    */
        MIB_READ_ONLY,		/* ifExtnsPromiscuous              */
	/* Generic Interface Test Table */
        MIB_NOT_SUPPORTED,	/* ifEXtnsTestCommunity */
        MIB_NOT_SUPPORTED,	/* ifEXtnsTestRequestId */
        MIB_NOT_SUPPORTED,	/* ifEXtnsTestType */
        MIB_NOT_SUPPORTED,	/* ifEXtnsTestResult */
        MIB_NOT_SUPPORTED,	/* ifEXtnsTestCode */
	/* Generic Receive Address Table Length */
	MIB_READ_ONLY,			/* RcvAddrTable                       */
	/* Ethernet-like Statistics group */
        MIB_READ_ONLY,		/* Alignment Errors: driver from RX Status */
        MIB_READ_ONLY,		/* FCS Errors: driver from RX Status */
        MIB_READ_ONLY,		/* Single Collision Frames: adapter stats */
        MIB_READ_ONLY,		/* Multiple Collision Frames: adapter stats */
        MIB_READ_ONLY,		/* SQE Test Errors: adapter stats */
        MIB_READ_ONLY,		/* Deferred Transmissions: adapter stats */
        MIB_READ_ONLY,		/* Late Collisions: adapter stats */
        MIB_READ_ONLY,		/* Excessive Collisions: driver from TX Status */
        MIB_READ_ONLY,		/* Internal Mac Transmit Errors */
        MIB_READ_ONLY,		/* Carrier Sense Errors: adapter stats */
        MIB_READ_ONLY,		/* Frame Too Longs: driver from RX Status */
        MIB_READ_ONLY,		/* Internal Mac Receive Errors  */
	/* Ethernet-like Collision Statistics group */
   	MIB_NOT_SUPPORTED,              /* dot3CollCount                      */
        MIB_NOT_SUPPORTED, MIB_NOT_SUPPORTED, MIB_NOT_SUPPORTED,
        MIB_NOT_SUPPORTED, MIB_NOT_SUPPORTED, MIB_NOT_SUPPORTED,
        MIB_NOT_SUPPORTED, MIB_NOT_SUPPORTED, MIB_NOT_SUPPORTED,
        MIB_NOT_SUPPORTED, MIB_NOT_SUPPORTED, MIB_NOT_SUPPORTED,
        MIB_NOT_SUPPORTED, MIB_NOT_SUPPORTED, MIB_NOT_SUPPORTED,
        MIB_NOT_SUPPORTED,              /* dot3CollFrequencies                */
        MIB_NOT_SUPPORTED, MIB_NOT_SUPPORTED, MIB_NOT_SUPPORTED,
        MIB_NOT_SUPPORTED, MIB_NOT_SUPPORTED, MIB_NOT_SUPPORTED,
        MIB_NOT_SUPPORTED, MIB_NOT_SUPPORTED, MIB_NOT_SUPPORTED,
        MIB_NOT_SUPPORTED, MIB_NOT_SUPPORTED, MIB_NOT_SUPPORTED,
        MIB_NOT_SUPPORTED, MIB_NOT_SUPPORTED, MIB_NOT_SUPPORTED
};

/*****************************************************************************/
/*
 * NAME:                  gm_aix_ip_ioctl
 *
 * FUNCTION:              ethernet driver ioctl routine.
 *
 * EXECUTION ENVIRONMENT: process only
 *
 * NOTES:
 *
 * RETURNS:               0 - OK
 *	                  ENETUNREACH - device is currently unreachable
 *	                  ENETDOWN - device is down
 *	                  EINVAL - invalid paramter
 *	                  ENOMEM - unable to allocate required memory
 *	                  EOPNOTSUPP - operation not supported
 */
/*****************************************************************************/
gm_aix_ip_ioctl(
  ndd_t *p_ndd,				/* ndd structure in device control */
  int cmd,				/* ioctl command */
  caddr_t arg,				/* ioctl command argument */
  int length				/* length of argument */
)
{
	struct gm_aix_ether_sc *sc;
	dev_ctrl_t *p_dev;		/* Device control structure ptr */
	int rc;				/* Function return code */
	int i;				/* Loop index */
	int ipri;			/* Current priority */
	ulong mapped_addr;

	sc = (struct gm_aix_ether_sc *) p_ndd->ndd_correlator;
	p_dev = sc->p_dev;

#if GM_AIX_DEBUG_IP
	/* conditional because function gm_aix_ip_cmdname() is conditional */
	GM_PRINT(GM_AIX_DEBUG_IP, ("gm_aix_ip_ioctl: p_dev 0x%x cmd %s\n",
				   p_dev, gm_aix_ip_cmdname(cmd)));
#endif
	/* TRC_FLOW_2(IOCTL_ENTRY, IN, *(uint*)DDS.lname, p_dev); */

	rc = 0;
	ipri = disable_lock(DDS.intr_priority, &p_dev->wrk.dev_lock);

	/*
	mapped_addr = (ulong) iomem_att (IO_MAP);
	GM_PRINT(GM_AIX_DEBUG_IP,
		 ("gm_aix_ip_ioctl: mapped_addr 0x%x\n", mapped_addr));
	if (mapped_addr != p_dev->mapped_addr) {
		GM_WARN(("gm_aix_ip_ioctl: invalid mapped addr\n"));
		iomem_det ((void *) mapped_addr);
		unlock_enable(ipri, &p_dev->wrk.dev_lock);
		return (ENXIO);

	}
	*/

	/* Branch on the ioctl command */
	switch(cmd) {

	/* Get generic and ethernet statistics: ndd_genstats + ent_genstats
	 * Note: there are no unique device stats; so NDD_GET_ALL_STATS logic
	 *       is identical to NDD_GET_STATS
	 */
	case NDD_GET_STATS:
	case NDD_GET_ALL_STATS:

		/* TRC_FLOW_0(IOCTL_ENTRY, NDD_GET_STATS); */

 		/* Make sure there is enough space. */
		if (length != sizeof(ent_ndd_stats_t)) {
			rc = EINVAL;
			break;
		}

		/* Update statistics */
		if (rc = update_stats(p_dev)) {
			break;
		}

		/* Copy the statistics to the user's buffer. */
		bcopy(&GENSTATS, arg, sizeof GENSTATS);
		bcopy(&ENTSTATS, arg + sizeof GENSTATS, sizeof ENTSTATS);

		break;

	case NDD_CLEAR_STATS:

		/* TRC_FLOW_0(IOCTL_ENTRY, NDD_CLEAR_STATS); */
		rc = clear_stats(p_dev);

		break;

#if ! GM_AIX_SUPPORT_MULTI
	/* No multicast support */
	case NDD_ENABLE_MULTICAST:
	case NDD_DISABLE_MULTICAST:

		rc = EOPNOTSUPP;
		break;
#else
	/* Enable the adapter to receive all multicast packets */
	case NDD_ENABLE_MULTICAST:

		/* TRC_FLOW_0(IOCTL_ENTRY, NDD_ENABLE_MULTICAST); */

		/* If the adapter is not open, punt */
		if (WRK.adap_state != OPEN_STATE) {
			if (WRK.adap_state == DEAD_STATE) {
				rc = ENETDOWN;
			} else {
				rc = ENETUNREACH;
			}
			break;
		}

		/* If the adapter is not already receiving multicast packets,
		 * enable that mode
		 */
		if (WRK.enable_multicast++ == 0) {

			NDD.ndd_flags |= NDD_MULTICAST;

			enable_multicast(p_dev);
		}

		break;


	/* Disable multicast reception */
	case NDD_DISABLE_MULTICAST:

		/* TRC_FLOW_0(IOCTL_ENTRY, NDD_DISABLE_MULTICAST); */

		/* If not already, error */
		if (WRK.enable_multicast == 0) {
			rc = EINVAL;
			break;
		}

		/* If counter now zero, disable multicast mode */
		if (--WRK.enable_multicast == 0) {

			NDD.ndd_flags &= ~NDD_MULTICAST;

			disable_multicast(p_dev);
		}
		break;

#endif /* GM_AIX_SUPPORT_MULTI */


	/* Enable the adapter to receive all packets (promiscuous mode) */
	case NDD_PROMISCUOUS_ON:

		/* TRC_FLOW_0(IOCTL_ENTRY, NDD_PROMISCUOUS_ON); */

		/* If the adapter is not open, punt */
		if (WRK.adap_state != OPEN_STATE) {
			if (WRK.adap_state == DEAD_STATE) {
				rc = ENETDOWN;
			} else {
				rc = ENETUNREACH;
			}
			break;
		}

		/* If the adapter is not already receiving all packets,
		 * enable that mode
		 */
		if (WRK.enable_promiscuous++ == 0) { 

			NDD.ndd_flags |= NDD_PROMISC;
			MIB.Generic_mib.ifExtnsEntry.promiscuous = TRUE;

			enable_promiscuous(p_dev);
		}

		break;


	/* Disable promiscuous mode */
	case NDD_PROMISCUOUS_OFF:

		/* TRC_FLOW_0(IOCTL_ENTRY, NDD_PROMISCUOUS_OFF); */

		/* If not already, error */
		if (!WRK.enable_promiscuous) {
			rc = EINVAL;
			break;
		}

		/* If counter now zero, disable promiscuous mode */
		if (--WRK.enable_promiscuous == 0) {

			NDD.ndd_flags &= ~NDD_PROMISC;
			MIB.Generic_mib.ifExtnsEntry.promiscuous = FALSE;

			disable_promiscuous(p_dev);
		}
		break;


	/* Add receive filter */
	case NDD_ADD_FILTER:

		rc = add_filter(p_dev, (ns_8022_t*)arg);

		break;


	/* Delete receive filter */
	case NDD_DEL_FILTER:

		rc = del_filter(p_dev, (ns_8022_t*)arg);

		break;


	/* Query the MIB support status on the driver */
	case NDD_MIB_QUERY:

		/* TRC_FLOW_0(IOCTL_ENTRY, NDD_MIB_QUERY); */

		/* Check the adapter state. */
		if (WRK.adap_state != OPEN_STATE) {
			if (WRK.adap_state == DEAD_STATE) {
				rc = ENETDOWN;
			} else {
				rc = ENETUNREACH;
			}
			break;
		}

		/* If destination buffer is too small, error */
		if (length < sizeof(ethernet_all_mib_t)) {
			rc = EINVAL;
			break;
		}

		/* Copy the static structure to the destination buffer in
		 * kernel space
		 */
		bcopy(&ent_mib_status, arg, sizeof(ethernet_all_mib_t));
		break;


	/* Get all MIB values */
	case NDD_MIB_GET:
      
		/* TRC_FLOW_0(IOCTL_ENTRY, NDD_MIB_GET); */

		/* If destination buffer is too small, error */
		if (length < sizeof(ethernet_all_mib_t)) {
			rc = EINVAL;
			break;
		}

		/* Get the mib statistics */
		if (rc = get_mib(p_dev)) {
			break;
		}

		/* Copy the mibs to the destination buffer in kernel space */
		bcopy(&MIB, arg, sizeof MIB);

		break;


	/* Return the receive address table, consisting of all valid
	 * addresses for the device, including: host network address,
	 * broadcast address, and any multicast addresses.
	 */
	case NDD_MIB_ADDR:

		rc = get_rx_addrs(p_dev, arg, length);

		break;


	/* Add a status request from the following list:
	 * 1. Send Bad RX Packet
	 * 2. it's a short list isn't it
	 *
	 * Note: This is a request for a special status; standard status
	 *       notifications are always sent
	 */
	case NDD_ADD_STATUS: {
		ns_com_status_t *p_stat = (ns_com_status_t *)arg;

		/* TRC_FLOW_0(IOCTL_ENTRY, NDD_ADD_STATUS); */

		/* If device is not open, punt */
		if (WRK.adap_state != OPEN_STATE) {
			if (WRK.adap_state == DEAD_STATE) {
				rc = ENETDOWN;
			} else {
				rc = ENETUNREACH;
			}
			break;
		}

		/* If not status request, ignore it */
		if (p_stat->filtertype != NS_STATUS_MASK) {
			rc = EINVAL;
			break;
		}

		/* If request to send bad rx packets, indicate */
		if (p_stat->mask & NDD_BAD_PKTS) {
			NDD.ndd_flags |= ENT_RCV_BAD_FRAME;
			WRK.send_bad_rx_pkts++;
		}

		/* If not recognized, keep track */
		else 	WRK.otherstatus++;

		break;
	}

	/* Delete a status request */
	case NDD_DEL_STATUS: {
		ns_com_status_t *p_stat = (ns_com_status_t *)arg;

		/* TRC_FLOW_0(IOCTL_ENTRY, NDD_DEL_STATUS); */

		/* If not status request, ignore it */
		if (p_stat->filtertype != NS_STATUS_MASK) {
			rc = EINVAL;
			break;
		}

		/* If request to stop sending bad rx packets, indicate */
		if (p_stat->mask & NDD_BAD_PKTS) {

			/* If not doing it, why stop */
			if (WRK.send_bad_rx_pkts == 0) {
				rc = EINVAL;
				break;
			}

			if (--WRK.send_bad_rx_pkts == 0)
				NDD.ndd_flags &= ~ENT_RCV_BAD_FRAME;

		}

		/* If not recognized, keep track */
		else {
			if (WRK.otherstatus == 0) {
				rc = EINVAL;
				break;
			}
			WRK.otherstatus--;
		}

		break;
	}

#if ! GM_AIX_SUPPORT_MULTI
	/* No multicast support */
	case NDD_ENABLE_ADDRESS:
	case NDD_DISABLE_ADDRESS:

		rc = EOPNOTSUPP;
		break;
#else
	/* Add a multicast address to the multicast table */
	case NDD_ENABLE_ADDRESS:
		
		/* Add the address referenced by arg */
		if ((rc = add_multi_addr(p_dev, arg)) != 0) {
			break;
		}

		/* If not already, set adapter to receive group addresses */
		if (WRK.enable_multicast++ == 0)
			enable_multicast(p_dev);

		break;


	/* Delete a multicast address from the multicast table */
	case NDD_DISABLE_ADDRESS:

		/* Delete the address referenced by arg */
		if ((rc = del_multi_addr(p_dev, arg)) != 0) {
			break;
		}

		/* If not zero already, decrement the enable multicast counter
		   and if now zero, disable adapter mulitcast reception
		 */
		if (WRK.enable_multicast == 0) break;
		if (--WRK.enable_multicast == 0)
			disable_multicast(p_dev);

		break;
#endif /* GM_AIX_SUPPORT_MULTI */

	default:
/*		TRC_ERR_2(IOCTL_ENTRY, CMD_NOT_RECOGNIZED, cmd, arg); */

		GM_WARN(("gm_aix_ip_ioctl: unsupported ioctl 0x%x\n", cmd));

		rc = EOPNOTSUPP;
		break;
	}

	/* iomem_det ((void *) mapped_addr); */
	unlock_enable(ipri, &p_dev->wrk.dev_lock);

	GM_PRINT(GM_AIX_DEBUG_IP, ("gm_aix_ip_ioctl: returning 0x%x\n", rc));
	/* TRC_FLOW_2(IOCTL_ENTRY, OUT, *(uint*)DDS.lname, rc); */

	return(rc);
}


/*****************************************************************************/
/*
 * NAME:     		  get_mib
 *
 * FUNCTION: 		  Gather the current MIB statistics
 *
 * EXECUTION ENVIRONMENT: process only
 *
 * NOTES:
 *
 * RETURNS:		  errno values
 */
/*****************************************************************************/
int 
get_mib(
  dev_ctrl_t *p_dev			/* Device control structure */
)
{
	update_stats(p_dev);

	bcopy(UNKNOWNCHIPSET, MIB_IFEXTNS.chipset, CHIPSETLENGTH);
	MIB_IFEXTNS.mcast_tx_ok = ENTSTATS.mcast_tx_ok;
	MIB_IFEXTNS.bcast_tx_ok = ENTSTATS.bcast_tx_ok;
	MIB_IFEXTNS.mcast_rx_ok = ENTSTATS.mcast_rx_ok;
	MIB_IFEXTNS.bcast_rx_ok = ENTSTATS.bcast_rx_ok;
	MIB_IFEXTNS.promiscuous = (NDD.ndd_flags & NDD_PROMISC) == NDD_PROMISC;
	MIB.Generic_mib.RcvAddrTable = WRK.multi_addr_count + 2;

	MIB_DOT3STATS.align_errs        = ENTSTATS.align_errs;
	MIB_DOT3STATS.fcs_errs          = ENTSTATS.fcs_errs;
	MIB_DOT3STATS.s_coll_frames     = ENTSTATS.s_coll_frames;
	MIB_DOT3STATS.m_coll_frames     = ENTSTATS.m_coll_frames;
	MIB_DOT3STATS.sqetest_errs      = ENTSTATS.sqetest;
	MIB_DOT3STATS.defer_tx          = ENTSTATS.defer_tx;
	MIB_DOT3STATS.late_collisions   = ENTSTATS.late_collisions;
	MIB_DOT3STATS.excess_collisions = ENTSTATS.excess_collisions;
	MIB_DOT3STATS.mac_tx_errs       = ENTSTATS.underrun +
							 ENTSTATS.tx_timeouts;
	MIB_DOT3STATS.carriers_sense    = ENTSTATS.carrier_sense;
	MIB_DOT3STATS.long_frames       = ENTSTATS.long_frames;
	MIB_DOT3STATS.mac_rx_errs       = ENTSTATS.rx_drop + ENTSTATS.overrun +
		               ENTSTATS.rx_collisions + ENTSTATS.no_resources;

	return(0);
} 

/*****************************************************************************/
/*
 * NAME:                  update_stats
 *
 * FUNCTION:              Update ndd_genstats + ent_genstats
 *                        
 * EXECUTION ENVIRONMENT: process only
 *
 * NOTES:		  This function updates standard statistics structures
 *			  located in the device control structure.  The source
 *			  of the statistics in these structures is as follows:
 *
 *			  1. The driver maintains some directly when tx and rx
 *			     events occur
 *			  2. This function snapshots some of the driver's
 *			     working variables for others
 *			  3. The rest are read from the adapter's statistics
 *			     registers and accumulated by the driver when
 *			     an UPDATE_STATS interrupt occurs
 *
 * RETURNS:               errno values
 */
/*****************************************************************************/
int
update_stats(
  dev_ctrl_t *p_dev			/* Device control structure */
)
{
	int   i;			/* Simple loop counter. */

	/* If not open, punt */
	if (WRK.adap_state != OPEN_STATE) {
printf("update: adapt_state = 0x%x\n",WRK.adap_state);
		if (WRK.adap_state == DEAD_STATE)
			return(ENETDOWN);

		return(ENETUNREACH);
	}

	/* Update from working variables */
	GENSTATS.ndd_elapsed_time = ELAPSED_TIME(WRK.ndd_stime);
	GENSTATS.ndd_xmitque_cur  = 0;

	ENTSTATS.device_type = ENT_GM_PCI;
	ENTSTATS.dev_elapsed_time = ELAPSED_TIME(WRK.dev_stime);
	ENTSTATS.ndd_flags = NDD.ndd_flags;

	for (i = 0; i < ENT_NADR_LENGTH; i++) {
		ENTSTATS.ent_nadr[i] = WRK.net_addr[i];
	}

	ENTSTATS.sw_txq_len = 0;
	ENTSTATS.hw_txq_len = 0;

	ENTSTATS.rx_collisions   += 0;
	ENTSTATS.carrier_sense   += ADAPSTATS.carrier_lost;
	ENTSTATS.cts_lost        += 0;
	ENTSTATS.late_collisions += ADAPSTATS.late_coll;
	ENTSTATS.sqetest         += ADAPSTATS.no_cd_beat;
	ENTSTATS.defer_tx        += ADAPSTATS.tx_defer;
	ENTSTATS.s_coll_frames   += ADAPSTATS.one_coll;
	ENTSTATS.m_coll_frames   += ADAPSTATS.multiple_coll;

	return(0);
}

 
/*****************************************************************************/
/*
 * NAME:     		  clear_stats
 *
 * FUNCTION:		  Clear all the statistics
 *
 * EXECUTION ENVIRONMENT: process only
 *
 * NOTES:
 *
 * RETURNS:		  errno values
 */
/*****************************************************************************/
int 
clear_stats(
  dev_ctrl_t *p_dev			/* Device control structure */
)
{
	/* If not open, punt */
	if (WRK.adap_state != OPEN_STATE) {
		if (WRK.adap_state == DEAD_STATE)
			return(ENETDOWN);

		return(ENETUNREACH);
	}

	/* Clear the statistics on the adapter by reading them */

	/* Reset the start time for both ndd and device */
	WRK.ndd_stime = WRK.dev_stime = lbolt;

	/* Clear all statistics structures */
	bzero(&GENSTATS, sizeof GENSTATS);
	bzero (&ENTSTATS, sizeof ENTSTATS);
	bzero (&ADAPSTATS, sizeof ADAPSTATS);

	return(0);
} 


/*****************************************************************************/
/*
 * NAME:     		  get_rx_addrs
 *
 * FUNCTION:		  Builds a receive address table and copies it to the
 *			  destination buffer. Table contains the following
 *			  addresses: network, broadcast, any multicast
 *
 * EXECUTION ENVIRONMENT: process only
 *
 * NOTES:		  
 *
 * RETURNS:		  errno values
 */
/*****************************************************************************/

int
get_rx_addrs(
  dev_ctrl_t *p_dev,			/* Device control structure */
  char *pbuf,				/* Destination buffer */
  uint bsize				/* Size of destination buffer */
)
{
	ndd_mib_addr_t *p_table = (ndd_mib_addr_t *)pbuf;
	ndd_mib_addr_elem_t *p_elem;
	int elem_len;
	int count = 0;
	ent_multi_t *mtab = &WRK.multi_table;
	int i;
	int rc = 0;
	
	/* TRC_FLOW_0(IOCTL_ENTRY, NDD_MIB_ADDR); */

	/* If not open, punt */
	if (WRK.adap_state != OPEN_STATE) {
		if (WRK.adap_state == DEAD_STATE)
			return ENETDOWN;

		return ENETUNREACH;
	}

	/* If not enough room for even first address, report error */
	if (bsize < sizeof(ndd_mib_addr_t)) {
		return E2BIG;
	}

	/* Since the element structure allows for a minimal address length,
	 * calculate the actual size of an element
	 */
	elem_len = sizeof(ndd_mib_addr_elem_t) +
				      ENT_NADR_LENGTH - sizeof p_elem->address;

	/* Adjust buffer by size of count element */
	bsize -= sizeof p_table->count;
	pbuf += sizeof p_table->count;

	/* If room, copy the network address */
	if (bsize < elem_len) {
		rc = E2BIG;
		goto exit;
	}

	p_elem = (ndd_mib_addr_elem_t *)pbuf;
	p_elem->status = NDD_MIB_VOLATILE;
	p_elem->addresslen = ENT_NADR_LENGTH;
	COPY_NADR(WRK.net_addr, p_elem->address);
	bsize -= elem_len;
	pbuf += elem_len;
	count++;
	
	/* If room, copy the broadcast address */
	if (bsize < elem_len) {
		rc = E2BIG;
		goto exit;
	}

	p_elem = (ndd_mib_addr_elem_t *)pbuf;
	p_elem->status = NDD_MIB_NONVOLATILE;
	p_elem->addresslen = ENT_NADR_LENGTH;
	COPY_NADR(ent_broad_adr, p_elem->address);
	bsize -= elem_len;
	pbuf += elem_len;
	count++;
		
	/* For each multicast address table, copy each multicast addresses */
	while (mtab != NULL) {
		for (i=0; i < mtab->in_use; i++) {

			/* If buffer not big enough, no more */
			if (bsize < elem_len) {
				rc = E2BIG;
				goto exit;
			}

			/* Copy the multicast address */
			p_elem = (ndd_mib_addr_elem_t *)pbuf;
			p_elem->status = NDD_MIB_VOLATILE;
			p_elem->addresslen = ENT_NADR_LENGTH;
			COPY_NADR(mtab->m_slot[i].m_addr, p_elem->address);
			bsize -= elem_len;
			pbuf += elem_len;
			count++;
		}

		mtab = mtab->next;
	}
 exit:
	/* Put the final count into the buffer */
	p_table->count = count;

	return rc;
}

 
/*****************************************************************************/
/*
 * NAME:     		  add_filter
 *
 * FUNCTION:		  Adds a receive filter to a device-specific list
 *
 * EXECUTION ENVIRONMENT: process only
 *
 * NOTES:		  Not implemented.  Demuxer will receive all addressed
 *			  packets.
 *
 * RETURNS:		  errno values
 */
/*****************************************************************************/
int 
add_filter(
	 dev_ctrl_t *p_dev,			/* Device control structure */
	 ns_8022_t *filter			/* Filter structure */
)
{
	/* TRC_FLOW_2(IOCTL_ENTRY, NDD_ADD_FILTER, filter->filtertype,
		   filter->ethertype); */

	return EOPNOTSUPP;
}

 
/*****************************************************************************/
/*
 * NAME:     		  del_filter
 *
 * FUNCTION:		  Removess a receive filter from a device-specific list
 *
 * EXECUTION ENVIRONMENT: process only
 *
 * NOTES:		  Not implemented.  Demuxer will receive all addressed
 *			  packets.
 *
 * RETURNS:		  errno values
 */
/*****************************************************************************/
int 
del_filter(
  dev_ctrl_t *p_dev,			/* Device control structure */
  ns_8022_t *filter			/* Filter structure */
)
{
	/* TRC_FLOW_2(IOCTL_ENTRY, NDD_DEL_FILTER, filter->filtertype,
					    filter->ethertype); */

	return EOPNOTSUPP;
}

#if GM_AIX_SUPPORT_MULTI

/*****************************************************************************/
/*
 * NAME:     		  add_multi_addr
 *
 * FUNCTION: 		  Add a multicast address to the multicast table
 *
 * EXECUTION ENVIRONMENT: process only
 *
 * NOTES:		  The multicast table consists of a base table in
 *			  the dev_ctrl structure and xmalloc()ed extensions.
 *
 * RETURNS:		  errno value
 */
/*****************************************************************************/

int
add_multi_addr(
  dev_ctrl_t *p_dev,			/* Device control structure ptr */
  char *addr				/* Multicast address */
)
{
	ent_multi_t *mtab;		/* Multicast addr table in dev_ctrl */
	ent_multi_t *p_new;		/* In case all tables are full */
	int i;

	/* TRC_FLOW_2(IOCTL_ENTRY, NDD_ENABLE_ADDRESS, *(uint*)addr,
					    *(uint*)(addr+4)); */

	GM_PRINT(GM_AIX_DEBUG_IP,
		 ("Add multi addr: 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x\n",
		  *addr, *(addr + 1), *(addr + 2),
		  *(addr + 3), *(addr + 4), *(addr + 5)));

	/* If not open, punt */
	if (WRK.adap_state != OPEN_STATE) {
		if (WRK.adap_state == DEAD_STATE)
			return ENETDOWN;

		return ENETUNREACH;
	}

	/* If not a valid multicast address, error */
	if (!(*addr & ENT_GROUP_BIT)) {
/*		TRC_ERR_2(IOCTL_ENTRY, BAD_MULTI_ADDRESS, *(uint*)addr,
			   				   *(uint*)(addr+4)); */
		return EINVAL;
	}
			
	/* If the address is a duplicate, increment the ref_count and return */
	for (mtab = &WRK.multi_table; mtab != NULL; mtab = mtab->next) {
		for (i=0; i < mtab->in_use; i++) {
			if (SAME_NADR(addr, mtab->m_slot[i].m_addr)) {
				mtab->m_slot[i].ref_count++;

				return 0;
			}
		}
	}

	/* Locate the first table with room */
	for (mtab = &WRK.multi_table; mtab->in_use >= MULTI_TABLE_SLOT;
	     						 mtab = mtab->next);

	/* If existing tables are full, allocate additional table */
	if (mtab == NULL) {
		if ((mtab =
		     xmalloc(sizeof(ent_multi_t), 3, pinned_heap)) == NULL) {
			return ENOMEM;
		}
		bzero(mtab, sizeof(ent_multi_t));

		/* Insert the new table just after the head table */
		mtab->next = WRK.multi_table.next;
		WRK.multi_table.next = mtab;
		mtab->in_use = 0;
	}

	/* Add the address into the table and increment counts */
	COPY_NADR(addr, mtab->m_slot[mtab->in_use].m_addr);
	mtab->m_slot[mtab->in_use].ref_count = 1;
	mtab->in_use++;
	WRK.multi_addr_count++;
	NDD.ndd_flags |= NDD_ALTADDRS;

	return 0;
}


/*****************************************************************************/
/*
 * NAME:		  del_multi_addr
 *
 * FUNCTION:		  Delete a multicast address from the multicast table
 *
 * EXECUTION ENVIRONMENT: process only
 *
 * NOTES:		  The multicast table consists of a base table in
 *			  the dev_ctrl structure and xmalloc()ed extensions.
 *
 * RETURNS:		  errno value
 */
/*****************************************************************************/

int
del_multi_addr(
  dev_ctrl_t *p_dev,			/* Device control structure ptr */
  char *addr				/* Multicast address */
)
{

	ent_multi_t *mtab;		/* Multicast addr table in dev_ctrl */
	ent_multi_t *p_prev = NULL;	/* In case all tables are full */
	int i;

	/* TRC_FLOW_2(IOCTL_ENTRY, NDD_DISABLE_ADDRESS, *(uint*)addr,
						    *(uint*)(addr+4)); */

	GM_PRINT(GM_AIX_DEBUG_IP, ("Delete multi addr\n"));

	/* If not open, punt */
	if (WRK.adap_state != OPEN_STATE) {
		if (WRK.adap_state == DEAD_STATE)
			return ENETDOWN;

		return ENETUNREACH;
	}

	/* If not a valid multicast address, error */
	if (!(*addr & ENT_GROUP_BIT)) {
/*		TRC_ERR_2(IOCTL_ENTRY, BAD_MULTI_ADDRESS, *(uint*)addr,
			   				   *(uint*)(addr+4)); */
		return EINVAL;
	}
			
	/* Search all multicast address tables for the arg address */
	for (mtab = &WRK.multi_table; mtab != NULL; mtab = mtab->next) {
		for (i = 0; i < mtab->in_use; i++) {
			if (SAME_NADR(addr, mtab->m_slot[i].m_addr))  {
				break;
			}
		}

		/* If found, search no more */
		if (i < mtab->in_use) 
			break;

		/* If not in base table, remember the previous table
		 * in case this table is deleted
		 */
		p_prev = mtab;
	}

	/* If address not found, error */
	if (mtab == NULL) return EINVAL;

	/* Decrement the ref_count for the address.  If now zero, move
	 * up addresses further down in table.
	 * Note: having a table with no holes, speeds up the search for
	 * matching addresses
	 */
	if (--mtab->m_slot[i].ref_count == 0) {

		for (; i < mtab->in_use - 1; i++) {
			COPY_NADR(mtab->m_slot[i+1].m_addr, 
					  	       mtab->m_slot[i].m_addr);
			mtab->m_slot[i].ref_count = mtab->m_slot[i+1].ref_count;
		}
		bzero(&mtab->m_slot[i], sizeof(multi_slot_t));

		/* Update counters and status */
		mtab->in_use--;
		if (--WRK.multi_addr_count == 0)
			NDD.ndd_flags &= ~NDD_ALTADDRS;

		/* If the table is now empty and it is an expansion,
		 * free the space
		 */
		if (mtab->in_use == 0) {
			if (p_prev != NULL) {
				p_prev->next = mtab->next;
				xmfree(mtab, pinned_heap);
			}
		}
	}
	return 0;
}


/*****************************************************************************/
/*
 * NAME:     		  enable_multicast
 *
 * FUNCTION: 		  Issues the Set RX Filter command to enable
 *			  reception of packets with group destination
 *			  addresses
 *
 * EXECUTION ENVIRONMENT: process
 *
 * NOTES:		  1. Assumes the disable_lock() is in effect
 *			  2. Assumes not in window 0
 *
 * RETURNS:  		  Nothing
 */
/*****************************************************************************/

void
enable_multicast(
  dev_ctrl_t *p_dev			/* Device control structure ptr */
)
{
	GM_PRINT(GM_AIX_DEBUG_IP, ("enable multicast\n"));
	WRK.filter |= ENABLE_MULTICAST;
}


/*****************************************************************************/
/*
 * NAME:     		  disable_multicast
 *
 * FUNCTION: 		  Issues the Set RX Filter command to disable
 *			  reception of packets with group destination
 *			  addresses
 *
 * EXECUTION ENVIRONMENT: process
 *
 * NOTES:		  1. Assumes the disable_lock() is in effect
 *			  2. Assumes not in window 0
 *
 * RETURNS:  		  Nothing
 */
/*****************************************************************************/

void
disable_multicast(
  dev_ctrl_t *p_dev			/* Device control structure ptr */
)
{
	GM_PRINT(GM_AIX_DEBUG_IP, ("disable multicast\n"));
	WRK.filter &= ~ENABLE_MULTICAST;
}
#endif /* GM_AIX_SUPPORT_MULTI */


/*****************************************************************************/
/*
 * NAME:     		  enable_promiscuous
 *
 * FUNCTION: 		  Issues the Set RX Filter command to enable
 *			  reception of packets with group destination
 *			  addresses
 *
 * EXECUTION ENVIRONMENT: process
 *
 * NOTES:		  1. Assumes the disable_lock() is in effect
 *			  2. Assumes not in window 0
 *
 * RETURNS:  		  Nothing
 */
/*****************************************************************************/

void
enable_promiscuous(
  dev_ctrl_t *p_dev			/* Device control structure ptr */
)
{
	GM_PRINT(GM_AIX_DEBUG_IP, ("enable promiscous\n"));
	WRK.filter |= ENABLE_PROMISCUOUS;
}


/*****************************************************************************/
/*
 * NAME:     		  disable_promiscuous
 *
 * FUNCTION: 		  Issues the Set RX Filter command to disable
 *			  reception of packets with group destination
 *			  addresses
 *
 * EXECUTION ENVIRONMENT: process
 *
 * NOTES:		  1. Assumes the disable_lock() is in effect
 *			  2. Assumes not in window 0
 *
 * RETURNS:  		  Nothing
 */
/*****************************************************************************/

void
disable_promiscuous(
  dev_ctrl_t *p_dev			/* Device control structure ptr */
)
{
	GM_PRINT(GM_AIX_DEBUG_IP, ("disable promiscuous\n"));
	WRK.filter &= ~ENABLE_PROMISCUOUS;
}
