
/******************************************************************-*-c-*-
 * Myricom GM networking software and documentation			 *
 * Copyright (c) 2001 by Myricom, Inc.					 *
 * All rights reserved.	 See the file `COPYING' for copyright notice.	 *
 *************************************************************************/

/* author: gallatin@myri.com */

#ifndef _gm_bsd_ether_arch_h_
#define _gm_bsd_ether_arch_h_

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/sockio.h>
#include <sys/mbuf.h>

#if GM_PREVENT_NATIVE_MALLOC_USE
#undef malloc
#undef free
#include <sys/malloc.h>
#define malloc GM_DO_NOT_USE (malloc)
#define free GM_DO_NOT_USE (free)
#else 
#include <sys/malloc.h>
#endif /* GM_PREVENT_MALLOC_USE */

#include <sys/kernel.h>
#include <sys/socket.h>

#include <net/if.h>
#include <net/if_arp.h>
#include <net/ethernet.h>
#include <net/if_dl.h>
#include <net/if_media.h>

#include <netinet/in.h>
#include <netinet/in_var.h>
#include <vm/vm.h>		/* for vtophys */
#include <vm/pmap.h>		/* for vtophys */
#include <machine/clock.h>	/* for DELAY */
#include <machine/bus_memio.h>
#include <machine/bus_pio.h>
#include <machine/bus.h>
#include <machine/resource.h>
#include <sys/bus.h>
#include <sys/rman.h>

/* for checksum offloading */
#define _IP_VHL
#include <netinet/in_systm.h>
#include <netinet/ip.h>
#include <machine/in_cksum.h>


#include "gm_debug.h"
#include "gm_internal.h"
#include "gm_ether.h"
#include "gm_debug_lanai_dma.h"

#if   (__FreeBSD_version >= 410000)
#define GM_BSD_CHECKSUM_OFFLOADING_SUPPORTED 1
#include <net/bpf.h>
#else
#define GM_BSD_CHECKSUM_OFFLOADING_SUPPORTED 0
#endif

#if   (__FreeBSD_version >= 500000)
#define IF_DROP(ifq)	_IF_DROP(ifq);
#endif

/* Later version of FreeBSD name the wait flags to the 
   mbuf allocator differently from the wait flags to malloc */

#ifdef M_DONTWAIT
#undef M_NOWAIT
#define M_NOWAIT M_DONTWAIT
#endif

#ifdef M_TRYWAIT
#undef M_WAITOK
#define M_WAITOK M_TRYWAIT
#endif

struct gm_bsd_ether_sc;	/* forward declaration; declared in gm_bsd_ether.h */

struct gm_bsd_ether_arch_dma_info
{
  gm_ethernet_segment_descriptor_t desc[GM_MAX_ETHERNET_SG_CNT];
  gm_dp_t dp[GM_MAX_ETHERNET_SG_CNT];
  int nsegs;
};


struct gm_bsd_ether_arch_sc
{
  struct arpcom arpcom;			/* the network (ethernet driver) */
#define is_if   arch.arpcom.ac_if	/* network-visible interface    */
#define is_addr arch.arpcom.ac_enaddr	/* hardware Ethernet address    */
#if   (__FreeBSD_version >= 500000)
  struct mtx mutex; 
#endif  
};

extern struct gm_bsd_ether_sc *gm_bsd_ether_sc[];
extern struct mbuf *bogus_m;
extern struct gm_bsd_ether_arch_dma_info bogus_dma;

static int gm_freebsd_ether_ioctl (struct ifnet *ifp, unsigned long cmd, caddr_t data);

static gm_inline void
gm_bsd_ether_arch_incr_tx_counters (struct gm_bsd_ether_arch_sc *sc,
			   struct mbuf *m)
{
  /* Not needed on FreeBSD */
}

static gm_inline void
gm_bsd_ether_arch_incr_rx_counters (struct gm_bsd_ether_arch_sc *sc, unsigned int ulen,
			   struct ether_header *eh)
{
  /* Not needed on FreeBSD */
}


static gm_inline void
gm_bsd_ether_arch_sc_lock (struct gm_bsd_ether_arch_sc *sc)
{
  /* Not needed on FreeBSD 4.x */
#if (__FreeBSD_version > 500000 )
  mtx_lock (&sc->mutex);
#endif
}

static gm_inline void
gm_bsd_ether_arch_sc_unlock (struct gm_bsd_ether_arch_sc *sc)
{
  /* Not needed on FreeBSD 4.x */
#if (__FreeBSD_version > 500000 )
  mtx_unlock (&sc->mutex);
#endif
}

static gm_inline void
gm_bsd_ether_arch_bpf_xmit (struct ifnet *ifp, struct mbuf *m_head)
{
#if (__FreeBSD_version >= 500000)
  BPF_MTAP((ifp), (m_head));
#elif (__FreeBSD_version >= 410000)
  if ((ifp)->if_bpf)
    {
      bpf_mtap ((ifp), (m_head));
    }
#endif /* __FreeBSD_version >= 410000 */
  /* bpf is too complicated to try to support for releases prior to
     4.1-RELEASE */

}

static gm_inline void
gm_bsd_ether_arch_unencap (struct gm_bsd_ether_sc *sc, struct gm_bsd_ether_arch_dma_info *dma)
{
  dma->nsegs = 0;
}

#ifdef __alpha__
#undef vtophys
#define vtophys(va)	alpha_XXX_dmamap((vm_offset_t)va)
#endif

#endif /*_gm_bsd_ether_arch_h_*/


/*
  This file uses GM standard indentation.

  Local Variables:
  c-file-style:"gnu"
  tab-width:8
  End:
*/
