/******************************************************************-*-c-*-
 * Myricom GM networking software and documentation			 *
 * Copyright (c) 2001 by Myricom, Inc.					 *
 * All rights reserved.	 See the file `COPYING' for copyright notice.	 *
 *************************************************************************/

/* author: glenn@myri.com */

#include "gm_arch.h"
#include "gm_call_trace.h"
#include "gm_debug_yp.h"
#include "gm_impl.h"
#include "gm_lanai_command.h"
#include "gm_pio.h"
#include "gm_types.h"

#define GM_DEBUG_LANAI_COMMAND 0

#if GM_DEBUG_LANAI_COMMAND || GM_DEBUG_YP
#undef GM_LOCALLY_ENABLE_CALL_TRACE
#define GM_LOCALLY_ENABLE_CALL_TRACE 1
#endif

/*#define GM_MAX_SPIN_US (1000 * 1000)*/
#define GM_USECS_PER_SEC 	1000000
#define GM_MAX_SPIN_US 		1000
#define GM_COMMAND_TIMEOUT_SECS	10

#define GM_DUMP_DMA_DESCRIPTOR(is, dma_descriptor) do {			\
  gm_lp_t next_with_flags;						\
									\
  if (gm_is_lX (is))							\
    {									\
      /* We don't yet support dumping LANaiX DMA descriptors. */	\
      break;								\
    }									\
									\
  next_with_flags							\
    = (gm_read_lanai_global_lp						\
       (is, dma_descriptor.format.pre_lX.next_with_flags));		\
  if (next_with_flags & GM_DMA_DESCRIPTOR_TERMINAL)			\
    {									\
      /* We don't care about completed descriptors. */			\
      break;								\
    }									\
  _GM_INFO (("{ next_with_flags = 0x%x%s%s%s,\n",			\
	     next_with_flags & ~GM_DMA_DESCRIPTOR_FLAG_MASK,		\
	     next_with_flags & GM_DMA_DESCRIPTOR_E2L			\
	     ? " + E2L" : " + L2E",					\
	     next_with_flags & GM_DMA_DESCRIPTOR_TERMINAL		\
	     ? " + TERMINAL" : "",					\
	     next_with_flags & GM_DMA_DESCRIPTOR_WAKE			\
	     ? " + WAKE" : ""));					\
  _GM_INFO (("  csum[] = {0x%04x, 0x%04x},\n",				\
	     gm_read_lanai_global_u16					\
	     (is, dma_descriptor.format.pre_lX.csum[0]),			\
	     gm_read_lanai_global_u16					\
	     (is, dma_descriptor.format.pre_lX.csum[1])));		\
  _GM_INFO (("  len = 0x%x,\n",						\
	     gm_read_lanai_global_u32					\
	     (is, dma_descriptor.format.pre_lX.len)));			\
  _GM_INFO (("  lanai_addr = 0x%x,\n",					\
	     gm_read_lanai_global_lp					\
	     (is, dma_descriptor.format.pre_lX.lanai_addr)));		\
  _GM_INFO (("  ebus_addr_high = 0x%x}\n",				\
	     gm_read_lanai_global_u32					\
	     (is, dma_descriptor.format.pre_lX.ebus_addr_high)));	\
  _GM_INFO (("  ebus_addr_low = 0x%x}\n",				\
	     gm_read_lanai_global_u32					\
	     (is, dma_descriptor.format.pre_lX.ebus_addr_low)));		\
} while (0)
	   
static void
gm_lanai_command_report_error (gm_instance_state_t * is, enum gm_lanai_command cmd)
{
  gm_lp_t lanai_current_handler;
  gm_lp_t lanai_last_dispatch;
  int i;

  GM_NOTE (("LANai[%d] interface not responding to command %d\n",
	    is->id, cmd));
  _GM_NOTE (("    isr=0x%x eimr=0x%x\n",
	     is->get_ISR (is), is->get_EIMR (is)));
  _GM_NOTE (("    sram[0] = 0x%x 0x%x 0x%x 0x%x \n",
	     *(unsigned int *) is->lanai.sram,
	     *(unsigned int *)&is->lanai.sram[4],
	     *(unsigned int *)&is->lanai.sram[8],
	     *(unsigned int *)&is->lanai.sram[12]));
  if (is->interrupt)
    {
      _GM_NOTE (("    is->interrupt.last_byte.ready = %u\n",
		 (unsigned int) gm_ntoh_u8 (is->interrupt->last_byte.ready)));
      _GM_NOTE (("    gm.interrupt.last_byte.ready = %u\n",
		 (unsigned int)
		 gm_read_lanai_global_u8 (is, interrupt.last_byte.ready)));
      _GM_NOTE (("    is->interrupt.type = %u\n",
		 (unsigned int) gm_ntoh_u32 (is->interrupt->type)));
      _GM_NOTE (("    gm.interrupt.type = %u\n",
		 ((unsigned int)
		  gm_read_lanai_global_u32 (is, interrupt.type))));
    }
  if (!gm_is_lX (is))
    {
      _GM_NOTE (("    SYNC DMA len=%d lar=0x%x ear=0x%x%08x\n",
		 gm_read_lanai_global_u32
		 (is, sync_dma.descriptor[0].format.pre_lX.len),
		 gm_read_lanai_global_lp
		 (is, sync_dma.descriptor[0].format.pre_lX.lanai_addr),
		 gm_read_lanai_global_u32
		 (is, sync_dma.descriptor[0].format.pre_lX.ebus_addr_high),
		 gm_read_lanai_global_u32
		 (is, sync_dma.descriptor[0].format.pre_lX.ebus_addr_low)));
    }
  
  
  lanai_last_dispatch = gm_read_lanai_global_lp (is, last_dispatch);
  if (lanai_last_dispatch)
    {
      char last_dispatch[GM_MAX_FIRMWARE_STRING_LEN + 1];
      
      __gm_ntok_strncpy (is, last_dispatch,
			 is->lanai.sram + lanai_last_dispatch,
			 GM_MAX_FIRMWARE_STRING_LEN);
      last_dispatch[GM_MAX_FIRMWARE_STRING_LEN] = 0;
      GM_INFO (("Last dispatch was at %s\n", last_dispatch));
    }
  lanai_current_handler = gm_read_lanai_global_lp (is, current_handler);
  if (lanai_current_handler)
    {
      char current_handler[GM_MAX_FIRMWARE_STRING_LEN + 1];
      char current_handler_extra[GM_MAX_FIRMWARE_STRING_LEN + 1];

      __gm_ntok_strncpy (is, current_handler,
			 is->lanai.sram + lanai_current_handler,
			 GM_MAX_FIRMWARE_STRING_LEN);
      current_handler[GM_MAX_FIRMWARE_STRING_LEN] = 0;
      __gm_ntok_strncpy
	(is, current_handler_extra,
	 is->lanai.sram + gm_read_lanai_global_lp (is, current_handler_extra),
	 GM_MAX_FIRMWARE_STRING_LEN);
      current_handler_extra[GM_MAX_FIRMWARE_STRING_LEN] = 0;

      GM_INFO (("current handler is %s (%s)\n",
		current_handler, current_handler_extra));
    }

  /* Dump the DMA descriptors */

  GM_INFO (("dma_descriptor:\n"));
  GM_DUMP_DMA_DESCRIPTOR (is, sync_dma.descriptor[0]);

  GM_INFO (("user_dma_descriptors:\n"));
  for (i=0; i<GM_MAX_NUM_DMA_SEGMENTS+1; i++)
    {
      GM_DUMP_DMA_DESCRIPTOR (is, user_dma_descriptor[i]);
    }
}


/* Perform a command, assuming that the caller has already written any
   parameters the command has to is->command.data. */

static
gm_status_t
_gm_lanai_command (struct gm_instance_state *is, enum gm_lanai_command cmd)
{
  int spin_quota = GM_MAX_SPIN_US;

  gm_assert (is);

  GM_CALLED_WITH_ARGS (("%p,%d", is, cmd));

  if (!is->lanai.running)
    GM_RETURN (GM_TIMED_OUT);
  
  gm_assert (gm_read_lanai_global_u32 (is, command.type)
	     == GM_LANAI_COMMAND_NONE);
  is->command_done = 0;
  GM_STBAR ();
  gm_write_lanai_global_u32 (is, command.type, cmd);
  GM_STBAR ();
  /* readback to flush write */
  gm_read_lanai_global_u32 (is, command.type);

  /* sleep until command completes or timeout occurs */
  if ((gm_arch_timed_sleep (&is->command_sync, 10) 
       == GM_SLEEP_TIMED_OUT)
      || (is->command_done == 0))
    {
      /* Timed out before the command finished */
      gm_lanai_command_report_error (is, cmd);
      gm_disable_lanai (is);
      GM_RETURN_STATUS (GM_TIMED_OUT);
    }

  /* Return the firmware-generated status code. */

 done:
  GM_RETURN_STATUS (gm_read_lanai_global_u32 (is, command.status));
}

/* Perform a simple LANai command that has no parameters 

   Unlike other routines in this file, The caller of this routine is
   expected to have acquired the per-instance command_sync mutex */

gm_status_t
gm_lanai_command (struct gm_instance_state *is, enum gm_lanai_command cmd)
{
  /* make sure this is not a command that requires extra args. */
  gm_assert (is);

  /* confirm that this entry point is not used for commands that require
     parameters. */
  gm_assert (cmd != GM_LANAI_COMMAND_CLOSE);
  gm_assert (cmd != GM_LANAI_COMMAND_DEREGISTER);
  gm_assert (cmd != GM_LANAI_COMMAND_OPEN);
  gm_assert (cmd != GM_LANAI_COMMAND_UNCACHE_PORT_PTES);

  return _gm_lanai_command (is, cmd);
}

/****************
 * Convenient interfaces for commands with parameters.
 ****************/

gm_status_t
gm_lanai_command_open (struct gm_instance_state *is, unsigned int port_num)
{
  gm_status_t status;
  gm_assert (is);
  gm_assert (port_num < GM_NUM_PORTS);

  gm_arch_mutex_enter (&is->command_sync);
  gm_write_lanai_global_u32 (is, command.data.open.port_num, (gm_u32_t) port_num);
  status = _gm_lanai_command (is, GM_LANAI_COMMAND_OPEN);
  gm_arch_mutex_exit (&is->command_sync);
  return status;
}

gm_status_t
gm_lanai_command_close (struct gm_instance_state *is, unsigned int port_num)
{
  gm_status_t status;

  gm_assert (is);
  gm_assert (port_num < GM_NUM_PORTS);
  gm_arch_mutex_enter (&is->command_sync);
  gm_write_lanai_global_u32 (is, command.data.close.port_num,
			     (gm_u32_t) port_num);
  status = _gm_lanai_command (is, GM_LANAI_COMMAND_CLOSE);
  gm_arch_mutex_exit (&is->command_sync);
  return status;
}

gm_status_t
gm_lanai_command_deregister (struct gm_instance_state *is,
			     unsigned int port_num, gm_up_t uvma, gm_up_t len)
{
  gm_status_t status;

  GM_CALLED ();
  gm_assert (is);
  gm_assert (uvma);
  gm_assert (len);
  gm_assert (GM_PAGE_ALIGNED (len));
  gm_assert (port_num < GM_NUM_PORTS);
  
  /* Break any huge deregistration in to reasonable sized deregistrations
     to prevent timeouts during the operation. Relock the command_sync
     each interation to give other threads access to the lanai */
  
  while (len > GM_MAX_LANAI_DEREG_LEN)
    {
      gm_arch_mutex_enter (&is->command_sync);
      gm_write_lanai_global_u32 (is, command.data.deregister.port_num,
				 (gm_u32_t) port_num);
      gm_write_lanai_global_up (is, command.data.deregister.uvma, uvma);
      gm_write_lanai_global_up (is, command.data.deregister.len,
				GM_MAX_LANAI_DEREG_LEN);
      status = _gm_lanai_command (is, GM_LANAI_COMMAND_DEREGISTER);
      gm_arch_mutex_exit (&is->command_sync);
      if (status != GM_SUCCESS)
	{
	  GM_RETURN_STATUS (status);
	}
      uvma += GM_MAX_LANAI_DEREG_LEN;
      len -= GM_MAX_LANAI_DEREG_LEN;
    }
  gm_arch_mutex_enter (&is->command_sync);
  gm_write_lanai_global_u32 (is, command.data.deregister.port_num,
			     (gm_u32_t) port_num);
  gm_write_lanai_global_up (is, command.data.deregister.uvma, uvma);
  gm_write_lanai_global_up (is, command.data.deregister.len, len);
  status = _gm_lanai_command (is, GM_LANAI_COMMAND_DEREGISTER);
  gm_arch_mutex_exit (&is->command_sync);
  GM_RETURN_STATUS (status);
}

gm_status_t
gm_lanai_command_uncache_port_ptes (struct gm_instance_state *is,
				    unsigned int port_num)
{
  gm_status_t status;

  gm_assert (is);
  gm_assert (port_num < GM_NUM_PORTS);
  
  gm_arch_mutex_enter (&is->command_sync);
  gm_write_lanai_global_u32 (is, command.data.uncache_port_ptes.port_num,
			     (gm_u32_t) port_num);
  status = _gm_lanai_command (is, GM_LANAI_COMMAND_UNCACHE_PORT_PTES);
  gm_arch_mutex_exit (&is->command_sync);
  return status;
}

gm_status_t
gm_lanai_command_set_route (struct gm_instance_state *is,
			    gm_u32_t global_id,
			    gm_u8_n_t route[],
			    unsigned int route_len)
{
  gm_status_t status;

  GM_CALLED_WITH_ARGS (("%p,0x%x,%p,0x%x", is, global_id, route, route_len));
  
  gm_assert (is);

  if (route_len > GM_MAX_NETWORK_DIAMETER)
    {
      GM_RETURN_STATUS (GM_INVALID_PARAMETER);
    }
  
  gm_arch_mutex_enter (&is->command_sync);
  gm_write_lanai_global_u32 (is, command.data.set_route.global_id, global_id);
  gm_write_lanai_global_u32 (is, command.data.set_route.route_len, route_len);
  gm_bcopy_to_lanai_globals (is, route, command.data.set_route.route,
			     route_len);
  status = _gm_lanai_command (is, GM_LANAI_COMMAND_SET_ROUTE);
  gm_arch_mutex_exit (&is->command_sync);
  GM_RETURN_STATUS (status);
}

/**
   Have the LANai perform a yellow pages lookup.

   @param is \b (IN) The board to perform the query.
   
   @param port_num \b (IN) The port number requesting the query.

   @param key (IN) The query key, which is the name of the attribute to query.

   @param value (IN) The attribute value to match, or "" to match any value.

   @param *node_id \b (IN) (OUT) The node ID to query, or 0 to broadcast the query.
   (OUT) The node ID that responded to the query.

   @param answer (OUT) Where to store the matching value, if any.

   @author Glenn Brown

   @version GM_API_VERSION (as defined in gm.h)
*/
gm_status_t
gm_lanai_command_yp (struct gm_port_state *ps,
		     unsigned int timeout_usecs,
		     const char *key,
		     const char *value,
		     unsigned int *node_id, /* *node_id = 0 to broadcast */
		     char (*answer)[GM_MAX_FIRMWARE_STRING_LEN+1])
{
  char query[2 * (GM_MAX_FIRMWARE_STRING_LEN + 1)];
  gm_instance_state_t *is;
  gm_status_t status;
  unsigned int port_num;

  GM_CALLED_WITH_ARGS
    (("%p,%p(%d),\"%s\",\"%s\",%p",
      ps, node_id, *node_id, key, value ? value : "(NULL)", answer));
  
  gm_assert (key);
  gm_assert (ps);
  is = ps->instance;
  gm_assert (is);
  port_num = ps->id;
  
  /* Check the input parameters. */
  
  if (!value)
    value = "";
  if (gm_strlen (key) > GM_MAX_FIRMWARE_STRING_LEN
      || gm_strlen (value) > GM_MAX_FIRMWARE_STRING_LEN)
    {
      status = GM_INVALID_PARAMETER;
      goto abort_with_nothing;
    }

  /* Build the query string by placing the null-terminated key and value
     strings after each other, with a NULL character in between. */

  {
    char *ptr;

    gm_strcpy (query, key);
    ptr = query + gm_strlen (query) + 1;
    gm_strcpy (ptr, value);
  }

  /* Acquire the port's sleep sync both because we need to sleep
     waiting for the result, and to limit YP queries to one per
     port at any time. */
  
  gm_arch_mutex_enter (&ps->yp_sync);
  
  /* Pass the port's query to the firmware, in the per-port storage. */

  gm_bcopy_to_lanai_globals (is, query, port[port_num].yp.in_out,
			     sizeof (query));
  gm_write_lanai_global_u32 (is, port[port_num].yp.node_id, *node_id);
  gm_write_lanai_global_s32 (is, port[port_num].yp.query_timeout_time,
			     timeout_usecs);
  
  /* Command the LANai to start processing the port's YP query. */

  gm_arch_mutex_enter (&is->command_sync);
  gm_write_lanai_global_u32 (is, command.data.yp.port_num, port_num);
  status = _gm_lanai_command (is, GM_LANAI_COMMAND_YP);
  gm_arch_mutex_exit (&is->command_sync);
  if (status != GM_SUCCESS)
    {
      goto abort_with_port_sleep_mutex;
    }

  /* Block until the query completes. */
  
  {
    gm_arch_sleep_status_t sleep_status;

    sleep_status = gm_arch_signal_sleep (&ps->yp_sync);
    if (sleep_status == GM_SLEEP_INTERRUPTED)
      {
	/* The sleep was interrupted before the command completed, so
	   cancel the YP command.  We do this via a LANai command
	   because it makes it easy to ensure that the LANai still
	   generates the single wake interrupt needed for proper sleep
	   accounting, even if the command completes just as it is
	   interrupted. */
	
	gm_arch_mutex_enter (&is->command_sync);
	gm_write_lanai_global_u32 (is, command.data.yp.port_num, port_num);
	status = _gm_lanai_command (is, GM_LANAI_COMMAND_YP_CANCEL);
	gm_arch_mutex_exit (&is->command_sync);
	
	/* Let the user-level caller know we were interrupted so they can try
	   again. */
	status = GM_INTERRUPTED;
	goto abort_with_port_sleep_mutex;
      }
  }
  
  /* Read the results. */

  status = gm_read_lanai_global_u32 (is, port[port_num].yp.status);
  if (status == GM_SUCCESS)
    {
      *node_id = gm_read_lanai_global_u32 (is, port[port_num].yp.node_id);
      _GM_PRINT (GM_DEBUG_YP, ("Answer from node %d\n", *node_id));
      if (answer)
	{
	  gm_bcopy_from_lanai_globals (is, port[port_num].yp.in_out,
				       answer, sizeof (*answer));
	  _GM_PRINT
	    (GM_DEBUG_YP, ("Answer value is \"%s\"\n", (char *) answer));
	}
    }
  
  /* Release the sleep mutex. */

  gm_arch_mutex_exit (&ps->yp_sync);

  GM_PRINT
    (GM_DEBUG_YP,
     ("answer from node %d is \"%s\"\n",
      *node_id, answer ? &(*answer)[0] : "<NULL>"));
  GM_RETURN_STATUS (status);
  
 abort_with_port_sleep_mutex:
  gm_arch_mutex_exit (&ps->yp_sync);
 abort_with_nothing:
  GM_RETURN_STATUS (status);
}


/*
  This file uses GM standard indentation.

  Local Variables:
  c-file-style:"gnu"
  tab-width:8
  End:
*/
