/*
  gm_arch.c
  architecture specific code for hpux 11.0 GM driver
  finucane@myri.com (David Finucane)
*/

#define __LP64__

#include "gm_internal.h"
#include "gm_page_hash.h"
#include "gm_lanai.h"
#include "gm_instance.h"
#include "gm_impl.h"

#include <_types.h>
#include <sys/types.h>
#include <sys/param.h>
#include <sys/errno.h>
#include <machine/reg.h>
#include <machine/vmparam.h>
#include "/usr/conf/machine/vm_prfmac.h"
#include <sys/malloc.h>
#include <sys/dma.h>
#include <mod_conf.h>
#include <sys/moddefs.h>
#include <sys/io.h>
#include <sys/pci.h>
#include <wsio/pci.h>
#include <limits.h>
#include <sys/iomap.h>
#include <sys/mman.h>
#include <machine/pde.h>
#include <wsio/wsio.h>	/* this must be the last #include */

#ifndef PAGESIZE
#define PAGESIZE 4096
#endif

#include "finucane.h"

extern struct mod_operations gio_mod_ops;

int gm_hpux_load (void*driver_info_arg);
int gm_hpux_unload (void*driver_info_arg);
int gm_hpux_open (dev_t dev, int flag);
int gm_hpux_close (dev_t dev);
int gm_hpux_ioctl (dev_t dev, int cmd, caddr_t arg, int flag);

static gm_arch_sync_t gm_hpux_debug_sync;

static int gm_hpux_unload_count;

#define MOD_WSIO_PCI 4

int gm_hpux_test;

static struct mod_type_data gm_hpux_driver_link =
{
  "gm",
  0
};

static struct modlink gm_hpux_mod_link [2] =
{
  { &gio_mod_ops,
    &gm_hpux_driver_link
  },
  {
    0, 0
  }
};

static struct mod_conf_data gm_hpux_mc =
{
  MCD_VERSION, 0
};

struct modwrapper gm_wrapper = 
{
  MODREV,
  gm_hpux_load,
  gm_hpux_unload,
  0,
  &gm_hpux_mc,
  gm_hpux_mod_link
};

static drv_info_t gm_hpux_driver_info =
{
  "gm",
  "ext_bus",
  DRV_MP_SAFE,
  -1,
  -1,
  0,
  0,
  0
};

static drv_ops_t gm_hpux_driver_ops =
{
  gm_hpux_open,
  gm_hpux_close,
  0,
  0,
  0,
  0,
  0,
  0,
  gm_hpux_ioctl,
  0,
  0,
  0,
  0,
  0
};

static wsio_drv_data_t gm_hpux_data =
{
  "myrinet_card",
  T_INTERFACE,
  DRV_CONVERGED,
  0,
  0
};

static wsio_drv_info_t gm_hpux_wsio_driver_info =
{
  &gm_hpux_driver_info,
  &gm_hpux_driver_ops,
  &gm_hpux_data
};

static struct gfsw gm_hpux_gfsw;

static int gm_hpux_dev_init ();
static int (*gm_hpux_saved_dev_init)(void);
static int (*gm_hpux_saved_attach)(uint32_t, struct isc_table_type*);

gm_instance_state_t* gm_hpux_instances [GM_ARCH_MAX_INSTANCE];
int gm_hpux_num_instances;

#define GM_HPUX_IS_STATIC_MINOR(n)((n) < gm_hpux_num_instances*2)
#define GM_HPUX_INVALID_DYMANIC_MINOR(n)(n > GM_ARCH_MAX_MINOR || n < gm_hpux_num_instances * 2)
#define GM_HPUX_STATIC_MINOR_TO_UNIT(n)((n)/2)
#define GM_HPUX_DYNAMIC_MINOR_TO_UNIT(n)((n / ((GM_ARCH_MAX_MINOR - 2 * GM_ARCH_MAX_INSTANCE) / GM_ARCH_MAX_INSTANCE)))
#define GM_HPUX_OS_MINOR_TO_GM_MINOR(n)(n - (GM_ARCH_MAX_INSTANCE * 2 - gm_hpux_num_instances * 2))
#define GM_HPUX_GM_MINOR_TO_OS_MINOR(n)((GM_ARCH_MAX_INSTANCE * 2 - gm_hpux_num_instances * 2) + n)

/*gm_status_t gm_disable_PCI_config_command_bit (gm_instance_state_t * is, gm_u16_t value);*/

static gm_hpux_mapped_chunk_t*gm_hpux_get_chunk (gm_instance_state_t*is, char*kaddr, int length)
{
  gm_hpux_mapped_chunk_t*chunk;
  for (chunk = is->arch.mapped_chunks; chunk; chunk = chunk->next)
  {
    gm_hpux_printf (("kaddr is %x, chunk->ka is %x, kaddr + length is %x, chunk->ka + chunk->length is %x\n",
		     kaddr, chunk->ka, kaddr + length, chunk->ka + chunk->length));
    if (kaddr >= chunk->ka && kaddr + length <= chunk->ka + chunk->length)
      return chunk;
  }
  return 0;
}

/*hp docs are inconsistent here*/
static int gm_hpux_isr (struct isc_table_type *isc, int unit)
{
  return gm_hpux_instances [unit] && gm_intr (gm_hpux_instances [unit]) == GM_ARCH_INTR_CLAIMED;
}

static int gm_hpux_init (struct isc_table_type*isc)
{
  gm_instance_state_t*is; 
  unsigned char val;

  gm_hpux_printf (("gm_hpux_init\n"));
  
  if (gm_hpux_num_instances > GM_ARCH_MAX_INSTANCE)
  {
    gm_hpux_printf (("gm_hpux: too many units\n"));
    return 0;
  }
  
  if (!(is = gm_malloc (sizeof (gm_instance_state_t), 0)))
  {
    gm_hpux_printf (("gm_hpux_init: failed to alloc instance state\n"));
    return 0;
  }
  
  gm_hpux_instances [gm_hpux_num_instances] = is;
  is->arch.unit = gm_hpux_num_instances;
  gm_hpux_num_instances++;
  is->arch.mapped_chunks = 0;
  
  is->arch.isc = isc;

  if (isc->if_reg_ptr)
    unmap_mem_from_host (isc, isc->if_reg_ptr, PAGESIZE);
  
  pci_read_cfg_uint32_isc (isc, 0x10, &is->arch.pci_base_address);

  is->arch.pci_base_address &= ~0xf;
  isc->if_reg_ptr = is->arch.pci_base_address;

  if (gm_instance_init (is, is->arch.unit, GM_MYRINET_BUS_PCI) != GM_SUCCESS)
  {
    gm_hpux_printf (("gm_hpux_init: gm_instance_init failed\n"));
    goto abort_with_malloc;
  }

  if (gm_minor_alloc(is, &is->clone_minor) != GM_SUCCESS) 
  {
    gm_hpux_printf (("gm_minor_alloc failed.\n"));
    goto abort_with_is;
  }

  gm_hpux_printf (("is->clone_minor is %d.\n", is->clone_minor));
  
  gm_minor_get_port_state(is->clone_minor)->instance = is;
  gm_minor_get_port_state(is->clone_minor)->privileged = 0;

  if (gm_minor_alloc(is, &is->privileged_clone_minor) != GM_SUCCESS) 
  {
    gm_hpux_printf (("gm_minor_alloc failed.\n"));
    goto abort_with_clone_minor;
  }

  gm_hpux_printf (("is->privileged_clone_minor is %d.\n", is->privileged_clone_minor));

  gm_minor_get_port_state (is->privileged_clone_minor)->instance = is;
  gm_minor_get_port_state (is->privileged_clone_minor)->privileged = 1;

  gm_always_assert (GM_HPUX_IS_STATIC_MINOR (is->privileged_clone_minor));
  gm_always_assert (GM_HPUX_STATIC_MINOR_TO_UNIT(is->privileged_clone_minor) == is->arch.unit);
  gm_always_assert (GM_HPUX_IS_STATIC_MINOR (is->clone_minor));
  gm_always_assert (GM_HPUX_STATIC_MINOR_TO_UNIT(is->clone_minor) == is->arch.unit);

  gm_hpux_printf (("gm_hpux_init returning normally.\n"));
  
  return 0;

  abort_with_clone_minor:
  gm_minor_free (is->clone_minor);

  abort_with_is:
  gm_instance_finalize (is);

  abort_with_malloc:
  gm_free (is);
  gm_hpux_instances [gm_hpux_num_instances] = 0;
  gm_hpux_num_instances--;

  return 1;
}

static int gm_hpux_detach (gm_instance_state_t*is)
{
  gm_hpux_printf (("gm_hpux_detach\n"));

  /*  gm_disable_PCI_config_command_bit (is, GM_PCI_COMMAND_MASTER);*/
  isrunlink (is->arch.isc, gm_hpux_isr, -1, (long) is->arch.isc, is->arch.unit);
  
  if (is->privileged_clone_minor)
  {
    gm_minor_free (is->privileged_clone_minor);
    is->privileged_clone_minor = 0;
  }
  if (is->clone_minor)
  {
    gm_minor_free (is->clone_minor);
    is->clone_minor = 0;
  }
  gm_instance_finalize (is);
  gm_always_assert (is->arch.unit >= 0 && is->arch.unit < gm_hpux_num_instances);
  gm_always_assert (is == gm_hpux_instances [is->arch.unit]);
  gm_free (is);
  gm_hpux_instances [is->arch.unit] = 0;
}

static int gm_hpux_attach (uint32_t product_id, struct isc_table_type*isc)
{
  PCI_ID*id = (PCI_ID*) &product_id;
  
  gm_hpux_printf (("gm_hpux_attach \n"));
  if ((id->vendor_id == GM_PCI_VENDOR_MYRICOM && id->device_id == GM_PCI_DEVICE_MYRINET || 
       id->vendor_id == GM_PCI_VENDOR_MYRICOM2 && id->device_id == GM_PCI_DEVICE_MYRINET))
  {
    printf ("found a card\n");

    gm_hpux_printf (("testing debug print\n"));
    gm_hpux_printf (("debug print worked\n"));
    
    isc->gfsw = &gm_hpux_gfsw;
    isc->gfsw->init = gm_hpux_init;
    isc->gfsw->diag = 0;
    isc->if_id = (int)(id->device_id & 0x0000ffffU);

    isrlink (isc, gm_hpux_isr, -1, (long) isc, gm_hpux_num_instances);

    isc_claim (isc, &gm_hpux_wsio_driver_info);
  }
  
  if (gm_hpux_saved_attach)
    return (*gm_hpux_saved_attach)(product_id, isc);
  else
    return 0;
}

static int gm_hpux_load (void*driver_info_arg)
{
  int i;
  
  gm_assert (driver_info_arg);
  gm_hpux_wsio_driver_info.drv_info = (struct drv_info*) driver_info_arg;

  printf ("hello. I was compiled on %s %s\n",__DATE__, __TIME__);

  if (!wsio_install_driver (&gm_hpux_wsio_driver_info))
  {
    printf ("wsio_install_driver failed\n");
    return ENXIO;
  }

  if (mod_wsio_attach_list_add (MOD_WSIO_PCI, &gm_hpux_attach))
  {
    printf ("gm_hpux_load: mod_wsio_attach_list_add failed\n");
    wsio_uninstall_driver (&gm_hpux_wsio_driver_info);
    return ENXIO;
  }
  
  gm_arch_sync_init (&gm_hpux_debug_sync, 0);
  gm_arch_get_page_len (&GM_PAGE_LEN);

  gm_hpux_dev_init ();
  gm_init ();
  return 0;
}

static int gm_hpux_unload (void*driver_info_arg)
{
  int i;

  gm_hpux_printf (("gm_hpux_unload.\n"));

  if (++gm_hpux_unload_count % 2)
  {
    gm_hpux_printf (("skipping this unload, saying we're busy\n"));
    return EBUSY;
  }
  
  for (i = 0; i < gm_hpux_num_instances; i++)
  {
    gm_always_assert (gm_hpux_instances [i]);
    gm_hpux_detach (gm_hpux_instances [i]);
  }

  if (mod_wsio_attach_list_remove (MOD_WSIO_PCI, &gm_hpux_attach))
  {
    gm_hpux_printf (("gm_hpux_attach: mod_wsio_attach_list_remove failed\n"));
    return ENXIO;
  }
  
  /*  gm_hpux_wsio_driver_info.drv_info = (struct drv_info*) driver_info_arg;*/
  wsio_uninstall_driver (&gm_hpux_wsio_driver_info);

  return 0;
}

static int gm_hpux_dev_init ()
{
  if (gm_hpux_saved_dev_init)
    return (*gm_hpux_saved_dev_init)();
  else
    return 1;
}


/*called if statically configured*/
int gm_install ()
{
  gm_hpux_saved_attach = pci_attach;
  pci_attach = gm_hpux_attach;
  
  gm_hpux_saved_dev_init = dev_init;
  dev_init = gm_hpux_dev_init;
  
  return wsio_install_driver (&gm_hpux_wsio_driver_info);  
}


static int gm_hpux_open (dev_t dev, int flag)
{
  int minor_id;
  minor_id  = minor (dev);

  gm_hpux_printf (("gm_hpux_open with minor number %d\n", minor_id));

  if (GM_HPUX_IS_STATIC_MINOR (minor_id))
  {
    gm_hpux_printf (("I think this is a static minor\n"));
    return gm_hpux_instances [GM_HPUX_STATIC_MINOR_TO_UNIT (minor_id)] ? 0 : ENODEV;
  }
  else if (GM_HPUX_INVALID_DYMANIC_MINOR (minor_id))
  {
    gm_hpux_printf (("I'm returning an error because I think minor_id is out of range\n"));
    return ENODEV;
  }
  gm_hpux_printf (("gm_hpux_open returning 0\n"));

  return 0;
}

static int gm_hpux_close (dev_t dev)
{
  int minor_id;
  gm_instance_state_t*is;
  gm_port_state_t*ps;
  
  minor_id  = minor (dev);

  gm_hpux_printf (("gm_hpux_close with minor number %d\n", minor_id));

  if (GM_HPUX_IS_STATIC_MINOR (minor_id))
  {
    gm_hpux_printf (("GM_HPUX_IS_STATIC_MINOR was true\n"));
    if (!(is = gm_hpux_instances [GM_HPUX_STATIC_MINOR_TO_UNIT (minor_id)]))
      return ENODEV;
  }
  else
  {
    if (GM_HPUX_INVALID_DYMANIC_MINOR (minor_id))
      return ENODEV;
    
    gm_hpux_printf (("GM_HPUX_IS_STATIC_MINOR was false. calling gm_minor_get_port_state\n"));
    
    minor_id = GM_HPUX_OS_MINOR_TO_GM_MINOR (minor_id);

    gm_hpux_printf (("didn't crash\n"));
    
    if (!(ps = gm_minor_get_port_state (minor_id)))
    {
      gm_hpux_printf (("gm_hpux_close: couldn't get port state for minor %d\n", minor_id));
      return ENODEV;
    }
    if (ps->opened)
      gm_port_state_close (ps);
    gm_minor_free (minor_id);
  }
  return 0;
}

static int gm_hpux_localize_status (gm_status_t status)
{
#define CASE(from,to) case from : return to
  switch (status) {
    CASE(GM_SUCCESS, 0);
    CASE(GM_INPUT_BUFFER_TOO_SMALL, EFAULT);
    CASE(GM_OUTPUT_BUFFER_TOO_SMALL, EFAULT);
    CASE(GM_TRY_AGAIN, EAGAIN);
    CASE(GM_BUSY, EBUSY);
    CASE(GM_MEMORY_FAULT, EFAULT);
    CASE(GM_INTERRUPTED, EINTR);
    CASE(GM_INVALID_PARAMETER, EINVAL);
    CASE(GM_OUT_OF_MEMORY, ENOMEM);
    CASE(GM_INVALID_COMMAND, EINVAL);
    CASE(GM_PERMISSION_DENIED, EPERM);
    default:
      CASE(GM_INTERNAL_ERROR, EPFNOSUPPORT);
      CASE(GM_UNATTACHED, -1);	/* ??? */
      CASE(GM_UNSUPPORTED_DEVICE, ENXIO);
  }
}
char*_gm_ioctl_cmd_name (unsigned int cmd);

static int gm_hpux_ioctl (dev_t dev, int cmd, caddr_t arg, int flag)
{
  int r, minor_id, unit;
  gm_instance_state_t*is;
  gm_port_state_t*cps, *ps;
  void*p;
  gm_mmap_info_t m;
  gm_hpux_mapped_chunk_t*chunk;
    
  gm_hpux_printf (("gm_hpux_ioctl\n"));
  
  minor_id = minor (dev);
  
  if (cmd == GM_MMAP)
  {
    gm_hpux_printf (("GM_MMAP\n"));

    if (gm_arch_copyin (*((void**)arg), &m, sizeof (gm_mmap_info_t)) != GM_SUCCESS)
    {
      gm_hpux_printf (("gm_arch_copyin failed\n"));
      return ENXIO;
    }
    
    gm_hpux_printf (("m.off is %x, m.len is %x\n", m.offset, m.len));

    if (!GM_HPUX_IS_STATIC_MINOR (minor_id))
      minor_id = GM_HPUX_OS_MINOR_TO_GM_MINOR (minor_id);
  
    gm_hpux_printf (("calling gm_minor_get_port_state\n"));
    
    if (!(cps = gm_minor_get_port_state (minor_id)))
    {
      gm_hpux_printf (("gm_hpux_ioctl: gm_minor_get_port_state failed\n"));
      return ENODEV;
    }

    gm_hpux_printf (("calling gm_prepare_to_mmap\n"));

    if (gm_prepare_to_mmap (cps, m.offset, m.len, m.flags) != GM_SUCCESS)
    {
      gm_hpux_printf (("gm_prepare_to_mmap failed\n"));
      return ENODEV;
    }
    gm_hpux_printf (("calling gm_mmap\n"));

    if (gm_mmap (cps, m.offset, &p) != GM_SUCCESS)
    {
      gm_hpux_printf (("gm_mmap failed, which it should never do\n"));
      return ENODEV;
    }
    
    if (gm_mapping_in_io_space (cps, m.offset))
    {
      m.flags = 0;
      
      gm_hpux_printf (("getting physical address for p %x\n", p));

      chunk = gm_hpux_get_chunk (cps->instance, p, m.len);
      if (!chunk)
      {
	gm_hpux_printf (("gm_hpux_get_chunk failed\n"));
	return ENODEV;
      }
      
      if (1 || !chunk->mapped)
      {
	gm_hpux_printf (("calling user_iomap\n"));
      
	chunk->ua = user_iomap (0, chunk->pa, chunk->length / PAGESIZE);
	if (!chunk->ua)
	{
	  gm_hpux_printf (("user_iomap failed\n"));
	  return ENODEV;
	}
	gm_hpux_printf (("user_iomap: ua is %x, pa is %x\n", chunk->ua, chunk->pa));
	chunk->mapped = 1;
      }
      gm_hpux_printf (("chunk->ua is %x, chunk->ua + (int)((char*) p - (char*) chunk->pa is %x\n",
		       chunk->ua,
		       chunk->ua + (int)((char*) p - (char*) chunk->pa)));
      
      m.va = chunk->ua + (int)((char*) p - (char*) chunk->pa);
    }
    else
    {
      gm_hpux_printf (("not in io space. Just returning kernel pointer %x\n", p));
      m.flags = GM_HPUX_KERNEL_SPACE;
      m.va = p;
    }
    
    gm_hpux_printf (("calling gm_finish_mmap\n"));
    gm_finish_mmap (cps, m.offset, m.len, (gm_up_t)m.va);

    gm_hpux_printf (("copying out mmap_info flags = %x, va = %x\n", m.flags, m.va));
    
    if (gm_arch_copyout (&m, *((void**)arg), sizeof (gm_mmap_info_t)) != GM_SUCCESS)
    {
      gm_hpux_printf (("gm_arch_copyout failed\n"));
      return ENXIO;
    }

    gm_hpux_printf (("returning 0 from ioctl\n"));
    
    return 0;
  }
  else if (cmd == GM_MUNMAP)
  {
    gm_hpux_printf (("GM_MUNMAP\n"));
    return ENXIO;
  }
  else if (cmd == GM_GET_DEV)
  {
    gm_hpux_printf (("GM_GET_DEV\n"));
    gm_hpux_printf (("os minor_id is %d\n", minor_id));

    if (!GM_HPUX_IS_STATIC_MINOR (minor_id))
    {
      gm_hpux_printf (("gm_hpux_ioctl: bad minor number %d for GM_GET_DEV\n", minor_id));
      return ENODEV;
    }
 
    unit = GM_HPUX_STATIC_MINOR_TO_UNIT (minor_id);
    if (unit < 0 || unit >= gm_hpux_num_instances)
    {
      gm_hpux_printf (("gm_hpux_ioctl: bad unit %d for GM_GET_DEV\n", unit));
      return ENODEV;
    }
    
    if (!(is = gm_hpux_instances [unit]))
      return ENODEV;
  
    if (gm_minor_alloc (is, &minor_id) != GM_SUCCESS)
    {
      gm_hpux_printf (("gm_hpux_ioctl: gm_minor_alloc failed\n"));
      return EINVAL;
    }
    gm_hpux_printf (("alloc'd minor_id is %d\n", minor_id));
    
    if (!(ps = gm_minor_get_port_state (minor_id)))
    {
      gm_hpux_printf (("gm_hpux_ioctl: gm_minor_get_port_state failed for new device\n"));
      return ENODEV;
    }
    ps->instance = is;
    ps->privileged = is;
  
    gm_hpux_printf (("new gm minor_id is %d\n", minor_id));

    minor_id = GM_HPUX_GM_MINOR_TO_OS_MINOR (minor_id);
    
    gm_hpux_printf (("to the os this will be minor id %d\n", minor_id));
    
    p = *((void**) arg);
    gm_arch_copyout (&minor_id, (void*) p, sizeof (int));
    return 0;
  }
  else
  {
    gm_hpux_printf (("%s\n", _gm_ioctl_cmd_name (cmd)));
    
    gm_hpux_printf (("minor_id is %d, GM_HPUX_OS_MINOR_TO_GM_MINOR (minor_id) is %d",
		     minor_id, GM_HPUX_OS_MINOR_TO_GM_MINOR (minor_id)));
    
    if (!(ps = gm_minor_get_port_state (GM_HPUX_OS_MINOR_TO_GM_MINOR (minor_id))))
    {
      gm_hpux_printf (("fuck! gm_minor_get_port_state failed\n"));
      return ENXIO;
    }
    
    gm_hpux_printf (("about to call gm_ioctl\n"));
    
    gm_arch_mutex_enter (&ps->sync);
    p = *((void**) arg);
    r = gm_ioctl (ps, cmd, p, INT_MAX, p, INT_MAX, 0);
    gm_arch_mutex_exit (&ps->sync);
    return gm_hpux_localize_status (r);
  }
}

void *__gm_arch_kernel_malloc (unsigned long len, int flags)
{
  char*p;
  p = kmalloc (len, M_DYNAMIC, M_NOWAIT);
  gm_hpux_printf (("kernel malloc'd range from %p to %p\n", p,  p + len));
  return p;
}

void __gm_arch_kernel_free (void *ptr)
{
  /* kfree (ptr, M_IOSYS);*/
  FREE (ptr, M_DYNAMIC);
}

void gm_arch_sync_init (gm_arch_sync_t *s, gm_instance_state_t *is)
{
  b_initsema (&s->semaphore, 1, REAL_DRV_SEMA_ORDER, "gm_arch_sync");
  return;
}

void gm_arch_mutex_enter (gm_arch_sync_t *s)
{
  b_psema (&s->semaphore);
}

void gm_arch_mutex_exit (gm_arch_sync_t *s)
{
  b_vsema (&s->semaphore);
}

void gm_arch_sync_destroy (gm_arch_sync_t *s)
{
}

void gm_arch_sync_reset (gm_arch_sync_t *s)
{
  gm_arch_sync_init (s, 0);
  return;
}

void gm_arch_wake (gm_arch_sync_t * s)
{
  wakeup (s);
}

gm_arch_sleep_status_t gm_arch_signal_sleep (gm_arch_sync_t *s)
{
  get_sleep_lock (s);
  return sleep (s, PZERO | PCATCH) == 0 ? GM_SLEEP_WOKE : GM_SLEEP_INTERRUPTED;
}

static int gm_hpux_timeout (gm_arch_sync_t*s)
{
  wakeup (s);
}

static gm_arch_sleep_status_t gm_hpux_timed_sleep (gm_arch_sync_t *s, int milliseconds)
{
  get_sleep_lock (s); 
  timeout (gm_hpux_timeout, s, milliseconds * HZ / 1000);
  if (sleep (s, PZERO))
  {
    untimeout (gm_hpux_timeout, s);
    return GM_SLEEP_INTERRUPTED;
  }
  return GM_SLEEP_WOKE;
}

gm_arch_sleep_status_t gm_arch_timed_sleep (gm_arch_sync_t *s, int seconds)
{
  return gm_hpux_timed_sleep (s, seconds * 1000);
}

void gm_hpux_debug_sleep ()
{
  gm_hpux_timed_sleep (&gm_hpux_debug_sync, 100);
}

gm_status_t gm_arch_dma_region_alloc (gm_instance_state_t * is, gm_arch_dma_region_t * r,gm_size_t len, gm_u32_t flags,gm_register_page_function_t register_page_function, void *user_arg)
{
  int i, num_pages;
 
  if (wsio_allocate_shared_memory (is->arch.isc, len, &r->iova, &r->va, 0) != SHMEM_OK)
    return GM_FAILURE;
  
  gm_hpux_printf (("gm_arch_dma_region_alloc\n"));
  gm_hpux_printf (("is->arch.isc is %p, len is %d, r->iova is %p, r->va is %p\n", is->arch.isc, len, r->iova, r->va));
  gm_hpux_printf (("and r is %p\n", r));
  
  r->ai = &is->arch;
  r->iova_advance = r->iova;
  r->length = len;
  
  num_pages = (GM_PAGE_ROUNDUP (u32, len) / GM_PAGE_LEN);
  
  gm_always_assert (num_pages >= 1);
  
  if (register_page_function)
  {
    for (i = 0; i < num_pages; i++)
    {
      gm_dp_t p = (gm_dp_t) (r->iova + PAGESIZE * i);
      register_page_function (user_arg, p, i);
    }
  }
  return GM_SUCCESS;
}

void gm_arch_dma_region_free (gm_arch_dma_region_t * r)
{
  /*  gm_hpux_printf (("gm_arch_dma_region_free\n"));
      gm_hpux_printf (("r is %p\n", r));
      gm_hpux_printf (("r->ai->isc is %p, len is %d, r->iova is %p, r->va is %p\n", r->ai->isc, r->length, r->iova, r->va));
  */
  wsio_free_shared_memory (r->ai->isc, r->length, r->iova, r->va, 0);
}

gm_dp_t gm_arch_dma_region_dma_addr (gm_arch_dma_region_t * r)
{
  gm_dp_t p;
  p = (gm_dp_t) r->iova;
  return p;
}

gm_dp_t gm_arch_dma_region_dma_addr_advance (gm_arch_dma_region_t * r)
{
  gm_dp_t p;
  p = (gm_dp_t) r->iova_advance;
  r->iova_advance += PAGESIZE;
  return p;
}

void * gm_arch_dma_region_kernel_addr (gm_arch_dma_region_t * r)
{
  return r->va;
}

gm_status_t gm_arch_dma_region_sync (gm_arch_dma_region_t * r, int command)
{
  dma_sync_IO (KERNELSPACE, r->va, r->length, 0);
  return GM_SUCCESS;
}

gm_status_t gm_arch_map_io_space (gm_instance_state_t *is, gm_u32_t offset, gm_u32_t len, void **kaddr)
{
  void*pa = EXTEND_ADDR ((is->arch.pci_base_address + offset));
  
  *kaddr = kernel_iomap_public (0, pa, len / PAGESIZE, PROT_URW);
  gm_hpux_printf (("gm_arch_map_io_space returning address %p, hex length %x\n", *kaddr, len));

  if (*kaddr)
  {
    gm_hpux_mapped_chunk_t* new_chunk = gm_malloc (sizeof (gm_hpux_mapped_chunk_t), 0);
    void*tp;
    
    if (!new_chunk)
    {
      gm_hpux_printf (("gm_arch_map_io_space: kernel malloc failed\n"));
      kernel_iounmap (*kaddr, len / PAGESIZE);
      *kaddr = 0;
      return GM_FAILURE;
    }
    gm_hpux_printf (("pa is %p and va is %p\n", pa, *kaddr));
    
    new_chunk->mapped = 0;
    new_chunk->pa = pa;
    new_chunk->ka = *kaddr;
    new_chunk->ua = 0;
    new_chunk->length = len;
    new_chunk->prev = 0;
    new_chunk->next = is->arch.mapped_chunks;
    if (is->arch.mapped_chunks)
      is->arch.mapped_chunks->prev = new_chunk;
    is->arch.mapped_chunks = new_chunk;

  }
  return *kaddr ? GM_SUCCESS : GM_FAILURE;
}

void gm_arch_unmap_io_space (gm_instance_state_t *is, gm_u32_t offset, gm_u32_t len, void **kaddr)
{
  gm_hpux_mapped_chunk_t*chunk;
  for (chunk = is->arch.mapped_chunks; chunk && chunk != *kaddr; chunk = chunk->next);  

  if (!chunk)
  {
    gm_hpux_printf (("gm_arch_unmap_io_space: couldn't find mapped chunk\n"))
      return;
  }

  if (chunk->next)
    chunk->next->prev = chunk->prev;
  if (chunk->prev)
    chunk->prev = chunk->next;
  else is->arch.mapped_chunks = chunk->next;

  gm_free (chunk);
  
  kernel_iounmap (*kaddr, len / PAGESIZE);
}

extern char*hostname;
gm_status_t gm_arch_gethostname (char *ptr, int len)
{
  /*  strncpy (ptr, hostname, len);*/
  return GM_FAILURE;
}

void gm_arch_spin (gm_instance_state_t *is, gm_u32_t usecs)
{
  busywait (usecs);
}

gm_status_t gm_arch_port_state_init (gm_port_state_t *ps)
{
  return GM_SUCCESS;
}

gm_status_t gm_arch_port_state_open (gm_port_state_t *ps)
{
  return GM_SUCCESS;
}

void gm_arch_port_state_close (gm_port_state_t *ps)
{
  return;
}

void gm_arch_port_state_fini (gm_port_state_t *ps)
{
  return;
}

gm_status_t gm_arch_read_pci_config_32 (gm_instance_state_t *is, gm_offset_t offset, gm_u32_t *value)
{
  pci_read_cfg_uint32_isc (is->arch.isc, offset, value);
  return GM_SUCCESS;
}

gm_status_t gm_arch_write_pci_config_32 (gm_instance_state_t *is, gm_offset_t offset, gm_u32_t value)
{
  pci_write_cfg_uint32_isc (is->arch.isc, offset, value);
  return GM_SUCCESS;
}

gm_status_t gm_arch_read_pci_config_16 (gm_instance_state_t *is, gm_offset_t offset, gm_u16_t *value)
{
  pci_read_cfg_uint16_isc (is->arch.isc, offset, value);
  return GM_SUCCESS;
}

gm_status_t gm_arch_write_pci_config_16 (gm_instance_state_t *is, gm_offset_t offset, gm_u16_t value)
{
  pci_write_cfg_uint16_isc (is->arch.isc, offset, value);
  return GM_SUCCESS;
}

gm_status_t gm_arch_read_pci_config_8 (gm_instance_state_t *is, gm_offset_t offset, gm_u8_t *value)
{
  pci_read_cfg_uint8_isc (is->arch.isc, offset, value);
  return GM_SUCCESS;
}

gm_status_t gm_arch_write_pci_config_8 (gm_instance_state_t *is, gm_offset_t offset, gm_u8_t value)
{
  pci_write_cfg_uint8_isc (is->arch.isc, offset, value);
  return GM_SUCCESS;
}

gm_port_state_t *gm_arch_port_for_id (gm_instance_state_t * is, unsigned port)
{
  return GM_FAILURE;
}

gm_status_t gm_arch_copyout (void *from, void *to, gm_size_t len)
{
  return copyout (from, to, len) == 0 ? GM_SUCCESS : GM_FAILURE;
}

gm_status_t gm_arch_copyin (void *from, void *to, gm_size_t len)
{
  return copyin (from, to, len) == 0 ? GM_SUCCESS : GM_FAILURE;
}

void gm_arch_abort (void)
{
  return;
}

void memcpy (char*a, char*b, int len)
{
  bcopy (b, a, len);
}

gm_status_t gm_arch_get_page_len (unsigned long*r)
{
  *r = PAGESIZE;
  return GM_SUCCESS;
}

/*don't care*/
gm_status_t gm_arch_mmap_contiguous_segment (gm_port_state_t *ps, void * kaddr, unsigned long blockSize, gm_up_t *vaddr)
{
  return GM_FAILURE;
}
void gm_arch_munmap_contiguous_segments (gm_port_state_t *ps)
{
  return;
}
gm_s32_t gm_arch_dma_region_status (gm_arch_dma_region_t *r)
{
  return 0; 
}

int gm_hpux_get_num_instances ()
{
  return gm_hpux_num_instances;
}


void gm_arch_unlock_user_buffer_page (gm_arch_page_lock_t*lock)
{
  gm_hpux_printf (("gm_arch_unlock_user_buffer_page\n"));
  
  (((io_map_cntl_t*) lock->isc->map_funcs)->unmap)
    (
     ((io_map_cntl_t*) lock->isc->map_funcs)->arg, &lock->io_vec);

  /*  wsio_unmap (lock->isc, &lock->io_vec);*/
}

extern space_t ldusid(caddr_t addr);

gm_status_t gm_arch_lock_user_buffer_page (gm_port_state_t*ps, gm_up_t in, gm_dp_t*dma_addr, gm_arch_page_lock_t*lock)
{
  gm_instance_state_t *is;

  init_map_context (&lock->map_cb);

  gm_hpux_printf (("gm_arch_lock_user_buffer_page\n"));

  gm_assert (ps);
  is = ps->instance;
  gm_assert (is);
  
  lock->isc = is->arch.isc;
  lock->host_vec.iov_base = (caddr_t) in;
  lock->host_vec.iov_len = PAGESIZE;

  
  if ((((io_map_cntl_t*) lock->isc->map_funcs)->map_contiguous)
      (
       ((io_map_cntl_t*) lock->isc->map_funcs)->arg,
       &lock->map_cb, IO_CONTIGUOUS | IO_SAFE | IO_LOCK,
       ldusid (lock->host_vec.iov_base),
       &lock->host_vec,
       &lock->io_vec) < 0)

    /*
      if (wsio_map
      (lock->isc,
      &lock->map_cb, 
      IO_CONTIGUOUS | IO_SAFE | IO_LOCK,
      ldsid (lock->host_vec.iov_base),
      &lock->host_vec,
      &lock->io_vec) < 0)*/
  {
    gm_hpux_printf (("wsio_map failed\n"));
    return GM_FAILURE;
  }

  /*
  if (lock->io_vec.iov_len != lock->host_vec.iov_len)
  {
    gm_hpux_printf (("wsio_map half failed\n"));
    gm_hpux_printf (("len %d vs len %d\n", lock->io_vec.iov_len, lock->host_vec.iov_len));
    gm_arch_unlock_user_buffer_page (lock);
    return GM_FAILURE;
  }
  */
  *dma_addr = lock->io_vec.iov_base;
  return GM_SUCCESS;
}

void*memset (void *s, int c, size_t n)
{
  while (n--)
    ((char *) s)[n] = c;
  return s;
}
