/******************************************************************-*-c-*-
 * Myricom GM networking software and documentation			 			 *
 * Copyright (c) 2000 by Myricom, Inc.					 				 *
 * All rights reserved.	 See the file `COPYING' for copyright notice.	 *
 *************************************************************************/

#include <sys/types.h>
#include "/usr/include/.unsupp/sys/_types.h"
#include <sys/errno.h>
#include <sys/io.h>
#include <sys/spinlock.h>
#include <sys/malloc.h>
#include <sys/socket.h>
#include <sys/mbuf.h>
#include <sys/param.h>
#include <sys/dlpi.h>
#include <sys/dlpi_ext.h>
#include <sys/mci.h>
#include <sys/mib.h>
#include <netinet/mib_kern.h>
#include <io/dma.h>
#include <wsio/pci.h>
#include <net/if.h>
#include <net/if_arp.h>
#include <netinet/in.h>
#include <netinet/if_ether.h>
#include <netinet/if_ieee.h>
#include <sio/lan_dlpikrn.h>
#include <sys/stream.h>
#include <sys/stropts.h>
#include <sys/netio.h>
#include <sys/net_diag.h>
#include <sys/netdiag1.h>
#include <sys/callout.h>
#include <sys/sysinfo.h>
#include <mod_conf.h>
#include <sys/moddefs.h>
#include <wsio/wsio.h>	/* this must be the last #include */

#include "gm_internal.h"
#include "gm_page_hash.h"
#include "gm_ether.h"
#include "gm_lanai.h"
#include "gm_instance.h"
#include "gm_impl.h"
#include "gmip.h"

#include "finucane.h"

static
void gmip_recv_interrupt (void *gmnetp, const unsigned int ulen, gm_u16_t csum)
{
}

static
void gmip_sent_interrupt(void *gmnetp)
{
}

static int gmip_unload_count = 0;

extern gm_instance_state_t* gm_hpux_instances [GM_ARCH_MAX_INSTANCE];
extern int gm_hpux_num_instances;

int gmip_load (void*driver_info_arg);
int gmip_unload (void*driver_info_arg);
int gmip_open (queue_t *q, dev_t *devp, int flag, int sflag, cred_t *credp);
int gmip_close (queue_t*q);
int gmip_rsrv (queue_t*q);
int gmip_wsrv (queue_t*q);
int gmip_wput (queue_t *q, mblk_t *mblkp);

extern struct mod_operations gio_mod_ops;
drv_ops_t gmip_driver_ops;
static struct mod_type_data gmip_driver_link =
{
  "gmip",
  0
};

static struct modlink gmip_mod_link [2] =
{
  { &gio_mod_ops,
    &gmip_driver_link
  },
  {
    0, 0
  }
};
 
static struct mod_conf_data gmip_mc =
{
  MCD_VERSION, 0
};

struct modwrapper gmip_wrapper = 
{
  MODREV,
  gmip_load,
  gmip_unload,
  0,
  &gmip_mc,
  gmip_mod_link
};


drv_info_t gmip_driver_info = 
{
  "gmip",
  "pseudo",
  DRV_CHAR | DRV_PSEUDO | DRV_MP_SAFE,
  -1,
  -1,
  0,0,0
};

static struct module_info gmip_rminfo =  
{
  5060, "gmip", 0, 65536, 65536, 1
};

static struct module_info gmip_wminfo =  
{
  5060, "gmip", 0, 65536, 1, 1
};

static struct qinit gmip_rinit = 
{
  0, gmip_rsrv, gmip_open, gmip_close, 0, &gmip_rminfo
};

static struct qinit gmip_winit =
{
  gmip_wput, gmip_wsrv, 0, 0, 0, &gmip_wminfo
};

struct streamtab gmip_info = { &gmip_rinit, &gmip_winit };

streams_info_t gmip_str_info =
{
  "gmip",
  -1,
  { 
    &gmip_rinit, &gmip_winit, NULL, NULL
  },

  STR_IS_DEVICE | MGR_IS_MP | STR_SYSV4_OPEN,
  SQLVL_QUEUE,
  ""
};

static int gmip_attach (struct isc_table_type*isc)
{
  char hw_path_str[100];
  hw_path_t hw_path;
  
  gmip_data_t*d = kmalloc (sizeof (gmip_data_t), M_DYNAMIC, M_NOWAIT);
  
  gm_always_assert (d);
  bzero (d, sizeof (gmip_data_t));

  d->hwift.mac_type = DEV_ETHER;
  d->hwift.llc_flags = ETHERTYPE;
  d->hwift.mjr_num = -1;
  d->hwift.nm_id = get_nmid ();
  d->hwift.instance_num = wsio_isc_to_instance (isc, 0);
  d->hwift.mtu = GM_IP_MTU;
  d->hwift.name = "gmip";
  if (!io_node_to_hw_path (isc->card_node, 0, &hw_path) && io_hw_path_to_str (hw_path_str, 0, &hw_path) != -1)
    strcpy (d->hwift.hdw_path, hw_path_str);
  d->hwift.hdw_state = 0;
  d->hwift.mac_addr_len = 6;
  gm_get_unique_board_id (d->port, d->hwift.mac_addr);
  d->hwift.features = DRV_MP | DRV_MBLK | DRV_NO_FAST_PATH;
  d->hwift.arpmod_name = "";
  d->hwift.ppa = isc->my_address;
  d->hwift.watch_timer = 0;
  d->hwift.reserved2 = 0;

  gm_hpux_printf (("calling hw_ift_attach, so help me God.\n"));
  hw_ift_attach (&d->hwift);
}

			
static int gmip_load (void*driver_info_arg)
{
  int i, r;
  drv_info_t*drv_info_p = (struct drv_info*) driver_info_arg;
  
  gm_assert (driver_info_arg);
  
  bzero (&gmip_driver_ops, sizeof (drv_ops_t));
  gmip_str_info.inst_major = drv_info_p->c_major;
  
  gm_hpux_printf (("gmip_load. I was compiled on %s %s\n",__DATE__, __TIME__));

  gm_hpux_printf (("calling install_driver\n"));
  
  if (r = install_driver (drv_info_p, &gmip_driver_ops))
  {
    gm_hpux_printf (("install_driver failed\n"));
    return r;
  }
  
  gm_hpux_printf (("calling str_install\n"));
  if (r = str_install (&gmip_str_info))
  {
    gm_hpux_printf (("str_install failed with return value %d \n", r));
    uninstall_driver (drv_info_p);
    return r;
  }
  gm_hpux_printf (("calling gmip_attach\n"));
  for (i = 0; i < gm_hpux_num_instances; i++)
    if (!gmip_attach (gm_hpux_instances [i]->arch.isc))
      return 1;
  
  return 0;
}

static int gmip_unload (void*driver_info_arg)
{
  gm_hpux_printf (("gmip_unload\n"));

  if (++gmip_unload_count % 2)
  {
    gm_hpux_printf (("skipping this unload, saying we're busy\n"));
    return EBUSY;
  }
  str_uninstall (&gmip_str_info);
  uninstall_driver (&gmip_driver_info);

  return 0;
}

static int gmip_open (queue_t *q, dev_t *devp, int flag, int sflag, cred_t *credp)
{
  gm_hpux_printf (("gmip_open\n"));
  return 0;
}

static int gmip_close (queue_t*q)
{
  gm_hpux_printf (("gmip_close\n"));
  return 0;
}

static int gmip_rsrv (queue_t*q)
{
  gm_hpux_printf (("gmip_rsrv\n"));
  return 0;
}

static int gmip_wsrv (queue_t*q)
{
  gm_hpux_printf (("gmip_wsrv\n"));
  return 0;
}

static int gmip_wput (queue_t *q, mblk_t *mblkp)
{
  gm_hpux_printf (("gmip_wput\n"));
  return 0;
}

