/******************************************************************-*-c-*-
 * Myricom GM networking software and documentation			 *
 * Copyright (c) 2001 by Myricom, Inc.					 *
 * All rights reserved.	 See the file `COPYING' for copyright notice.	 *
 *************************************************************************/

#ifndef _gm_arch_debug_h_
#define _gm_arch_debug_h_


/**********************************************************************
 * This file defines debugging flags used only by IRIX-specific code.
 * In general, this file should be included only within the ./drivers/irix
 * subtree, but debugging-type test programs like gm_debug.c may also use it.
 *
 * NOTE: For a Table of Contents of this subdirectory, see README_TOC.
 **********************************************************************/

/**********************************************************************
 * 
 * Debugging stuff
 *
 * NOTE: Turning any of these DEBUG or DISPLAY flags on without having
 *       GM_DEBUG set is likely to cause compile-time diagnostics, and
 *       as a rule is a futile exercise.  GM_DEBUG is required to make
 *       these flags function correctly.
 *
 * If you add a flag here, add a corresponding GM_IRIX_PRINT_DEBUG_FLAG()
 * macro invocation in gm_irix.c.
 *
 **********************************************************************/

/* Turn this on to cause no more than <n> instances to be initialized */
#define GM_IRIX_INSTANCE_LIMIT 0   /* (<n> == 0) ==> initialize all of them */

/* Turn this on to enable debugging of the device-hash algorithms */
#define GM_IRIX_DEBUG_DEVHASH 0

/* Turn this on to enable debugging of the device-naming algorithm */
#define GM_IRIX_DEBUG_DEVNAME 0

/* Turn this on to enable debugging of DMA maps */
#define GM_IRIX_DEBUG_DMAMAPS 0

/* Turn this on to enable ifnet debugging */
#define GM_IRIX_DEBUG_IP 0

/* Turn this on to enable "poor man's memory leak detection" */
#define GM_IRIX_DEBUG_MEMLEAKS 0

/* Turn this on to enable debugging of mem allocs and frees */
#define GM_IRIX_DEBUG_MEMORY 0

/* Turn this on to enable debugging of mutexes */
#define GM_IRIX_DEBUG_MUTEX 0

/* Turn this on to enable PIO debugging */
#define GM_IRIX_DEBUG_PIO 0

/* Turn this on to enable sync debugging */
#define GM_IRIX_DEBUG_SYNC 0

/* Turn this on to enable sleep/wake debugging */
#define GM_IRIX_DEBUG_WAKE 0

/* Turn this on to enable display of values from the config space */
#define GM_IRIX_DISPLAY_CONFIG_VALUES    0

/* Turn this on to enable display of the IRIX device descriptor */
#define GM_IRIX_DISPLAY_DEV_DESC         0

/* Turn this on to enable display of the IRIX 'devt' */
#define GM_IRIX_DISPLAY_DEVT             0

/* Turn this on to enable display of global values at startup */
#define GM_IRIX_DISPLAY_GLOBAL_PARMS     GM_DEBUG

/* Turn this on to enable display of the IRIX interrupt handle */
#define GM_IRIX_DISPLAY_INTERRUPT_HANDLE 0

/* Turn this on to enable display of the IRIX device inventory data */
#define GM_IRIX_DISPLAY_INVENTORY        0

/* Turn this on to enable display of the special SET_PORT_NUM args struct */
#define GM_IRIX_DISPLAY_SETPN_ARGS       0

/* Turn this on to enable display of the IRIX vertex handle */
#define GM_IRIX_REPORT_VHANDL            0

/* Turn this on to force all DMA allocs to use PCIIO_DMA_CMD  */
#define GM_IRIX_FORCE_DMA_CMD            1

/* Turn this on to force all DMA allocs to use PCIIO_DMA_DATA */
#define GM_IRIX_FORCE_DMA_DATA           0

#if GM_IRIX_FORCE_DMA_CMD && GM_IRIX_FORCE_DMA_DATA
#warning Trying to force DMA_CMD and DMA_DATA ... illogical
#endif

#endif /* _gm_arch_debug_h_ */
