/******************************************************************-*-c-*-
 * Myricom GM networking software and documentation			 *
 * Copyright (c) 2000 by Myricom, Inc.	      				 *
 * All rights reserved.	 See the file `COPYING' for copyright notice.	 *
 *************************************************************************/

#include "gm_internal.h"
#include "gm_arch_def.h"
#include "gm_impl.h"

#include <asm/page.h>
#include <linux/mm.h>
#include <asm/pgtable.h>
#include <linux/pagemap.h>

/* Linux 2.2.x compatibility stuff, will eventually disappear */
#if LINUX_XX >= 24
#include <asm/pgalloc.h>
#include <linux/highmem.h>
#define l24_pte_page pte_page
#define l24_mk_pte mk_pte

#if GM_CPU_powerpc || GM_CPU_powerpc64
#undef set_pte
#define set_pte gm_linux_set_pte
void gm_linux_set_pte(pte_t *ptep, pte_t pte)
{
#if _PAGE_HASHPTE != 0
	pte_update(ptep, ~_PAGE_HASHPTE, pte_val(pte) & ~_PAGE_HASHPTE);
#else
	*ptep = pte;
#endif
}
#endif /* GM_CPU_powerpc */

#else
#define l24_pte_page(a) virt_to_page (pte_page (a))
#define GFP_HIGHUSER GFP_USER
#define alloc_page(a)  virt_to_page (__get_free_page(a))
#define kmap(a) ((void*)(PAGE_OFFSET + (((a)-mem_map) << PAGE_SHIFT)))
#define kunmap(a)
#define copy_user_page(new,old,addr) copy_page(new,old)
#define l24_mk_pte(a,fl) mk_pte((unsigned long)kmap(a),(fl))
#endif

static int
gm_linux_cow (struct vm_area_struct *vma, pmd_t * src_pmd, pmd_t * dst_pmd,
             unsigned long addr_base)
{
  void *new, *old;
  struct page *new_page;
  pte_t *src_pte, *dst_pte;
  struct page *old_page;
  unsigned long pfn;

  src_pte = pte_offset_map (src_pmd, addr_base);
  dst_pte = pte_offset_map_nested (dst_pmd, addr_base);
  if (!pte_present(*src_pte))
    {
      goto cow_end;
    }
  pfn = gm_linux_pfn (GM_LINUX_PHYS_FROM_PTE (*src_pte));
  if (!gm_hash_find (gm_linux_pfn_hash, &pfn))
    {
      goto cow_end;
    }
  old_page = l24_pte_page (*src_pte);
  gm_always_assert (pte_present (*src_pte) && pte_present (*dst_pte) &&
		    !pte_write (*src_pte) && !pte_write (*dst_pte) &&
		    l24_pte_page (*dst_pte) == old_page);
  /* page count should be: 1 in parent, 1 in child, 1 in page_hash */
  gm_always_assert (atomic_read (&old_page->count) >= 2);
  if (atomic_read (&old_page->count) != 3)
    {
      GM_PRINT (GM_PRINT_LEVEL >= 2,
		("after_fork:old_page->count(%d) != 3 \n",
		 atomic_read (&old_page->count)));
      goto cow_end;
    }
  /* we need to release the pte mapping before forcing the cow that
     can sleep */
  pte_unmap_nested (dst_pte);
  pte_unmap (src_pte);
  if (!gm_linux_get_user_pages)
    {
#ifdef HAVE_ASM_RMAP_H
      GM_WARN (("fork() not supported on this kernel when GM ports are open"));
      return -1;
#endif
      new_page = alloc_page (GFP_HIGHUSER);
      if (!new_page)
	{
	  GM_WARN (("copy-on-write error: not enough memory!\n"));
	  return -1;
	}
      /* copy new_page */
      new = kmap (new_page);
      old = kmap (old_page);
#if GM_CPU_ia64 || GM_CPU_powerpc || GM_CPU_powerpc64 || GM_CPU_x86_64
      memcpy(new,old,PAGE_SIZE);
#else
      copy_user_page (new, old, addr_base);
      /* would be necessary for archs that do not like virtual
	 aliasing, none of the GM/Linux supported archs are affected
	 sparc64 might */
#endif
      kunmap (old_page);
      kunmap (new_page);
#if !GM_CPU_powerpc
      flush_page_to_ram (old_page);
      flush_page_to_ram (new_page);
#endif
      flush_cache_page (vma, addr_base);
      src_pte = pte_offset_map (src_pmd, addr_base);
      dst_pte = pte_offset_map_nested (dst_pmd, addr_base);
      set_pte (dst_pte,
	       pte_mkwrite (pte_mkdirty (l24_mk_pte (new_page, vma->vm_page_prot))));
      page_cache_release (old_page);	/* remove child reference */
      /* put again original page in rw mode */
      set_pte (src_pte, pte_mkwrite (*src_pte));
#if LINUX_XX >= 24 && !GM_CPU_ia64 && !GM_CPU_powerpc && !GM_CPU_powerpc64
      /* we do not need this as we are in fork, we only modified the
	 child tlb which has never been active and there will be a
	 flush_tlb_page before that happens, we are in dup_mmap in the
	 parent and there is a flush_tlb_mm at the end of this
	 function */
#if GM_CPU_alpha && !defined CONFIG_ALPHA_GENERIC && !defined CONFIG_SMP
      /* __load_new_mm_context() is used by the non-generic, non SMP
	 flush_tlb_page() and is not exported, so we need to call
	 the  function through the machine vector */
      if (vma->vm_mm == current->active_mm)
	      alpha_mv.mv_flush_tlb_current_page(vma->vm_mm, vma, addr_base);
      else
	      flush_tlb_other(vma->vm_mm);
#else
      flush_tlb_page (vma, addr_base);
#endif /* GM_CPU_alpha */
#endif /* LINUX_XX >= 24 */
    }
  else /* fonction get_user_pages available */
    {
      int res;
      vma->vm_mm->mmap_cache = vma;
      res = (*gm_linux_get_user_pages) (current,vma->vm_mm, addr_base, 1, 
					1, 0, &new_page, NULL);
      if (res != 1)
	{
	  GM_WARN (("copy-on-write error(%d): not enough memory?\n",res));
	  return -1;
	}
      gm_always_assert (new_page != old_page);
      src_pte = pte_offset_map (src_pmd, addr_base);
      dst_pte = pte_offset_map_nested (dst_pmd, addr_base);
      gm_always_assert (pte_present (*src_pte) && pte_present (*dst_pte) &&
			!pte_write (*src_pte) && pte_write (*dst_pte) &&
			l24_pte_page (*dst_pte) == new_page &&
			l24_pte_page (*src_pte) == old_page);
      /* page count should be: 1 in parent, 1 in child, 1 in page_hash */
      gm_always_assert (atomic_read (&old_page->count) == 2);
      gm_always_assert (atomic_read (&new_page->count) == 2);
      page_cache_release (new_page);
      set_pte (src_pte, pte_mkwrite (*src_pte));
    }
 cow_end:
  pte_unmap_nested (dst_pte);
  pte_unmap (src_pte);
}

void
gm_linux_fork (struct vm_area_struct *vma)
{
  unsigned long addr;
  pgd_t *src_pgd, *dst_pgd;

  gm_assert (vma->vm_mm != current->mm);

  /* OK we must be probably in a fork call */
  addr = vma->vm_start;
  src_pgd = pgd_offset (current->mm, addr);
  dst_pgd = pgd_offset (vma->vm_mm, addr);

  /* loop over pgd entries */
  while (addr < vma->vm_end)
    {
      unsigned long pmd_off = addr & ~PGDIR_MASK;
      unsigned long pmd_end = pmd_off + vma->vm_end - addr;
      pmd_t *src_pmd, *dst_pmd;
      /* check  valid pgd entries */
      if (pgd_none (*src_pgd) || pgd_bad (*src_pgd) ||
          pgd_none (*dst_pgd) || pgd_bad (*dst_pgd))
        {
          goto continue_pgd_loop;
        }
      src_pmd = pmd_offset (src_pgd, pmd_off);
      dst_pmd = pmd_offset (dst_pgd, pmd_off);
      if (pmd_end > PGDIR_SIZE)
	pmd_end = PGDIR_SIZE;

      /* loop over pmd entries */
      while (pmd_off < pmd_end)
	{
	  unsigned long pte_off = pmd_off & ~PMD_MASK;
	  unsigned long pte_end = pte_off + pmd_end - pmd_off;
	  /* check valid pmd entries */
	  if (pmd_none (*src_pmd) || pmd_bad (*src_pmd) ||
              pmd_none (*dst_pmd) || pmd_bad (*dst_pmd))
            {
              goto continue_pmd_loop;
            }
	  if (pte_end > PMD_SIZE)
	    pte_end = PMD_SIZE;

	  /* loop over pte entries */
	  for (; pte_off < pte_end; pte_off += PAGE_SIZE)
	    {
	      /* see code inside linux/mm/memory.c in cow case : do_wp_page */
	      unsigned long addr_base = ((addr & PGDIR_MASK)
					 + (pmd_off & PMD_MASK) + pte_off);
	      gm_linux_cow (vma, src_pmd, dst_pmd, addr_base);
	    }			/* end pte loop */
        continue_pmd_loop:
	  pmd_off = (pmd_off + PMD_SIZE) & PMD_MASK;
	  src_pmd++;
	  dst_pmd++;
	}			/* end pmd loop */
    continue_pgd_loop:
      addr = (addr + PGDIR_SIZE) & PGDIR_MASK;
      src_pgd++;
      dst_pgd++;
    }				/* end pgd loop */
}



/*
  This file uses GM standard indentation:

  Local Variables:
  c-file-style:"gnu"
  tab-width:8
  c-backslash-column:72
  End:
*/
