/******************************************************************-*-c-*-
 * Myricom GM networking software and documentation			 *
 * Copyright (c) 2002 by Myricom, Inc.					 *
 * All rights reserved.	 See the file `COPYING' for copyright notice.	 *
 *************************************************************************/

/* author: gallatin@myri.com */
extern "C"
{

#include <sys/types.h>

#include <kern/kern_types.h>
#include <kern/host.h>
#include <mach/mach_types.h>
#include <mach/vm_types.h>
#include <mach/vm_map.h>
#include <mach/kmod.h>

#include <sys/types.h>
#include <sys/malloc.h>
#include <sys/conf.h>
#include <sys/systm.h>
#include <sys/kernel.h>
#if (GM_MACOSX_VERSION == 101)
#include <sys/signalvar.h>
#else
/* header file seems to have disappeared in Jaguar.  Sigh. */
void	psignal __P((struct proc *p, int sig));
#endif

#include <miscfs/devfs/devfs.h>
#include "gm_cpp.h"
#include "gm_arch.h"
#include "gm_instance.h"
#include "gm_debug.h"
#include "gm_bsd_ether.h"
}

/* IOKit includes for a driver which talks to the PCI bus */

#include <IOKit/pci/IOPCIDevice.h>
#include <IOKit/IODeviceMemory.h>
#include <IOKit/IOInterruptEventSource.h>
#include <IOKit/network/IOMbufMemoryCursor.h>

static int gm_macos_open (dev_t dev, int flags, int devtyple, struct proc *p);
static int gm_macos_close (dev_t dev, int flags, int devtyple, struct proc *p);
static int gm_macos_ioctl (dev_t dev, u_long cmd, caddr_t data, int fflag,
			   struct proc *p);
static int gm_macos_stop (struct tty *tp, int rw);
static int gm_macos_special_create ();
static void gm_macos_special_destroy ();
static int gm_macos_seltrue (dev_t dev, int which, void *wql, struct proc *p);

static struct cdevsw gm_macos_cdevsw = {
  gm_macos_open,		/* d_open */
  gm_macos_close,		/* d_close */
  eno_rdwrt,			/* d_read */
  eno_rdwrt,			/* d_write */
  gm_macos_ioctl,		/* d_ioctl */
  gm_macos_stop,		/* d_stop */
  eno_reset,			/* d_reset */
  0,				/* d_ttys */
  gm_macos_seltrue,		/* d_select */
  eno_mmap,			/* d_mmap */
  eno_strat,			/* strategy */
  eno_getc,			/* d_getc */
  eno_putc,			/* d_putc */
  0				/* d_type */
};

/*
 * This is our IO-Kit "Class".  We interact with IOKit as little as
 * possible.. 
 */


class gm:public IOService
{
  OSDeclareDefaultStructors (gm);

 private:
  int loaded_ok;
  IOWorkLoop *workLoop;
  IOInterruptEventSource *interruptSource;

  void interruptOccurred (OSObject *obj, IOInterruptEventSource *src,
			  int count);
 public:
  int irq;
  IOPCIDevice *pciNub;
  gm_instance_state_t *is;

  /* override methods */
  virtual bool start (IOService *provider);
  virtual void stop (IOService *provider);
  virtual void free ();
  virtual bool createWorkLoop ();
  virtual IOWorkLoop *getWorkLoop ();

};

static IOMbufNaturalMemoryCursor *gm_macos_mbufCursor;
static gm_instance_state_t *gm_instances[GM_ARCH_MAX_INSTANCE];
static SInt32 gm_macos_num_instance = 0;
static int gm_macos_major = -1;
static long gm_macos_max_user_locked_pages;
static SInt32 gm_macos_use_count = 0;
IOLock *gm_macos_open_lock = 0;
static int gm_macos_order_start (int);
static void gm_macos_start_next_unit (void);
static pid_t gm_macos_proc_table[GM_ARCH_MAX_MINOR + 1];

#define GET_MINOR(minor_id) 	((minor_id) - (GM_ARCH_MAX_INSTANCE*2 - gm_macos_num_instance*2))
#define EXTERN_MINOR(minor_id) 	((minor_id) + (GM_ARCH_MAX_INSTANCE*2 - gm_macos_num_instance*2))
#define INVALID_MINOR(minor_id) \
 (((minor_id) +(GM_ARCH_MAX_INSTANCE*2 - gm_macos_num_instance*2)) >= GM_ARCH_MAX_MINOR)
void *gm_macos_specials[GM_ARCH_MAX_MINOR];
#define MS_PER_SEC 1000

#define super IOService 
OSDefineMetaClassAndStructors (gm, IOService);

static int gm_macos_port_reaper_exited = 0;
static int gm_macos_port_reaper_should_exit = 0;
int gm_page_lock_failed = 0;

/*
 * This thread looks for processes which have exited without
 * deregistering their memory & deregisters it for them.  
 * 
 * At exit, OS-X removes the process from the process table, so
 * pfind() won't find it.  Then it attempts to deallocate all memory &
 * close all file descriptors.  Since vm_deallocate() will block when
 * memory its trying to free is wired, there's a deadlock.  Luckily,
 * we can detect this scenario by noticing that we've got a process
 * which isn't in the process table, but which has an open port. If
 * this ever happens, we close the port on its behalf.
 * vm_deallocate() then completes, and the process exits as normal.
 */

void gm_macos_port_reaper (void *argument)
{
  int port;
  pid_t pid;
  struct proc *p;
  dev_t fake_dev;

  while (!gm_macos_port_reaper_should_exit)
    {
      /* sleep for 5 seconds */
      IOSleep (5 * MS_PER_SEC);

      /* grab the open lock so we don't race with an opener or closer
       * while we're manipulating the gm_macos_proc_table[] (and calling
       * close 
       */
      IOLockLock (gm_macos_open_lock);
      for (port = 0; port < GM_ARCH_MAX_MINOR; port++)
	{
	  if (gm_macos_proc_table[port] != 0)
	    {
	      pid = gm_macos_proc_table[port];
	      /*
	       * if we cannot find the process in the process table,
	       * that means that it is exiting & may be blocked trying
	       * to free registered memory
	       */
	      if ((p = pfind (pid)) == NULL)
		{
		  fake_dev = makedev (gm_macos_major, port);
		  gm_macos_close (fake_dev, 0, 0, NULL);
		  gm_macos_proc_table[port] = 0;
		}
	    }
	}
      IOLockUnlock (gm_macos_open_lock);
    }
  gm_macos_port_reaper_exited = 1;
  IOExitThread ();
}

/*
 * Similar to above.  This is called at module unload time.  It simply
 * walks the list of open ports & closes them all down & kill'ing the
 * process 
 */

void
gm_macos_killall (void)
{
  int port;
  pid_t pid;
  struct proc *p;
  dev_t fake_dev;

  IOLockLock (gm_macos_open_lock);
  for (port = 0; port < GM_ARCH_MAX_MINOR; port++)
    {
      if (gm_macos_proc_table[port] != 0)
	{
	  pid = gm_macos_proc_table[port];
	  if ((p = pfind (pid)) != NULL)
	    {
	      GM_NOTE (("killing pid %d so module can unload\n", pid));
	      psignal (p, SIGKILL);
	    }
	  fake_dev = makedev (gm_macos_major, port);
	  gm_macos_close (fake_dev, 0, 0, NULL);
	  gm_macos_proc_table[port] = 0;
	}
    }
  IOLockUnlock (gm_macos_open_lock);
}

/*
 *  Ordering code and related cruft.  This is used to ensure that the
 *  same board gets mapped to the same unit number each time the
 *  driver is loaded.
 */

struct gm_macos_order
{
  thread_t thread;
  int position;
};

/*
 * Can you beleive they don't even let you use qsort??? 
 * This is a simple bubble sort..
 */

static void
gm_macos_sort_order (int size, struct gm_macos_order *array)
{
  int i, j;
  struct gm_macos_order temp;

  for (i = size - 1; i >= 0; --i)
    {
      for (j = 0; j < i; j++)
	{
	  if (array[j].position > array[j + 1].position)
	    {
	      bcopy (&array[j], &temp, sizeof (temp));
	      bcopy (&array[j + 1], &array[j], sizeof (temp));
	      bcopy (&temp, &array[j + 1], sizeof (temp));
	    }
	}
    }
}

static struct gm_macos_order gm_macos_ordering[GM_ARCH_MAX_INSTANCE];
static int gm_macos_next_unit = 0;
static SInt32 gm_macos_num_units = 0;

/*
 * IOKit calls ::start() non-deterministically.  For each device, a
 * thread is created to probe/attach it.  These threads can start()
 * our driver in any order, resulting in different boards ending up as
 * unit X each time.  We'd like to have the same board be unit X each time.
 * So we pass in our position on the pci bus ((bus * 1000) + slot) and
 * use this to sort ourselves.  The device with the lowest number goes
 * first.  What a pain!
 */


static int
gm_macos_order_start (int position)
{
  int my_unit;
  struct gm_macos_order my_ordering;

  my_unit = OSIncrementAtomic (&gm_macos_num_units);

  if (my_unit >= GM_ARCH_MAX_INSTANCE)
    {
      return -1;
    }

  my_ordering.thread = current_thread ();
  my_ordering.position = position;
  bcopy (&my_ordering, &gm_macos_ordering[my_unit], sizeof (my_ordering));

  /* the first thread to get here will sort things out.
   * wait for it to do that */
  if (my_unit != 0)
    {
      assert_wait ((event_t) my_ordering.thread, THREAD_ABORTSAFE);
      (void) thread_block (NULL);
    }
  else
    {
      /* let other threads enter themselves into the array 
       * 1 second should be long enough to wait..
       */
      IOSleep (1 * MS_PER_SEC);
      /* sort array */
      gm_macos_sort_order (gm_macos_num_units, gm_macos_ordering);

      /* if we're not the first unit, then wake the first unit up & 
         wait our turn */
      if (position != gm_macos_ordering[0].position)
	{
	  assert_wait ((event_t) my_ordering.thread, THREAD_ABORTSAFE);
	  gm_macos_start_next_unit ();
	  (void) thread_block (NULL);
	}
    }
  gm_macos_next_unit += 1;
  return (gm_macos_next_unit - 1);
}


/*
 * wake up the next thread to attach.
 */
static void
gm_macos_start_next_unit ()
{
  /* wake up the next thread in line */
  if (gm_macos_next_unit < gm_macos_num_units)
    {
      wakeup (gm_macos_ordering[gm_macos_next_unit].thread);
    }
  else /* or wake up the first thread so it can release the
	  gm_macos_open_lock and allow access to all the boards */
    {
      if (gm_macos_num_units > 0)
	{
	  wakeup (gm_macos_ordering[0].thread);
	}
    }
}

#define GM_DEBUG_ATTACH 0

/* 
 * This is the IOKit "attach" routine.  Its called at module load time
 * and sets up the driver's class.
 */

bool gm::start (IOService *provider)
{
  IOWorkLoop *myWorkLoop;
  IOMemoryMap *map;
  int unit = gm_macos_num_instance;
  int has_lock;
  int instance_number;
  gm_status_t status;

  /* 
   * cache our instance number;
   */

  instance_number = OSIncrementAtomic (&gm_macos_num_instance);

  if (instance_number >= GM_ARCH_MAX_INSTANCE)
    {
      GM_WARN (("Too many boards too attach all of them"
		"only %d boards configured, increase GM_ARCH_MAX_INSTANCE",
		GM_ARCH_MAX_INSTANCE));
      goto abort_with_instance_number;
    }
  /* Start our superclass first. */
  if (super::start (provider) == false)
    {
      goto abort_with_instance_number;
    }
  /* Cache our provider to an instance variable. */
  pciNub = OSDynamicCast (IOPCIDevice, provider);
  if (pciNub == 0)
    {
      goto abort_with_super;
    }
  /* Retain provider, released in free(). */
  pciNub->retain ();

  /* Open our provider. */

  if (pciNub->open (this) == false)
    {
      goto abort_with_provider;
    }

  
  /* Request power.  If the machine was asleep before our driver was
     loaded, PCI config space may not be valid */
  if (pciNub->requestPowerDomainState(kIOPMPowerOn,
				      (IOPowerConnection *)getParentEntry (gIOPowerPlane),
				      IOPMLowestState) != IOPMNoErr)
    {
      GM_WARN(("Failed to power-on board\n"));
      goto abort_with_provider_open;
    }
				      
  /* now we need to order our attach.  IOKit calls 
     ::start() in a non-deterministic matter, so its up to us
  */

  unit = gm_macos_order_start ((pciNub->getDeviceNumber ()) +
			       (pciNub->getBusNumber () * 1000));
  if (unit < 0)
    {
      goto abort_with_provider_open;
    }

  GM_INFO (("GM: Board at bus %d slot %d now attaching\n",
	    pciNub->getBusNumber (), pciNub->getDeviceNumber ()));

  /*
   * Create a "workloop" -- this is basically an interrupt thread
   * which is awakended for every hardware interrupt. 
   */

  myWorkLoop = getWorkLoop ();
  if (myWorkLoop == NULL)
    {
      GM_WARN (("GM:  Couldn't allocate workloop\n"));
      goto abort_with_provider_open;
    }

  interruptSource = IOInterruptEventSource::
    interruptEventSource ((OSObject *) this,
			  (IOInterruptEventAction) & gm::interruptOccurred,
			  (IOService *) provider, (int) 0);

  if (interruptSource == NULL)
    {
      GM_WARN (("GM:  Couldn't allocate Interrupt handler\n"));
      goto abort_with_workloop;

    }

  if (myWorkLoop->addEventSource (interruptSource) != kIOReturnSuccess)
    {
      GM_WARN (("Couldn't add Interrupt event source\n"));
      goto abort_with_interrupt_source;
    }

  /* enable interrupts from the OS's point of view, GM still has them
     disabled */
  interruptSource->enable ();
  getWorkLoop ()->enableAllInterrupts ();

  /* map memory ( unmap ..magically.. taken care of by OS-X ) */
  map = provider->mapDeviceMemoryWithIndex (0);
  if (map == 0)
    {
      GM_WARN (("Couldn't map PCI memory\n"));
      goto abort_with_workloop_interrupt_source;
    }

  has_lock = 0;
  if (unit == 0)
    {
      gm_macos_open_lock = IOLockAlloc ();
      if (gm_macos_open_lock == NULL)
	{
	  GM_NOTE (("Failed to allocate kernel lock\n"));
	  goto abort_with_workloop_interrupt_source;
	}

      /*
       * Grab the open lock prior to creating the device nodes..  
       * If somebody tries to open the device before GM is up, then
       * we'll panic the machine.
       */
      IOLockLock (gm_macos_open_lock);
      has_lock = 1;
      status = gm_init ();
      GM_INFO (("GM: driver version %s\n", _gm_build_id));
      if (status != GM_SUCCESS)
	{
	  GM_WARN (("could not init gm\n"));
	  goto abort_with_provider_open;
	}

      /*
       * Be very conservative: with the amount of ram we can lock.
       * At least in 10.1, the OS-X VM system seems rather fragile
       * and cannot gracefully handle more than 1/2 physical memory
       * being wired.
       */
      gm_macos_max_user_locked_pages = atop (mem_size) / 2;

      /* The IOMMU on G5s is limited to mapping 2GB of RAM
	 To be fair, we limit GM to 1.75GB.  This isn't a big problem
	 on G4's, since they can't take more than 3GB anyway.
      */
#define IOMMU_LIMIT 1024*1024*1792

      if (gm_macos_max_user_locked_pages > atop (IOMMU_LIMIT))
	gm_macos_max_user_locked_pages = atop (IOMMU_LIMIT);

      GM_INFO (("GM: gm_register_memory will be able to lock "
		"%d pages (%d MBytes)\n", gm_macos_max_user_locked_pages,
		ptoa (gm_macos_max_user_locked_pages) / (1024 * 1024)));

      /*
       * Create minor devices.  /dev/gm%d is the clone device for normal users,
       * and /dev/gmp%d is the clone device for privileged users.
       */
      gm_macos_major = gm_macos_special_create ();
      if (gm_macos_major == -1)
	{
	  goto abort_with_gm;
	}
    }

  MALLOC (is, gm_instance_state_t *, sizeof *is, M_DEVBUF, M_NOWAIT);
  bzero (is, sizeof (*is));
  if (is == 0)
    {
      GM_WARN (("couldn't allocate is\n"));
      goto abort_with_gm;
    }

  gm_instances[unit] = is;
  is->arch.io_kva = (caddr_t) map->getVirtualAddress ();

  /* save our class so that we can call C++ IOKit functions from C */
  is->arch.cpp_class = (void *) this;

  if (gm_arch_get_page_len (&GM_PAGE_LEN) != GM_SUCCESS)
    {
      GM_WARN (("cannot determine page size\n"));
      goto abort_with_is;
    }

  if (gm_instance_init (is, unit, GM_MYRINET_BUS_PCI) != GM_SUCCESS)
    {
      GM_WARN (("gm_instance_init failed\n"));
      goto abort_with_is;
    }

  /* enable interrupts.  HACK: This must be done only after it is safe
     to call the interrupt handler for this device, which we installed
     earlier. */

  GM_PRINT (GM_DEBUG_ATTACH,
	    ("Enabling interrupts for instance %d.\n", is->id));
  if (gm_initialize_interrupts (is) != GM_SUCCESS)
    {
      GM_NOTE (("Cannot enable interrupts.\n"));
      goto abort_with_initialized_instance;
    }

  /* Determine minor node number to use for /dev/gm? . */

  GM_PRINT (GM_DEBUG_ATTACH, ("Allocating minor number.\n"));
  if (gm_minor_alloc (is, &is->clone_minor) != GM_SUCCESS)
    {
      GM_NOTE (("Failed to allocate minor number.\n"));
      goto abort_with_enabled_interrupts;
    }
  GM_PRINT (GM_DEBUG_ATTACH,
	    ("using minor number %ld.\n", (long) is->clone_minor));
  gm_minor_get_port_state (is->clone_minor)->instance = is;
  gm_minor_get_port_state (is->clone_minor)->privileged = 0;

  /* Determine minor node number to use for /dev/gmp? . */
  GM_PRINT (GM_DEBUG_ATTACH, ("Allocating minor number.\n"));
  if (gm_minor_alloc (is, &is->privileged_clone_minor) != GM_SUCCESS)
    {
      GM_NOTE (("Failed to allocate minor number.\n"));
      goto abort_with_minor;
    }


  GM_PRINT (GM_DEBUG_ATTACH, ("using minor number %ld.\n",
			      (long) is->privileged_clone_minor));
  gm_minor_get_port_state (is->privileged_clone_minor)->instance = is;
  gm_minor_get_port_state (is->privileged_clone_minor)->privileged = 1;

  if (unit == 0)
    {
      IOCreateThread (gm_macos_port_reaper, NULL);
      gm_macos_mbufCursor = 
	IOMbufNaturalMemoryCursor::withSpecification( GM_ETHERNET_MTU, GM_MAX_ETHERNET_SG_CNT );
    }

  if ((gm_macos_mbufCursor == NULL) || gm_macos_ether_attach (is))
    {
      GM_NOTE (("GM: Unit %d IP interface attach FAILED\n", unit));
    }
  else
    {
      GM_INFO (("GM: Unit %d IP interface attach ok\n", unit));
    }
  GM_INFO (("GM: Unit %d Loaded ok\n", unit));

  loaded_ok = 1;
  gm_macos_start_next_unit ();
  if (unit == 0)
    {
      /* wait for all boards to finish or fail attaching before we
	 allow access to any board */
      if (gm_macos_num_units > 1)
	{
	  assert_wait ((event_t)current_thread (), THREAD_ABORTSAFE);
	  (void) thread_block (NULL);
	}
      IOLockUnlock (gm_macos_open_lock);
    }	

  return (true);

 abort_with_privileged_clone_minor:
  gm_minor_free (is->privileged_clone_minor);

 abort_with_minor:
  gm_minor_free (is->clone_minor);

 abort_with_enabled_interrupts:
  gm_disable_interrupts (is);

 abort_with_initialized_instance:
  gm_instance_finalize (is);

 abort_with_is:
  FREE (is, M_DEVBUF);
  gm_instances[unit] = 0;

 abort_with_gm:
  if (gm_macos_num_instance == 1)
    {
      gm_finalize ();
      if (gm_macos_major != -1)
	{
	  gm_macos_special_destroy ();
	}
    }

 abort_with_workloop_interrupt_source:
  getWorkLoop ()->removeEventSource (interruptSource);

 abort_with_interrupt_source:
  interruptSource->release ();

 abort_with_workloop:
  myWorkLoop->release ();

 abort_with_provider_open:
  pciNub->close (this);

  if (has_lock)
    IOLockUnlock (gm_macos_open_lock);

 abort_with_provider:
  pciNub->release ();
  pciNub = 0;

 abort_with_super:
  super::stop (provider);

 abort_with_instance_number:
  OSDecrementAtomic (&gm_macos_num_instance);
  GM_INFO (("NOT Loaded OK!\n"));
  loaded_ok = 0;
  gm_macos_start_next_unit ();
  return false;
}


/*
 * Function: stop <IOService>
 *
 *Stop all activities and prepare for termination.
 */

void
gm::stop (IOService *provider)
{
  super::stop (provider);
}

/*
 * This is the module unload routine.  We cannot veto unload; the best
 * we can do is to sleep..
 */

void
gm::free ()
{
  int unit, instance_number, reaper_waiting_time;

  if (loaded_ok == 0)
    {
      /* 
       *  if we have not been loaded OK, all teardown is done by the
       *  start function 
       */
      super::free ();
      return;
    }

  /* Kill any procs that have gm ports open prior to exit */

  if (gm_macos_use_count != 0)
    {
      GM_NOTE (("Waiting for all /dev/gm* special files to be closed:"
		"%d outstanding reference(s)\n", gm_macos_use_count));
      gm_macos_killall ();
      IOSleep (1 * 1000);
    }
  gm_macos_port_reaper_should_exit = 1;
  reaper_waiting_time = 0;
  while (!gm_macos_port_reaper_exited)
    {
      reaper_waiting_time++;
      if ((reaper_waiting_time % 20) == 0)
	{
	  GM_WARN(("GM reaper thread hasn't exited after %d secs.\n",
		   reaper_waiting_time ));
	}
      IOSleep (1 * MS_PER_SEC);
    }
  unit = is->id;
  /* detach the ethernet interface -- this may block */
  gm_macos_ether_detach (is);
  gm_minor_free (is->privileged_clone_minor);
  gm_minor_free (is->clone_minor);
  gm_disable_interrupts (is);
  gm_instance_finalize (is);
  FREE (is, M_DEVBUF);
  gm_instances[unit] = 0;
  instance_number = OSDecrementAtomic (&gm_macos_num_instance);
  if (instance_number == 1)
    {
      IOLockFree (gm_macos_open_lock);
      gm_finalize ();
      if (gm_macos_major != -1)
	{
	  gm_macos_special_destroy ();
	}
      if (gm_macos_mbufCursor)
	{
	  gm_macos_mbufCursor->release();
	}
    }
  getWorkLoop ()->removeEventSource (interruptSource);
  interruptSource->release ();
  workLoop->release ();
  pciNub->close (this);
  pciNub->release ();
  pciNub = 0;
  super::free ();		/* pass free to our superclass */
  GM_INFO (("GM: Unit %d unloaded!\n", unit));
}

void
gm::interruptOccurred (OSObject *obj, IOInterruptEventSource *src, int count)
{
  if (gm_interrupting (is) == GM_ARCH_INTR_CLAIMED)
    {
      gm_handle_claimed_interrupt (is);
    }
}

bool
gm::createWorkLoop ()
{
  workLoop = IOWorkLoop::workLoop ();

  return (workLoop != 0);
}

/* Override IOService::getWorkLoop() method to return our workloop. */

IOWorkLoop *
gm::getWorkLoop ()
{

  if (!workLoop)
    {
      workLoop = IOWorkLoop::workLoop ();
    }

  return workLoop;
}

#define GM_DEBUG_KMALLOC 0

/************************************************************************
 * GM OS abstraction functions
 ************************************************************************/

/*
  We need to use our own malloc because the OSX malloc allocates from
  a tiny (16MB) kalloc_map submap for large allocations.  To get
  around this, we must call kmem_alloc() directly so that we can
  specify that we want our allocation to come from the larger
  kernel_map.  For smaller allocations, we can call the Mach kalloc
  directly (for small allocations, it uses the zone allocator, and
  that grabs memory from the kernel_map, not the kmem_alloc submap).

  All of this means that when we free memory, we must know what the
  size of the allocation was.

*/   

struct gm_macos_mem_hint {
  unsigned long len;
  char buffer[0];
};

void *
__gm_arch_kernel_malloc (unsigned long len, int flags)
{
  unsigned long alloc_len;
  struct gm_macos_mem_hint *hint;
  int status;
  void *ret;

  alloc_len = len + sizeof(*hint);
  if (alloc_len < 8192)
    {
      hint = (struct gm_macos_mem_hint *)kalloc(alloc_len);
    }
  else
    {
      vm_offset_t addr;
      status = kmem_alloc(kernel_map, &addr, alloc_len);
      if (status == KERN_SUCCESS)
	{
	  hint = (struct gm_macos_mem_hint *)addr;
	}
      else
	{
	  hint = NULL;
	}
    }
  if (hint == NULL)
    return NULL;
  
  hint->len = alloc_len;
  hint++;
  return hint;
}

void
__gm_arch_kernel_free (void *ptr)
{
  struct gm_macos_mem_hint *hint;

  hint = (struct gm_macos_mem_hint *)ptr;
  hint--;
  if (hint->len < 8192)
    {
      kfree ((vm_offset_t)hint, hint->len);
    }
  else
    {
      kmem_free (kernel_map, (vm_offset_t)hint, hint->len);
    }
}

/****************
 * synchronization
 ****************/

/* The following functions must implement both mutex and sleep/wake
   functionality.  The mutex and sleep/wake functionality should be
   independent.  (That is, we really should have implemented this as
   independent gm_arch_mutex_t and gm_arch_sleep_t structures, but we
   did not.) */

/* initialize the sync structure */

void
gm_arch_sync_init (gm_arch_sync_t *s, gm_instance_state_t *is)
{
  s->lock = IOLockAlloc ();
  s->wake_lock = IOLockAlloc ();
  s->wake_cnt = 0;
}

/* acquire mutually exclusive access */

void
gm_arch_mutex_enter (gm_arch_sync_t *s)
{
  IOLockLock (s->lock);
}

/* release mutually exclusive access */

void
gm_arch_mutex_exit (gm_arch_sync_t *s)
{
  IOLockUnlock (s->lock);
}

/* free any associated storage allocated in gm_arch_sync_init() */

void
gm_arch_sync_destroy (gm_arch_sync_t *s)
{
  IOLockFree (s->lock);
  IOLockFree (s->wake_lock);
  s->lock = 0;
  s->wake_lock = 0;
}

/* Reset the sync variable. */

void
gm_arch_sync_reset (gm_arch_sync_t *s)
{
  IOLockFree (s->lock);
  IOLockFree (s->wake_lock);
  s->wake_cnt = 0;
  s->lock = IOLockAlloc ();
  s->wake_lock = IOLockAlloc ();
}

/* Wake a process sleeping on the sync, if any.  It is acceptable to
   wake more than one process sleeping on the sync, but performance will
   be adversely effected. */

void
gm_arch_wake (gm_arch_sync_t *s)
{
  OSIncrementAtomic (&s->wake_cnt);
  thread_wakeup_one (s);
}

/* sleep until wake or signal */

gm_arch_sleep_status_t
gm_arch_signal_sleep (gm_arch_sync_t *s)
{
  int result;

  while (s->wake_cnt <= 0)
    {
      assert_wait (s, THREAD_ABORTSAFE);
      if (s->wake_cnt > 0)
	{
	  clear_wait (current_thread (), 0);
	  goto woke;
	}
      result = thread_block (NULL);
      if (result != THREAD_AWAKENED)
	{
	  return GM_SLEEP_INTERRUPTED;
	}
    }
 woke:
  OSDecrementAtomic (&s->wake_cnt);
  return GM_SLEEP_WOKE;

}

/* sleep until wake or timeout */

gm_arch_sleep_status_t
gm_arch_timed_sleep (gm_arch_sync_t *s, int seconds)
{
  int result;

  while (s->wake_cnt <= 0)
    {
      assert_wait (s, THREAD_UNINT);
      if (s->wake_cnt > 0)
	{
	  clear_wait (current_thread (), 0);
	  goto woke;
	}
      thread_set_timer (hz * seconds, NSEC_PER_SEC / hz);
      result = thread_block (NULL);
      thread_cancel_timer ();
      if (result != THREAD_AWAKENED)
	{
	  return GM_SLEEP_TIMED_OUT;
	}
    }
 woke:
  OSDecrementAtomic (&s->wake_cnt);
  return GM_SLEEP_WOKE;
}


/****************
 * DMA regions
 ****************/

/* The following functions allocate regions of DMAable memory.  It must
   be possible to map the memory in the region into user address space. */

/* Allocate a page aligned region of DMAable memory.  LEN will a
   multiple of the page length.  FLAGS will be some combination of the
   bits GM_ARCH_DMA_READ, GM_ARCH_DMA_WRITE, GM_ARCH_DMA_STREAMING,
   and GM_ARCH_DMA_CONSISTENT.  These flags can be used to improve
   performance, or can be ignored where no behaviourly conflict
   exists.

   If REGISTER_PAGE_FUNC is non-null, REGISTER_PAGE_FUNC (USER_ARG,
   dma_addr, page_num) should be called for each allocated DMA page,
   to allow GM to register the individual pages in its page table, if
   necessary.

   For compatibility with crazy OS's like Solaris that only allow the
   DMA addresses for the individual pages to be obtained by scanning
   the region from start to end, this function does not return the
   address of the allocated region, but merely stores state in the
   region to allow the region to be so scanned later, with
   gm_arch_dma_region_dma_addr() and
   gm_arch_dma_region_dma_addr_advance().  */

static gm_dp_t
region_dma_addr (gm_arch_dma_region_t *r, char *addr)
{
  IOPhysicalAddress pa;
  IOByteCount offset;
  IOMemoryDescriptor *md;
  IOByteCount len;
  
  offset = (IOByteCount)addr - (IOByteCount)r->md_addr;
  md = (IOMemoryDescriptor *) r->md;
  len = GM_PAGE_LEN;
  pa = md->getPhysicalSegment (offset, &len);

  return (gm_dp_t)pa;
}

gm_status_t
gm_arch_dma_region_alloc (gm_instance_state_t *is,
			  gm_arch_dma_region_t *r,
			  gm_size_t len, gm_u32_t flags,
			  gm_register_page_function_t reg, void *user_arg)
{

  char *addr;
  int pages, page_num;
  IOMemoryDescriptor *md;

  pages = round_page (len) / GM_PAGE_LEN;
  gm_always_assert (pages >= 1);
  /* 
   * grab the memory and remember how we allocated it 
   */
  r->is = is;
  r->flags = flags;
  r->len = round_page (len);

  if (flags & GM_ARCH_DMA_CONTIGUOUS) 
    {
      /*
       * use contigmalloc
       */
      r->type = GM_ARCH_CONTIGMALLOC_REGION;
      r->alloc_addr = IOMallocContiguous (r->len, PAGE_SIZE, 0);
      r->addr = r->alloc_addr;
    }
  else
    {
      /*
       * use normal malloc
       */
      r->type = GM_ARCH_MALLOC_REGION;
      r->alloc_addr = gm_malloc (r->len + GM_PAGE_LEN);
      r->addr = (void *) round_page ((vm_offset_t) r->alloc_addr);

    }
  if (r->alloc_addr == 0)
    {
      r->type = 0;
      return GM_FAILURE;
    }
  r->addr2 = (char *) r->addr;
  r->md_addr = r->addr;
  md = IOMemoryDescriptor::withAddress ((vm_address_t) r->addr,
					(IOByteCount) r->len,
					kIODirectionOutIn, kernel_task);
  if (md == NULL)
    {
      goto abort_with_alloc;
    }
  
  if (md->prepare (kIODirectionOutIn) != kIOReturnSuccess)
    {
      goto abort_with_md;
    }
  r->md = (void *)md;
  
  if (reg)
    {
      addr = (char *) r->addr;
      for (page_num = 0; page_num < pages; page_num++, addr += GM_PAGE_LEN)
	{
	  gm_dp_t bus_addr = region_dma_addr (r, addr);
	  reg (user_arg, bus_addr, page_num);
	}
    }

  return (GM_SUCCESS);

 abort_with_md:
  md->release ();

 abort_with_alloc:
  switch (r->type) 
    {
    case GM_ARCH_MALLOC_REGION:
      gm_free(r->alloc_addr);
      break;

    case GM_ARCH_CONTIGMALLOC_REGION:
      IOFreeContiguous (r->alloc_addr, r->len);
      break;
      
    default:
      GM_PANIC (("gm_arch_dma_region_alloc:bad type\n"));
      break;
    }

  return (GM_FAILURE);
}

/* free the region and associated resources */

void
gm_arch_dma_region_free (gm_arch_dma_region_t *r)
{
  IOMemoryDescriptor *md;

  md = (IOMemoryDescriptor *)r->md;
  md->complete (kIODirectionOutIn);	/* unwire */
  md->release ();		/* free the decriptor */


  switch (r->type) 
    {
    case GM_ARCH_MALLOC_REGION:
      gm_free(r->alloc_addr);
      break;

    case GM_ARCH_CONTIGMALLOC_REGION:
      IOFreeContiguous (r->alloc_addr, r->len);
      break;
      
    default:
      GM_PANIC (("gm_arch_dma_region_free:bad type\n"));
      break;
    }
}

/* return the DMA address for the region at the current position, originally
   set to the beginning of the region. */

gm_dp_t
gm_arch_dma_region_dma_addr (gm_arch_dma_region_t *r)
{
  return (region_dma_addr (r, (caddr_t) r->addr));
}

/* return the DMA address for the region at the current position,
   originally set to the beginning of the region, and advance the
   position by a GM_PAGE_LEN. */

gm_dp_t
gm_arch_dma_region_dma_addr_advance (gm_arch_dma_region_t *r)
{
  char *previous = r->addr2;
  r->addr2 += PAGE_SIZE;
  return region_dma_addr (r, previous);
}

/* return the kernel address for the region at the current position, originally
   set to the beginning of the region. */

void *
gm_arch_dma_region_kernel_addr (gm_arch_dma_region_t *r)
{
  return r->addr;
}

/* Synchronize the CPU and DMA views of the region */

gm_status_t
gm_arch_dma_region_sync (gm_arch_dma_region_t *r, int command)
{
  return GM_SUCCESS;
}

/****************
 * memory mapping
 ****************/

/* Memory map the part of the board IS starting at OFFSET and of length LEN
   into kernel memory, and return the kernel address at KADDR. */

gm_status_t
gm_arch_map_io_space (gm_instance_state_t *is, gm_u32_t offset,
		      gm_u32_t len, void **kaddr)
{
  *kaddr = (void *) (is->arch.io_kva + offset);
  return GM_SUCCESS;
}

void
gm_arch_unmap_io_space (gm_instance_state_t *is, gm_u32_t offset,
			gm_u32_t len, void **kaddr)
{
}

/* memory mapping (kernel memory into user space) */

gm_status_t
gm_arch_mmap_contiguous_segment (gm_port_state_t *ps,
				 void *kaddr,
				 unsigned long blockSize, gm_up_t *vaddr)
{
  return GM_FAILURE;
}

void
gm_arch_munmap_contiguous_segments (gm_port_state_t *ps)
{
}

/****************
 * misc
 ****************/

/* copy the hostname to PTR, failing if its length is more than LEN. */

gm_status_t
gm_arch_gethostname (char *ptr, int len)
{
  bzero (ptr, len);
  if (len > MAXHOSTNAMELEN)
    {
      len = MAXHOSTNAMELEN;
    }
  strncpy (ptr, hostname, len - 1);
  ptr[len - 1] = 0;
  return GM_SUCCESS;
}

/* spin for the indicated number of microseconds. */

void
gm_arch_spin (gm_instance_state_t *is, gm_u32_t usecs)
{
  IODelay (usecs);
}

/* return the page length */

gm_status_t
gm_arch_get_page_len (unsigned long *page_len)
{
  static vm_size_t vm_size = 0;
  kern_return_t status;
  host_t host;

  host = host_self ();

  if (vm_size == 0)
    {
      status = host_page_size (host, &vm_size);
      if (status != KERN_SUCCESS)
	return GM_FAILURE;
    }
  *page_len = (unsigned long) vm_size;
  return GM_SUCCESS;
}

/****************
 * port state initialization/finalization
 ****************/

/* ??? */

gm_status_t
gm_arch_port_state_init (gm_port_state_t *ps)
{
  return GM_SUCCESS;
}

/* Perform any necessary initialization just after the port has been opened. */

gm_status_t
gm_arch_port_state_open (gm_port_state_t *ps)
{
  return GM_SUCCESS;
}

/* Perform any necessary cleanup just before the port is closed */

void
gm_arch_port_state_close (gm_port_state_t *ps)
{
}

/*
 * Since OS-X is broken & doesn't supprt device mmap, when we close a
 * port, we must unmap any kernel or device memory that was mapped
 * into user space.
 */

void
gm_arch_port_state_fini (gm_port_state_t *ps)
{
  gm_macosx_md_list_entry_t *entry;
  IOMemoryDescriptor *md;
  IOMemoryMap *map;

  while (ps->arch.entry)
    {
      entry = ps->arch.entry;
      ps->arch.entry = entry->next;
      map = (IOMemoryMap *) entry->map;
      md = (IOMemoryDescriptor *) entry->md;
      map->unmap ();
      map->release ();
      md->release ();
      gm_free (entry);
    }
}


/****************************************************************
 * PCI
 ****************************************************************/

/* These functions need be supported only if PCI support is desired. */

#ifdef GM_SUPPORT_PCI
/* read and write the PCI configuration space */

gm_status_t
gm_arch_read_pci_config_32 (gm_instance_state_t *is, gm_offset_t offset,
			    gm_u32_t *value)
{
  gm *cpp_class;

  cpp_class = (gm *) is->arch.cpp_class;
  *value = cpp_class->pciNub->configRead32 (offset);
  return GM_SUCCESS;
}

gm_status_t
gm_arch_write_pci_config_32 (gm_instance_state_t *is, gm_offset_t offset,
			     gm_u32_t value)
{
  gm *cpp_class;

  cpp_class = (gm *) is->arch.cpp_class;
  cpp_class->pciNub->configWrite32 (offset, value);
  return GM_SUCCESS;
}

gm_status_t
gm_arch_read_pci_config_16 (gm_instance_state_t *is, gm_offset_t offset,
			    gm_u16_t *value)
{
  gm *cpp_class;

  cpp_class = (gm *) is->arch.cpp_class;
  *value = cpp_class->pciNub->configRead16 (offset);
  return GM_SUCCESS;
}

gm_status_t
gm_arch_write_pci_config_16 (gm_instance_state_t *is, gm_offset_t offset,
			     gm_u16_t value)
{
  gm *cpp_class;

  cpp_class = (gm *) is->arch.cpp_class;
  cpp_class->pciNub->configWrite16 (offset, value);
  return GM_SUCCESS;
}

gm_status_t
gm_arch_read_pci_config_8 (gm_instance_state_t *is, gm_offset_t offset,
			   gm_u8_t *value)
{
  gm *cpp_class;

  cpp_class = (gm *) is->arch.cpp_class;
  *value = cpp_class->pciNub->configRead8 (offset);
  return GM_SUCCESS;
}

gm_status_t
gm_arch_write_pci_config_8 (gm_instance_state_t *is, gm_offset_t offset,
			    gm_u8_t value)
{
  gm *cpp_class;

  cpp_class = (gm *) is->arch.cpp_class;
  cpp_class->pciNub->configWrite8 (offset, value);
  return GM_SUCCESS;
}
#endif /* GM_SUPPORT_PCI */

/****************************************************************
 * memory registration
 ****************************************************************/

/* These functions need be supported only if memory registration support
   is desired.  When implementing drivers that support memory registration,
   leave GM_CAN_REGISTER_MEMORY undefined until the rest of GM is working,
   including all test programs. */

#ifdef GM_CAN_REGISTER_MEMORY
/* lock (a.k.a. "pin", a.k.a. "wire") the indicated user page. */

gm_status_t
gm_arch_lock_user_buffer_page (gm_port_state_t *ps,
			       gm_up_t uaddr,
			       gm_dp_t *dma_addr, gm_arch_page_lock_t *lock)
{
  IOMemoryDescriptor *md;
  IOByteCount len;
  IOReturn ret;
  task_t task;

  if (gm_macos_max_user_locked_pages <= 0)
    {
      gm_page_lock_failed++;
      return GM_OUT_OF_MEMORY;
    }
  task = current_task ();
  if (task == NULL)
    {
      goto abort_with_nothing;
    }

  /*
   * Create an "IOMemoryDescriptor" to specify the
   * user's page.
   */

  md = IOMemoryDescriptor::withAddress ((vm_address_t) uaddr,
					(IOByteCount) GM_PAGE_LEN,
					kIODirectionOutIn, task);
  if (md == NULL)
    {
      return GM_MEMORY_FAULT;
    }
  /*  
   * This involves paging in the memory, if necessary, and wiring it
   * down for the duration of the transfer.  The memory can be unwired
   * using md->complete().  This is totally atomic from the point of
   * view of our thread; no locks need or should be taken.
   */
  if (md->prepare (kIODirectionOutIn) != kIOReturnSuccess)
    {
      goto abort_with_md;
    }

  len = GM_PAGE_LEN;
  *dma_addr = md->getPhysicalSegment (0, &len);
  if (len != GM_PAGE_LEN)
    {
      GM_WARN (("md->getPhysicalSegment ret len %d, lock failed\n", len));
      goto abort_with_md_wired;
    }

  OSDecrementAtomic (&gm_macos_max_user_locked_pages);
  lock->md = (void *) md;
  return GM_SUCCESS;

 abort_with_md_wired:
  md->complete (kIODirectionOutIn);

 abort_with_md:
  md->release ();

 abort_with_nothing:
  return GM_FAILURE;
}

/* unlock the indicated user page. */

void
gm_arch_unlock_user_buffer_page (gm_arch_page_lock_t *lock)
{
  IOMemoryDescriptor *md;

  md = (IOMemoryDescriptor *) lock->md;
  md->complete (kIODirectionOutIn);	/* unwire */
  md->release ();		/* free the decriptor */
  OSIncrementAtomic (&gm_macos_max_user_locked_pages);
}

/* lock the indicated page of kernel memory.
 * all kernel memory in the kernel map is non-pagable, so this is a
 * nop 
 */

gm_status_t
gm_arch_lock_kernel_buffer_page (gm_instance_state_t *is,
				 void *kaddr,
				 gm_dp_t *dma_addr,
				 gm_arch_page_lock_t *lock)
{
  return GM_SUCCESS;
}

/* unlock the indicated page of kernel memory. */

void
gm_arch_unlock_kernel_buffer_page (gm_arch_page_lock_t *lock)
{
}
#endif /* GM_CAN_REGISTER_MEMORY */

/****************************************************************
 * ioctl I/O
 ****************************************************************/

/* architecture-specific functions to copy data into and out of the kernel. */

#define GM_DEBUG_IOCTL 0

gm_status_t
gm_arch_copyout (const void *from, void *to, gm_size_t len)
{
  /* TO is a user buffer address corresponding to an ioctl()-like buffer */
  int errno;

  GM_PRINT (GM_DEBUG_IOCTL,
	    ("copyout: from %p to %p len %ld\n", from, to, len));
  errno = copyout ((void *) from, to, len);
  return errno ? GM_FAILURE : GM_SUCCESS;
}

gm_status_t
gm_arch_copyin (const void *from, void *to, gm_size_t len)
{
  /* FROM is a user buffer address corresponding to an ioctl()-like buffer */
  int errno;

  GM_PRINT (GM_DEBUG_IOCTL,
	    ("copyin: from %p to %p len %ld\n", from, to, len));
  errno = copyin ((void *) from, to, len);
  return errno ? GM_FAILURE : GM_SUCCESS;

}


/****************************************************************
 * BSD driver
 ****************************************************************/

static void
gm_macos_insert_proctable (u_int minor)
{
  struct proc *p;

  p = current_proc ();
  gm_macos_proc_table[minor] = p->p_pid;
}

static void
gm_macos_remove_proctable (u_int minor)
{
  gm_macos_proc_table[minor] = 0;
}

#define GM_DEBUG_OPEN 0

static int
gm_macos_open (dev_t dev, int flags, int devtyple, struct proc *p)
{
  gm_status_t status;
  gm_port_state_t *ps;
  gm_instance_state_t *is;
  int unit, error;
  u_int minor_id;

  error = 0;

  IOLockLock (gm_macos_open_lock);

  GM_PRINT (GM_DEBUG_OPEN,
	    ("gm_macos_open: open called gm dev 0x%x minor=%d\n", dev,
	     minor (dev)));

  minor_id = minor (dev);
  if (minor_id >= (GM_ARCH_MAX_INSTANCE *2))
    {
      GM_PRINT (GM_DEBUG_OPEN, ("gm_macos_open: open for 'dynamic' device\n"));
    }
  else
    {
      if (minor_id < 0)
	{
	  GM_PRINT (GM_DEBUG_OPEN,
		    ("gm_macos_open: open for bad minor_id=%d\n", minor_id));
	  error = ENODEV;
	  goto abort_with_lock;
	}

      unit = minor_id / 2;
      is = gm_instances[unit];
      if (!is)
	{
	  GM_PRINT (GM_DEBUG_OPEN,
		    ("gm_macos_open: open for bad instance=%d\n", unit));
	  error = ENODEV;
	  goto abort_with_lock;
	}
    }
  /* success, so commit */

  GM_PRINT (GM_DEBUG_OPEN,
	    ("gm_macos_open: User opened dev 0x%x with minor num %d.\n",
	     dev, minor (dev)));
  OSIncrementAtomic (&gm_macos_use_count);
  gm_macos_insert_proctable (minor (dev));

 abort_with_lock:
  IOLockUnlock (gm_macos_open_lock);
  return error;

}

#define GM_DEBUG_CLOSE 0

static int
gm_macos_close (dev_t dev, int flags, int devtyple, struct proc *p)
{
  gm_port_state_t *ps;
  gm_instance_state_t *is;
  int unit, error, locked;

  error = 0;
  if (p != NULL)
    {
      /* called from ioct, not locked */
      IOLockLock (gm_macos_open_lock);
      locked = 1;
    }
  else
    {
      /* called from reaper thread, already locked */
      locked = 0;
    }
  GM_PRINT (GM_DEBUG_CLOSE, ("gm_macos_close called  dev = 0x%x  minor = %d\n",
			     dev, minor (dev)));

  if (minor (dev) >= (GM_ARCH_MAX_INSTANCE *2))
    {
      GM_PRINT (GM_DEBUG_CLOSE,
		("closing 'dynamic' minor = %d\n", GET_MINOR (minor (dev))));
      ps = gm_minor_get_port_state (GET_MINOR (minor (dev)));
      /* this can happen when exit calls close after the port reaper
         has closed the port.  This is "normal"
      */
      if (!ps)
	{
	  GM_PRINT (GM_DEBUG_CLOSE,
		    (": could not get port_state for minor = %d\n",
		     GET_MINOR (minor (dev))));
	  error = ENODEV;
	  goto abort_with_lock;
	}
      if (ps->opened)
	{
	  gm_port_state_close (ps);
	}
      gm_minor_free (GET_MINOR (minor (dev)));
    }
  else
    {
      unit = minor (dev) / 2;
      is = gm_instances[unit];
      if (!is)
	{
	  GM_WARN (("gm_macos_close: instance[%d] not found\n", unit));
	  error = ENODEV;
	  goto abort_with_lock;
	}
    }

  gm_macos_remove_proctable (minor (dev));

  OSDecrementAtomic (&gm_macos_use_count);

 abort_with_lock:
  if (locked)
    {
      IOLockUnlock (gm_macos_open_lock);
    }
  return error;
}


/*
 * Grr.. OS-X does not support device mmap, so we need to do it
 * ourselves using IOKit primitives.
 */

#define GM_DEBUG_MMAP 0

static gm_status_t
gm_macos_mmap (gm_port_state_t *ps, void *arg)
{
  gm_mmap_info_t i;
  gm_status_t status;
  void *kva;
  task_t task;
  IOMemoryDescriptor *md;
  IOMemoryMap *user_map;
  gm_macosx_md_list_entry_t *entry;

  GM_PRINT (GM_DEBUG_MMAP, ("gm_macos_mmap: called\n"));
  if (gm_arch_copyin (arg, &i, sizeof (gm_mmap_info_t)) != GM_SUCCESS)
    {
      GM_PRINT (GM_DEBUG_MMAP,
		("gm_macos_mmap: copyin failed -  addr %p\n", arg));
      return GM_MEMORY_FAULT;
    }

  GM_PRINT (GM_DEBUG_MMAP, ("gm_macosx_mmap: offset 0x%x\n", i.offset));
  GM_PRINT (GM_DEBUG_MMAP, ("gm_macosx_mmap: len 0x%x\n", i.len));
  GM_PRINT (GM_DEBUG_MMAP, ("gm_macosx_mmap: perms 0x%x\n",
			    i.requested_permissions));
  GM_PRINT (GM_DEBUG_MMAP, ("gm_macosx_mmap: va 0x%x\n", i.va));

  /* validate mapping */

  status = gm_prepare_to_mmap (ps, i.offset, i.len, i.requested_permissions);
  if (status != GM_SUCCESS)
    {
      GM_WARN (("gm_macosx_mmap: gm_prepare_to_mmap failed\n"));
      return status;
    }

  /* map -- this routine returns a kernel virtual address */
  status = gm_mmap (ps, i.offset, &kva);
  if (status != GM_SUCCESS)
    {
      GM_WARN (("gm_macosx_mmap: gm_mmap failed\n"));
      return status;
    }

  entry = (gm_macosx_md_list_entry_t *) gm_malloc (sizeof (*entry));
  if (entry == NULL)
    {
      GM_WARN (("gm_malloc failed for md list entry\n"));
      goto abort_with_nothing;
    }

  task = current_task ();
  if (task == NULL)
    {
      goto abort_with_nothing;
    }

  md = IOMemoryDescriptor::withAddress (kva, i.len, kIODirectionOutIn);
  if (!md)
    {
      GM_WARN (("IOMemoryDescriptor::withAddress failed for %p len  %d\n",
		kva, i.len));
      goto abort_with_entry;
    }

  /* create a region i.len bytes long in this process's address space */
  user_map = md->map (task, (unsigned int) 0, kIOMapAnywhere, 0, i.len);
  if (!user_map)
    {
      GM_WARN (("IOMemoryDescriptor::withAddress failed for %p len  %d\n",
		kva, i.len));
      goto abort_with_md;
    }

  i.va = (void *) user_map->getVirtualAddress ();

  /* Tell the user where to find it */
  if (gm_arch_copyout (&i, arg, sizeof (gm_mmap_info_t)) != GM_SUCCESS)
    {
      GM_PRINT (GM_DEBUG_MMAP,
		("gm_macos_mmap: copyout failed -  addr %p\n", arg));
      goto abort_with_user_map;
    }

  /* save the map entry on the arch-specific port "entry" list.  It
   * will be freed when the port is closed in gm_arch_port_state_fini()
   */
  entry->md = (void *) md;
  entry->map = (void *) user_map;
  entry->next = ps->arch.entry;
  ps->arch.entry = entry;

  return GM_SUCCESS;


 abort_with_user_map:
  user_map->release ();

 abort_with_md:
  md->release ();

 abort_with_entry:
  gm_free (entry);

 abort_with_nothing:
  return GM_FAILURE;



}
static int
gm_macos_localize_status (gm_status_t status)
{
#define CASE(from,to) case from : return to
  switch (status)
    {
      CASE (GM_SUCCESS, 0);
      CASE (GM_INPUT_BUFFER_TOO_SMALL, EFAULT);
      CASE (GM_OUTPUT_BUFFER_TOO_SMALL, EFAULT);
      CASE (GM_TRY_AGAIN, EAGAIN);
      CASE (GM_BUSY, EBUSY);
      CASE (GM_MEMORY_FAULT, EFAULT);
      CASE (GM_INTERRUPTED, EINTR);
      CASE (GM_INVALID_PARAMETER, EINVAL);
      CASE (GM_OUT_OF_MEMORY, ENOMEM);
      CASE (GM_INVALID_COMMAND, EINVAL);
      CASE (GM_PERMISSION_DENIED, EPERM);
    default:
      CASE (GM_INTERNAL_ERROR, EPFNOSUPPORT);
      CASE (GM_UNATTACHED, -1);	/* ??? */
      CASE (GM_UNSUPPORTED_DEVICE, ENXIO);
    }
}

/*
 * dumb-down the BSD ioctl method (where copyins are done by the
 * generic specfs ioctl hander) & have the GM ioctl function call
 * copyin/copyout as if we were a sysv style system 
 */
static int
gm_macos_ioctl (dev_t dev, u_long cmd, caddr_t data, int fflag, struct proc *p)
{
  gm_port_state_t *ps;
  gm_status_t status = GM_SUCCESS;
  vm_offset_t user_addr;
  int error = ENOTTY;
  unsigned copied_out;


  //  IOLockLock(gm_macos_open_lock);

  GM_PRINT (GM_DEBUG_IOCTL, ("ioctl: gm dev=0x%x minor=%d ptr = %p %s\n",
			     dev, minor (dev), data,
			     _gm_ioctl_cmd_name (cmd)));

  if (cmd == GM_GET_DEV)
    {
      gm_instance_state_t *is;
      gm_port_state_t *clone_ps;
      gm_status_t status = GM_SUCCESS;
      gm_arch_minor_t minor_id = minor (dev);

      GM_PRINT (GM_DEBUG_IOCTL, ("gm_macos_ioctl: command is GM_GET_DEV\n"));

      if (minor_id >= (GM_ARCH_MAX_INSTANCE * 2))
	{
	  GM_WARN (("gm_macos_ioctl() WRONG %d\n", minor_id));
	  error = ENODEV;
	  goto abort_with_lock;
	}

      is = gm_instances[(u_int) minor_id / 2];

      if (!is)
	{
	  error = ENODEV;
	  goto abort_with_lock;
	}

      clone_ps = gm_minor_get_port_state (minor (dev));

      if (!clone_ps)
	{
	  GM_NOTE (("gm_macos_ioctl: gm_minor_get_port_state(%d) failed\n",
		    minor (dev)));
	  error = ENODEV;
	  goto abort_with_lock;
	}

      GM_PRINT (GM_DEBUG_IOCTL,
		("gm_macos_ioctl: Got clone_ps 0x%p %d.\n",
		 clone_ps, minor (dev)));

      /* Determine an unused minor number for the opened device. */

      if (gm_minor_alloc (clone_ps->instance, &minor_id) != GM_SUCCESS)
	{
	  GM_NOTE (("gm_macos_ioctl: Could not create new minor number.\n"));
	  error = ENOMEM;
	  goto abort_with_lock;
	}

      if (INVALID_MINOR (minor_id))
	{
	  gm_minor_free (minor_id);
	  goto abort_with_lock;
	}

      GM_PRINT (GM_DEBUG_IOCTL,
		("gm_macos_ioctl: Allocated a new minor number %d.\n",
		 minor_id));
      ps = gm_minor_get_port_state (minor_id);
      if (!ps)
	{
	  GM_NOTE ((__GM_FUNCTION__
		    ": gm_minor_get_port_state(%d) for new device failed\n",
		    minor_id));
	  error = ENODEV;
	  goto abort_with_lock;
	}

      GM_PRINT (GM_DEBUG_IOCTL, ("got new port state = %p for minor=%d\n",
				 ps, minor_id));

      /* Record the instance associated with this minor number. */

      ps->instance = clone_ps->instance;
      gm_assert (ps->instance);

      /* Record the privileges from the clone device */

      ps->privileged = clone_ps->privileged;

      {
	int extern_minor = EXTERN_MINOR (minor_id);
	user_addr = *(vm_offset_t *) data;
	gm_arch_copyout (&extern_minor, (void *) user_addr,
			 sizeof (int));
	GM_PRINT (GM_DEBUG_IOCTL,
		  ("gm_macos_ioctl: Copyout minor=%d to user_addr = 0x%lx\n",
		   EXTERN_MINOR (minor_id), user_addr));
      }
      error = 0;
    }
  else
    {
      u_int minor_id = GET_MINOR (minor (dev));
      GM_PRINT (GM_DEBUG_IOCTL, ("ioctl using new minor = %d\n", minor_id));

      ps = gm_minor_get_port_state (minor_id);
      gm_assert (ps);
      if (!ps)
	{
	  error = ENXIO;
	  goto abort_with_lock;
	}
      user_addr = *(vm_offset_t *) data;

      if (cmd == GM_MMAP)
	{
	  status = gm_macos_mmap (ps, (void *) user_addr);
	}
      else
	{
	  gm_arch_mutex_enter (&ps->sync);
	  GM_PRINT (GM_DEBUG_IOCTL,
		    ("ioctl: minor=%d  ps=0x%p user_addr=0x%lx\n",
		     minor_id, ps, user_addr));
	  status = gm_ioctl (ps, cmd, (void *) user_addr, INT_MAX,
			     (void *) user_addr, INT_MAX, 0);
	  gm_arch_mutex_exit (&ps->sync);
	}

      GM_PRINT (GM_DEBUG_IOCTL, ("ioctl returns %d\n",
				 gm_macos_localize_status (status)));
      error = gm_macos_localize_status (status);
    }

 abort_with_lock:
  //  IOLockUnlock(gm_macos_open_lock);
  return (error);
}

static int
gm_macos_stop (struct tty *tp, int rw)
{
  return -1;
}


/* XXX HACK -- sys/systm.h's seltru doesn't match cdevsw proto */
static int
gm_macos_seltrue (dev_t dev, int which, void *wql, struct proc *p)
{
  return seltrue (dev, which, p);
}


/*
 *  OS-X uses devfs, and it will create our device special nodes in /dev.  
 *  An administrator can always change the permissions and ownership
 *  via chmod/chown if he doesn't like the defaults.
 */

static int
gm_macos_special_create ()
{
  int i, perms;
  char name[16];

  gm_macos_major = cdevsw_add (-1, &gm_macos_cdevsw);
  if (gm_macos_major == -1)
    {
      return gm_macos_major;
    }
  /* got a cdevsw_entry, now create device nodes via devfs */

  bzero (gm_macos_specials, GM_ARCH_MAX_MINOR * sizeof (void *));
  for (i = 0; i < GM_ARCH_MAX_MINOR; i++)
    {
      if (i < 2 * GM_ARCH_MAX_INSTANCE)
	{
	  if (i & 1)
	    {
	      /* priv. port, so use secure permissions */
	      sprintf (name, "gmp%d", i / 2);
	      perms = 0600;
	    }
	  else
	    {
	      sprintf (name, "gm%d", i / 2);
	      perms = 0666;
	    }
	}
      else
	{
	  sprintf (name, "gm%d", i);
	  perms = 0666;
	}

      /* create the special file */
      gm_macos_specials[i] = devfs_make_node (makedev (gm_macos_major, i),
					      DEVFS_CHAR, UID_ROOT, GID_WHEEL,
					      perms, name, i);
      if (gm_macos_specials[i] == NULL)
	{
	  gm_macos_special_destroy ();
	  return -1;
	}
    }
  return gm_macos_major;
}


/*
 * Undo the above. Remove the cdevsw, and free our special devices..
 */

static void
gm_macos_special_destroy ()
{
  int i;

  cdevsw_remove (gm_macos_major, &gm_macos_cdevsw);
  for (i = 0; i < GM_ARCH_MAX_MINOR; i++)
    {
      if (gm_macos_specials[i] != NULL)
	{
	  devfs_remove (gm_macos_specials[i]);
	  gm_macos_specials[i] = 0;
	}
    }
  gm_macos_major = -1;
}

void
gm_arch_abort (void)
{
  GM_WARN (("don't panic, but gm_arch_abort () was called."));
}

/* 
 * Encapsulate an mbuf chain into an ethernet segment descriptor.
 * This routine returns the number of segments in the ethernet segment
 * descriptor.
 *
 * This routine is in this file, rather than in gm_macosx_ether.c 
 * because we must use IOKit C++ clases to get the DMA addresses.
 */

int
gm_bsd_ether_arch_encap (struct gm_bsd_ether_sc *sc,
			 struct mbuf *m, 
			 struct gm_bsd_ether_arch_dma_info *dma)
{
  struct IOPhysicalSegment segments[GM_MAX_ETHERNET_SG_CNT];
  int count, i;

  count = gm_macos_mbufCursor->getPhysicalSegments (m, segments, 
						    GM_MAX_ETHERNET_SG_CNT);
  
  for (i = 0; i < count; i++)
    {
      dma->desc[i].ptr = segments[i].location;
      dma->desc[i].len = segments[i].length;
    }
  dma->nsegs = count;

  return count;
}

/*
  This file uses GM standard indentation:

  Local Variables:
  c-file-style:"gnu"
  tab-width:8
  c-backslash-column:72
  End:
*/
