/******************************************************************-*-c-*-
 * Myricom GM networking software and documentation			 *
 * Copyright (c) 2002 by Myricom, Inc.					 *
 * All rights reserved.	 See the file `COPYING' for copyright notice.	 *
 *************************************************************************/

/* author: glenn@myri.com */

#ifndef _gm_arch_types_h_
#define _gm_arch_types_h_

#include <sys/types.h>
#include <sys/errno.h>
#include <IOKit/IOLocks.h>

typedef struct gm_arch_instance_info
{
	void *cpp_class;
	caddr_t io_kva;
}
gm_arch_instance_info_t;

/* This structure defines the layout of the "arch" field of the
   gm_port_state_t type.  Use it to define storage for any
   per-port OS-specific data. */

/*
 * Note that the md and map fields are void * because the real types
 * are defined in c++ header files that the rest of gm would choke on
 * if it had to include them
*/

typedef struct gm_macosx_md_list_entry
{
	void *md;	
	void *map;
	struct gm_macosx_md_list_entry *next;
}
gm_macosx_md_list_entry_t;

typedef struct gm_arch_port_info
{
	/* insert architecture-specific per-device information here. */
	gm_macosx_md_list_entry_t *entry;
}
gm_arch_port_info_t;

/* This structure defines any storage needed to implement the
   gm_arch_dma_region*() functions. */

#define GM_ARCH_MALLOC_REGION 1
#define GM_ARCH_CONTIGMALLOC_REGION 2

typedef struct gm_arch_dma_region
{
	struct gm_instance_state *is;	/* for debugging purpose */
	void *addr;			/* aligned address */
	char *addr2;			/* for gm_arch_dma_region_advance */
	void *md_addr;
	void *md;
	void *alloc_addr;
	unsigned len;
	int flags;
	unsigned char type;
	unsigned char order;
}
gm_arch_dma_region_t;

/* This structure defines any storage needed to implement the
   gm_arch_mutex_*() and gm_arch_sync_*() functions.  (The functionality
   of these two sets of functions should be independent, even when given
   the same gm_arch_sync_t parameter. */

#define GM_SLEEP_WOKE 0
#define GM_SLEEP_TIMED_OUT EWOULDBLOCK
#define GM_SLEEP_INTERRUPTED ERESTART

typedef struct gm_arch_sync
{
	/* insert architecture-specific definition here.  Include whatever is
	   needed by gm_arch_mutex_enter(), gm_arch_mutex_exit(),
	   gm_arch_wake(), gm_arch_timed_sleep(), and gm_arch_signal_sleep(). */
	IOLock *lock;
	IOLock *wake_lock;
	SInt32 wake_cnt;
}
gm_arch_sync_t;

/* A minor device number type for this architecture.  A value of this type
   must be supplied by the OS for each per-device driver entry point,
   uniquely identifying the device corresponding to the call. */

typedef int gm_arch_minor_t;

/* Insert definition of a page lock here.  Include whatever is
   needed by gm_arch_lock_user_buffer_page() and
   gm_arch_unlock_user_buffer_page(). */

typedef struct gm_arch_page_lock
{
	void *user_map;
	void *md;
	void *user_iomap;
	void *user_addr;
	char *temp_page;
}
gm_arch_page_lock_t;

#endif /* _gm_arch_types_h_ */
