/******************************************************************-*-c-*-
 * Myricom GM networking software and documentation			 *
 * Copyright (c) 2001 by Myricom, Inc.					 *
 * All rights reserved.	 See the file `COPYING' for copyright notice.	 *
 *************************************************************************/

/* author: gallatin@myri.com */

#ifndef _gm_bsd_ether_arch_h_
#define _gm_bsd_ether_arch_h_

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/sockio.h>
#include <sys/mbuf.h>
#include <sys/malloc.h>
#include <sys/kernel.h>
#include <sys/socket.h>

#include <net/if.h>
#include <net/if_types.h>

#include <netinet/in.h>		/* for in_addr */
#include <net/if_arp.h>
#include <net/ethernet.h>
#include <net/if_dl.h>
#include <net/dlil.h>
#include <net/if_media.h>
#include <net/route.h>
#include <net/bpf.h>

#include <vm/pmap.h>		/* for vtophys */

/* for checksum offloading */
#define _IP_VHL
#include <netinet/in_systm.h>
#include <netinet/ip.h>
/*#include <machine/in_cksum.h>*/


#include "gm_debug.h"
#include "gm_internal.h"
#include "gm_ether.h"
#include "gm_debug_lanai_dma.h"

#define GM_BSD_CHECKSUM_OFFLOADING_SUPPORTED 1

struct gm_bsd_ether_sc;		/* fwd decl; declared in gm_bsd_ether.h */

struct gm_bsd_ether_arch_dma_info
{
  gm_ethernet_segment_descriptor_t desc[GM_MAX_ETHERNET_SG_CNT];
  int nsegs;
};

struct gm_bsd_ether_arch_sc
{
  struct arpcom arpcom;			/* the network (ethernet driver) */
#define is_if   arch.arpcom.ac_if	/* network-visible interface    */
#define is_addr arch.arpcom.ac_enaddr	/* hardware Ethernet address    */
  IOSimpleLock *lock;
};

extern struct gm_bsd_ether_sc *gm_bsd_ether_sc[];
extern struct mbuf *bogus_m;
extern struct gm_bsd_ether_arch_dma_info bogus_dma;
extern int myri_base_unit;

static int gm_macos_ether_ioctl (struct ifnet *ifp, unsigned long cmd,
				 void *data);
void gm_macos_ether_detach (gm_instance_state_t *is);
int gm_macos_ether_attach (gm_instance_state_t *is);


static gm_inline void
gm_bsd_ether_arch_incr_tx_counters (struct gm_bsd_ether_arch_sc *arch_sc,
				    struct mbuf *m)
{
  arch_sc->arpcom.ac_if.if_obytes += m->m_pkthdr.len;
}

static gm_inline void
gm_bsd_ether_arch_incr_rx_counters (struct gm_bsd_ether_arch_sc *arch_sc,
				    unsigned int ulen, struct ether_header *eh)
{
  arch_sc->arpcom.ac_if.if_ibytes += ulen;
}


static gm_inline void
gm_bsd_ether_arch_sc_lock (struct gm_bsd_ether_arch_sc *arch_sc)
{
  IOSimpleLockLock (arch_sc->lock);
}

static gm_inline void
gm_bsd_ether_arch_sc_unlock (struct gm_bsd_ether_arch_sc *arch_sc)
{
  IOSimpleLockUnlock (arch_sc->lock);
}

static gm_inline void
gm_bsd_ether_arch_bpf_xmit (struct ifnet *ifp, struct mbuf *m_head)
{
  if ((ifp)->if_bpf)
    {
      bpf_mtap ((ifp), (m_head));
    }
}

static gm_inline void
gm_bsd_ether_arch_bpf_rx (struct ifnet *ifp, struct mbuf *m_head)
{
  if ((ifp)->if_bpf)
    {
      ((char *)m_head->m_data) -= sizeof (struct ether_header);
      m_head->m_len += sizeof (struct ether_header);
      bpf_mtap ((ifp), (m_head));
      ((char *)m_head->m_data) += sizeof (struct ether_header);
      m_head->m_len -= sizeof (struct ether_header);
    }
}

static gm_inline void
gm_bsd_ether_arch_unencap (struct gm_bsd_ether_sc *sc,
			   struct gm_bsd_ether_arch_dma_info *dma)
{
  dma->nsegs = 0;
}

#endif /*_gm_bsd_ether_arch_h_*/


/*
  This file uses GM standard indentation.

  Local Variables:
  c-file-style:"gnu"
  tab-width:8
  End:
*/
