/******************************************************************-*-c-*-
 * Myricom GM networking software and documentation			 *
 * Copyright (c) 2001 by Myricom, Inc.					 *
 * All rights reserved.	 See the file `COPYING' for copyright notice.	 *
 *************************************************************************/

/* author: gallatin@myri.com */

#include "gm_bsd_ether.h"
#include <sys/queue.h>
#include <iokit/IOLib.h>
#include <sys/sysctl.h>

extern int gm_bsd_ether_xmit_copies;
extern IOLock *gm_macos_open_lock;

struct gm_bsd_ether_sc *gm_bsd_ether_sc[GM_ARCH_MAX_INSTANCE] = { 0 };
int gm_macos_ether_count = 0;
static int gm_max_ether_gather = GM_MAX_ETHERNET_GATHER_CNT;
static int gm_max_ether_scatter = GM_MAX_ETHERNET_SCATTER_CNT;
static int gm_macos_ether_base_en_unit = 0;
extern int gm_page_lock_failed;

 SYSCTL_NODE (_net, OID_AUTO, gm, CTLFLAG_RW, 0, "GM");
 struct sysctl_oid_list sysctl__net_newcat_children;
 SYSCTL_INT (_net_gm, OID_AUTO, ethernet_xmit_copies, CTLFLAG_RW,
	    &gm_bsd_ether_xmit_copies, 0,
	    "Mbuf chain copies caused by misalgined or over length mbuf chains");
 SYSCTL_INT (_net_gm, OID_AUTO, ethernet_max_scatter, CTLFLAG_RD,
	    &gm_max_ether_scatter, 0,
	    "Max mbufs that can be mapped for recv.");
 SYSCTL_INT (_net_gm, OID_AUTO, ethernet_max_gather, CTLFLAG_RD,
	    &gm_max_ether_gather, 0,
	    "Unit number of 1st GM ethernet interface");
 SYSCTL_INT (_net_gm, OID_AUTO, gm_base_en_unit, CTLFLAG_RD,
	    &gm_macos_ether_base_en_unit, 0,
	    "Max mbufs that can be mapped for xmit.");
 SYSCTL_INT (_net_gm, OID_AUTO, page_lock_failed, CTLFLAG_RW,
	    &gm_page_lock_failed, 0, "failed attempts to lock memory");


struct mbuf *bogus_m = NULL;
struct gm_bsd_ether_arch_dma_info bogus_dma;

/* 
 *  allocates a new "softc" struct for the gm ether device
 */

static struct gm_bsd_ether_sc *
gm_macos_ether_alloc_sc (int unit)
{
  struct gm_bsd_ether_sc *sc;

  if (gm_bsd_ether_sc[unit] != NULL)
    {
      GM_WARN (("attempt to attach myri%d twice failed\n", unit));
      return NULL;
    }

  MALLOC (sc, struct gm_bsd_ether_sc *, sizeof (*sc),
	  M_DEVBUF, M_WAITOK);

  if (sc != NULL)
    {
      bzero (sc, sizeof (struct gm_bsd_ether_sc));
      gm_bsd_ether_sc[unit] = sc;
      sc->arch.lock = IOSimpleLockAlloc ();
    }
  gm_macos_ether_count++;
  return sc;
}

/* 
 * free a softc
 */

static void
gm_macos_ether_dealloc_sc (int unit)
{
  struct gm_bsd_ether_sc *sc;

  sc = gm_bsd_ether_sc[unit];
  gm_bsd_ether_sc[unit] = NULL;
  IOSimpleLockFree (sc->arch.lock);
  FREE (sc, M_DEVBUF);
}

/*
 * called by ether_ioctl to "initialize" an interface.  We must
 * supply the function pointer, but we do not use this feature.
 */

static void
gm_macos_ether_init (void *sc_ptr)
{
}

/*
 * Callback if dlil_detatch() blocks
 */

static int
gm_macos_ether_if_free (struct ifnet *ifp)
{
  ifp->if_free = 0;
  return 0;
}

/*
 * HACK: Find the lowest numbered ethernet interface in the system.
 * We really should be able to ask for a number & to be able to free
 * up a number at unload, but there is no interface for this.
 * If a user loads another, non-gm "ethernet" interface after gm is 
 * loaded, it could try to steal our unit number(!).
 */


static int
gm_macos_ether_find_base_en_unit ()
{
  struct ifnet *ifp;

  /* do this only once */
  if (gm_macos_ether_base_en_unit != 0)
    {
      return gm_macos_ether_base_en_unit;
    }

  TAILQ_FOREACH (ifp, &ifnet, if_link)
    {
      if (ifp->if_name != NULL)
	{
	  if (strncmp (ifp->if_name, "en", 2) == 0)
	    {
	      gm_macos_ether_base_en_unit++;
	    }
	}
    }
  return gm_macos_ether_base_en_unit;
}

/*
 * This is the transmit routine.  Obtain the sleep lock
 * for the driver and call the common transmit routine.
 * Note that this interface provides queuing which is visible
 * to the upper layers, unlike smelly IOKit..
 */

static int
gm_macos_ether_start (struct ifnet *ifp, struct mbuf *m)
{
  struct gm_ether_softc *sc = ifp->if_softc;

  gm_bsd_ether_arch_sc_lock (ifp->if_softc);

  if (IF_QFULL (&ifp->if_snd))
    {
      IF_DROP (&ifp->if_snd);
      m_freem (m);
      gm_bsd_ether_arch_sc_unlock (ifp->if_softc);
      return (ENOBUFS);
    }

  IF_ENQUEUE (&ifp->if_snd, m);
  if ((ifp->if_flags & IFF_OACTIVE) == 0)
    {
      gm_bsd_ether_start (ifp);
    }

  gm_bsd_ether_arch_sc_unlock (ifp->if_softc);
  return 0;
}

/*
 * Ick! We need to register and unregister our sysctls manually.
 *
 * We also need to do it here, because the c++ compiler will puke
 * all over the header files required for sysctls, even if they're
 * wrapped in extern "C" {}
 */

static void
gm_macosx_register_sysctls (void)
{
  sysctl_register_oid (&sysctl__net_gm);
  sysctl_register_oid (&sysctl__net_gm_ethernet_xmit_copies);
  sysctl_register_oid (&sysctl__net_gm_ethernet_max_scatter);
  sysctl_register_oid (&sysctl__net_gm_ethernet_max_gather);
  sysctl_register_oid (&sysctl__net_gm_gm_base_en_unit);
  sysctl_register_oid (&sysctl__net_gm_page_lock_failed);
}

static void
gm_macosx_unregister_sysctls (void)
{
  sysctl_unregister_oid (&sysctl__net_gm_page_lock_failed);
  sysctl_unregister_oid (&sysctl__net_gm_gm_base_en_unit);
  sysctl_unregister_oid (&sysctl__net_gm_ethernet_max_gather);
  sysctl_unregister_oid (&sysctl__net_gm_ethernet_max_scatter);
  sysctl_unregister_oid (&sysctl__net_gm_ethernet_xmit_copies);
  sysctl_unregister_oid (&sysctl__net_gm);
}

/*
 * creates a new ethernet driver, initialzes it, and attaches it to the
 * network stack
 */
int
gm_macos_ether_attach (gm_instance_state_t *is)
{
  int i, unit, myri_base_unit;
  struct ifnet *ifp;
  struct ifaddr *ifa;
  struct sockaddr_dl *sdl;
  struct gm_bsd_ether_sc *sc;

  if (GM_MAX_ETHERNET_GATHER_CNT * MCLBYTES <= GM_ETHERNET_MTU)
    {
      GM_WARN (("GM_MAX_ETHERNET_GATHER_CNT (%d) is too small to hold a\n"
		"%d byte long mbuf chain consisting of %d byte mbufs\n"
		"Each transmit of %d bytes or more will have to be copied\n"
		"Consider rebuilding GM using a 32-bit gm_dp_t\n",
		GM_MAX_ETHERNET_GATHER_CNT, GM_ETHERNET_MTU, MCLBYTES,
		(GM_MAX_ETHERNET_GATHER_CNT - 1) * MCLBYTES));
    }
  unit = is->id;
  sc = gm_macos_ether_alloc_sc (unit);
  if (sc == NULL)
    {
      return ENOMEM;
    }

  sc->is = is;
  if (bogus_m == NULL)
    {
      bogus_m = gm_bsd_ether_arch_rxalloc (M_NOWAIT, GM_MAX_ETHERNET_SCATTER_CNT);
      if (bogus_m == NULL)
	{
	  gm_macos_ether_dealloc_sc (unit);
	  return ENOMEM;
	}
      if (!gm_bsd_ether_arch_encap (sc, bogus_m, &bogus_dma))
	{
	  m_freem (bogus_m);
	  bogus_m = NULL;
	  gm_macos_ether_dealloc_sc (unit);
	  return ENOMEM;
	}
    }
  if (unit == 0)
    {
      gm_macosx_register_sysctls ();
    }
  ifp = &sc->is_if;
  ifp->if_softc = sc;
  ifp->if_addrlen = 6;
  for (i = 0; i < 6; i++)
    {
      sc->is_addr[i] = is->lanai.eeprom.lanai_board_id[i];
      sc->current_mac[i] = sc->is_addr[i];      
    }

  ifp->if_name = "myri";
  ifp->if_unit = unit;

  /* leave extra space to prepend myritype */
  ifp->if_addrlen = 6;
  ifp->if_hdrlen = sizeof (struct ether_header) + 8;
  ifp->if_mtu = GM_IP_MTU;
  ifp->if_timer = 0;
  ifp->if_ioctl = gm_macos_ether_ioctl;
  ifp->if_init = gm_macos_ether_init;
  ifp->if_output = gm_macos_ether_start;
  ifp->if_flags = IFF_BROADCAST | IFF_MULTICAST | IFF_SIMPLEX;
  ifp->if_baudrate = 2000000000;
  ifp->if_snd.ifq_maxlen = 64;
  ifp->if_family = APPLE_IF_FAM_ETHERNET;
  myri_base_unit = gm_macos_ether_find_base_en_unit ();
  ifp->if_unit = myri_base_unit + unit;
  ether_ifattach (ifp);
  ifp->if_mtu = GM_IP_MTU;
  ifp->if_free = gm_macos_ether_if_free;

  /* HACK!! OS-X is broken! They discard checksums on packets from
   *  devices whose if_hwassist is 0.  However, if_hwassist is meant
   *  only to tell what sort of transmit offloading a device will do.
   *  To make recv. checksum offloading work, set it to a non-zero
   *  value which is nonsense wrt to xmit offload capabilities.  Since
   *  we don't support transmit checksum offload, we can't set it to
   *  anything which makes sense..  Sigh..
   */
  ifp->if_hwassist = CSUM_DATA_VALID;

  dlil_if_attach(ifp);
  bpfattach (ifp, DLT_EN10MB, sizeof (struct ether_header));

  return 0;

}


#if (GM_MACOSX_VERSION == 101)
static int
gm_macos_ether_inet_prmod_ioctl (u_long dl_tag, struct ifnet *ifp,
				 u_long command, caddr_t data)
{
  struct ifreq *ifr = (struct ifreq *) data;

  if (command == SIOCSIFMTU)
    {
      if (ifr->ifr_mtu > GM_IP_MTU)
	{
	  return EINVAL;
	}
      else
	{
	  ifp->if_mtu = ifr->ifr_mtu;
	  return 0;
	}
    }
  else
    {
      return ether_inet_prmod_ioctl (dl_tag, ifp, command, data);
    }

}

#endif /* GM_MACOSX_VERSION == 101 */

static int
gm_macos_ether_ioctl (struct ifnet *ifp, unsigned long cmd, void *data)
{
  struct ifreq *ifr = (struct ifreq *) data;
  struct gm_bsd_ether_sc *sc = (struct gm_bsd_ether_sc *)ifp->if_softc;
  int s, error;
  struct if_proto *proto;
  int retval;
  u_long dl_tag;

#if (GM_MACOSX_VERSION == 101)
  /* 
   *  XXX Broken MacOS 10.1 -- we need to override the "prmod_ioctl" so that
   *  we can set an mtu of > 1500 bytes after we give the ethernet
   *  interface an ip address.  This has been fixed in 10.2
   */

  retval = dlil_find_dltag (APPLE_IF_FAM_ETHERNET, ifp->if_unit,
			    PF_INET, &dl_tag);
  if (retval == 0)
    {
      proto = (struct if_proto *) dlttoproto (dl_tag);
    }

  if (retval == 0 && proto)
    {
      proto->dl_ioctl = gm_macos_ether_inet_prmod_ioctl;
    }
#endif /* GM_MACOSX_VERSION == 101 */
  
  /* MacOS 10.2 calls this routine in parallel, so we need to lock it.
     This did not happen in 10.1, but locking this part of the routine
     should be safe in 10.1 */
     
  IOLockLock(gm_macos_open_lock);

  s = splimp ();

  switch (cmd)
    {

    case SIOCSIFADDR:
      /* we don't need to do anything here, but we cannot return an
         error */
      error = 0;
      break;

    case SIOCSIFMTU:
      if (ifr->ifr_mtu > GM_IP_MTU)
	{
	  error = EINVAL;
	}
      else
	{
	  ifp->if_mtu = ifr->ifr_mtu;
	  error = 0;
	}
      break;
    case SIOCSIFFLAGS:
      if (ifp->if_flags & IFF_UP)
	{
	  if (sc->open == 1) 
	    {
	      error = 0;
	    }
	  else
	    {
	      error = gm_bsd_ether_open (ifp->if_softc);
	    }
	  if (error == 0)
	    {
	      ifp->if_flags |= IFF_RUNNING;
	    }
	  else
	    {
	      ifp->if_flags &= ~IFF_UP;
	    }
	  gm_bsd_ether_set_promisc (ifp, sc);
	  gm_bsd_ether_set_mac (ifp, sc);
	}
      else
	{
	  ifp->if_flags &= ~IFF_RUNNING;
	  if (sc->open) 
	    {
	      error = gm_bsd_ether_close (ifp->if_softc);
	    }
	  else
	    {
	      error = 0;
	    }
	}
      break;

    case SIOCADDMULTI:
      /* fall through */
    case SIOCDELMULTI:
      /* fall through */
    case SIOCPROTODETACH:
      /* fall through */
    case SIOCDIFADDR:
      error = 0;
      break;

      /* These 2 are needed to allow /usr/sbin/ip6 to detach the ip6
	 network address */
    case SIOCGIFPDSTADDR:
    case SIOCGIFPSRCADDR:
      /* and this hack is required to prevent ifconfig from
	 printing "tunnel ---> " */
    case 0xc1206940:
      error = ENOTTY;
      break;

    default:
      error = EOPNOTSUPP;
    }

  splx (s);
  IOLockUnlock(gm_macos_open_lock);
  return error;
}


int fail_test = 0;

struct mbuf *
gm_bsd_ether_arch_rxalloc (int how, int max_segs)
{
  struct mbuf *m_head, *mp, *m_tail;
  int len;
  unsigned long *ref_count;
  char *ptr;
  struct gm_jpool_entry *entry;

  /* get the first header mbuf */
  MGETHDR (mp, how, MT_DATA);
  if (mp == NULL)
    {
      return NULL;
    }

  /* make it into a 2K cluster mbuf */
  MCLGET (mp, how);
  if ((mp->m_flags & M_EXT) == 0)
    {
      m_free (mp);
      return NULL;
    }
  mp->m_len = MCLBYTES;
  
  /* start building the chain of 2K mbufs */
  m_head = m_tail = mp;
  len = GM_ETHERNET_MTU - MCLBYTES;
  
  while (len > 0)
    {
      /* get the next header mbuf */
      MGET (mp, how, MT_DATA);
      if (mp == NULL)
	{
	  goto abort_with_cluster;
	}
      /* attach to chain */
      m_tail->m_next = mp;
      m_tail = mp;
      /* use a 2K "cluster" */
      MCLGET (mp, how);
      if ((mp->m_flags & M_EXT) == 0)
	{
	  goto abort_with_cluster;
	}
      mp->m_len = MCLBYTES;
      len -= MCLBYTES;
    }

  return m_head;

 abort_with_cluster:
  m_freem (m_head);
  return (NULL);
}

void
gm_macos_ether_detach (gm_instance_state_t *is)
{
  struct gm_bsd_ether_sc *sc;
  struct ifnet *ifp;

  sc = gm_bsd_ether_sc[is->id];
  /* if we failed to attach, then there's nothing to detach */
  if (sc == NULL)
    {
      return;
    }
  ifp = &sc->is_if;

  /* force broken MacOS-X to unload our module */
#if (GM_MACOSX_VERSION == 101)
  ifp->if_eflags &= ~IFEF_DETACH_DISABLED;
#endif
  ifp->refcnt = 1;
  if (dlil_if_detach (ifp) == DLIL_WAIT_FOR_FREE)
    {
      GM_NOTE (("dlil_if_detach() told us to wait\n"));
      while (ifp->if_free)
	{
	  IOSleep (10 * 1000);
	}
    }

  /* 
   * the system no longer knows about the interface, close it &
   * deallocate it 
   */

  if (sc->open)
    {
      gm_bsd_ether_close (sc);
    }

  if (bogus_m)
    {
      gm_bsd_ether_arch_unencap (sc, &bogus_dma);
      m_freem (bogus_m);
      bogus_m = 0;
    }
  gm_macos_ether_dealloc_sc (is->id);
  if (is->id == 0)
    {
      gm_macosx_unregister_sysctls ();
    }
}




/*
  This file uses GM standard indentation.

  Local Variables:
  c-file-style:"gnu"
  tab-width:8
  End:
*/
