/*
  miniport specific code for the gm nt driver
*/

#include <ndis.h>
#include "gmm.h"


/* External symbols that are needed for initilization and cleanup. */
extern PDRIVER_OBJECT the_one_and_only_driver_object;
extern gm_u32_t unit_count;

void gm_nt_destroy_minor_node (gm_arch_minor_t minor);
VOID gm_nt_DpcForIsr (IN PKDPC Dpc, IN PVOID DeferredContext,
		      IN PVOID SystemArgument1, IN PVOID SystemArgument2);
BOOLEAN gm_nt_Isr (IN PKINTERRUPT intr, IN PVOID context);
gm_status_t gm_nt_create_minor_node (gm_instance_state_t * is,
				     PCWSTR _basename,
				     gm_arch_minor_t * minorp,
				     int is_clone_device,
				     int is_priv, unsigned long port_id);
void gm_nt_enter ();
void gm_nt_exit ();
extern gm_arch_sync_t _gm_arch_serialize_sync;
extern struct gm_hash *gm_device_hash;
extern struct gm_hash *gm_nt_phys_hash;


void gmm_allocate_shared_memory (struct gm_port *p, int size,
				 void **virtualAddress,
				 ULONG * physicalLow, ULONG * physicalHigh);
void gmm_free_shared_memory (struct gm_port *p, void *virtualAddress);

/* Tell NDIS that we're ethernet. */
NDIS_STATUS
get_medium (PUINT medium_index, PNDIS_MEDIUM medium_array,
	    UINT medium_array_size)
{
  UINT ui;
  for (ui = 0; ui < medium_array_size; ++ui)
    {
      if (medium_array[ui] == NdisMedium802_3)
	{
	  *medium_index = ui;
	  return NDIS_STATUS_SUCCESS;
	}
    }
  return NDIS_STATUS_UNSUPPORTED_MEDIA;
}

/* Fill in pci config space information. */
NDIS_STATUS
get_pci_config (gm_instance_state_t * is)
{
  ULONG ul;
  ul = NdisReadPciSlotInformation (is->arch.miniportAdapterHandle, 0, 0,
				   &is->ifc.pci.config,
				   sizeof (is->ifc.pci.config));
  if (ul != sizeof (is->ifc.pci.config))
    {
      return NDIS_STATUS_FAILURE;
    }
  return NDIS_STATUS_SUCCESS;
}

NDIS_STATUS
get_bus_and_slot (gm_instance_state_t * is)
{
  PDEVICE_OBJECT pdo;
  ULONG ul;
  NDIS_STATUS status;
  ULONG length;

  NdisMGetDeviceProperty (is->arch.miniportAdapterHandle, &pdo, NULL, NULL,
			  NULL, NULL);
  status = IoGetDeviceProperty (pdo, DevicePropertyBusNumber, sizeof (ul),
				&ul, &length);
  if (status != STATUS_SUCCESS)
    {
      return NDIS_STATUS_FAILURE;
    }
  is->arch.bus_number = ul;
  status = IoGetDeviceProperty (pdo, DevicePropertyAddress, sizeof (ul), &ul,
				&length);
  if (status != STATUS_SUCCESS)
    {
      return NDIS_STATUS_FAILURE;
    }
  is->arch.slot_number = ul >> 16;
  return NDIS_STATUS_SUCCESS;
}

NDIS_STATUS
get_resources (gm_instance_state_t * is)
{
  PCM_RESOURCE_LIST resources;
  PCM_FULL_RESOURCE_DESCRIPTOR full_descriptor;
  PCM_PARTIAL_RESOURCE_LIST partial_list;
  PCM_PARTIAL_RESOURCE_DESCRIPTOR partial_descriptor;
  ULONG ul;
  BOOLEAN have_interrupt = FALSE;
  BOOLEAN have_memory = FALSE;

  NdisMGetDeviceProperty (is->arch.miniportAdapterHandle, NULL, NULL, NULL,
			  &resources, NULL);
  full_descriptor = resources->List;
  partial_list = &full_descriptor->PartialResourceList;
  for (ul = 0; ul < partial_list->Count; ++ul)
    {
      partial_descriptor = &partial_list->PartialDescriptors[ul];
      switch (partial_descriptor->Type)
	{
	case CmResourceTypeInterrupt:
	  is->arch.interrupt_level = partial_descriptor->u.Interrupt.Level;
	  is->arch.interrupt_vector = partial_descriptor->u.Interrupt.Vector;
	  is->arch.interrupt_affinity =
	    partial_descriptor->u.Interrupt.Affinity;
	  have_interrupt = TRUE;
	  break;
	case CmResourceTypeMemory:
	  is->arch.board_phys_addr.QuadPart =
	    partial_descriptor->u.Memory.Start.QuadPart;
	  is->board_span = partial_descriptor->u.Memory.Length;
	  have_memory = TRUE;
	  break;
	}
    }
  if (!have_interrupt || !have_memory)
    {
      return NDIS_STATUS_FAILURE;
    }
  return NDIS_STATUS_SUCCESS;
}

NDIS_STATUS
get_translated_resources (gm_instance_state_t * is)
{
  PCM_RESOURCE_LIST resources;
  PCM_FULL_RESOURCE_DESCRIPTOR full_descriptor;
  PCM_PARTIAL_RESOURCE_LIST partial_list;
  PCM_PARTIAL_RESOURCE_DESCRIPTOR partial_descriptor;
  ULONG ul;

  NdisMGetDeviceProperty (is->arch.miniportAdapterHandle, NULL, NULL, NULL,
			  NULL, &resources);
  full_descriptor = resources->List;
  partial_list = &full_descriptor->PartialResourceList;
  for (ul = 0; ul < partial_list->Count; ++ul)
    {
      partial_descriptor = &partial_list->PartialDescriptors[ul];
      switch (partial_descriptor->Type)
	{
	case CmResourceTypeInterrupt:
	  is->arch.irql = (KIRQL) partial_descriptor->u.Interrupt.Level;
	  is->arch.translated_interrupt_vector =
	    partial_descriptor->u.Interrupt.Vector;
	  is->arch.affinity = partial_descriptor->u.Interrupt.Affinity;
	  return NDIS_STATUS_SUCCESS;
	}
    }
  return NDIS_STATUS_FAILURE;
}

NDIS_STATUS
create_instance (gm_instance_state_t ** is,
		 NDIS_HANDLE miniport_adapter_handle)
{
  NDIS_STATUS status;

  *is = gm_malloc (GM_PAGE_ROUNDUP (u32, sizeof (**is)));
  if (*is == NULL)
    {
      return NDIS_STATUS_RESOURCES;
    }
  RtlZeroMemory (*is, sizeof (**is));
  (*is)->arch.miniportAdapterHandle = miniport_adapter_handle;
  (*is)->lanai.eeprom.bus_type = GM_MYRINET_BUS_PCI;
  status = get_pci_config (*is);
  if (status != NDIS_STATUS_SUCCESS)
    {
      goto abort_with_is;
    }
  status = get_bus_and_slot (*is);
  if (status != NDIS_STATUS_SUCCESS)
    {
      goto abort_with_is;
    }
  (*is)->arch.driver_object = the_one_and_only_driver_object;
  status = get_resources (*is);
  if (status != NDIS_STATUS_SUCCESS)
    {
      goto abort_with_is;
    }
  status = get_translated_resources (*is);
  if (status != NDIS_STATUS_SUCCESS)
    {
      goto abort_with_is;
    }

  KeInitializeSpinLock (&(*is)->arch.intr_spin_lock);
  KeInitializeDpc (&(*is)->arch.dpc_for_isr, gm_nt_DpcForIsr, (*is));
  ExInitializeFastMutex (&(*is)->arch.spin_fast_mutex);
  KeInitializeTimer (&(*is)->arch.spin_timer);

  if (gm_instance_init (*is, unit_count, GM_MYRINET_BUS_PCI) != GM_SUCCESS)
    {
      status = NDIS_STATUS_FAILURE;
      goto abort_with_dpc;
    }
  if (gm_nt_create_minor_node (*is, L"gm", &(*is)->clone_minor, 1, 0, 0)
      != GM_SUCCESS)
    {
      status = NDIS_STATUS_FAILURE;
      goto abort_with_instance_init;
    }
  if (gm_nt_create_minor_node (*is, L"gmp", &(*is)->privileged_clone_minor, 1,
			       1, 0) != GM_SUCCESS)
    {
      status = NDIS_STATUS_FAILURE;
      goto abort_with_minor_node;
    }

  status = IoConnectInterrupt (&(*is)->arch.intr_obj, gm_nt_Isr, *is,
			       &(*is)->arch.intr_spin_lock,
			       (*is)->arch.translated_interrupt_vector,
			       (*is)->arch.irql, (*is)->arch.irql,
			       LevelSensitive, TRUE, (*is)->arch.affinity,
			       FALSE);
  if (status != STATUS_SUCCESS)
    {
      status = NDIS_STATUS_FAILURE;
      goto abort_with_privileged_minor_node;
    }
  gm_initialize_interrupts (*is);
  return NDIS_STATUS_SUCCESS;

 abort_with_privileged_minor_node:
  gm_nt_destroy_minor_node ((*is)->privileged_clone_minor);
 abort_with_minor_node:
  gm_nt_destroy_minor_node ((*is)->clone_minor);
 abort_with_instance_init:
  gm_instance_finalize (*is);
 abort_with_dpc:
  KeRemoveQueueDpc (&(*is)->arch.dpc_for_isr);
 abort_with_is:
  gm_free (*is);
  return status;
}

void
destroy_instance (gm_instance_state_t * is)
{
  gm_finalize_interrupts (is);
  IoDisconnectInterrupt (is->arch.intr_obj);
  KeRemoveQueueDpc (&is->arch.dpc_for_isr);
  gm_nt_destroy_minor_node (is->privileged_clone_minor);
  gm_nt_destroy_minor_node (is->clone_minor);
  gm_instance_finalize (is);
  gm_free (is);
}

ULONG
get_dma_alignment (NDIS_HANDLE miniport_adapter_handle)
{
  ULONG ul;
#if GM_CPU_x86
  ul = NdisGetCacheFillSize ();
#elif GM_CPU_ia64 || GM_CPU_x86_64
  ul = NdisMGetDmaAlignment (miniport_adapter_handle);
#else
#error unknown architecture
#endif
  return ul;
}

NDIS_STATUS
create_context (gmm_Context ** context, gm_instance_state_t * is,
		NDIS_HANDLE miniport_adapter_handle)
{
  NDIS_STATUS status;
  void *p;

  status = NdisAllocateMemoryWithTag (&p, sizeof (**context), 'CE');
  if (status != NDIS_STATUS_SUCCESS)
    {
      goto abort_with_nothing;
    }
  *context = p;
  NdisZeroMemory (*context, sizeof (**context));
  (*context)->miniportAdapterHandle = miniport_adapter_handle;
  (*context)->is = is;
  NdisMSetAttributesEx (miniport_adapter_handle, *context, 0,
			(NDIS_ATTRIBUTE_BUS_MASTER
			 | NDIS_ATTRIBUTE_DESERIALIZE), NdisInterfacePci);

  status = NdisMAllocateMapRegisters (miniport_adapter_handle, 0, TRUE,
				      GMM_NUM_MAP_REGISTERS,
				      GMM_MAX_BYTES_PER_DMA);
  if (status != NDIS_STATUS_SUCCESS)
    {
      goto abort_with_context;
    }
  (*context)->cacheFillSize = get_dma_alignment (miniport_adapter_handle);
  status = gmm_allocate_send_ring (*context);
  if (status != NDIS_STATUS_SUCCESS)
    {
      goto abort_with_map_registers;
    }
  status = gmm_allocate_recv_ring (*context);
  if (status != NDIS_STATUS_SUCCESS)
    {
      goto abort_with_send_ring;
    }
  if (gm_open (&(*context)->port, unit_count, GM_ETHERNET_PORT_ID,
	       GMM_PORT_NAME, GM_API_VERSION_1_0) != GM_SUCCESS)
    {
      status = NDIS_STATUS_FAILURE;
      goto abort_with_recv_ring;
    }

  status = gm_enable_ethernet ((*context)->port, GMM_NUM_RECV_BUFFERS, 7);


  gm_pause_lanai (is);
  gm_ethernet_set_recv_intr_callback ((*context)->port, gmm_receiveInterrupt,
				      *context);
  gm_ethernet_set_sent_intr_callback ((*context)->port, gmm_sendInterrupt,
				      *context);
  gmm_provide_recv_ring (*context);
  gm_unpause_lanai (is);

  return NDIS_STATUS_SUCCESS;
 abort_with_recv_ring:
  gmm_free_recv_ring (*context);
 abort_with_send_ring:
  gmm_free_send_ring (*context);
 abort_with_map_registers:
  NdisMFreeMapRegisters (miniport_adapter_handle);
 abort_with_context:
  NdisFreeMemory (*context, sizeof (**context), 0);
 abort_with_nothing:
  return status;
}

void
destroy_context (gmm_Context * context)
{
  gm_pause_lanai (context->is);
  gm_ethernet_set_recv_intr_callback (context->port, NULL, NULL);
  gm_ethernet_set_sent_intr_callback (context->port, NULL, NULL);
  gm_unpause_lanai (context->is);
  gm_close (context->port);
  gmm_free_recv_ring (context);
  gmm_free_send_ring (context);
  NdisMFreeMapRegisters (context->miniportAdapterHandle);
  NdisFreeMemory (context, sizeof (*context), 0);
}

NDIS_STATUS
gmm_Initialize (PNDIS_STATUS openErrorStatus,
		PUINT SelectedMediumIndex,
		PNDIS_MEDIUM MediumArray,
		UINT MediumArraySize,
		NDIS_HANDLE MiniportAdapterHandle,
		NDIS_HANDLE WrapperConfigurationContext)
{
  NDIS_STATUS status;
  gm_instance_state_t *is;
  gmm_Context *context;

  status = get_medium (SelectedMediumIndex, MediumArray, MediumArraySize);
  if (status != NDIS_STATUS_SUCCESS)
    {
      goto abort_with_nothing;
    }
  status = create_instance (&is, MiniportAdapterHandle);
  if (status != NDIS_STATUS_SUCCESS)
    {
      goto abort_with_nothing;
    }
  status = create_context (&context, is, MiniportAdapterHandle);
  if (status != NDIS_STATUS_SUCCESS)
    {
      goto abort_with_instance;
    }
  ++unit_count;
  return NDIS_STATUS_SUCCESS;

 abort_with_instance:
  destroy_instance (is);
 abort_with_nothing:
  return status;
}

VOID
gmm_Halt (NDIS_HANDLE MiniportAdapterContext)
{
  gmm_Context *context;
  gm_instance_state_t *is;
  PNDIS_PACKET packet;

  context = (gmm_Context *) MiniportAdapterContext;
  is = context->is;

  NdisAcquireSpinLock (&context->send_lock);
  while (context->firstTxQueue != NULL)
    {
      packet = context->firstTxQueue;
      DEQUEUE_PACKET (context->firstTxQueue, context->lastTxQueue);
      NdisReleaseSpinLock (&context->send_lock);
      NdisMSendComplete (context->miniportAdapterHandle, packet,
			 NDIS_STATUS_FAILURE);
      NdisAcquireSpinLock (&context->send_lock);
    }
  NdisReleaseSpinLock (&context->send_lock);
  while ((context->send_ring_count > 0) || (context->recv_ring_count > 0))
    {
      NdisMSleep (2000);
    }

  destroy_context (context);
  destroy_instance (is);
  --unit_count;
}

NDIS_STATUS
gmm_Reset (PBOOLEAN addressingReset, NDIS_HANDLE miniportAdapterContext)
{
  return NDIS_STATUS_SUCCESS;
#if 0
  gmm_Context *context = (gmm_Context *) miniportAdapterContext;

  GMM_ALWAYS_PRINT (("gmm_Reset called.\n"));

  NdisAcquireSpinLock (&context->receiveQueueSpinLock);
  NdisAcquireSpinLock (&context->send_lock);

  *addressingReset = TRUE;

  if (gmm_queueEmpty (&context->sendPendQueue))
    {
      NdisReleaseSpinLock (&context->send_lock);
      NdisReleaseSpinLock (&context->receiveQueueSpinLock);
      return NDIS_STATUS_SUCCESS;
    }

  GMM_ALWAYS_PRINT (("gmm_Reset to be completed later\n"));

  context->flushingSendQueue = 1;

  NdisReleaseSpinLock (&context->send_lock);
  NdisReleaseSpinLock (&context->receiveQueueSpinLock);
  return NDIS_STATUS_PENDING;
#endif
}

BOOLEAN
gmm_CheckForHang (IN NDIS_HANDLE miniportAdapterContext)
{
  return FALSE;
}

VOID
gmm_Unload (IN PDRIVER_OBJECT DriverObject)
{
  gm_nt_enter ();
  gm_destroy_hash (gm_nt_phys_hash);
  gm_destroy_hash (gm_device_hash);
  gm_nt_exit ();
  gm_arch_sync_destroy (&_gm_arch_serialize_sync);
  gm_finalize ();
}
