#include <ndis.h>
#include "gmm.h"

NDIS_STATUS
gmm_allocate_send_ring (gmm_Context * context)
{
  int i;
  int j;
  gmm_send_item *item;

  NdisAllocateSpinLock (&context->send_lock);

  item = context->send_ring;
  for (i = 0; i < GMM_NUM_SEND_BUFFERS; ++i)
    {
      NdisMAllocateSharedMemory (context->miniportAdapterHandle,
				 GM_ETHERNET_OVER_MYRINET_MTU, FALSE,
				 &item->va, &item->pa);
      if (item->va == NULL)
	{
	  goto abort_with_items;
	}
      ++item;
    }
  return NDIS_STATUS_SUCCESS;

 abort_with_items:
  item = context->send_ring;
  for (j = 0; j < i; ++j)
    {
      NdisMFreeSharedMemory (context->miniportAdapterHandle,
			     GM_ETHERNET_OVER_MYRINET_MTU, FALSE,
			     item->va, item->pa);
      ++item;
    }
  NdisFreeSpinLock (&context->send_lock);
  return NDIS_STATUS_FAILURE;
}

void
gmm_free_send_ring (gmm_Context * context)
{
  int i;
  gmm_send_item *item;

  item = context->send_ring;
  for (i = 0; i < GMM_NUM_SEND_BUFFERS; ++i)
    {
      NdisMFreeSharedMemory (context->miniportAdapterHandle,
			     GM_ETHERNET_OVER_MYRINET_MTU, FALSE,
			     item->va, item->pa);
      ++item;
    }

  NdisFreeSpinLock (&context->send_lock);
}

NDIS_STATUS
gmm_send_packet (gmm_Context * context, PNDIS_PACKET packet)
{
  gmm_send_item *item;
  UINT physical_buffer_count;
  UINT buffer_count;
  PNDIS_BUFFER buffer;
  UINT total_packet_length;
  UINT length;
  UINT ui;
  char *p;
  char *q;
  gm_ethernet_segment_descriptor_t segment;
  NDIS_PHYSICAL_ADDRESS_UNIT physical_address_array[10];
  UINT array_size;

  if (context->send_ring_count == GMM_NUM_SEND_BUFFERS)
    {
      return NDIS_STATUS_RESOURCES;
    }

  item = &context->send_ring[context->send_ring_head];
  context->send_ring_head = ((context->send_ring_head + 1)
			     % GMM_NUM_SEND_BUFFERS);
  context->send_ring_count++;

  p = item->va;
  NdisQueryPacket (packet, &physical_buffer_count, &buffer_count, &buffer,
		   &total_packet_length);
  for (ui = 0; ui < buffer_count; ++ui)
    {
      NdisQueryBuffer (buffer, &q, &length);
#if 0
      NdisMStartBufferPhysicalMapping (context->miniportAdapterHandle,
				       buffer, ui, TRUE,
				       &physical_address_array[0],
				       &array_size);
      NdisMCompleteBufferPhysicalMapping (context->miniportAdapterHandle,
					  buffer, ui);
#endif
      NdisMoveMemory (p, q, length);
      NdisGetNextBuffer (buffer, &buffer);
      p += length;
    }

  segment.ptr = item->pa.QuadPart;
  segment.len = total_packet_length;

  item->packet = packet;
  if (item->va[0] & 0x01)
    {
      gm_ethernet_broadcast (context->port, 1, &segment);
    }
  else
    {
      gm_ethernet_send (context->port, 1, &segment, item->va);
    }
  context->xmit_ok++;

  return NDIS_STATUS_SUCCESS;
}

void
gmm_SendPackets (NDIS_HANDLE miniportAdapterContext,
		 PPNDIS_PACKET packets, UINT numPackets)
{
  gmm_Context *context = (gmm_Context *) miniportAdapterContext;
  UINT ui;
  NDIS_STATUS status;

  NdisAcquireSpinLock (&context->send_lock);
  for (ui = 0; ui < numPackets; ++ui)
    {
      status = gmm_send_packet (context, packets[ui]);
      if (status == NDIS_STATUS_RESOURCES)
	{
	  context->numPacketsQueued++;
	  ENQUEUE_PACKET (context->firstTxQueue, context->lastTxQueue,
			  packets[ui]);
	  status = NDIS_STATUS_PENDING;
	}
      NDIS_SET_PACKET_STATUS (packets[ui], status);
    }
  NdisReleaseSpinLock (&context->send_lock);
}

void
gmm_sendInterrupt (gmm_Context * context)
{
  PNDIS_PACKET packet;
  int head;
  int count;
  gmm_send_item *item;

  NdisAcquireSpinLock (&context->send_lock);
  head = context->send_ring_head;
  count = context->send_ring_count;
  item = &context->send_ring[((head + GMM_NUM_SEND_BUFFERS - count)
			      % GMM_NUM_SEND_BUFFERS)];
  context->send_ring_count--;
  NdisReleaseSpinLock (&context->send_lock);
  NdisMSendComplete (context->miniportAdapterHandle, item->packet,
		     NDIS_STATUS_SUCCESS);
  NdisAcquireSpinLock (&context->send_lock);
  if (context->firstTxQueue != NULL)
    {
      packet = context->firstTxQueue;
      DEQUEUE_PACKET (context->firstTxQueue, context->lastTxQueue);
      gmm_send_packet (context, packet);
    }
  NdisReleaseSpinLock (&context->send_lock);
}
