
/******************************************************************-*-c-*-
 * Myricom GM networking software and documentation			 *
 * Copyright (c) 2001 by Myricom, Inc.					 *
 * All rights reserved.	 See the file `COPYING' for copyright notice.	 *
 *************************************************************************/

/* author: gallatin@myri.com */

#ifndef _gm_bsd_ether_arch_h_
#define _gm_bsd_ether_arch_h_

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/mbuf.h>
#include <sys/buf.h>
#include <sys/protosw.h>
#include <sys/socket.h>
#include <sys/vmmac.h>
#include <vm/vm_kern.h>
#include <sys/ioctl.h>
#include <sys/errno.h>
#include <sys/time.h>
#include <sys/kernel.h>
#include <sys/proc.h>
#include <sys/sysconfig.h>
#include <sys/malloc.h>

/* this is needed to resolve problems w/ipv6 headers that are
   apparently not shipped w/binary systems */
#if (GM_TRU64_VERSION>0x0500)
#undef _OSF_SOURCE
#endif

#include <net/if.h>
#include <net/netisr.h>
#include <net/route.h>
#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <netinet/in_var.h>
#include <netinet/ip.h>
#include <netinet/ip_var.h>
#include <netinet/if_ether.h>
#include <net/ether_driver.h>
#include <io/common/devdriver.h>
#include <hal/cpuconf.h>
#include <io/dec/pci/pci.h>

#include "gm_debug.h"
#include "gm_internal.h"
#include "gm_ether.h"
#include "gm_debug_lanai_dma.h"
#include "gm_arch_def.h"

#if   (GM_TRU64_VERSION >= 0x501) 
#define GM_BSD_CHECKSUM_OFFLOADING_SUPPORTED 1
#else
#define GM_BSD_CHECKSUM_OFFLOADING_SUPPORTED 0
#endif

struct gm_bsd_ether_sc;	/* forward declaration; declared in gm_bsd_ether.h */
struct gm_bsd_ether_pkt;/* forward declaration; declared in gm_bsd_ether.h */

struct gm_bsd_ether_arch_dma_info
{
  gm_ethernet_segment_descriptor_t desc[GM_MAX_ETHERNET_SG_CNT];
  gm_dp_t dp[GM_MAX_ETHERNET_SG_CNT];
  dma_handle_t sg[GM_MAX_ETHERNET_SG_CNT];	/* OS S/G resources */
  int nsegs;
};

struct gm_bsd_ether_arch_sc
{
  struct ether_driver is_ed;		/* the network (ethernet driver) */
#define is_ac   arch.is_ed.ess_ac	/* Ethernet common part */
#define ctrblk  is_ed.ess_ctrblk	/* Counter block */
#define ztime   arch.is_ed.ess_ztime	/* Time counters last zeroed */
#define is_if   is_ac.ac_if		/* network-visible interface    */
#define is_addr is_ac.ac_enaddr		/* hardware Ethernet address    */
  simple_lock_data_t lk;		/* SMP lock UP */
#define is_ctlr	is->arch.ctlr		/* controller struct in gm instance state */
  int	dma_flags;			/* type of DMA this device can do */
};

static void gm_osf1_ether_destroy_ring (struct gm_bsd_ether_pkt *ring,
				   int entries);
static struct gm_bsd_ether_sc *gm_osf1_ether_alloc_sc (int unit);
static void gm_osf1_ether_dealloc_sc (struct gm_bsd_ether_sc *sc);
static int gm_osf1_ether_init_ring (struct gm_bsd_ether_sc *sc,
			       struct gm_bsd_ether_pkt *ring, int entries);
static int gm_osf1_ether_dma_init (struct gm_bsd_ether_sc *sc);
static int gm_osf1_ether_dma_uninit (struct gm_bsd_ether_sc *sc);
int gm_osf1_ether_attach (gm_instance_state_t * is);
void gm_osf1_ether_detach (int unit);
static int gm_osf1_ether_start_wrapper (struct ifnet *ifp);
static int gm_osf1_ether_ioctl (struct ifnet *ifp, unsigned int cmd, caddr_t data);

extern struct gm_bsd_ether_sc *gm_bsd_ether_sc[];
extern struct mbuf *bogus_m;
extern struct gm_bsd_ether_arch_dma_info bogus_dma;

static gm_inline void
gm_bsd_ether_arch_unencap (struct gm_bsd_ether_sc *sc, struct gm_bsd_ether_arch_dma_info *dma)
{
  int i;

  for (i = 0; i < dma->nsegs; i++)
    dma_map_unload (0, dma->sg[i]);

  dma->nsegs = 0;
}

/* take care of slightly "odd" ethernet counters that are specific to 
   Tru64 
*/

static gm_inline void
gm_bsd_ether_arch_incr_tx_counters (struct gm_bsd_ether_arch_sc *arch_sc,
			   struct mbuf *m)
{
  arch_sc->ctrblk.est_bloksent++;
  arch_sc->ctrblk.est_bytesent += m->m_pkthdr.len;
}

static gm_inline void
gm_bsd_ether_arch_incr_rx_counters (struct gm_bsd_ether_arch_sc *arch_sc, unsigned int ulen,
			   struct ether_header *eh)
{
  arch_sc->ctrblk.est_blokrcvd++;
  arch_sc->ctrblk.est_bytercvd += ulen;
  if ((eh)->ether_dhost[0] & 1)
    {
      arch_sc->ctrblk.est_mblokrcvd++;
      arch_sc->ctrblk.est_mbytercvd += (ulen);
    }
}
static gm_inline void
gm_bsd_ether_arch_sc_lock (struct gm_bsd_ether_arch_sc *arch_sc)
{
  simple_lock (&arch_sc->lk);
}

static gm_inline void
gm_bsd_ether_arch_sc_unlock (struct gm_bsd_ether_arch_sc *arch_sc)
{
  simple_unlock (&arch_sc->lk);
}


static gm_inline void
gm_bsd_ether_arch_bpf_xmit (struct ifnet *ifp, struct mbuf *m_head)
{
  /* bpf is taken care of for us by ether_in/output on Tru64 */
}

#if (GM_BSD_CHECKSUM_OFFLOADING_SUPPORTED == 1)
#include "in_pseudo.h"
#endif /* GM_BSD_CHECKSUM_OFFLOADING_SUPPORTED */

#endif /*_gm_bsd_ether_arch_h_*/


/*
  This file uses GM standard indentation.

  Local Variables:
  c-file-style:"gnu"
  tab-width:8
  End:
*/
