
/******************************************************************-*-c-*-
 * Myricom GM networking software and documentation			 *
 * Copyright (c) 2001 by Myricom, Inc.					 *
 * All rights reserved.	 See the file `COPYING' for copyright notice.	 *
 *************************************************************************/

/* author: gallatin@myri.com */

#include "gm_bsd_ether.h"

/*
 * For more information regarding the structure of network driver
 * interfaces in Tru64, refer to the section 9 manual pages 
 * which are included the Tru64 DDK.  The Tru64 DDK is available
 * http://www.tru64unix.compaq.com/docs/dev_doc/DOCUMENTATION/HTML/dev_docs_r2.html
 * The comments in their sample tulip NIC PCI bus driver are also helpful.
 * Note that their sample network driver is not very good and should
 * be avoided.
 * 
 * Kernel debugging is described in detail in the kdbx(8) man page.
 * 
 * One can also use the builtin kdebug ddb debugger for live kernel
 * debugging.   See the ikdebug(8) man page.
 */

/*  
 *  All externally visable data
 */

decl_simple_lock_info (static, gm_softc_lockinfo);
extern struct timeval time;

struct gm_bsd_ether_sc *gm_bsd_ether_sc[GM_ARCH_MAX_INSTANCE] = { 0 };
int gm_ether_count = 0;

struct mbuf *bogus_m = NULL;
struct gm_bsd_ether_arch_dma_info bogus_dma;

static void
gm_osf1_ether_destroy_ring (struct gm_bsd_ether_pkt *ring, int entries)
{
  int entry, i;

  for (entry = 0; entry < entries; entry++)
    {
      for (i = 0; i < GM_MAX_ETHERNET_SG_CNT; i++)
	{
	  dma_map_dealloc (ring[entry].dma.sg[i]);
	}
    }

}

/* 
 *  allocates a new "softc" struct for the gm ether device
 *  Note that the ether_driver portion must be leaked on
 *  driver exit, so we do not use gm_malloc() to allocate it
 */

static struct gm_bsd_ether_sc *
gm_osf1_ether_alloc_sc (int unit)
{
  struct gm_bsd_ether_sc *sc;

  if (gm_bsd_ether_sc[unit] != NULL)
    {
      GM_WARN (("attempt to attach myri%d twice failed\n", unit));
      return NULL;
    }

  sc = gm_calloc (sizeof (*sc), 1);

  if (sc != NULL)
    {
      gm_bsd_ether_sc[unit] = sc;
    }
  gm_ether_count++;
  return sc;
}

/* 
 * free a softc; note that this can only be called prior
 * to calling if_attach() 
 */

static void
gm_osf1_ether_dealloc_sc (struct gm_bsd_ether_sc *sc)
{
  gm_bsd_ether_sc[sc->is_ctlr->ctlr_num] = NULL;
  gm_free (sc);
}


static int
gm_osf1_ether_init_ring (struct gm_bsd_ether_sc *sc, struct gm_bsd_ether_pkt *ring,
		    int entries)
{
  int status, entry, flags, i;

  for (entry = 0; entry < entries; entry++)
    {
      for (i = 0; i < GM_MAX_ETHERNET_SG_CNT; i++)
	{
	  status =
	    dma_map_alloc (GM_ETHER_MAXDMA, sc->is_ctlr, &ring[entry].dma.sg[i],
			   sc->arch.dma_flags);
	  if (status == 0)
	    {
	      goto abort_with_entries;
	    }
	}

    }
  return ESUCCESS;

 abort_with_entries:

  while (i)
    {
      i--;
      dma_map_dealloc (ring[entry].dma.sg[i]);
    }
  gm_osf1_ether_destroy_ring (ring, entry);
  return ENOMEM;
}

static int
gm_osf1_ether_dma_init (struct gm_bsd_ether_sc *sc)
{
  int i, status;

  sc->arch.dma_flags = DMA_ALL | DMA_IN | DMA_OUT;

  /* use a monster window for DMA if we can */
#ifdef DMA_64BIT
  if (sizeof (gm_dp_t) == 8 && sc->is->flags & GM_INSTANCE_64BIT_DMA_ADDR_OK)
    {
      sc->arch.dma_flags |= DMA_64BIT;
    }
#endif

  status = gm_osf1_ether_init_ring (sc, sc->send_ring, GM_SEND_RING_CNT);
  if (status != ESUCCESS)
    return status;
  status = gm_osf1_ether_init_ring (sc, sc->recv_ring, GM_RECV_RING_CNT);
  if (status != ESUCCESS)
    {
      gm_osf1_ether_destroy_ring (sc->send_ring, GM_SEND_RING_CNT);
      return status;
    }

  /* setup bogus_m -- bogus_m is used in order to be able to give
   * _something_ to the lanai if a dma_map_load() in gm_ether_encap()
   * fails.  In this case, the bogus_m will be handed to the lanai &
   * all data received into it will be discarded (since it could be
   * in use multiple times, by multiple instances.)  It simply acts as
   * a placeholder.  In practice, this should never need to be used;
   * this exists in order to cover all the error handling bases.
   */

  if (bogus_m != NULL)
    {				/* previously allocated by another instance */
      return ESUCCESS;
    }

  bogus_m = gm_bsd_ether_arch_rxalloc (M_NOWAIT, GM_MAX_ETHERNET_SCATTER_CNT);
  if (bogus_m == NULL)
    {
      goto abort_with_recv_ring;
    }
  for (i = 0; i < GM_MAX_ETHERNET_SG_CNT; i++)
    {
      status =
	dma_map_alloc (GM_ETHER_MAXDMA, sc->is_ctlr, &bogus_dma.sg[i],
		       sc->arch.dma_flags);
      if (status == 0)
	{
	  goto abort_with_bogus_m;
	}
    }

  if (!gm_bsd_ether_arch_encap (sc, bogus_m, &bogus_dma))
    {
      goto abort_with_bogus_m;
    }
  return ESUCCESS;


 abort_with_bogus_m:
  m_freem (bogus_m);
  bogus_m = NULL;

  while (i)
    {
      i--;
      dma_map_dealloc (bogus_dma.sg[i]);
    }

 abort_with_recv_ring:
  gm_osf1_ether_destroy_ring (sc->recv_ring, GM_RECV_RING_CNT);
 abort_with_send_ring:
  gm_osf1_ether_destroy_ring (sc->send_ring, GM_SEND_RING_CNT);

  return ENOMEM;

}

static int
gm_osf1_ether_dma_uninit (struct gm_bsd_ether_sc *sc)
{
  int i;

  gm_osf1_ether_destroy_ring (sc->recv_ring, GM_RECV_RING_CNT);
  gm_osf1_ether_destroy_ring (sc->send_ring, GM_SEND_RING_CNT);
  if (gm_ether_count == 1)
    {
      /* we're the last interface, so destroy bogus_m and its DMA map */
      gm_bsd_ether_arch_unencap (sc, &bogus_dma);
      for (i = 0; i < GM_MAX_ETHERNET_SG_CNT; i++)
	{
	  dma_map_dealloc (bogus_dma.sg[i]);
	}
      m_freem (bogus_m);
      bogus_m = NULL;
    }
  return ESUCCESS;
}

/*
 * creates a new ethernet driver, initialzes it, at attaches it to the
 * network stack
 */
int
gm_osf1_ether_attach (gm_instance_state_t * is)
{
  int i, r, unit, status;
  struct ifnet *ifp;
  struct sockaddr_in *sin;
  struct gm_bsd_ether_sc *sc;
  struct controller *ctlr;
  char *name;

  ctlr = is->arch.ctlr;
  unit = is->id;
  sc = gm_osf1_ether_alloc_sc (unit);
  if (sc == NULL)
    {
      printf("failed to alloc sc\n");
      return ENOMEM;
    }
  sc->is = is;

  status = gm_osf1_ether_dma_init (sc);
  if (status != ESUCCESS)
    {
      GM_NOTE (("gm_osf1_ether_dma_init failed, status %d\n", status));
      gm_osf1_ether_dealloc_sc (sc);
      return status;
    }


  ifp = &sc->is_if;
  ifp->if_addrlen = 6;
  for (i = 0; i < 6; i++)
    {
      sc->is_addr[i] = is->lanai.eeprom.lanai_board_id[i];
    }

  /* allocate an intentionally leaked name; attachpfilter() caches old
     names and provides no way to remove them, so we must give it a
     pointer which is valid even after the device has been unloaded */

  name = gm_malloc (IFNAMSIZ);
  if (name == NULL) 
    {
      GM_NOTE (("gm_osf1_ether_attach: malloc failed\n"));
      gm_osf1_ether_dealloc_sc (sc);
      return ENOMEM;
    }
  strncpy (name, "myri", IFNAMSIZ);
  ifp->if_name = name;
  ifp->if_unit = unit;

  /* leave extra space to prepend myritype */
  ifp->if_hdrlen = sizeof (struct ether_header) + 8;
  ifp->if_mtu = GM_IP_MTU;
  ifp->if_mediamtu = GM_ETHERNET_MTU;
  ifp->if_type = IFT_ETHER;
  ifp->if_timer = 0;

  ((struct arpcom *) ifp)->ac_flag = 0;
  sin = (struct sockaddr_in *) &ifp->if_addr;
  sin->sin_family = AF_INET;

  sc->is_ac.ac_bcastaddr = (u_char *) etherbroadcastaddr;
  sc->is_ac.ac_arphrd = ARPHRD_ETHER;
  ifp->if_affinity = NETMASTERCPU;	/* NETALLCPU or NETMASTERCPU to funnel */
  ifp->lk_softc = &sc->arch.lk;
  ifp->if_version = "Myricom 1280/2000 Myrinet Interface";
  simple_lock_setup (&sc->arch.lk, gm_softc_lockinfo);

  ifp->if_ioctl = gm_osf1_ether_ioctl;
  ifp->if_watchdog = /*gm_ether_watchdog */ NULL;
  ifp->if_start = gm_osf1_ether_start_wrapper;
  mb ();
  ifp->if_output = ether_output;
  mb ();
  ifp->if_flags =
    IFF_BROADCAST | IFF_MULTICAST | IFF_NOTRAILERS | IFF_SIMPLEX;

  IFQ_LOCKINIT_SPLIMP (&ifp->if_snd);
  ifp->if_snd.ifq_maxlen = GM_SEND_RING_CNT;

  /*
   * Add the interface to the list of configured ifnet structures,
   * and attach it to the packet filter.
   */
  attachpfilter (&sc->arch.is_ed);
  if_attach (ifp);
  ifp->if_flags |= IFF_RUNNING;
  return ESUCCESS;
}

void
gm_osf1_ether_detach (int unit)
{
  struct gm_bsd_ether_sc *sc;
  struct ifnet *ifp, *prev_ifp, *cur_ifp;
  int s, status;

  sc = gm_bsd_ether_sc[unit];
  /* if we failed to attach, then there's nothing to detach */
  if (sc == NULL)
    {
      return;
    }
  ifp = &sc->is_if;
  if (sc->open)
    {
      s = splnet ();
      /* must be called at splnet without lock held */
      if_down (ifp);
      splx (s);
    }
  s = splimp ();
  gm_bsd_ether_arch_sc_lock (&sc->arch);
  if (sc->open)
    {
      status = gm_bsd_ether_close (sc);
      if (status != GM_SUCCESS)
	GM_WARN(("gm_bsd_ether_close failed with status %d\n", status));
    }

  gm_osf1_ether_dma_uninit (sc);
  /* remove the interface from the global list of interfaces so we
   * don't see multiple copies of myriX if driver is reloaded.
   */
  prev_ifp = ifnet;
  for (cur_ifp = ifnet; cur_ifp && (cur_ifp != ifp);
       cur_ifp = cur_ifp->if_next)
    {
      prev_ifp = cur_ifp;
    }
  if (cur_ifp == ifnet)
    {
      ifnet = ifp->if_next;
    }
  else
    {
      prev_ifp->if_next = ifp->if_next;
    }

  ifp->if_next = NULL;
#if (GM_TRU64_VERSION>=0x0501)
  ifp->if_flags = IFF_DETACHED;
#else
  ifp->if_flags = 0;
#endif
  gm_bsd_ether_arch_sc_unlock (&sc->arch);
  simple_lock_terminate (&sc->arch.lk);
  splx (s);
}

/* 
 * Encapsulate an mbuf chain into an ethernet segment descriptor.  As
 * a side-effect, it calls os-specific routines to map mbuf data
 * pointers to DMA address.  This routine returns the number of
 * segments in the ethernet segment descriptor.
 */

int
gm_bsd_ether_arch_encap (struct gm_bsd_ether_sc *sc, struct mbuf *m_head,
		struct gm_bsd_ether_arch_dma_info *dma)
{
  int count, i, flags, realloc_needed, rv;
  struct mbuf *m;
  struct sglist *sglistp;
  sg_entry_t entry;

  count = 0;

  for (m = m_head; m != NULL; m = m->m_next)
    {
      if (m->m_len == 0)
	continue;
      flags = sc->arch.dma_flags;
      realloc_needed = 0;
      sglistp = dma->sg[count];

      /* Don't ask for a contig mapping if we don't need one -- 
       * Tru64 will be happy to give us a cheap direct mapping
       * rather than an expensive S/G chipset mapping.
       * In practice, is extremely rare to encounter an mbuf which
       * crosses page boundaries & actually requires a contig mapping
       *
       * S/G mappings are very expenive -- on a pyxis, updating the SG
       * TLB hardware is more expensive than memory copies!
       */
      if ((mtod (m, vm_offset_t) + (vm_offset_t) m->m_len) >
	  trunc_page (mtod (m, vm_offset_t) + PAGE_SIZE))
	{
	  flags |= DMA_CONTIG;
	  if ((sglistp->flags & DMA_HW_SG) == 0)
	    realloc_needed = 1;
	}
      else
	{
	  if ((sglistp->flags & DMA_HW_SG) != 0)
	    realloc_needed = 1;
	}

      /* if we need to reallocate a new dma handle, hang on to the old
       * one until we are certain that we're able to get a new one.
       * Having an invalid dma handle will leave things in an
       * inconsistant state
       */
      if (realloc_needed)
	{
	  dma->sg[count] = NULL;	/* sglistp holds old value */
	  rv = dma_map_alloc (GM_ETHER_MAXDMA, sc->is_ctlr, &dma->sg[count],
			      flags);
	  if (rv == 0)
	    {
	      GM_NOTE (("gm_bsd_ether_arch_encap: dma_map_alloc failed\n"));
	      dma->sg[count] = sglistp;
	      goto abort_with_map;
	    }
	  dma_map_dealloc (sglistp);
	}
      rv = dma_map_load (m->m_len, mtod (m, vm_offset_t), NULL, sc->is_ctlr,
			 &dma->sg[count], 0, flags);
      if (rv == 0)
	{
	  GM_NOTE (("gm_bsd_ether_arch_encap: dma_map_load failed\n"));
	  goto abort_with_map;
	}
      entry = dma_get_curr_sgentry (dma->sg[count]);
      if ((entry == NULL) || (entry->bc < m->m_len))
	{
	  GM_NOTE (("gm_bsd_ether_arch_encap: dma_map_load fragmented\n"));
	  count++;
	  goto abort_with_map;
	}
      dma->desc[count].ptr = (gm_dp_t) entry->ba;
      dma->desc[count].len = (gm_u32_t) entry->bc;
      count++;
    }

  dma->nsegs = count;
  return count;

 abort_with_map:
  /* unload any mappings we may have loaded */
  for (i = 0; i < count; i++)
    {
      dma_map_unload (0, dma->sg[i]);
    }
  return 0;
}


static int
gm_osf1_ether_start_wrapper (struct ifnet *ifp)
{
  struct gm_bsd_ether_sc *sc = gm_bsd_ether_sc[ifp->if_unit];
  int s = splimp ();

  /* 
   * try to obtain our lock, if we're already busy, don't block the
   * stack
   */

  if (simple_lock_try (&sc->arch.lk))
    {
      gm_bsd_ether_start (ifp);
      gm_bsd_ether_arch_sc_unlock (&sc->arch);
    }
  splx (s);
  return (0);
}


struct mbuf *
gm_bsd_ether_arch_rxalloc (int how, int max_segs)
{
  struct mbuf *m1, *m2;

  /* get the first header mbuf */
  MGETHDR (m1, how, MT_DATA);
  if (m1 == NULL)
    {
      return NULL;
    }
  /* make it into a 2K cluster mbuf */
  MCLGET2 (m1, how);
  if ((m1->m_flags & M_EXT) == 0)
    {
      goto fail;
    }
  m1->m_len = MCL2KBYTES;

  /* get the second mbuf */
  MGET (m2, how, MT_DATA);
  if (m2 == NULL)
    {
      goto fail;
    }

  /* attach the second mbuf to the first mbuf cluster */
  m1->m_next = m2;

  /* make it into an 8K cluster mbuf */
  MCLGET (m2, how);
  if ((m2->m_flags & M_EXT) == 0)
    {
      goto fail;
    }
  m2->m_len = MCLBYTES;

  /* return the chain */
  return m1;

 fail:
  /* free the entire chain */
  m_freem (m1);
  return NULL;

}


static int
gm_osf1_ether_ioctl (struct ifnet *ifp, unsigned int cmd, caddr_t data)
{
  struct gm_bsd_ether_sc *sc = gm_bsd_ether_sc[ifp->if_unit];
  struct ifdevea *ifd = (struct ifdevea *) data;
  struct ifreq *ifr = (struct ifreq *) data;
  struct ctrreq *ctr = (struct ctrreq *) data;
  unsigned short ifmtu;
  int i, locked, s, status;

  s = splimp ();
  gm_bsd_ether_arch_sc_lock (&sc->arch);
  locked = 1;

  status = ESUCCESS;

  switch (cmd)
    {
    case SIOCENABLBACK:
      ifp->if_flags |= IFF_LOOPBACK;
      break;
    case SIOCDISABLBACK:
      ifp->if_flags &= ~IFF_LOOPBACK;
      break;

    case SIOCRPHYSADDR:
      /*
       * Return the current and default physical addresses.
       */
      bcopy (sc->is_addr, ifd->current_pa, 6);
      for (i = 0; i < 6; i++)
	{
	  ifd->default_pa[i] = sc->is->lanai.eeprom.lanai_board_id[i];
	}
      break;

    case SIOCSPHYSADDR:
      /*
       * Set (override) the hardware physical address
       */
      bcopy (ifr->ifr_addr.sa_data, sc->is_addr, 6);
      pfilt_newaddress (sc->arch.is_ed.ess_enetunit, sc->is_addr);

      /* 
       * If an IP address has been configured then an ARP 
       * packet must be broadcast to tell all hosts which 
       * currently have our address in their ARP tables to 
       * update their information.
       */
      gm_bsd_ether_arch_sc_unlock (&sc->arch);
      splx (s);

      if (((struct arpcom *) ifp)->ac_flag & AC_IPUP)
	{
	  rearpwhohas ((struct arpcom *) ifp);
	}

      /* Also call in to update the af_link address. */
      if_sphyaddr (ifp, ifr);
      locked = 0;
      break;

    case SIOCRDCTRS:
    case SIOCRDZCTRS:
      ctr->ctr_ether = sc->arch.ctrblk;
      ctr->ctr_type = CTR_ETHER;
      ctr->ctr_ether.est_seconds =
	(time.tv_sec - sc->ztime) >
	0xfffe ? 0xffff : (time.tv_sec - sc->ztime);
      if (cmd == SIOCRDZCTRS)
	{
	  sc->ztime = time.tv_sec;
	  bzero (&sc->arch.ctrblk, sizeof (struct estat));
	}
      break;

    case SIOCSIFADDR:
      ifp->if_flags |= IFF_UP;
      status = gm_bsd_ether_open (sc);
      break;

    case SIOCSIFFLAGS:
      if (ifr->ifr_flags & IFF_UP)
	{
	  status = gm_bsd_ether_open (sc);
	}
      else
	{
	  status = gm_bsd_ether_close (sc);
	}
      break;

    case SIOCSIPMTU:
      bcopy (ifr->ifr_data, (u_char *) & ifmtu, sizeof (u_short));
      if (ifmtu > GM_IP_MTU || ifmtu < IP_MINMTU)
	{
	  status = EINVAL;
	}
      else
	{
	  ifp->if_mtu = ifmtu;
	}
      break;

    case SIOCADDMULTI:
      /* fall through */
    case SIOCDELMULTI:
      status = ESUCCESS;
      break;

    default:
      status = EINVAL;

    }

  if (locked)
    {
      gm_bsd_ether_arch_sc_unlock (&sc->arch);
      splx (s);
    }
  return status;


}




/*
  This file uses GM standard indentation.

  Local Variables:
  c-file-style:"gnu"
  tab-width:8
  End:
*/
