/******************************************************************-*-c-*-
 * Myricom GM networking software and documentation			 *
 * Copyright (c) 1999 by Myricom, Inc.					 *
 * All rights reserved.	 See the file `COPYING' for copyright notice.	 *
 *************************************************************************/

#include    <sys/types.h>
#include    <sys/errno.h>
#include    <sys/debug.h>
#include    <sys/stropts.h> 
#include    <sys/stream.h>
#include    <sys/strlog.h>
#include    <sys/cmn_err.h>
#include    <sys/kmem.h>
#include    <sys/conf.h> 
#include    <sys/ksynch.h>
#include    <sys/stat.h>
#include    <sys/dlpi.h>
#include    <sys/modctl.h>
#ifdef  KSTAT
#include        <sys/kstat.h>
#endif
#include    <sys/time.h>
#include    <sys/ddi.h>
#include    <sys/sunddi.h>

#include "gm_internal.h"
#include "gm_ether.h"
#include "myri.h"

static char *myri_compile_flags = MYRI_CFLAGS;

/* this needs to be defined for host_ifc.c */
extern struct myri *myridev;
extern int myridebug;

extern struct timeval time_start[32], time_stop[32];

#ifdef sparc_solaris
void
usleep(unsigned usec)
{
  unsigned i;
  for (i = 0; i < usec; i++) {
  }
}
#endif

/* increment (x) by (inc) and modulo (max) */
gm_inline unsigned
MODINC(int x, int inc, int max)
{

  register int tmp;
  register int j;

  tmp = x + inc;
  if ((j = (tmp - max)) >= 0) {
    return j;
  }
  return (tmp);
}

gm_inline int gm_get_receive_unused(struct myri *myrip)
{
  return (myrip->recv_buffer_count - NUM_RECV_BUFF1);
}

gm_inline
void
mybcopy(char *src, char *dst, int len)
{
	register int count;
	register char *rsrc = src;
	register char *rdst = dst;

	for (count=len; count; count--) {
		*dst++ = *src++;
	}
}

/******************  copy_mblk_to_buffer()  ***************************/
/*
 * Note: length is the exact length of the packet after padding with zeros
 *       The true length of the packet before padding comes from the mblk itself
 *       as we copy it.
 *
 */
gm_u32_t
copy_mblk_to_buffer(char *vptr, char *lptr2, int max_len)
{
  mblk_t *m = (mblk_t *) vptr;
  unsigned char *ptr;
  unsigned short *lptr = (unsigned short *) lptr2;
  int buf_len;
  int add;
  gm_u32_t len = 2;

  ptr = (unsigned char *) &lptr[0];

  while (m) {
    buf_len = (m->b_wptr - m->b_rptr);
	if ((len+buf_len)>= max_len) {
		CMN_ERR(0,(CE_CONT, "copy_to_mblk tried to overrun buffer\n"));
		break;
	}
    bcopy( (void *)m->b_rptr, (void *)ptr, buf_len);
    ptr += buf_len;
    len += buf_len;
    /* 
     * gm_hex_dump((caddr_t) m->b_rptr, buf_len); 
     */
    m = m->b_cont;
  }

  return len;
}

#define MYRI_COPY_THRESHOLD 256

static void
myri_unmap_xmit_dma (struct myri_dma_state *dma)
{
  while (dma->nsegs_bound > 0) {
    dma->nsegs_bound--;
    ddi_dma_unbind_handle(dma->handle[dma->nsegs_bound]);
  }
}

static int
myri_map_xmit_dma (struct myri_dma_state *dma, mblk_t *mp, struct myri *myrip)
{
  int nsegs, err, i;
  unsigned int count;
  size_t len;
  ddi_dma_cookie_t cookies[GM_MAX_ETHERNET_SCATTER_CNT];
  ddi_dma_handle_t handle;
  mblk_t *mp_tmp;

  if (dma->nsegs_bound)
    CMN_ERR_PRINT(0,(CE_CONT, "active segs!\n"));

  if (dma->mp)
    CMN_ERR_PRINT(0,(CE_CONT, "mblk attached!\n"));

  /* if the total amount of data in the chain is smaller than 
     an aritrary copy threashold, then just copy  */

  if ((len = msgdsize(mp)) < MYRI_COPY_THRESHOLD)
    return (0);

  /* make sure we (think) we have enough scatters for the chain.  This
     is just an estimate, as on systems without an IOMMU there could
     be an mblk which crosses a page boundary, and which would require
     multiple scatters */

  nsegs = 0;
  for (mp_tmp = mp; mp_tmp; mp_tmp = mp_tmp->b_cont)
    nsegs++;

  if (nsegs > GM_MAX_ETHERNET_SCATTER_CNT) {
    goto abort_with_nothing;
  }

  /* now map the buffer for DMA */
  nsegs = 0;
  for (mp_tmp = mp; mp_tmp; mp_tmp = mp_tmp->b_cont) {
    len = mp_tmp->b_wptr - mp_tmp->b_rptr;

    if (len == 0) /* skip zero-length segments */
      continue;

    handle = dma->handle[dma->nsegs_bound];
    err = ddi_dma_addr_bind_handle(handle, NULL, 
				   (char *)mp_tmp->b_rptr, len,
				   DDI_DMA_WRITE|DDI_DMA_STREAMING,
				   DDI_DMA_DONTWAIT, NULL, 
				   &cookies[0], &count);
    if (err != DDI_SUCCESS) {
      CMN_ERR_PRINT(0,(CE_CONT, "bind handle failed\n"));
      goto abort_with_dma_bound;
    }
    dma->nsegs_bound++;

    if (nsegs + count > GM_MAX_ETHERNET_SCATTER_CNT) {
      /* too many segments, time to give up and copy */
      CMN_ERR_PRINT(BUFF_PRINT,
		    (CE_CONT, "too many segs: %d, len %d\n",
		     nsegs, len));
      goto abort_with_dma_bound;
    }

    /* we could have received multiple DMA cookies if an mblk crossed
       a page boundary, so iterate over them */
    for (i = 0; i < count; i++) {

      /* we get the first DMA cookie for free.  Subsequent ones must
	 be retrieved here */
      if (i > 0)
	ddi_dma_nextcookie(handle, &cookies[i]);

      dma->desc[nsegs].ptr = cookies[i].dmac_laddress;
      dma->desc[nsegs].len = cookies[i].dmac_size;
      nsegs++;
    }
    if (ddi_dma_sync(handle, 0, len, DDI_DMA_SYNC_FORDEV) != DDI_SUCCESS) {
      CMN_ERR_PRINT (0, (CE_CONT, "DDI_DMA_SYNC_FORDEV FAILED\n"));
    }
  }
  dma->mp = mp;
  return nsegs;

 abort_with_dma_bound:
  myri_unmap_xmit_dma(dma);
 abort_with_nothing:
  myrip->myri_xmit_copies++;
  return (0);
}


/***********************  myri_tolanai()  *****************************/

int
myri_tolanai(queue_t * wq, mblk_t * mp, struct myri *myrip)
{
  int rv = 1;
  gm_u8_t myri_addr[8];
  mblk_t * mp_save = mp;
  struct myri *myrip_save=myrip;
  int index, nsegs;
  gm_u32_t len;
  register gm_ethernet_segment_descriptor_t *gather;

  CMN_ERR(3,  (CE_CONT, "        myri_tolanai() called\n"));

  if (!mp) 
    return(0);

  if ((unsigned long)wq <= (unsigned long)0x10000) {
    CMN_ERR(0,(CE_CONT,"Bad wq passed to myri_tolanai = 0x%lx\n",
		(unsigned long)wq));
    CMN_ERR(0,(CE_CONT,"    0x%lx   0x%lx   0x%lx\n",
            (unsigned long)wq, (unsigned long)mp, (unsigned long)myrip));
    return(0);
  }

  bcopy((gm_u8_t *)mp->b_rptr, myri_addr, 6);


  MUTEX_ENTER(&myrip->myri_xmitlock);
  myrip->xmit_count++;

#ifdef MYRI_DEBUG
  ASSERT(myrip->xmit_count == 1);
#endif

  /* send the one message pointed to by mp */
  /* start with the next one */
  index = CYCLE(myrip->last_send_index, NUM_SEND_BUFF);
  
  if (myrip->sendinfo[index].status != BUF_UNUSED) {
    CMN_ERR_PRINT (1, (CE_CONT, "myri_tolanai: no send buffers free!\n"));
    goto send_error;
  }

  /* found a buffer */
  myrip->last_send_index = index;
  myrip->send_buffer_count++;
    
  nsegs = myri_map_xmit_dma(&myrip->xmit_handles[index], mp, myrip);
  
  if (nsegs != 0) {
    /* send directly from DMA-mapped mbkl */
    myrip->sendinfo[index].status = BUF_LOANED;
    gather = myrip->xmit_handles[index].desc;
  } else {
    /* send by copying into pre-mapped DMAable memory */
    myrip->sendinfo[index].status = BUF_LANAI;
      
    /* copy the chain to the buffer */
    len = copy_mblk_to_buffer((char *) mp,
			      (char *) myrip->sendinfo[index].kptr, SEND_LEN);
    
    if (mp != mp_save) {
      CMN_ERR(0,(CE_CONT, "yikes - mp is corrupted?? 0x%lx 0x%lx\n",
		 mp, mp_save));
    }
    else {
      /* free chain */
      freemsg(mp);
      mp = NULL;
    }

    CMN_ERR(3, (CE_CONT, "tolanai: packet before handing to LANai\n"));
    /* gm_hex_dump((char *) myrip->sendinfo[index].kptr, 16); */ 
    
    /* hand it to the LANai */
    gather = &myrip->sendinfo[index].ether_pkt;
    gather->len = len;
    nsegs = 1;
    if (ddi_dma_sync(myrip->sendinfo[index].handle, 0,
		     len, DDI_DMA_SYNC_FORDEV) != DDI_SUCCESS) {
      CMN_ERR_PRINT (0, (CE_CONT, "DDI_DMA_SYNC_FORDEV FAILED index = %d len = %d\n",
			 index, len));
    }
  }
  CMN_ERR(2,(CE_CONT,"ethernet_send idx=%d ptr=0x%lx len=%d\n",
	     index, gather->ptr,gm_hton_u32(gather->len)));
  
  if (myri_addr[0] != 0xff) {
    gm_ethernet_send(myrip->gm_port, nsegs, gather, myri_addr);
  } else {		  
    gm_ethernet_broadcast(myrip->gm_port, nsegs, gather);
  }
    
  myrip->myri_opackets++;
  
  myrip->xmit_count--;
  MUTEX_EXIT(&myrip->myri_xmitlock);
    
  if (myrip != myrip_save) {
    CMN_ERR(0,(CE_CONT, "yikes - myrip is corrupt 0x%lx 0x%lx\n",
	       myrip, myrip_save));
  }
  else {
    /* check for stuff that got blocked */
    if (myrip->myri_wantw) {
      CMN_ERR(1,(CE_CONT,"wantw is set - reenabling - 1\n"));
      myriwenable(myrip);
    }
  }
  return (rv);

send_error:
  myrip->myri_oerrors++;
  if (mp) {
    freemsg(mp);
  }
  mp = NULL;

  myrip->xmit_count--;
  MUTEX_EXIT(&myrip->myri_xmitlock);

  return (0);
}


/******************   myri_mutex_enter()   *********************/
/*
 * Grab all the semaphores.
 *
 */
gm_inline void  
myri_mutex_enter(struct myri *myrip)
{
  MUTEX_ENTER(&myrip->myri_xmitlock);
  ++myrip->xmit_count;
#ifdef MYRI_DEBUG
  ASSERT(myrip->xmit_count == 1);
#endif
  MUTEX_ENTER(&myrip->myri_recvlock);
  ++myrip->recv_count;
#ifdef MYRI_DEBUG
  ASSERT(myrip->recv_count == 1);
#endif
}


/******************   myri_mutex_exit()   *********************/
/*
 * Release all the semaphores.
 *
 */
gm_inline void
myri_mutex_exit(struct myri *myrip)
{
  /* MUTEX_EXIT(&myrip->myri_bufflock); */
  myrip->recv_count--;
  MUTEX_EXIT(&myrip->myri_recvlock);
  myrip->xmit_count--;
  MUTEX_EXIT(&myrip->myri_xmitlock);
  MUTEX_EXIT(0);
}

/***********************  myri_dma_error()   *****************************/
static char *dma_error_names[16] = {"DDI_DMA_MAPPED",
				    "DDI_DMA_NORESOURCES",
				    "DDI_DMA_NOMAPPING",
				    "DDI_DMA_TOOBIG",
				    "DDI_DMA_TOOSMALL",
				    "DDI_DMA_LOCKED",
				    "DDI_DMA_BADLIMITS",
				    "DDI_DMA_STALE",
				    "DDI_DMA_BADATTR",
				    "DDI_DMA_INUSE"};

void
myri_dma_error(int rv)
{
  if ((rv<=0) && (rv>=-15)){
    rv = -rv;
    cmn_err(CE_CONT, "dma_setup %s (%d)\n",
	    dma_error_names[rv],-rv);
  }
  else {
    cmn_err(CE_CONT, "dma_setup rv=%d not understood!\n", rv);
  }
}
  
/*****************   myri_handle_lanai_intr()  ************/
/*
 * Called from myriintr.
 *
 */

void
myri_recv_intr(void *context, gm_ethernet_recv_t *e)
{
  struct myri *myrip = (struct myri *) context;
  gm_u8_t index = gm_ntoh_u8 (e->tag);
  unsigned int len = gm_ntoh_u32 (e->length);
  mblk_t *mp;

  MUTEX_ENTER(&myrip->myri_recvlock);
  myrip->recv_count++;

#ifdef MYRI_DEBUG
  ASSERT(myrip->recv_count == 1);
#endif
  mp = myri_do_recv(myrip, len, index);
  if (mp == NULL) {
    MUTEX_EXIT(&myrip->myri_recvlock);
    return;
  }
  
  myrip->recv_count--;
  MUTEX_EXIT(&myrip->myri_recvlock);
  
#ifdef LATER
  if (had_to_copy || gm_get_receive_unused(myrip) <= ( NUM_RECV_BUFF1 / 4)) {
    /* now try to find more bufs */
    struct buff_info binfo;
    
    binfo.myrip = myrip;
    binfo.index = -1;
    binfo.queue = RECV1_Q;
    myri_add_buff(&binfo);
    
    had_to_copy = 0;
  }
#endif

#ifdef IPQ
  {
    queue_t *ipq = myrip->myri_ipq;
    struct medium_header *ehp;
    
    ehp = (struct medium_header *)mp->b_rptr;
    
    if ((get_ether_type(ehp) == ETHERTYPE_IP) &&
	((ehp->med_dhost.medaddr[0] & 01) == 0) &&
	(ipq) &&
	canput(ipq->q_next)) {
      CMN_ERR (3, (CE_CONT, "receive fast path\n"));
      mp->b_rptr += sizeof (struct medium_header);
      (void) putnext(ipq, mp);
    } else
      myrisendup(myrip, mp);
  }						       
#else
  myrisendup(myrip, mp);
#endif
  CMN_ERR(10, (CE_CONT, "myriINTERRUPT exiting\n"));
}

void
myri_sent_intr(void *context)
{
    struct myri *myrip = (struct myri *) context;

    CMN_ERR(10, (CE_CONT, "myri_recv_intr called \n"));
    CMN_ERR(3, (CE_CONT, "SEND_EVENT \n"));
    MUTEX_ENTER(&myrip->myri_xmitlock);
    if (myrip->sendinfo[myrip->head_send_index].status == BUF_LOANED) {
      /* we mapped the packed for DMA, so we need to tear down those
	 resources */
      myri_unmap_xmit_dma(&myrip->xmit_handles[myrip->head_send_index]);
      freemsg(myrip->xmit_handles[myrip->head_send_index].mp);
      myrip->xmit_handles[myrip->head_send_index].mp = NULL;
    }
    myrip->sendinfo[myrip->head_send_index].status = BUF_UNUSED;
    myrip->head_send_index = CYCLE(myrip->head_send_index, NUM_SEND_BUFF);
    myrip->send_buffer_count--;
    MUTEX_EXIT(&myrip->myri_xmitlock);
#ifdef SET_WANTW
    /* check for pending send messages */
    if (myrip->myri_wantw) {
      CMN_ERR(1,(CE_CONT,"wantw is set - reenabling - 2\n"));
      myriwenable(myrip);
    }
#endif	/* SET_WANTW */
}

/***************** myri_do_recv()  **********************/
/*
 * Check the recv_ack queue.
 * called while already holding the recv_lock mutex
 * ASSERT(mutex_owned(&myrip->recvlock));
 *
 */

mblk_t *
myri_do_recv(struct myri * myrip, unsigned int length, int index)
{
  mblk_t *m = NULL;
  struct buff_info *infop;

#ifdef MYRI_DEBUG
  ASSERT(mutex_owned(&myrip->myri_recvlock));  
#endif

  CMN_ERR (3, (CE_CONT, "recv'ed message len %d\n", length));

  if (length > 0) {
    infop = (struct buff_info *)myrip->recv_buf_index[index].infop;
    MUTEX_ENTER(&myrip->myri_bufflock);
    myrip->recv_buffer_count--;
    MUTEX_EXIT(&myrip->myri_bufflock);

#ifdef MYRI_DEBUG    
    ASSERT (index >= 0);
    ASSERT (index < NUM_RECV_BUFF1);
#endif

    if (infop->status == BUF_LANAI) {
      /* FIX - SYNC?? */
      if (ddi_dma_sync(infop->handle, 0, length, 
		       DDI_DMA_SYNC_FORKERNEL) != DDI_SUCCESS) {
	CMN_ERR_PRINT(0, (CE_CONT, "DDI_DMA_SYNC_FORKERNEL failed index = %d len = %d\n",
			  infop->index, length));
      }
      
      /* loan it out if possible */
    
    
      if ((length < 128) || 
	  (gm_get_receive_unused(myrip) <= MYRI_LBUF_MIN)) {
      
	if (length >= 128) {
	  myrip->myri_rxbuf_low++;
	}
	/*
	 * not enough buffers at the LANai, will copy
	 * msg
	 */
	m = allocb(length + 2, BPRI_HI);
	if (m) {	  
	  m->b_rptr += 2; /* misalign start of ether header */
	  bcopy((void *) infop->kptr, (void *) m->b_rptr, length);
	  m->b_wptr = (u_char *) ((u_char *) m->b_rptr + length);   
	  /*  m->b_datap->db_type = M_DATA; */
	  
	  infop->status = BUF_UNUSED;
	  myri_add_buff(infop);
	} else {
	  CMN_ERR_PRINT(BUFF_PRINT, (CE_CONT, "do_ecv: failed allocb\n"));
	  infop->status = BUF_UNUSED;
	  myri_add_buff(infop);
	  myrip->myri_rx_allocb_failed++;
	  goto recv_error;
	}
      } else { /* Loan buffer */
	/*
	 * lanai has enough buffers, will loan this
	 * one to higher layers
	 */
	
	
	CMN_ERR (BUFF_PRINT, (CE_CONT, "myri: NO copy (%d>%d)\n",
			     gm_get_receive_unused(myrip),
			     MYRI_LBUF_MIN));

	
	/* this avoids a copy */
	if ((m = esballoc((unsigned char *) infop->kptr, length,
			  BPRI_HI,
			  &infop->free_func))) {

	  m->b_wptr = (u_char *) ((u_char *) m->b_rptr + length);
	  /* m->b_datap->db_type = M_DATA; */
	  infop->status = BUF_LOANED;
	} else {
	  CMN_ERR_PRINT(BUFF_PRINT, (CE_CONT, "do_recv: failed esballoc\n"));
	  myrip->myri_rx_esballoc_failed++;
	  infop->status = BUF_UNUSED;
	  myri_add_buff(infop);
	  goto recv_error;
	}
      }					/* else lanai count OK */
    } else {
      CMN_ERR_PRINT (0, (CE_CONT, "do_recv: bad INDEX  index = %d \n", index));

#if (BUFF_PRINT<=GM_PRINT_LEVEL)
      myri_dump_buff_info(&myrip->recv1info[0]);
#endif							/* BUFF_PRINT */
      goto recv_error;
    }
    
    m->b_cont = NULL;

  } else if ((length == 0) || (!m)) {
    return (NULL);
  }
  
  myrip->myri_ipackets++;

  /* m->b_rptr should now be pointing at the ether header */
  
  return (m);


recv_error:
  myrip->myri_ierrors++;
  
  if (m) {
    freemsg(m);
  }

  CMN_ERR (1, (CE_CONT, "Got to recv_error.\n"));
  return (NULL);
}



void
myri_dump_msg(mblk_t * mp)
{
  if (mp) {
    int len = 0, cnt = 0;
    mblk_t *mp_tmp;
    
    len = msgdsize(mp);

    CMN_ERR(0, (CE_CONT, "   MSG: len = %d bytes\n", len));
    
    mp_tmp = mp;
    while (mp_tmp) {
      CMN_ERR(0, (CE_CONT, "  %d: rptr = 0x%x  len = %d   type = %d\n",
		  cnt++, mp_tmp->b_rptr,
		  (mp_tmp->b_wptr - mp_tmp->b_rptr),
		  mp_tmp->b_datap->db_type));
      if (GM_PRINT_LEVEL >= 0)
	gm_hex_dump(mp_tmp->b_rptr, (mp_tmp->b_wptr - mp_tmp->b_rptr));
      mp_tmp = mp_tmp->b_cont;
    }						/* while */
  }
}

/* end of myri.c */
