/******************************************************************-*-c-*-
 * Myricom GM networking software and documentation			 *
 * Copyright (c) 1999 by Myricom, Inc.					 *
 * All rights reserved.	 See the file `COPYING' for copyright notice.	 *
 *************************************************************************/

/* author: glenn@myri.com */

/************************************************************************
 * This file includes the OS-specific driver code for Solaris.
 ************************************************************************/

/* this is here to avoid gcc-2.8.1 and up problems with varargs */
#ifdef __GNUC__
#define _SYS_VARARGS_H
#include <stdarg.h>
#endif

#include "gm_arch.h"
#include "gm_call_trace.h"
#include "gm_debug_open.h"
#include "gm_lanai.h"
#include "gm_page_hash.h"
#include "gm_pio.h"

#if GM_SUPPORT_PCI
/* #warning Compiled with PCI support */
#include <sys/pci.h>
#endif

#include <sys/mkdev.h>
#include <sys/kmem.h>


#ifndef __GNUC__
#include <sys/varargs.h>
#endif

#include <sys/cmn_err.h>
#include <sys/ddi.h>
#include <sys/sunddi.h>
#include <sys/vnode.h>

#include <sys/utsname.h>

#if GM_SUPPORT_L4 || GM_SUPPORT_L5 
#define GM_ARCH_SOLARIS_ALWAYS_CONSISTENT 1
#else
#define GM_ARCH_SOLARIS_ALWAYS_CONSISTENT !GM_ENABLE_SPARC_STREAMING
#endif

/************************************************************************
 * Forward declarations (used to build tables)
 ************************************************************************/

static int gm_solaris_open (dev_t * devp, int flags, int otyp,
			    cred_t * credp);
static int gm_solaris_close (dev_t dev, int flags, int otyp, cred_t * credp);
#if GM_BC_DRIVER
static int gm_solaris_read (dev_t dev, struct uio *uio_ptr,
			    cred_t * cred_ptr);
#endif
static int gm_solaris_mmap (dev_t dev, off_t off, int prot);
static int gm_solaris_ioctl (dev_t dev, int cmd, intptr_t arg,
			     int mode, cred_t * cred_p, int *rval_p);
static int gm_solaris_segmap (dev_t dev, off_t off, struct as *asp,
			      caddr_t * addrp, off_t len, unsigned int prot,
			      unsigned int maxprot, unsigned int flags,
			      cred_t * cred_p);

static int gm_solaris_identify (dev_info_t * dip);
static int gm_solaris_attach (dev_info_t * dip, ddi_attach_cmd_t cmd);
static int gm_solaris_detach (dev_info_t * dip, ddi_detach_cmd_t cmd);
static int gm_solaris_getinfo (dev_info_t * dip, ddi_info_cmd_t infocmd,
			       void *arg, void **result);

/***********************************************************************
 * Globals
 ***********************************************************************/

#define GM_DRIVER_VERSION       "Myrinet GM Driver "
static char gm_driver_version[40] = {0};

/************
 * Loadable module required data structures
 ************/

static struct cb_ops gm_solaris_cb_ops = {
  gm_solaris_open,		/* open */
  gm_solaris_close,		/* close */
  nodev,			/* not a block driver */
  nodev,			/* no print routine */
  nodev,			/* no dump routine */
#if GM_BC_DRIVER
  gm_solaris_read,		/* read --- for compatibility only */
#else
  nodev,			/* no read routine */
#endif
  nodev,			/* no write routine */
  gm_solaris_ioctl,		/* ioctl */
  nodev,			/* no devmap routine */
  gm_solaris_mmap,		/* mmap */
  gm_solaris_segmap,		/* segmap -- using default */
  nochpoll,			/* no chpoll routine */
  ddi_prop_op,			/* prop_op -- using default */
  NULL,				/* cb_str not a STREAMS driver */
#if GM_SIZEOF_UP_T == 8
  D_64BIT |			/* device handles 64-bit offsets */
#endif
  D_NEW | D_MP,			/* safe for multi-thread/multi-processor */
  CB_REV,			/* revision */
  nodev,			/* no aread routine */
  nodev				/* no awrite routine */
};

static struct dev_ops gm_solaris_ops = {
  DEVO_REV,			/* DEVO_REV indicated by manual  */
  0,				/* device reference count */
  gm_solaris_getinfo,		/* getinfo */
  gm_solaris_identify,		/* identify (contrary to docs, IS needed) */
  nulldev,			/* probe (nulldev self-identifying devices) */
  gm_solaris_attach,		/* attach */
  gm_solaris_detach,		/* detatch */
  nodev,			/* device reset routine -- set to nodev */
  &gm_solaris_cb_ops,
  (struct bus_ops *) NULL,	/* bus operations */
};

static struct modldrv modldrv = {
  &mod_driverops,
  gm_driver_version,
  &gm_solaris_ops,
};

static struct modlinkage modlinkage = {
  MODREV_1,
  {&modldrv, NULL},
};

/************
 * Other globals
 ************/

void *gm_instancep;		/* For per-device structures */

ddi_dma_attr_t gm_dma_attr = {
#ifdef ORIG_ONE
  DMA_ATTR_V0,			/* version */
  0x0,				/* addr_lo */
  0xFFFFFFFF,			/* addr_hi */
  0x7FFFFFFF,			/* count_max */
  0 /* set before used */ ,	/* align (for DMA resources) */
  0x7F,				/* burstsizes */
  1,				/* minxfer */
  0x7FFFFFFF,			/* maxxfer */
  0x7FFFFFFF,			/* seg */
  -1,				/* sgllen */
  1,				/* dma granularity */
  0				/* flags -- must be 0 */
#elif defined BOBS
  DMA_ATTR_V0,			/* version */
  (unsigned long long) 0x0,	/* addr lower */
  (unsigned long long) 0xFFFFFFFF,	/* addr upper */
  (unsigned long long) 0x00FFFFFF,	/* counter */
  (unsigned long long) 0x04,	/* 4-byte word alignment */
  0x7F,				/* burst size */
  4,				/* min-xfer size */
  (unsigned long long) 0x00FFFFFF,	/* max_xfer size */
  (unsigned long long) 0x00FFFFFF,	/* max_seg size */
  20,				/* Gather count */
  4,				/* dma granularity */
  0				/* flags -- must be 0 */
#else
  DMA_ATTR_V0,			/* version */
  0x0,				/* addr_lo */
  0xFFFFFFFF,			/* addr_hi */
  0x7FFFFFFF,			/* count_max */
  0 /* set before used */ ,	/* align (for DMA resources) */
  0x7F,				/* burstsizes */
  1,				/* minxfer */
  0x7FFFFFFF,			/* maxxfer */
  0x7FFFFFFF,			/* seg */
  1,				/* sgllen */
  0,				/* set before used *//* dma granularity */
  0				/* flags -- must be 0 */
#endif
};

struct ddi_device_acc_attr gm_dev_access_attr = {
  DDI_DEVICE_ATTR_V0,		/* version */
  DDI_NEVERSWAP_ACC,		/* endian flash */
  DDI_STRICTORDER_ACC		/* data order */
};

unsigned int gm_arch_solaris_instance_ok[32] = { 0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0
};

/**********************************************************************
 * GM OS-abstraction required functions.
 **********************************************************************/

#if GM_SUPPORT_PCI

/************
 * PCI configuration space access functions.
 ************/

gm_status_t
gm_arch_read_pci_config_32 (gm_instance_state_t * is,
			    gm_offset_t offset, gm_u32_t * value)
{
  *value = pci_config_get32 (is->arch.pci_acc_handle, offset);
  return GM_SUCCESS;
}

gm_status_t
gm_arch_write_pci_config_32 (gm_instance_state_t * is,
			     gm_offset_t offset, gm_u32_t value)
{
  pci_config_put32 (is->arch.pci_acc_handle, offset, value);
  return GM_SUCCESS;
}

/* 16-bit accesses */

gm_status_t
gm_arch_read_pci_config_16 (gm_instance_state_t * is,
			    gm_offset_t offset, gm_u16_t * value)
{
  *value = pci_config_get16 (is->arch.pci_acc_handle, offset);
  return GM_SUCCESS;
}

gm_status_t
gm_arch_write_pci_config_16 (gm_instance_state_t * is,
			     gm_offset_t offset, gm_u16_t value)
{
  pci_config_put16 (is->arch.pci_acc_handle, offset, value);
  return GM_SUCCESS;
}

/* 8-bit accesses */

gm_status_t
gm_arch_read_pci_config_8 (gm_instance_state_t * is,
			   gm_offset_t offset, gm_u8_t * value)
{
  *value = pci_config_get8 (is->arch.pci_acc_handle, offset);
  return GM_SUCCESS;
}

gm_status_t
gm_arch_write_pci_config_8 (gm_instance_state_t * is,
			    gm_offset_t offset, gm_u8_t value)
{
  pci_config_put8 (is->arch.pci_acc_handle, offset, value);
  return GM_SUCCESS;
}
#endif

/************
 * GM synchronization functions
 ************/

/* For solaris, we must be careful to pass an appropriate iblock
   cookie to gm_mutex_init if the mutex is ever user by the interrupt
   handler. */

void
gm_arch_sync_init (gm_arch_sync_t * s, gm_instance_state_t * is)
{
  ddi_iblock_cookie_t _mu_cookie;

#if 1
  if (s == &is->command_sync)
    {
      _mu_cookie = is->arch.iblock_cookie;
      gm_assert (is->arch.iblock_cookie);
    }
  else
    {
      _mu_cookie = 0;
    }
#elif 0
  _mu_cookie = is ? is->arch.iblock_cookie : NULL;
  /* gm_assert (_mu_cookie || !is); */
#else
  _mu_cookie = NULL;
#endif

  gm_bzero (s, sizeof (*s));
  mutex_init (&s->mu, "GM mutex", MUTEX_DRIVER, 0);
  mutex_init (&s->_mu, "GM condvar mutex", MUTEX_DRIVER, _mu_cookie);
  cv_init (&s->cv, "GM condvar", CV_DRIVER, NULL);
  s->tag = GM_ARCH_SYNC_TAG;
}

void
gm_arch_sync_reset (gm_arch_sync_t * s)
{
  s->wake_cnt = 0;
}

static inline
gm_status_t
gm_solaris_check_sync (gm_arch_sync_t *s, const char *function_name)
{
  gm_status_t status = GM_SUCCESS;

#if 0
  /* Check for problems */
  
  if (!s)
    {
      GM_WARN (("%s() called with null pointer\n", function_name));
      status = GM_FAILURE;
    }
  if (s->tag != GM_ARCH_SYNC_TAG)
    {
      GM_WARN (("%s() called with bad pointer %p\n", function_name, s));
      status = GM_FAILURE;
    }

  /* If there was a problem, print the call stack. */

  if (status != GM_SUCCESS)
    {
      GM_WARN (("gm_solaris_check_sync(%p,\"%s\") failed\n"));
    }
#endif

  return status;
}


void
gm_arch_sync_destroy (gm_arch_sync_t * s)
{
  if (gm_solaris_check_sync (s, "gm_arch_sync_destroy") != GM_SUCCESS)
    {
      return;
    }
  cv_destroy (&s->cv);
  mutex_destroy (&s->_mu);
  mutex_destroy (&s->mu);
  s->tag = 0;
}

void
gm_arch_mutex_enter (gm_arch_sync_t * s)
{
  if (gm_solaris_check_sync (s, "gm_arch_mutex_enter") != GM_SUCCESS)
    {
      return;
    }
  mutex_enter (&s->mu);
}

void
gm_arch_mutex_exit (gm_arch_sync_t * s)
{
  if (gm_solaris_check_sync (s, "gm_arch_mutex_exit") != GM_SUCCESS)
    {
      return;
    }
  mutex_exit (&s->mu);
}

/* Sleep functions. Return 0 on wake, -1 on timeout, and 1 on signal. */

/* sleep until awakened or get a signal */

#define GM_DEBUG_SOLARIS_SLEEP 0

gm_arch_sleep_status_t
gm_arch_signal_sleep (gm_arch_sync_t * s)
{
  gm_arch_sleep_status_t ret = GM_SLEEP_WOKE;

  GM_PRINT (GM_DEBUG_SOLARIS_SLEEP,
	    ("gm_arch_signal_sleep() called.\n"));

  mutex_enter (&s->_mu);
  GM_PRINT (GM_DEBUG_SOLARIS_SLEEP,
	    ("gm_arch_signal_sleep() entered mutex.\n"));
  while (s->wake_cnt <= 0)
    {
      if (cv_wait_sig (&s->cv, &s->_mu) == 0)
	{
	  GM_PRINT (GM_DEBUG_SOLARIS_SLEEP,
		    ("gm_arch_signal_sleep returning GM_SLEEP_INTERRUPTED\n"));
	  ret = GM_SLEEP_INTERRUPTED;
	  break;
	}
    }
  if (ret == GM_SLEEP_WOKE)
    s->wake_cnt--;
  mutex_exit (&s->_mu);

  GM_PRINT (GM_DEBUG_SOLARIS_SLEEP,
	    ("gm_arch_signal_sleep() returning.\n"));

  return ret;
}

/* utility function to do either type of timed sleep */

gm_arch_sleep_status_t
gm_arch_timed_sleep (gm_arch_sync_t * s, int seconds)
{
  long timeout_time;
  gm_arch_sleep_status_t ret = GM_SLEEP_WOKE;

  GM_PRINT (GM_DEBUG_SOLARIS_SLEEP,
	    ("gm_arch_timed_sleep() called.\n"));

  drv_getparm (LBOLT, &timeout_time);
  timeout_time += seconds * drv_usectohz (1000000);

  mutex_enter (&s->_mu);
  GM_PRINT (GM_DEBUG_SOLARIS_SLEEP,
	    ("gm_arch_timed_sleep() entered mutex.\n"));
  while (s->wake_cnt <= 0)
    {
      if (cv_timedwait (&s->cv, &s->_mu, timeout_time) == -1)
	{
	  GM_PRINT (GM_DEBUG_SOLARIS_SLEEP,
		    ("gm_arch_timed_sleep returning GM_SLEEP_TIMED_OUT\n"));
	  ret = GM_SLEEP_TIMED_OUT;
	  break;
	}
    }
  s->wake_cnt--;
  mutex_exit (&s->_mu);

  GM_PRINT (GM_DEBUG_SOLARIS_SLEEP,
	    ("gm_arch_timed_sleep() returning.\n"));

  return ret;
}

/* Wake the thread sleeping on the synchronization variable. */

#define GM_DEBUG_SOLARIS_WAKE 0

void
gm_arch_wake (gm_arch_sync_t * s)
{
  GM_PRINT (GM_DEBUG_SOLARIS_WAKE, ("gm_arch_wake() called.\n"));

  mutex_enter (&s->_mu);
  GM_PRINT (GM_DEBUG_SOLARIS_WAKE, ("gm_arch_wake() entered mutex.\n"));
  s->wake_cnt++;
  cv_signal (&s->cv);
  mutex_exit (&s->_mu);

  GM_PRINT (GM_DEBUG_SOLARIS_WAKE, ("gm_arch_wake() returning.\n"));
}

/* Sleep for the indicated duration */

void
gm_arch_spin (gm_instance_state_t * is, gm_u32_t usecs)
{
#if GM_SOLARIS_COMPLICATED_SPIN
  long timeout_time;
  drv_getparm (LBOLT, &timeout_time);
  timeout_time += drv_usectohz (usecs);

  mutex_enter (&is->arch.spin_sync._mu);
  cv_timedwait_sig (&is->arch.spin_sync.cv, &is->arch.spin_sync._mu,
		    timeout_time);
  mutex_exit (&is->arch.spin_sync._mu);
#else
  drv_usecwait (usecs);
#endif
}

gm_status_t gm_arch_gethostname (char *ptr, int len)
{
  gm_strncpy(ptr, utsname.nodename, len - 1);
  ptr[len - 1] = 0;
#if GM_SHORT_HOSTNAME
  {
    char *s;
    /* drop everything from first '.' on to end */
    if (s = strchr(ptr, '.'))
      *s = 0;
  }
#endif
  return GM_SUCCESS;
}

#if GM_CAN_REGISTER_MEMORY

/************
 * Page locking
 ************/

int
gm_solaris_iodone (struct buf *b)
{
  return 0;
}

#define GM_DEBUG_LOCK_USER_BUFFER_PAGE 0

gm_status_t
gm_arch_lock_user_buffer_page (gm_port_state_t *ps, gm_up_t in,
			       gm_dp_t * dma_addr, gm_arch_page_lock_t * lock)
{
  gm_instance_state_t *is;
  uint_t cnt;
  int ddi_flags;
  int status;

  GM_CALLED ();

  gm_assert (GM_PAGE_LEN);
  gm_assert (GM_PAGE_ALIGNED (in));

  gm_assert (ps);
  is = ps->instance;
  gm_assert (is);
  
  /* lock down the user virtual memory addressed by in */
  status = ddi_umem_lock ((caddr_t)in,
			  GM_PAGE_LEN,
			  DDI_UMEMLOCK_READ | DDI_UMEMLOCK_WRITE,
			  &lock->umem_cookie);
  if (status) {
    GM_NOTE (("ddi_umem_lock failed for user memory at 0x%p, %d is returned.\n",
	     in, status));
    goto abort_with_nothing;
  }
  
  /* Map the pages into the kernel, get buf corresponding to umem_cookie. */
  
  lock->buf = ddi_umem_iosetup (lock->umem_cookie, 0, GM_PAGE_LEN, B_READ,
				ps->arch.dev, 0, gm_solaris_iodone,
				DDI_UMEM_NOSLEEP);
  if (lock->buf == NULL) {
    GM_NOTE (("Could not allocate buf for lock page\n"));
    goto abort_with_umem_unlock;
  }

  GM_PRINT (GM_DEBUG_LOCK_USER_BUFFER_PAGE, ("Allocating DMA handle.\n"));

  gm_dma_attr.dma_attr_align = GM_PAGE_LEN;
  gm_dma_attr.dma_attr_granular = GM_PAGE_LEN;
  status = ddi_dma_alloc_handle (is->arch.dip, &gm_dma_attr,
				 DDI_DMA_DONTWAIT, 0, &lock->dma_handle);
  if (status != DDI_SUCCESS) {
    GM_NOTE (("Could not allocate dma handle.\n"));
    goto abort_with_buf;
  }
  
  GM_PRINT (GM_DEBUG_LOCK_USER_BUFFER_PAGE, ("Binding DMA handle.\n"));
  
  ddi_flags = DDI_DMA_RDWR | 
    (GM_ARCH_SOLARIS_ALWAYS_CONSISTENT ? DDI_DMA_CONSISTENT : DDI_DMA_STREAMING);
  status = ddi_dma_buf_bind_handle (lock->dma_handle, lock->buf, ddi_flags,
				    DDI_DMA_DONTWAIT, 0, &lock->dma_cookie, &cnt) ;

  if (status != DDI_SUCCESS) {
    GM_NOTE (("Could not bind DMA handle for vmem at 0x%p.\n", in));
    goto abort_with_dma_handle;
  } 
  GM_PRINT (GM_DEBUG_LOCK_USER_BUFFER_PAGE, ("Performing sanity checks.\n"));
  if (cnt != 1) {
    GM_NOTE (("internal error: too many cookies.\n"));
    goto abort_with_binding;
  }
  if (lock->dma_cookie.dmac_size != GM_PAGE_LEN) {
    GM_NOTE (("internal error: unexpected cookie size.\n"));
    goto abort_with_binding;
  }

  /* Store the lock where we can find it when we want to unlock. */

  GM_PRINT (GM_DEBUG_LOCK_USER_BUFFER_PAGE, ("Storing lock.\n"));
 
  /* Get the DMA addr, by hacking with DMA cookies. */

  GM_PRINT (GM_DEBUG_LOCK_USER_BUFFER_PAGE, ("Reporting success.\n"));
  if (sizeof (gm_dp_t) == 4)
    *dma_addr = (gm_dp_t) lock->dma_cookie.dmac_address;
  else
    *dma_addr = (gm_dp_t) lock->dma_cookie.dmac_laddress;
  GM_RETURN_STATUS (GM_SUCCESS);

abort_with_binding:
  ddi_dma_unbind_handle (lock->dma_handle);
abort_with_dma_handle:
  ddi_dma_free_handle (&lock->dma_handle);
abort_with_buf:
  freerbuf (lock->buf);
abort_with_umem_unlock:
  ddi_umem_unlock (lock->umem_cookie);
abort_with_nothing:
  gm_bzero (lock, sizeof (*lock));
  GM_RETURN_STATUS (GM_FAILURE);
}

void
gm_arch_unlock_user_buffer_page (gm_arch_page_lock_t * lock)
{
  GM_CALLED_WITH_ARGS_EX (GM_DEBUG_LOCK_USER_BUFFER_PAGE,
			  ("%p", lock));
  ddi_dma_unbind_handle (lock->dma_handle);
  ddi_dma_free_handle (&lock->dma_handle);
  freerbuf(lock->buf);
  ddi_umem_unlock (lock->umem_cookie);
  gm_bzero (lock, sizeof (*lock));
  GM_RETURN_NOTHING ();
}

#endif /* GM_CAN_REGISTER_MEMORY */

/************************************************************************
 * Architecture-specific required functions
 ************************************************************************/

void
gm_arch_abort (void)
{
	cmn_err (CE_CONT, "***** don't panic, but gm_arch_abort () was called.****\n");
}


/* Return the value of GM_PAGE_LEN set during driver initialization. */

gm_status_t
gm_arch_get_page_len (unsigned long *page_len)
{
  if (!GM_PAGE_LEN)
    {
      GM_NOTE (("gm_solaris internal error: GM_PAGE_LEN unknown in call\n"
		"to gm_arch_get_page_len().\n"));
      return GM_FAILURE;
    }
  *page_len = GM_PAGE_LEN;
  return GM_SUCCESS;
}

/************
 * gm_port_state initialization 
 ************/

/* This is called just after the port state is created (in gm_minor.c)
   to perform architecture-specific initialization. */

gm_status_t gm_arch_port_state_init (gm_port_state_t * ps)
{
  return GM_SUCCESS;
}

/* This is called just before the port state is destroyed (in
   gm_minor.c) to perform architecture-specific finalization. */

void
gm_arch_port_state_fini (gm_port_state_t * ps)
{
  return;
}


/* This is called just after the port state is initialized (in gm_minor.c)
   to perform architecture-specific initialization. */

gm_status_t
gm_arch_port_state_open (gm_port_state_t * ps)
{
  return GM_SUCCESS;
}

/* This is called just before the port state is finalized (in
   gm_minor.c) to perform architecture-specific finalization. */

void
gm_arch_port_state_close (gm_port_state_t * ps)
{
  return;
}

/************
 * gm_ioctl support
 ************/

/* Copy user data to a kernel buffer. */

gm_status_t
gm_arch_copyin (const void *what, void *where, gm_size_t amount)
{
  int errno;

  errno = copyin ((caddr_t) what, (caddr_t) where, amount);
  return errno ? GM_FAILURE : GM_SUCCESS;
}

/* Copy kernel data to a user buffer */

gm_status_t
gm_arch_copyout (const void *what, void *where, gm_size_t amount)
{
  int errno;

  errno = copyout ((caddr_t) what, (caddr_t) where, amount);
  return errno ? GM_FAILURE : GM_SUCCESS;
}

/************
 * Kernel Memory Allocation
 ************/

/* malloc() analogue for the driver */

void *
__gm_arch_kernel_malloc (unsigned long len, int flags)
{
  gm_u32_t *p;

  /* Allocate extra space at beginning to store length for kmem_free call, and
     return an 8-byte aligned buffer. */
  gm_assert (sizeof (long) <= 8);
  p = (gm_u32_t *) kmem_alloc (8 + len, KM_NOSLEEP);
  if (!p)
    return p;
  p[0] = 8 + len;
  return &p[2];
}

/* free() analogue for the driver */

void
__gm_arch_kernel_free (void *ptr)
{
  gm_u32_t *p;

  p = ptr;
  kmem_free (&p[-2], p[-2]);
}

/************
 * DMA region manipulation
 ************/

/************
 * ??? Required for non cache-coherent I/O architectures, which we
 * do not yet support.
 ************/

gm_status_t gm_arch_dma_region_sync (gm_arch_dma_region_t * r, int command)
{
  int errno;

  errno = ddi_dma_sync (r->dma_handle, 0, 0, command);
  return errno ? GM_FAILURE : GM_SUCCESS;
}

/* Allocate LEN bytes of DMA memory that is contiguous in kernal space
   by possibly segmented in DMA space.

   If r->register_function is non-null, call r->register_page_function
   (r, dma_addr) for each page. */

#define GM_DEBUG_ARCH_DMA_REGION_ALLOC 0

gm_status_t
gm_arch_dma_region_alloc (gm_instance_state_t * is, gm_arch_dma_region_t * r,
			  gm_size_t len, gm_u32_t flags,
			  gm_status_t (*register_page_func) (void *,
							     gm_dp_t,
							     gm_u32_t),
			  void *arg)
{
  gm_size_t alloc_len;
  gm_status_t status;
  int ddi_flags;
  int error;
  size_t trash;
  
  GM_CALLED_WITH_ARGS (("%p,%r,"GM_U64_TMPL",0x%x,%p,%p", is, r,
			GM_U64_ARG (len), flags, register_page_func, arg));

  if (!GM_PAGE_LEN)
    {
      GM_PRINT (GM_DEBUG_ARCH_DMA_REGION_ALLOC, ("GM_PAGE_LEN == 0"));
      status = GM_INTERNAL_ERROR;
      goto abort_with_nothing;
    }

  gm_dma_attr.dma_attr_align = GM_PAGE_LEN;
  gm_dma_attr.dma_attr_granular = GM_PAGE_LEN;
  error = ddi_dma_alloc_handle (is->arch.dip, &gm_dma_attr, DDI_DMA_DONTWAIT,
				NULL, &r->dma_handle);
  if (error != DDI_SUCCESS)
    {
      GM_NOTE (("Could not allocate DMA handle; error = %d.", error));
      if (error == DDI_DMA_BADATTR)
	{
	  GM_NOTE (("Bad DMA attributes."));
	  status = GM_INVALID_PARAMETER;
	}
      else
	{
	  GM_PRINT (GM_DEBUG_ARCH_DMA_REGION_ALLOC,
		    ("could not alloc DMA hanlde"));
	  status = GM_FAILURE;
	}
      goto abort_with_nothing;
    }

  ddi_flags = 0;
  if (flags & GM_ARCH_DMA_READ)
    {
      ddi_flags |= DDI_DMA_READ;
      flags &= ~GM_ARCH_DMA_READ;
    }
  if (flags & GM_ARCH_DMA_WRITE)
    {
      ddi_flags |= DDI_DMA_WRITE;
      flags &= ~GM_ARCH_DMA_WRITE;
    }
  if (flags & GM_ARCH_DMA_CONSISTENT)
    {
      ddi_flags |= DDI_DMA_CONSISTENT;
      flags &= ~GM_ARCH_DMA_CONSISTENT;
    }
  if (flags & GM_ARCH_DMA_STREAMING)
    {
      ddi_flags |= DDI_DMA_STREAMING;
      flags &= ~GM_ARCH_DMA_STREAMING;
    }
  if (flags != 0)
    {
      GM_NOTE (("ddi_dma_mem_alloc: warning - got an unexpected flag\n"));
      error = GM_INTERNAL_ERROR;
      goto abort_with_nothing;
    }

#if GM_ARCH_SOLARIS_ALWAYS_CONSISTENT
  ddi_flags &= ~DDI_DMA_STREAMING;
  ddi_flags |= DDI_DMA_CONSISTENT;
#endif

  alloc_len = GM_PAGE_ROUNDUP (u32, len);
  error = ddi_dma_mem_alloc (r->dma_handle, alloc_len,
			     &gm_dev_access_attr,
			     ddi_flags & ~(DDI_DMA_READ | DDI_DMA_WRITE),
			     DDI_DMA_DONTWAIT, NULL, (caddr_t *) & r->addr,
			     &trash, &r->acc_handle);
  if (error != DDI_SUCCESS)
    {
      GM_NOTE (("Could not allocate DMAable memory.\n"));
      status = GM_OUT_OF_MEMORY;
      goto abort_with_dma_handle;
    }

  r->addr = GM_PAGE_ROUNDUP (ptr, r->addr);

  error = ddi_dma_addr_bind_handle (r->dma_handle, NULL, (caddr_t) r->addr,
				    alloc_len,
				    ddi_flags,
				    DDI_DMA_DONTWAIT, NULL,
				    &r->cookie, &r->cookie_cnt);
  if (error != DDI_SUCCESS)
    {
      GM_NOTE (("Could not bind dma handle."));
      status = GM_UNTRANSLATED_SYSTEM_ERROR;
      goto abort_with_dma_mem;
    }

  /* Compute the DMA status bits for the dma region. */

  r->sts = ddi_dma_burstsizes (r->dma_handle) >> 3 & 0x1ff;

#if GM_SUPPORT_PCI
  /* Verify computed PCI burstsize is legit. */

  if ((is->lanai.eeprom.bus_type == GM_MYRINET_BUS_PCI) && (r->sts != 0xf))
    {
      GM_NOTE (("PCI burst size not computed correctly.\n"));
      status = GM_INTERNAL_ERROR;
      goto abort_with_bound_handle;
    }
#endif

  if (ddi_dma_sync (r->dma_handle, 0, alloc_len, DDI_DMA_SYNC_FORDEV)
      != DDI_SUCCESS)
    {
      GM_NOTE (("Could not sync DMA region for device\n"));
      status = GM_INTERNAL_ERROR;
      goto abort_with_bound_handle;
    }
  
  if (ddi_dma_sync (r->dma_handle, 0, alloc_len, DDI_DMA_SYNC_FORCPU)
      != DDI_SUCCESS)
    {
      GM_NOTE (("Could not sync DMA region for CPU\n"));
      status = GM_INTERNAL_ERROR;
      goto abort_with_sync_for_device;
    }
  
  r->dma_addr = gm_calloc (alloc_len / GM_PAGE_LEN + 1, sizeof (gm_dp_t));
  if (!r->dma_addr)
    {
      status = GM_OUT_OF_MEMORY;
      goto abort_with_sync_for_cpu;
    }
  
  /* Compute the DMA addr of each page and register each page. */

  {
    unsigned cookie_cnt, page_num = 0;
    ddi_dma_cookie_t cookie;
    gm_size_t remaining_len = alloc_len;
    
    /* copy the DMA address for each page into a simple table
       called "r->dma_addr" */

    cookie_cnt = r->cookie_cnt;
    cookie = r->cookie;
    while (cookie_cnt--)
      {
	uint32_t cookie_dmac_address;
	size_t cookie_dmac_size;

	if (remaining_len == 0)
	  {
	    GM_WARN (("internal error parsing DMA cookie"));
	    status = GM_INTERNAL_ERROR;
	    goto abort_with_dma_addrs;
	  }
	
	cookie_dmac_address = cookie.dmac_address;
	cookie_dmac_size = cookie.dmac_size;
	if (!cookie_dmac_address)
	  {
	    GM_PRINT (GM_DEBUG_ARCH_DMA_REGION_ALLOC,
		      ("dmac_address is NULL"));
	    status = GM_INTERNAL_ERROR;
	    goto abort_with_dma_addrs;
	  }
	if (!cookie_dmac_size)
	  {
	    GM_PRINT (GM_DEBUG_ARCH_DMA_REGION_ALLOC,
		      ("dmac_size is zero"));
	    status = GM_INTERNAL_ERROR;
	    goto abort_with_dma_addrs;
	  }
	gm_assert (GM_PAGE_ALIGNED (cookie_dmac_address));
	gm_assert (GM_PAGE_ALIGNED (cookie_dmac_size));
	while (cookie_dmac_size)
	  {
	    r->dma_addr[page_num++] = cookie_dmac_address;
	    cookie_dmac_address += GM_PAGE_LEN;
	    cookie_dmac_size -= GM_PAGE_LEN;
	    remaining_len -= GM_PAGE_LEN;
	  }
	if (cookie_cnt)
	  ddi_dma_nextcookie (r->dma_handle, &cookie);
      }
    if (remaining_len != 0)
      {
	GM_WARN (("internal error parsing DMA cookie"));
	status = GM_INTERNAL_ERROR;
	goto abort_with_dma_addrs;
      }
    
    /* perform the registration callback for each page */

    if (register_page_func != NULL)
      {
	unsigned i;

	for (i=0; i<page_num; i++)
	  {
	    status = register_page_func (arg, r->dma_addr[i], i);
	    if (status != GM_SUCCESS)
	      {
		GM_PRINT (GM_DEBUG_ARCH_DMA_REGION_ALLOC,
			  ("could not register page"));
		goto abort_with_dma_addrs;
	      }
	  }
      }
  }
  
  /* Set this for gm_arch_dma_region_dma_addr_advance */
  
  r->dma_addr_pos = 0;
  
  GM_RETURN_STATUS (GM_SUCCESS);

abort_with_dma_addrs:
  gm_free (r->dma_addr);
abort_with_sync_for_cpu:
abort_with_sync_for_device:
abort_with_bound_handle:
  ddi_dma_unbind_handle (r->dma_handle);
abort_with_dma_mem:
  ddi_dma_mem_free (&r->acc_handle);
abort_with_dma_handle:
  ddi_dma_free_handle (&r->dma_handle);
abort_with_nothing:
  GM_RETURN_STATUS (status);
}

void
gm_arch_dma_region_free (gm_arch_dma_region_t * r)
{
  GM_CALLED ();

  if (!r) {
    cmn_err (CE_CONT, "GM: gm_arch_dma_region_free called with NULL region\n");
    GM_RETURN_NOTHING ();
  }
  gm_free (r->dma_addr);
  GM_PRINT (GM_DEBUG_ARCH_DMA_REGION_ALLOC, ("unbinding\n"));
  ddi_dma_unbind_handle (r->dma_handle);
  GM_PRINT (GM_DEBUG_ARCH_DMA_REGION_ALLOC, ("mem freeing\n"));
  ddi_dma_mem_free (&r->acc_handle);
  GM_PRINT (GM_DEBUG_ARCH_DMA_REGION_ALLOC, ("handle freeing\n"));
  ddi_dma_free_handle (&r->dma_handle);
  GM_PRINT (GM_DEBUG_ARCH_DMA_REGION_ALLOC, ("bzeroing\n"));
  bzero ((void *) r, sizeof (*r));
  GM_RETURN_NOTHING ();
}

/* Return the DMA address of the next page in the DMA region. */

static gm_dp_t
_gm_dma_region_dma_addr (gm_arch_dma_region_t * r, int advance)
{
  gm_dp_t ret;

  gm_assert (r->dma_addr);
  ret = r->dma_addr[r->dma_addr_pos];
  if (advance)
    {
      /* Advance only if we have not advanced to the end, which is marked
	 with a zero sentinel value.  This is done only for safety, since
         this case should never be encountered. */
      if (ret)
	{
	  r->dma_addr_pos++;
	}
      else
	{
	  /*  gm_log_call_trace (); */
	  GM_WARN (("advanced too far: position is %d\n", r->dma_addr_pos));
	}
    }
  return ret;
}

gm_dp_t gm_arch_dma_region_dma_addr (gm_arch_dma_region_t * r)
{
  return _gm_dma_region_dma_addr (r, 0);
}

gm_dp_t
gm_arch_dma_region_dma_addr_advance (gm_arch_dma_region_t * r)
{
  GM_CALLED_WITH_ARGS (("%p", r));
  GM_RETURN (_gm_dma_region_dma_addr (r, 1));
}

/* Return the kernal virtual address for a DMA region */

void *
gm_arch_dma_region_kernel_addr (gm_arch_dma_region_t * r)
{
  return r->addr;
}

/* Return the STS bits for a DMA region. */

gm_s32_t gm_arch_dma_region_status (gm_arch_dma_region_t * r)
{
  return r->sts;
}

/************
 * I/O space mapping
 ************/

#if GM_SUPPORT_PCI

#define GM_DEBUG_PCI_BUS_MASTER 0

gm_status_t
gm_arch_enable_PCI_bus_master_DMA (gm_instance_state_t * is)
{
  GM_PRINT (GM_DEBUG_PCI_BUS_MASTER, ("Enabling PCI bus master DMA.\n"));

  if (!is->arch.pci_acc_handle)
    {
      GM_NOTE (("internal error: is->arch.pci_acc_handle not set.\n"));
      return GM_FAILURE;
    }

  is->ifc.pci.config.Command
    =
    pci_config_get16 (is->arch.pci_acc_handle,
		      GM_OFFSETOF (gm_pci_config_t, Command));
  is->ifc.pci.config.Command |= GM_PCI_COMMAND_MASTER;
  pci_config_put16 (is->arch.pci_acc_handle,
		    GM_OFFSETOF (gm_pci_config_t, Command),
		    is->ifc.pci.config.Command);

  /* Pause for at least 10ms */
  gm_arch_spin (is, 15000);

  return GM_SUCCESS;
}

gm_status_t
gm_arch_disable_PCI_bus_master_DMA (gm_instance_state_t * is)
{
  GM_PRINT (GM_DEBUG_PCI_BUS_MASTER, ("Disabling PCI bus master DMA.\n"));

  if (!is->arch.pci_acc_handle)
    {
      GM_NOTE (("internal error: is->arch.pci_acc_handle not set.\n"));
      return GM_FAILURE;
    }

  is->ifc.pci.config.Command
    =
    pci_config_get16 (is->arch.pci_acc_handle,
		      GM_OFFSETOF (gm_pci_config_t, Command));
  is->ifc.pci.config.Command &= ~GM_PCI_COMMAND_MASTER;
  pci_config_put16 (is->arch.pci_acc_handle,
		    GM_OFFSETOF (gm_pci_config_t, Command),
		    is->ifc.pci.config.Command);

  /* Pause for at least 10ms */
  gm_arch_spin (is, 15000);

  return GM_SUCCESS;
}
#endif /* GM_SUPPORT_PCI */

/* utility function for the following functions.  Returns the Solaris
   register set to use for the board */

#define GM_DEBUG_SOLARIS_REG_SET 0

gm_status_t
gm_solaris_reg_set (gm_instance_state_t * is, int *reg_set)
{
  /* return cached value, if any. */
  if (is->arch.reg_set_cached)
    {
      *reg_set = is->arch.reg_set;
      return GM_SUCCESS;
    }

#if GM_SUPPORT_PCI
#  if GM_SUPPORT_64_BIT_PCI
#    error not yet supported
#  endif
  if (is->lanai.eeprom.bus_type == GM_MYRINET_BUS_PCI)
    {
      /* Determine the register set containing the PCI resource we
         want to map (the memory-mappable part of the interface). We do
         this by scanning the DDI "reg" property of the interface,
         which is an array of gm_ddi_reg_set structures.  */

#define REGISTER_NUMBER(ip) (ip[0] >>  0 & 0xff)
#define FUNCTION_NUMBER(ip) (ip[0] >>  8 & 0x07)
#define DEVICE_NUMBER(ip)   (ip[0] >> 11 & 0x1f)
#define BUS_NUMBER(ip)      (ip[0] >> 16 & 0xff)
#define ADDRESS_SPACE(ip)   (ip[0] >> 24 & 0x03)
#define PCI_ADDR_HIGH(ip)   (ip[1])
#define PCI_ADDR_LOW(ip)    (ip[2])
#define PCI_SPAN_HIGH(ip)   (ip[3])
#define PCI_SPAN_LOW(ip)    (ip[4])

#define GM_DDI_REG_SET_32_BIT_MEMORY_SPACE 2
#define GM_DDI_REG_SET_64_BIT_MEMORY_SPACE 3

      int *data, i, *rs;
      u_int nelementsp;
      char *address_space_name[] = { "Configuration Space",
	"I/O Space",
	"32-bit Memory Space",
	"64-bit Memory Space"
      };


      if (ddi_prop_lookup_int_array (DDI_DEV_T_ANY, is->arch.dip,
				     DDI_PROP_DONTPASS, "reg", &data,
				     &nelementsp) != DDI_SUCCESS)
	{
	  GM_NOTE (("Could not determine register set.\n"));
	  return GM_FAILURE;
	}

      GM_PRINT (GM_DEBUG_SOLARIS_REG_SET,
		("There are %d register sets.\n", nelementsp / 5));
      if (!nelementsp)
	{
	  GM_NOTE (("Didn't find any \"reg\" properties.\n"));
	  ddi_prop_free (data);
	  return GM_FAILURE;
	}

      /* Scan for the register number. */

      GM_PRINT (GM_DEBUG_SOLARIS_REG_SET,
		("*** Scanning for register number.\n"));
      for (i = 0; i < nelementsp / 5; i++)
	{
	  rs = &data[5 * i];

	  GM_PRINT (GM_DEBUG_SOLARIS_REG_SET, ("Examining register set %d:\n", i));
	  GM_PRINT (GM_DEBUG_SOLARIS_REG_SET,
		    ("  Register number = %d.\n", REGISTER_NUMBER (rs)));
	  GM_PRINT (GM_DEBUG_SOLARIS_REG_SET,
		    ("  Function number = %d.\n", FUNCTION_NUMBER (rs)));
	  GM_PRINT (GM_DEBUG_SOLARIS_REG_SET,
		    ("  Device number   = %d.\n", DEVICE_NUMBER (rs)));
	  GM_PRINT (GM_DEBUG_SOLARIS_REG_SET,
		    ("  Bus number      = %d.\n", BUS_NUMBER (rs)));
	  GM_PRINT (GM_DEBUG_SOLARIS_REG_SET,
		    ("  Address space   = %d (%s ).\n", ADDRESS_SPACE (rs),
		     address_space_name[ADDRESS_SPACE (rs)]));
	  GM_PRINT (GM_DEBUG_SOLARIS_REG_SET,
		    ("  pci address 0x%08x %08x\n", PCI_ADDR_HIGH (rs),
		     PCI_ADDR_LOW (rs)));
	  GM_PRINT (GM_DEBUG_SOLARIS_REG_SET,
		    ("  pci span 0x%08x %08x\n", PCI_SPAN_HIGH (rs),
		     PCI_SPAN_LOW (rs)));

	  /* We are looking for a memory property. */

	  if (ADDRESS_SPACE (rs) == GM_DDI_REG_SET_64_BIT_MEMORY_SPACE
	      || ADDRESS_SPACE (rs) == GM_DDI_REG_SET_32_BIT_MEMORY_SPACE)
	    {
	      *reg_set = i;
	      is->arch.reg_set = i;
	      is->arch.reg_set_cached = 1;
	      GM_PRINT (GM_DEBUG_SOLARIS_REG_SET,
			("Memory uses register set %d.\n", *reg_set));

	      /* Round board span up to a power of two to fix a bug */
	      /* in the Ultrasparc version of Solaris 2.6. */
	      is->board_span = 1 << gm_log2_roundup (PCI_SPAN_LOW (rs));

	      is->arch.board_span_cached = 1;
	      GM_PRINT (GM_DEBUG_SOLARIS_REG_SET,
			("Board span is 0x%x\n", (int) is->board_span));
	      break;
	    }
	}

      ddi_prop_free (data);

      /* If no match, fail. */
      if (i >= nelementsp / 5)
	{
	  GM_PRINT (GM_DEBUG_SOLARIS_REG_SET,
		    ("gm_solaris_reg_set returning GM_FAILURE\n"));
	  return GM_FAILURE;
	}

      GM_PRINT (GM_DEBUG_SOLARIS_REG_SET,
		("gm_solaris_reg_set returning GM_SUCCESS\n"));
      return GM_SUCCESS;
    }
#endif

#if GM_SUPPORT_SBUS
  if (is->lanai.eeprom.bus_type == GM_MYRINET_BUS_SBUS)
    {
      *reg_set = is->arch.reg_set = 0;
      is->arch.reg_set_cached = 1;
      return GM_SUCCESS;
    }
#endif

  return GM_FAILURE;
}

gm_status_t
gm_arch_map_io_space (gm_instance_state_t * is, gm_u32_t offset, gm_u32_t len,
		      void **kaddr)
{
  int reg_set;

  if (gm_solaris_reg_set (is, &reg_set) != GM_SUCCESS)
    return GM_FAILURE;

  /* Map the entire board into memory. */

  if (ddi_regs_map_setup (is->arch.dip, reg_set,
			  (caddr_t *) kaddr,
			  offset, len, &gm_dev_access_attr,
			  &is->arch.acc_handle) != DDI_SUCCESS)
    {
      return GM_FAILURE;
    }
  
  return GM_SUCCESS;
}

#define GM_DEBUG_SOLARIS_UNMAP_IO_SPACE 0

void
gm_arch_unmap_io_space (gm_instance_state_t * is, gm_u32_t offset,
			gm_u32_t len, void **kaddr)
{
  int reg_set;

  if (gm_solaris_reg_set (is, &reg_set) != GM_SUCCESS)
    return;

  GM_PRINT (GM_DEBUG_SOLARIS_UNMAP_IO_SPACE,
	    ("Unmapping reg_set = %d  at kaddr 0x%x offset 0x%x, len 0x%x.\n",
	     reg_set, (unsigned int) *kaddr, offset, len));

#if 0
  ddi_unmap_regs (is->arch.dip, reg_set, (caddr_t *) kaddr, offset, len);
#else
  ddi_regs_map_free (&is->arch.acc_handle);
#endif
}

#if GM_SUPPORT_SBUS

/************
 * SBUS eeprom copying
 ************/

/* Platform-dependent code for copying the eeprom information off of a
   SBUS board.  Note that the information stored in the eeprom on SBUS
   boards is accessed as individual SBUS properties in older boards
   and as a byte array in newer cards. */

gm_status_t
gm_arch_sbus_copy_eeprom (gm_instance_state_t * is)
{
  int unit;
  u_int eeprom_len;
  dev_info_t *dip;
  gm_myrinet_eeprom_t *eeprom_ptr = NULL;

  GM_CALLED ();

  gm_assert (is);
  gm_assert (is->arch.dip);
  gm_assert (is->lanai.eeprom.bus_type == GM_MYRINET_BUS_SBUS);

  unit = is->id;
  dip = is->arch.dip;

  /* Try to read the EEPROM as a byte array.  This works for newer boards. */

  if (ddi_prop_lookup_byte_array (DDI_DEV_T_ANY, dip,
				  DDI_PROP_DONTPASS, "myrinet-eeprom-info",
				  (u_char **) & eeprom_ptr, &eeprom_len)
      == DDI_PROP_SUCCESS)
    {
      is->lanai.eeprom = *eeprom_ptr;
      ddi_prop_free (eeprom_ptr);
    }
  else
    {
      /* Could not read the EEPROM as a byte array, so try to
         construct an eeprom from the board properties.  This works
         for older boards. */

      unsigned char *board_id = 0;
      int addrlen = 10;

      GM_INFO (("[%d]: looking for older EEPROM\n", unit));

      is->lanai.eeprom.board_type = 0xffff;
      is->lanai.eeprom.bus_type = GM_MYRINET_BUS_SBUS;

      is->lanai.eeprom.lanai_cpu_version =
	ddi_prop_get_int (DDI_DEV_T_ANY, dip, DDI_PROP_DONTPASS,
			  "cpu_version", 0);
      if (is->lanai.eeprom.lanai_cpu_version == 0)
	{
	  GM_NOTE (("[%d]: cpu_version property not found\n", unit));
	  goto abort_with_nothing;
	}

      is->lanai.eeprom.lanai_clockval =
	ddi_prop_get_int (DDI_DEV_T_ANY, dip, DDI_PROP_DONTPASS,
			  "clock_value", 0);
      if (is->lanai.eeprom.lanai_clockval == 0)
	{
	  GM_NOTE (("[%d]: clockval property not found\n", unit));
	  goto abort_with_nothing;
	}

      is->lanai.eeprom.lanai_sram_size =
	ddi_getprop (DDI_DEV_T_ANY, dip, DDI_PROP_DONTPASS, "sram_size", 0);
      if (is->lanai.eeprom.lanai_sram_size == 0)
	{
	  GM_NOTE (("[%d]: sram_size property not found\n", unit));
	  goto abort_with_nothing;
	}

      if (ddi_prop_lookup_byte_array (DDI_DEV_T_ANY, dip, DDI_PROP_DONTPASS,
				      "myrinet-board-id",
				      &board_id,
				      (u_int *) & addrlen)
	  != DDI_PROP_SUCCESS)
	{
	  GM_NOTE (("[%d]: board_id property not found\n", unit));
	  goto abort_with_nothing;
	}

      bcopy (&board_id[0], &is->lanai.eeprom.lanai_board_id[0], 6);
      ddi_prop_free (board_id);
    }

  GM_RETURN_STATUS (GM_SUCCESS);

abort_with_nothing:
  GM_RETURN_STATUS (GM_FAILURE);
}
#endif

/************************************************************************
 * Solaris-specific functions (not required by GM OS-abstraction)
 ************************************************************************/

#if 0
/* Get an iblock cookie just like ddi_get_iblock_cookie, only
   compatible with earlier versions of Solaris. */

gm_status_t
gm_solaris_get_iblock_cookie (dev_info_t * dip, u_int inumber,
			      ddi_iblock_cookie_t * iblock_cookie)
{
// #define GM_SOLARIS_NEW_STYLE_INTERRUPTS
#if GM_SOLARIS_NEW_STYLE_INTERRUPTS
  return ddi_get_iblock_cookie (dip, inumber, iblock_cookie);
#else
  ddi_idevice_cookie_t foo;

  if (ddi_add_ntr (dip, 0, iblock_cookie, &foo,
		   (u_int (*)(caddr_t)) nulldev, NULL) != DDI_SUCCESS)
    {
      GM_NOTE (("could not add null interrupt handler."));
      goto abort_with_nothing;
    }
  return GM_SUCCESS;

abort_with_nothing:
  return GM_FAILURE;
#endif
}
#endif

/************
 * Utility functions for generic error printing
 ************/

/* initialized in _init. */
kmutex_t gm_solaris_print_mu;
kmutex_t gm_solaris_driver_mutex;

void
gm_solaris_print (char *format, ...)
{
  va_list ap;
  /* mutex_enter (&gm_solaris_print_mu); */
  va_start (ap, format);
  vcmn_err (CE_CONT, format, ap);
  va_end (ap);

  /* mutex_exit (&gm_solaris_print_mu); */
/*   drv_usecwait (100000); */
}

void
gm_solaris_info (char *format, ...)
{
  va_list ap;
  /* mutex_enter (&gm_solaris_print_mu); */
  va_start (ap, format);
  vcmn_err (CE_CONT, format, ap);
  va_end (ap);

  /* mutex_exit (&gm_solaris_print_mu); */
/*   drv_usecwait (100000); */
}

void
gm_solaris_warn (char *format, ...)
{
  va_list ap;
  /* mutex_enter (&gm_solaris_print_mu); */
  va_start (ap, format);
  vcmn_err (CE_WARN, format, ap);
  va_end (ap);
  /* mutex_exit (&gm_solaris_print_mu); */
/*   drv_usecwait (100000); */
}

void
gm_solaris_note (char *format, ...)
{
  va_list ap;
  /* mutex_enter (&gm_solaris_print_mu); */
  va_start (ap, format);
  vcmn_err (CE_NOTE, format, ap);
  va_end (ap);
  /* mutex_exit (&gm_solaris_print_mu); */
/*   drv_usecwait (100000); */
}

void
gm_solaris_panic (char *format, ...)
{
  va_list ap;
/*   mutex_enter (&gm_solaris_print_mu); */
  va_start (ap, format);
  vcmn_err (CE_CONT, format, ap);
  va_end (ap);
  cmn_err (CE_CONT, "YIKES - gm_solaris_panic called");
  cmn_err (CE_CONT, "      - decided not to panic this time, system will be unstable!");
  gm_log_call_trace ();
/*   mutex_exit (&gm_solaris_print_mu); */
/*cmn_err (CE_PANIC, "(gm panic)"); */
/*   drv_usecwait (100000); */
}


/************************************************************************
 * Module entry points    (called via dev_ops structure)
 ************************************************************************/

#define GM_DEBUG_SOLARIS_ATTACH 0

static int
gm_solaris_attach (dev_info_t * dip, ddi_attach_cmd_t cmd)
{
  gm_instance_state_t *is;
  int instance;
  enum gm_bus_type bus_type;

  GM_CALLED ();
  mutex_enter (&gm_solaris_driver_mutex);
  
  GM_INFO (("GM driver version %s build %s\n", 
	_gm_version, _gm_build_id));

  /* Verify command. */

  if (cmd != DDI_ATTACH)
    goto abort_with_driver_mutex;

   if (gm_driver_version[0] == 0) {
    (void) strcpy (gm_driver_version, GM_DRIVER_VERSION);   
    (void) gm_strncpy (gm_driver_version + sizeof (GM_DRIVER_VERSION) - 1, 
		       _gm_version, 40 - sizeof(GM_DRIVER_VERSION) - 1);
    gm_driver_version[39] = '\0';
  }

  /* set the page length ASAP, as it is used allover. */

  if (!GM_PAGE_LEN)
    for (GM_PAGE_LEN = 4096; !ddi_btop (dip, GM_PAGE_LEN); GM_PAGE_LEN <<= 1);
  GM_INFO (("GM_PAGE_LEN (page size) is %d\n", GM_PAGE_LEN));

  instance = ddi_get_instance (dip);

  /* Verify that board can be DMA master. */

  if (ddi_slaveonly (dip) == DDI_SUCCESS)
    {
      GM_NOTE (("Myrinet board #%d in slave-only slot cannot be used.",
		instance));
      goto abort_with_driver_mutex;
    }

  /* Allocate the "instance state" for the interface (as Solaris soft state) */

  if (ddi_soft_state_zalloc (gm_instancep, instance) != DDI_SUCCESS)
    {
      GM_NOTE (("myrinet myri[%d]: soft state alloc failed", instance));
      goto abort_with_driver_mutex;
    }
  is = ddi_get_soft_state (gm_instancep, instance);

  /* Add the interrupt handler now that we know what IS is.  This will
     set is->arch.iblock_cookie so that we can properly initialize any
     mutex that depends on it.

     HACK: The interrupt handler will crash the machine if it called
     before the instance state is initialized.  We install it here
     anyway to get the iblock cookie, and then enable interrupts only
     after the instance is initialized. */


  /* Verify that high-level interrupts are supported */

  if (ddi_intr_hilevel (dip, 0) != 0)
    {
      GM_NOTE (("High level interrupts not supported"));
      goto abort_with_instance_state;
    }

  ddi_get_iblock_cookie (dip, 0, &is->arch.iblock_cookie);

  is->id = instance;
  is->arch.dip = dip;
#if GM_SOLARIS_COMPLICATED_SPIN
  gm_arch_sync_init (&is->arch.spin_sync, is);
#endif

  /* Determine the ifc type. */

  do
    {
#if GM_SUPPORT_PCI
      if ((gm_strcmp (ddi_get_name (dip), "pci10e8,8043") == 0)
	  || (gm_strcmp (ddi_get_name (dip), "pci14c1,8043") == 0)
	  || (gm_strcmp (ddi_get_name (dip), "pci103c,106b") == 0)
	  || (gm_strcmp (ddi_get_name (dip), "pci14c1,0") == 0))
	{
	  bus_type = GM_MYRINET_BUS_PCI;
	  break;
	}
#endif /* GM_SUPPORT_PCI */

#if GM_SUPPORT_SBUS
      if (gm_strcmp (ddi_get_name (dip), "MYRICOM,mlanai") == 0)
	{
	  bus_type = GM_MYRINET_BUS_SBUS;
	  break;
	}
#endif

      GM_NOTE (("[%d]: name NOT matched (%s)", instance, ddi_get_name (dip)));

      /* "if (1)" suppresses Solaris "end-of-loop not reached" warning */
      if (1) goto abort_with_spin_sync;
    }
  while (0);

  GM_PRINT (GM_DEBUG_SOLARIS_ATTACH,
	    ("[%d]: name matched (%s)\n", instance, ddi_get_name (dip)));

#if GM_SUPPORT_PCI
  /* Setup access to PCI configurations space before calling
     gm_instance_init(). */

  if (bus_type == GM_MYRINET_BUS_PCI)
    {
      /* Enable access to PCI configuration space */
      if (pci_config_setup (dip, &is->arch.pci_acc_handle) != DDI_SUCCESS)
	{
	  GM_NOTE (("Could not setup access to PCI config space.\n"));
	  goto abort_with_spin_sync;
	}
    }
#endif

  /* Initialize the interface, including mapping the board regions,
     copying eeprom, and starting the LANai control program (with
     interrupts disabled.) */


  if (gm_instance_init (is, instance, bus_type) != GM_SUCCESS)
    {
      GM_NOTE (("[%d]: Could not initialize instance.\n", instance));
      goto abort_with_initialized_config_state;
    }
  GM_INFO (("gm_instance_init instance=%d completed successfully\n",
	     instance));
  gm_arch_solaris_instance_ok[instance]++;

  /*** Set up interrupt handler for the device. ***/

#if GM_SUPPORT_SBUS

  /* Set the SBUS ifc interrupt level */

  if (is->lanai.eeprom.bus_type == GM_MYRINET_BUS_SBUS)
    {
      int int_level;

      int_level = ddi_getprop (DDI_DEV_T_ANY, dip, DDI_PROP_DONTPASS,
			       "interrupts", 0);
      if (int_level == 0)
	int_level = 4;

      if (is->lanai.eeprom.lanai_cpu_version >= 0x400)
	{
	  /* set the board interrupt level from the PROM value */

	  GM_INFO (("[%d]: Setting LANai board[%d] to "
		    "interrupt level = %d  0x%x\n",
		    instance, instance, int_level, 1 << int_level));

#if GM_NO_PARTWORD_PIO_WRITE
#error SBUS card and GM_NO_PARTWORD_PIO_WRITE are incompatible.
#endif
	  ((unsigned short *) is->lanai.control_regs)[1]
	    = GM_N(gm_hton_u16(1 << int_level));

	  GM_PRINT (GM_DEBUG_SOLARIS_ATTACH,
		    ("[%d]: LANAI_CONTROL[instance][1] = (short)0x%x\n",
		     instance,
		     ((unsigned short *) is->lanai.control_regs)[1]));

	  GM_PRINT (GM_DEBUG_SOLARIS_ATTACH,
		    ("[%d]: Setting LANai board[%d] to interrupt "
		     "level = %d  0x%x AGAIN \n", instance, instance,
		     int_level, 1 << int_level));

#if GM_NO_PARTWORD_PIO_WRITE
#error SBUS card and GM_NO_PARTWORD_PIO_WRITE are incompatible.
#endif
	  ((unsigned short *) is->lanai.control_regs)[1]
	    = GM_N(gm_hton_u16(1 << int_level));

	  GM_INFO (("[%d]: LANAI_CONTROL[instance] = (int)0x%x \n", instance,
		    *(unsigned int *) &((unsigned short *) is->lanai.
					control_regs)[0]));
	}
    }
#endif /* GM_SUPPORT_SBUS */

  /* enable interrupts.  HACK: This must be done only after it is safe
     to call the interrupt handler for this device, which we installed
     earlier. */

  if (ddi_add_intr (dip, 0, &is->arch.iblock_cookie, 0,
		    (u_int (*)(caddr_t)) gm_intr, (void *) is) != GM_SUCCESS)
    {
      GM_NOTE (("cannot add interrupt handler."));
      goto abort_with_initialized_instance;
    }

  GM_PRINT (GM_DEBUG_SOLARIS_ATTACH,
	    ("Enabling interrupts for instance %d.\n", is->id));
  if (gm_initialize_interrupts (is) != GM_SUCCESS)
    {
      GM_NOTE (("Cannot enable interrupts.\n"));
      goto abort_with_initialized_instance;
    }

#if 0
  GM_PRINT (1, ("debug spinning\n"));
  gm_arch_spin (is, 10000000);
  GM_PRINT (1, ("debug pausing\n"));
  gm_pause_lanai (is);
  GM_PRINT (1, ("debug unpausing\n"));
  gm_unpause_lanai (is);
  GM_PRINT (1, ("debug done (un)pausing\n"));
#endif

  /************
   * Create minor devices.  /dev/gm%d is the clone device for normal users,
   * and /dev/gmp%d is the clone device for privileged users.
   ************/

  /* Determine minor node number to use for /dev/gm? . */

  GM_PRINT (GM_DEBUG_SOLARIS_ATTACH, ("Allocating minor number.\n"));
  if (gm_minor_alloc (is, &is->clone_minor) != GM_SUCCESS)
    {
      GM_NOTE (("Failed to allocate minor number.\n"));
      goto abort_with_enabled_interrupts;
    }
  GM_PRINT (GM_DEBUG_SOLARIS_ATTACH,
	    ("using minor number %ld.\n", (long) is->clone_minor));
  gm_minor_get_port_state (is->clone_minor)->instance = is;
  gm_minor_get_port_state (is->clone_minor)->privileged = 0;

  /* Determine minor node number to use for /dev/gmp? . */

  GM_PRINT (GM_DEBUG_SOLARIS_ATTACH, ("Allocating minor number.\n"));
  if (gm_minor_alloc (is, &is->privileged_clone_minor) != GM_SUCCESS)
    {
      GM_NOTE (("Failed to allocate minor number.\n"));
      goto abort_with_minor;
    }
  GM_PRINT (GM_DEBUG_SOLARIS_ATTACH, ("using minor number %ld.\n",
				      (long) is->privileged_clone_minor));
  gm_minor_get_port_state (is->privileged_clone_minor)->instance = is;
  gm_minor_get_port_state (is->privileged_clone_minor)->privileged = 1;

  /* Create the minor node (/dev/gm?) */
  {
    char name[20];

    /* Create /devices/gm* clone device. */

    sprintf (name, "gm%d", instance);
    if (ddi_create_minor_node (dip, name, S_IFCHR, is->clone_minor,
			       DDI_PSEUDO, 0) != DDI_SUCCESS)
      {
	GM_NOTE (("ddi_create_minor_node_failed for unit %d.", instance));
	goto abort_with_minor2;
      }

    /* Create /devices/gmp* privileged clone device. */

    sprintf (name, "gmp%d", instance);
    if (ddi_create_minor_node (dip, name, S_IFCHR, is->privileged_clone_minor,
			       DDI_PSEUDO, 0) != DDI_SUCCESS)
      {
	GM_NOTE (("ddi_create_minor_node_failed for unit %d.", instance));
	goto abort_with_clone_minor_node;
      }

#if GM_BC_DRIVER
#error broken minor number details
    /* Create /devices/mlanai* for backwards-compatibility. */

    sprintf (name, "mlanai%x", instance);
    if (ddi_create_minor_node
	(dip, name, S_IFCHR,
	 GM_INSTANCE_TO_COMPATIBILITY_DEVICE_MINOR (instance),
	 DDI_PSEUDO, 0) != DDI_SUCCESS)
      {
	GM_NOTE (("ddi_create_minor_node_failed for unit %d.", instance));
	goto abort_with_privileged_minor_node;
      }
#endif
  }

  GM_PRINT (GM_DEBUG_SOLARIS_ATTACH, ("created minor nodes\n"));

  /* Put pointer to private info into dip. */

  ddi_set_driver_private (dip, (caddr_t) is);

  /* Report attachment */

  ddi_report_dev (dip);

  GM_PRINT (GM_DEBUG_SOLARIS_ATTACH, ("Reported attachment.  Done.\n"));

  mutex_exit (&gm_solaris_driver_mutex);
  GM_RETURN_INT (DDI_SUCCESS);

  /********************* Error handling */

abort_with_privileged_clone_minor_node:
abort_with_clone_minor_node:
  ddi_remove_minor_node (dip, NULL);
abort_with_minor2:
  gm_minor_free (is->privileged_clone_minor);
abort_with_minor:
  gm_minor_free (is->clone_minor);
abort_with_enabled_interrupts:
  gm_finalize_interrupts (is);
  ddi_remove_intr (dip, 0, is->arch.iblock_cookie);
abort_with_initialized_instance:
  gm_instance_finalize (is);
abort_with_initialized_config_state:
  if (is->lanai.eeprom.bus_type == GM_MYRINET_BUS_PCI)
    pci_config_teardown (&is->arch.pci_acc_handle);
abort_with_spin_sync:
#if GM_SOLARIS_COMPLICATED_SPIN
  gm_arch_sync_destroy (&is->arch.spin_sync);
#endif
abort_with_instance_state:
  ddi_soft_state_free (gm_instancep, instance);
abort_with_driver_mutex:
  mutex_exit (&gm_solaris_driver_mutex);
  GM_RETURN_INT (DDI_FAILURE);
}

/* Unattach everything in reverse order */

#define GM_DEBUG_SOLARIS_DETACH 0

static int
gm_solaris_detach (dev_info_t * dip, ddi_detach_cmd_t cmd)
{
  gm_instance_state_t *is;
  int instance;

  GM_CALLED ();
  if (cmd != DDI_DETACH)
    GM_RETURN_INT (DDI_FAILURE);

  mutex_enter (&gm_solaris_driver_mutex);
  GM_PRINT (GM_DEBUG_SOLARIS_DETACH, ("Calling ddi_get_instance()\n"));
  instance = ddi_get_instance (dip);
  GM_PRINT (GM_DEBUG_SOLARIS_DETACH, ("Called ddi_get_instance()\n"));
  GM_PRINT (GM_DEBUG_SOLARIS_DETACH, ("Calling ddi_get_soft_state()\n"));
  is = ddi_get_soft_state (gm_instancep, instance);
  GM_PRINT (GM_DEBUG_SOLARIS_DETACH, ("Called ddi_get_soft_state()\n"));

  GM_PRINT (GM_DEBUG_SOLARIS_DETACH, ("Calling ddi_remove_minor_node\n"));
  ddi_remove_minor_node (dip, NULL);
  GM_PRINT (GM_DEBUG_SOLARIS_DETACH, ("Called ddi_remove_minor_node\n"));
  gm_minor_free (is->privileged_clone_minor);
  gm_minor_free (is->clone_minor);
  gm_finalize_interrupts (is);
  GM_PRINT (GM_DEBUG_SOLARIS_DETACH, ("Calling ddi_remove_intr()\n"));
  ddi_remove_intr (dip, 0, is->arch.iblock_cookie);
  GM_PRINT (GM_DEBUG_SOLARIS_DETACH, ("Called ddi_remove_intr()\n"));
  gm_instance_finalize (is);
  if (is->lanai.eeprom.bus_type == GM_MYRINET_BUS_PCI)
    {
      GM_PRINT (GM_DEBUG_SOLARIS_DETACH, ("Calling pci_config_teardown()\n"));
      pci_config_teardown (&is->arch.pci_acc_handle);
      GM_PRINT (GM_DEBUG_SOLARIS_DETACH, ("Called pci_config_teardown()\n"));
    }
#if GM_SOLARIS_COMPLICATED_SPIN
  gm_arch_sync_destroy (&is->arch.spin_sync);
#endif
  GM_PRINT (GM_DEBUG_SOLARIS_DETACH, ("Calling ddi_soft_state_free()\n"));
  ddi_soft_state_free (gm_instancep, instance);
  GM_PRINT (GM_DEBUG_SOLARIS_DETACH, ("Called ddi_soft_state_free()\n"));
  gm_arch_solaris_instance_ok[instance] = 0;

  mutex_exit (&gm_solaris_driver_mutex);
  GM_RETURN_INT (DDI_SUCCESS);
}

/* Perform minor device number to device instance mapping for the
   kernel. */

static int
gm_solaris_getinfo (dev_info_t * dip, ddi_info_cmd_t infocmd, void *arg,
		    void **result)
{
  gm_instance_state_t *is;
  gm_port_state_t *ps;
  dev_t dev;
  int status = DDI_FAILURE;

  GM_CALLED ();

  mutex_enter (&gm_solaris_driver_mutex);
  switch (infocmd)
    {
    case DDI_INFO_DEVT2INSTANCE:
      dev = (dev_t) arg;
      ps = gm_minor_get_port_state (getminor (dev));
      if (!ps)
	goto abort_with_driver_mutex;
      is = ps->instance;
      if (!is)
	goto abort_with_driver_mutex;
      *result = (void *) is->id;
      mutex_exit (&gm_solaris_driver_mutex);
      GM_RETURN_INT (DDI_SUCCESS);

    case DDI_INFO_DEVT2DEVINFO:
      dev = (dev_t) arg;
      ps = gm_minor_get_port_state (getminor (dev));
      if (!ps)
	goto abort_with_driver_mutex;
      is = ps->instance;
      if (!is)
	goto abort_with_driver_mutex;
      *result = (void *) is->arch.dip;
      mutex_exit (&gm_solaris_driver_mutex);
      GM_RETURN_INT (DDI_SUCCESS);
    }
 abort_with_driver_mutex:
  mutex_exit (&gm_solaris_driver_mutex);
  GM_RETURN_INT (status);
}

static int
gm_solaris_localize_status (gm_status_t status)
{
#define CASE(from,to) case from : return to
  switch (status)
    {
      CASE (GM_SUCCESS, 0);
      CASE (GM_INPUT_BUFFER_TOO_SMALL, EFAULT);
      CASE (GM_OUTPUT_BUFFER_TOO_SMALL, EFAULT);
      CASE (GM_TRY_AGAIN, EAGAIN);
      CASE (GM_BUSY, EBUSY);
      CASE (GM_MEMORY_FAULT, EFAULT);
      CASE (GM_INTERRUPTED, EINTR);
      CASE (GM_INVALID_PARAMETER, EINVAL);
      CASE (GM_OUT_OF_MEMORY, ENOMEM);
      CASE (GM_INVALID_COMMAND, EINVAL);
      CASE (GM_PERMISSION_DENIED, EPERM);
      CASE (GM_INTERNAL_ERROR, EPROTO);
      CASE (GM_UNATTACHED, EUNATCH);
      CASE (GM_UNSUPPORTED_DEVICE, ENXIO);
      CASE (GM_NO_SUCH_DEVICE, ENODEV);
    default:
      return gm_solaris_localize_status (GM_INTERNAL_ERROR);
    }
}

/****************************************************************
 * Character device driver entry points.
 ****************************************************************/

/* open a device.  For GM, this is a clone device, and we don't
   do much initialization here, other than to allocate a new minor number
   and uninitialized gm_port_state.  Later, a GM_SET_PORT_NUMBER ioctl
   will allow the gm_port_state to be fully initialized. */

#define GM_DEBUG_SOLARIS_OPEN 0

static int
gm_solaris_open (dev_t *devp, int flags, int otyp, cred_t * credp)
{
  dev_t dev;
  gm_port_state_t *clone_ps;
  gm_port_state_t *ps;
  gm_status_t status;
  int error;
  minor_t minor;
  
  GM_CALLED ();

  if (otyp != OTYP_CHR)
    {
      GM_NOTE (("gm_solaris_open() called with bad otyp.\n"));
      status = GM_INVALID_PARAMETER;
      goto abort_with_nothing;
    }

  mutex_enter(&gm_solaris_driver_mutex);
  
  /* Determine which port state has been opened. */

  clone_ps = gm_minor_get_port_state (getminor (*devp));
  if (!clone_ps)
    {
      GM_NOTE (("GM: open: gm_minor_get_port_state_1 returns NULL.\n"));
      status = GM_NO_SUCH_DEVICE;
      goto abort_with_mutex;
    }

  /* Do nothing when opening the slave device. */
  
  if (GM_SOLARIS_IS_SLAVE (clone_ps))
    {
      status = GM_SUCCESS;
      goto abort_with_mutex;
    }

  /* Determine an unused minor number for the opened device. */

  if (gm_minor_alloc (clone_ps->instance, &minor) != GM_SUCCESS)
    {
      GM_NOTE (("Could not create new minor number.\n"));
      status = GM_NO_SUCH_DEVICE;
      goto abort_with_mutex;
    }
  GM_PRINT (GM_DEBUG_SOLARIS_OPEN, ("Using minor number %d.\n", minor));
  ps = gm_minor_get_port_state (minor);
  if (!ps) {
      GM_NOTE (("GM: open: gm_minor_get_port_state_2 returns NULL.\n"));
      status = GM_NO_SUCH_DEVICE;
      goto abort_with_minor;
  }
  gm_assert (ps);

  /* Record the instance associated with this minor number. */

  ps->instance = clone_ps->instance;
  if (!ps->instance) {
      GM_NOTE (("GM: open: ps->instance is NULL.\n"));
      status = GM_NO_SUCH_DEVICE;
      goto abort_with_minor;
  }
  gm_assert (ps->instance);

  /* Record the privileges from the clone device */

  ps->privileged = clone_ps->privileged;

  /* Signal OS to make new device with the same major number and a
     new minor number. */

  dev = makedevice (getmajor (*devp), minor);
  if (dev == NODEV)
    {
      GM_NOTE (("Could not make device number.\n"));
      status = GM_OUT_OF_MEMORY;
      goto abort_with_minor;
    }
  *devp = dev;

  /* Allocate a port state for the a slave vnode with which to
     associate all memory mappings.  This association is required
     because when we have memory registered via ddi_umem_lock() as
     well as PCI mapped memory in Solaris-2.8, the OS will deadlock
     trying to unmap the registered memory before calling our
     gm_solaris_close() entry point.  By associating the PCI mapping
     with another device, we trick Solaris into closing the original
     device early enough for a "unclean" exit: before it will look at
     registered memory (or PCI mapping).
     This gives us a hook (the gm_solaris_close() call for the master
     vnode) where we can unpin all registered memory.  Only after we
     have unregisted all registered memory can the final removal of
     mappings (part of process exiting) be done by Solaris-2.8 without
     deadlocking. The gm_solaris_close() call for the slave device
     will be done after all PCI mappings are removed, and the master
     device has closed the slave device */

  status = gm_minor_alloc (ps->instance, &ps->arch.slave_minor);
  if (status != GM_SUCCESS)
    {
      goto abort_with_minor;
    }
  
  /* Mark the port state as a slave by setting its arch.master_ps.
     This must be done before opening the slave device, since
     gm_solaris_open() checks for this marking to determine if the
     opened device is a slave device. */
     
  gm_minor_get_port_state (ps->arch.slave_minor)->arch.master_ps = ps;

  /* Create a slave device with wich to associate memory mappings.  This
     handy slave device has the same dispatch table as normal devices,
     so the OS will call gm_solaris_mmap() when we use a hack to
     mmap() using this device, and will call gm_solaris_close() to free
     these mappings when the application closes. */

  ps->arch.dev = dev;
  ps->arch.slave_vp = makespecvp (makedevice (getmajor (dev),
					      ps->arch.slave_minor),
				  VCHR);
  if (ps->arch.slave_vp == 0)
    {
      status = GM_FAILURE;
      goto abort_with_minor2;
    }

  /* Open the slave device so that we can use it safely with ddi_segmap().
     We must free the driver_mutex while we do this since VOP_OPEN() calls
     gm_solaris_open() for the slave device. */
  
  mutex_exit (&gm_solaris_driver_mutex);
  if (VOP_OPEN (&ps->arch.slave_vp, flags, credp) != 0)
    {
      mutex_enter (&gm_solaris_driver_mutex);
      status = GM_UNTRANSLATED_SYSTEM_ERROR;
      goto abort_with_slave_vp;
    }
  mutex_enter (&gm_solaris_driver_mutex);

  /* Success, so commit */

  GM_PRINT (GM_DEBUG_SOLARIS_OPEN,
	    ("User opened device with minor number 0x%x.\n", getminor (dev)));
  mutex_exit (&gm_solaris_driver_mutex);
  GM_RETURN_INT (gm_solaris_localize_status (GM_SUCCESS));

 abort_with_slave_vp:
  VN_RELE (ps->arch.slave_vp);
 abort_with_minor2:
  gm_minor_free (ps->arch.slave_minor);
 abort_with_minor:
  gm_minor_free (minor);
 abort_with_mutex:
  mutex_exit(&gm_solaris_driver_mutex);
 abort_with_nothing:
  GM_RETURN_INT (gm_solaris_localize_status (status));
}

#define GM_DEBUG_SOLARIS_CLOSE 0

static int
gm_solaris_close (dev_t dev, int flags, int otyp, cred_t * credp)
{
  gm_port_state_t *ps;
  struct vnode *slave_vp;
  int error;
  
  GM_CALLED ();

  /* Make sure no other thead is accessing the global structures used
     by gm_minor_get_port_state() */
  
  mutex_enter (&gm_solaris_driver_mutex);

  /* Get a pointer to the port state for the port being closed. */

  ps = gm_minor_get_port_state (getminor (dev));
  if (!ps)
    {
      GM_PRINT (GM_DEBUG_SOLARIS_CLOSE,
		("Closing slave device after master.\n"));
      mutex_exit (&gm_solaris_driver_mutex);
      GM_RETURN_INT (0);
    }

  /* Get exclusive access to the port state. */
  
  gm_arch_mutex_enter (&ps->sync);

  /****************
   * Catch slave closure , will occur either during the end of master
   * closing (when calling VOP_CLOSE on the slave), or even later, if
   * the PCI mappings are still present (in this case the slave close
   * occurs when the OS automatically remove all mappings).
   ****************/
  
  if (GM_SOLARIS_IS_SLAVE (ps))
    {
      gm_port_state_t *master_ps;

      GM_PRINT (GM_DEBUG_SOLARIS_CLOSE, ("Closing slave device.\n"));
      
      master_ps = ps->arch.master_ps;
      gm_assert (master_ps);

      /* closing the port should have been done by the master, we just
	 need to free the master and slave minors, and exit */

      gm_arch_mutex_exit (&ps->sync);

      /* Destroy the port_state structure */

      gm_assert (getminor (dev) == ps->minor);
      gm_assert (master_ps->arch.slave_minor == ps->minor);
      gm_minor_free (master_ps->arch.slave_minor);
      gm_minor_free (master_ps->minor);

      mutex_exit (&gm_solaris_driver_mutex);
      GM_RETURN_INT (0);
    }

  /****************
   * Close a master device.  This just undoes gm_solaris_open().
   ****************/

  GM_PRINT (GM_DEBUG_SOLARIS_CLOSE,
	    ("Closing master device.\n"));
  gm_assert (ps);

  /* First disable DMAs for the port, this will unregister memory
     which will allow the process to complete the exit steps
     successfully (in particular the destruction of the address space) */

  gm_port_state_close (ps);
  
  /* drop the port state lock. */
     
  gm_arch_mutex_exit (&ps->sync);

  /* Destroy the slave device. */

  GM_PRINT (GM_DEBUG_SOLARIS_CLOSE, ("Master is closing slave device.\n"));
  slave_vp = ps->arch.slave_vp;
  gm_assert (slave_vp);

  /* release mutex  now that we are finished accessing the port
     and to avoid deadlocking while trying to close the slave */

  mutex_exit (&gm_solaris_driver_mutex);

  /* we will now close the slave */

  error = VOP_CLOSE (slave_vp, flags, 1, otyp, credp);
  if (error)
    {
      GM_RETURN_INT (error);
    }

  /* potentially the master and slave port_state structures have been
     freed so do not dereference the pointer anymore now */

  GM_PRINT (GM_DEBUG_SOLARIS_CLOSE, ("Master closed slave device.\n"));

  /* Release the slave vnode. */
  
  VN_RELE (slave_vp);
  
  GM_RETURN_INT (0);
}

/* Map interface memory into the process.  This can be used to map the
   LANai control register, special registers, SRAM, or copy block into
   the user space.  The offsets and sizes of these various regions can
   be obtained using the GM_GET_MAPPING_SPECS ioctl, which returns
   a description of the mappings in a "gm_mapping_specs"
   structure.

   The pages of the copy block must be mapped in order and without
   skipping any offsets.

   gm_segmap ensures that all offsets passed to gm_mmap are
   legitimate, so there is no need to range check offsets here.
   However, checks must be performed here for "promiscuous" mappings
   of the device registers an SRAM areas other than the "send queue"
   and "activation" pages. */

#define GM_DEBUG_SOLARIS_MMAP 0

static int
gm_solaris_mmap (dev_t dev, off_t off, int prot)
{
  void *result;
  gm_port_state_t *ps;
  gm_port_state_t *slave_ps;
  int ret;

  GM_CALLED ();

  GM_PRINT (GM_DEBUG_SOLARIS_MMAP,
	    ("Mapping offset 0x%x for minor device %d\n", off,
	     getminor (dev)));

  mutex_enter (&gm_solaris_driver_mutex); 
  slave_ps = gm_minor_get_port_state (getminor (dev));
  
  if (!slave_ps)
    {
      GM_NOTE (("GM: mmap: ps NULL.\n"));
      ret = -1;
      goto abort_with_mutex;
    }

  /* Slave_ps is a fake port state.  Get the real port state for the
     device. */

  if (!GM_SOLARIS_IS_SLAVE (slave_ps))
    {
      GM_NOTE (("GM: in gm_solaris_mmap: but I am not a slave device?!\n"));
      ret = -1;
      goto abort_with_mutex;
    }
  ps = slave_ps->arch.master_ps;
  gm_assert (ps);

  GM_PRINT (GM_DEBUG_SOLARIS_MMAP, ("Calling gm_mmap().\n"));
  if (gm_mmap (ps, off, &result) != GM_SUCCESS)
    {
      ret = -1;
      goto abort_with_mutex;
    }
  gm_assert (result);

  GM_PRINT (GM_DEBUG_SOLARIS_MMAP, ("calling hat_getkpfnum(%p).\n", result));
  ret = hat_getkpfnum ((caddr_t) result);

 abort_with_mutex:
  mutex_exit (&gm_solaris_driver_mutex);
  GM_PRINT (GM_DEBUG_SOLARIS_MMAP, ("returning 0x%x.\n", ret));
  GM_RETURN_INT (ret);
}

/* Simply call gm_ioctl. */

#define GM_DEBUG_SOLARIS_IOCTL 0

static int
gm_solaris_ioctl (dev_t dev, int cmd, intptr_t arg,
		  int mode, cred_t * cred_p, int *rval_p)
{
  gm_port_state_t *ps;
  gm_status_t gm_status;

  GM_CALLED_WITH_ARGS (("0x%x,0x%x,%p,0x%x,%p,%p", dev, cmd, arg, mode,
			cred_p, rval_p));

  GM_PRINT (GM_DEBUG_SOLARIS_IOCTL,
	    ("gm_solaris_ioctl() called for dev with minor number 0x%x.\n",
	     getminor (dev)));

  /* Make sure ioctl is not being called by the kernel, since we cannot
     handle that yet. */

  if ((mode & FKIOCTL))
    {
      gm_status = GM_INTERNAL_ERROR;
      goto abort_with_nothing;
    }

  /* Split this function into 2 cases to try to avoid mutex confusion.
     Care must be take to always acquire the global mutex before the
     port mutex. */

  if (cmd != GM_SLEEP)
    {
      GM_PRINT (GM_DEBUG_SOLARIS_IOCTL, ("cmd == GM_SET_PORT_NUM.\n"));

      /* For the GM_SET_PORT_NUM command, we need to call gm_ioctl()
	 while holding the global mutex since we will be creating a
	 new port.  We also need to hold the port mutex for the port
	 we are using. */

      mutex_enter (&gm_solaris_driver_mutex);
      ps = gm_minor_get_port_state (getminor (dev));
      if (!ps)
	{
	  GM_NOTE (("ps is NULL\n"));
	  gm_status = GM_INTERNAL_ERROR;
	  goto abort_with_driver_mutex;
	}
      gm_arch_mutex_enter (&ps->sync);
      
      /* Catch ioctls on slave devices, which should never happen */

      if (GM_SOLARIS_IS_SLAVE (ps))
	{
	  GM_NOTE (("mmap-only slave got an ioctl\n"));
	  gm_status = GM_INTERNAL_ERROR;
	  goto abort_with_port_and_driver_mutex;
	}

      /* Call the portable IOCTL. */
      gm_status
	= gm_ioctl (ps, cmd, (void *) arg, INT_MAX, (void *) arg, INT_MAX, 0);
      /* fall through */
    abort_with_port_and_driver_mutex:
      gm_arch_mutex_exit (&ps->sync);
    abort_with_driver_mutex:
      mutex_exit (&gm_solaris_driver_mutex);
    }
  else
    {
      /* For all other ioctls, we need to hold the gm_solaris_driver_mutex
	 only until we have acquired the port mutex. */
      
      GM_PRINT (GM_DEBUG_SOLARIS_IOCTL, ("cmd != GM_SET_PORT_NUM\n"));
      
      /* Atomically get the port state and acquire the port mutex,
         using the global mutex as a short term lock. */

      mutex_enter (&gm_solaris_driver_mutex);
      ps = gm_minor_get_port_state (getminor (dev));
      if (!ps)
	{
	  mutex_exit (&gm_solaris_driver_mutex);
	  GM_NOTE (("ps is NULL\n"));
	  gm_status = GM_INTERNAL_ERROR;
	  goto abort_with_nothing;
	}
      gm_arch_mutex_enter (&ps->sync);
      mutex_exit (&gm_solaris_driver_mutex);
      
      /* Catch ioctls on slave devices, which should never happen */

      if (GM_SOLARIS_IS_SLAVE (ps))
	{
	  GM_NOTE (("mmap-only slave got an ioctl\n"));
	  gm_status = GM_INTERNAL_ERROR;
	  goto abort_with_port_mutex;
	}

      /* Call the portable IOCTL. */
      gm_status
	= gm_ioctl (ps, cmd, (void *) arg, INT_MAX, (void *) arg, INT_MAX, 0);
      /* fall through */
    abort_with_port_mutex:
      gm_arch_mutex_exit (&ps->sync);
    }
 abort_with_nothing:
  return (gm_solaris_localize_status (gm_status));
}

/* Perform segment mapping.

   We basically use the default ddi_segmap routine, but dynamically
   allocate copy block memory by calling gm_prepare_for_mmap before
   calling ddi_segmap so gm_mmap( ) (which is called by
   ddi_segmap( ) via gm_mmap( )) can simply assume assume the copy
   block is fully allocated.

   Also, after calling ddi_segmap successfully, the address of the
   mappings is passed to gm_finish_mmap( ), which uses this info to
   add the mapped pages to the page hash table for the device, which
   records the mappings between user virtual addresses and dma
   addresses.  This must be done here because this is the only place
   the user virtual addresses are available.

   Additionally, this segmap( ) implementation ensures that all mapping
   requests are properly page-aligned, and that mappings of the copy
   block memory are performed in order without any gaps. */

#define GM_DEBUG_SOLARIS_SEGMAP 0

static int
gm_solaris_segmap (dev_t dev,
		   off_t off,
		   struct as *asp,
		   caddr_t * addrp,
		   off_t len,
		   unsigned int prot,
		   unsigned int maxprot, unsigned int flags, cred_t * cred_p)
{
  gm_port_state_t *ps;
  int error;
  gm_status_t gm_status;

  GM_CALLED_WITH_ARGS (("?,%ld,%p,%p,%ld,0x%x,0x%x,0x%x,%p", (long) off, asp,
			addrp, (long) len, prot, maxprot, flags, cred_p));

  /* Atomically acquire the port and the port mutex */

  mutex_enter (&gm_solaris_driver_mutex);
  ps = gm_minor_get_port_state (getminor (dev));
  if (!ps) 
    {
      GM_PRINT (GM_DEBUG_SOLARIS_SEGMAP,
		("ps NULL!?\n"));
      error = gm_solaris_localize_status (GM_INTERNAL_ERROR);
      goto abort_with_driver_mutex;
    }
  gm_assert (ps);
  gm_arch_mutex_enter (&ps->sync);
  
  /* Catch segmaps on slave devices, which should never happen */
  
  if (GM_SOLARIS_IS_SLAVE (ps))
    {
      GM_NOTE (("mmap-only slave device asked to segmap\n"));
      error = gm_solaris_localize_status (GM_INTERNAL_ERROR);
      goto abort_with_ps_mutex;
    }
  
  /* Verify the mapping is legal and ensure all resources are
     available, allocating them if needed. */

  gm_status = gm_prepare_to_mmap (ps, off, len,
				  ((prot & PROT_WRITE ? GM_MAP_WRITE : 0)
				   | (prot & PROT_READ ? GM_MAP_READ : 0)));
  if (gm_status != GM_SUCCESS)
    {
      error = gm_solaris_localize_status (gm_status);
      GM_NOTE (("Error preparing to mmap.\n"));
      goto abort_with_ps_mutex;
    }

  /* Perform the mapping using the default segmap implementation,
     which calls mmap to map the individual pages. */

  /* Verify that the device has a slave device */
  
  if (!ps->arch.slave_minor)
    {
      error = gm_solaris_localize_status (GM_FAILURE);
      GM_NOTE (("Master device has no slave device to handle mapping.\n"));
      goto abort_with_ps_mutex;
    }
  
  /* Associate memory mappings with the slave device. */
  
  dev = makedevice (getmajor (dev), ps->arch.slave_minor);

  /* Need to release mutexes here since ddi_segmap will call gm_solaris_mmap()
     which will want to acquire them */
  mutex_exit (&gm_solaris_driver_mutex);
  gm_arch_mutex_exit (&ps->sync);

  error =
    ddi_segmap (dev, off, asp, addrp, len, prot, maxprot, flags, cred_p);
  if (error != DDI_SUCCESS)
    {
      GM_WARN (("Unanticipated memory mapping error %d", error));
      goto abort_with_nothing;
    }
  gm_assert (GM_PAGE_ALIGNED (*addrp));

  /* No need to gm_finish_mmap() here.  This is now done by a user ioctl. */

  GM_RETURN_INT (GM_SUCCESS);

 abort_with_ps_mutex:
  gm_arch_mutex_exit (&ps->sync);
 abort_with_driver_mutex:
  mutex_exit (&gm_solaris_driver_mutex);
 abort_with_nothing:
  GM_RETURN_INT (error);
}

#if GM_BC_DRIVER
#error broken minor number details

/* Implement gm_read so we can return EEPROM when emulating mlanai
   driver. */

static int
gm_solaris_read (dev_t dev, struct uio *uio, cred_t * credp)
{
  gm_instance_state_t *is;

  GM_CALLED ();

  if (!GM_COMPATIBILITY_DEVICE_MINOR_P (getminor (dev)))
    return EACCES;

  is = ddi_get_soft_state (gm_instancep,
			   GM_MINOR_TO_INSTANCE (getminor (dev)));

  GM_RETURN_INT (uiomove ((caddr_t) & is->lanai.eeprom,
			  min (sizeof (gm_myrinet_eeprom_t),
			       uio->uio_iov->iov_len), UIO_READ, uio));
}
#endif

#define GM_DEBUG_SOLARIS_IDENTIFY 0

static int
gm_solaris_identify (dev_info_t * dip)
{
  char *name;

  GM_CALLED ();

  name = ddi_get_name (dip);

  GM_PRINT (GM_DEBUG_SOLARIS_IDENTIFY, ("identifying name \"%s\"\n", name));

  if (gm_strcmp (name, "MYRICOM,mlanai") == 0
      || gm_strcmp (name, "pci10e8,8043") == 0
      || gm_strcmp (name, "pci14c1,8043") == 0
      || gm_strcmp (name, "pci14c1,0") == 0)
    GM_RETURN_INT (DDI_IDENTIFIED);
  else
    GM_RETURN_INT (DDI_NOT_IDENTIFIED);
}

/************************************************************************
 * Dynamic loading entry points.
 ************************************************************************/

int
_init (void)
{
  int error;

  mutex_init (&gm_solaris_print_mu, "GM print mu", MUTEX_DRIVER, NULL);
  mutex_init (&gm_solaris_driver_mutex, "GM driver mu", MUTEX_DRIVER, NULL);

  GM_CALLED ();

  /* preallocate state for 1 Myrinet card */

  error
    = ddi_soft_state_init (&gm_instancep, sizeof (gm_instance_state_t), 1);
  if (error != 0)
    goto abort_with_nothing;

  error = mod_install (&modlinkage);
  if (error != 0)
    goto abort_with_device_state;

  GM_RETURN_INT (DDI_SUCCESS);

abort_with_device_state:
  ddi_soft_state_fini (&gm_instancep);
abort_with_nothing:
  GM_NOTE (("gm device driver initialization failed."));
  GM_RETURN_INT (error);
}

#define GM_DEBUG_SOLARIS_FINI 0

int
_fini (void)
{
  int error;

  GM_CALLED ();

  GM_PRINT (GM_DEBUG_SOLARIS_FINI, ("Calling mod_remove()\n"));
  error = mod_remove (&modlinkage);
  GM_PRINT (GM_DEBUG_SOLARIS_FINI, ("Called mod_remove()\n"));
  if (error)
    {
      GM_NOTE (("error calling mod_remove.\n"));
      GM_RETURN_INT (error);
    }
  GM_PRINT (GM_DEBUG_SOLARIS_FINI, ("Calling ddi_soft_state_fini()\n"));
  gm_assert (gm_instancep);
  ddi_soft_state_fini (&gm_instancep);
  GM_PRINT (GM_DEBUG_SOLARIS_FINI, ("Called ddi_soft_state_fini()\n"));

  /* Don't use "RETURN" macro, since it might print something after the
     print mutex has been destroyed. */
  GM_PRINT (GM_DEBUG_SOLARIS_FINI, ("_fini  returning\n"));

  mutex_destroy (&gm_solaris_driver_mutex);
  mutex_destroy (&gm_solaris_print_mu);
  return 0;
}

int
_info (struct modinfo *modinfop)
{
  GM_CALLED ();
  GM_RETURN_INT (mod_info (&modlinkage, modinfop));
}



/*
  This file uses GM standard indentation:

  Local Variables:
  c-file-style:"gnu"
  tab-width:8
  End:
*/
