/******************************************************************-*-c-*-
 * Myricom GM networking software and documentation			 *
 * Copyright (c) 2003 by Myricom, Inc.					 *
 * All rights reserved.  See the file `COPYING' for copyright notice.	 *
 *************************************************************************/

/* author: glenn@myri.com */

#ifndef _gm_gateway_h_
#define _gm_gateway_h_

#include "gm_ether.h"
#include "gm_simple_types.h"

/* The timeout (in LANai ticks) before invalidating cached gateway
   information. */
#define GM_GATEWAY_CACHE_TIMEOUT 300 /* seconds */

/* Should be a power of 2 for speed. */
#define GM_GATEWAY_CACHE_ENTRIES_PER_CONNECTION 1

GM_TYPEDEF_LP_N_T (struct, gm_gateway_cache_entry);

/* A gateway cache entry.  These will be double-linked into an
   age list using the next & prev pointers.  These also act as hash
   table bins, keyed on "destination". */

typedef struct gm_gateway_cache_entry
{
  gm_gateway_cache_entry_lp_t next;
  gm_gateway_cache_entry_lp_t prev;
  /* 8 */
  gm_ethernet_mac_addr_t destination;
  gm_u16_n_t gateway_node_id;
  /* 8 */
  gm_u32_n_t valid_time;	/* Last known valid time, in seconds
				   since firmware startup.  Won't wrap
				   for 137 years. */
  gm_u32_n_t reserved;
} gm_gateway_cache_entry_t;

/* Global gateway cache state, for inclusion in the GM globals. */

typedef struct gm_gateway_cache
{
  gm_u32_n_t max_bin_id;
  gm_gateway_cache_entry_lp_t oldest_in_use;
  /* 8 */
} gm_gateway_cache_t;

/****************************************************************
 * Function prototypes
 ****************************************************************/

#if GM_BUILDING_FIRMWARE
extern gm_u16_t gm_gateway (gm_ethernet_mac_addr_t destination);
extern void gm_notice_gateway (gm_u16_t gw, gm_ethernet_mac_addr_t mac);
gm_status_t gm_gateway_cache_init (void);
void gm_debug_dump_gateway_cache (void);
#endif

#endif /* _gm_gateway_h_ */
