/******************************************************************-*-c-*-
 * Myricom GM networking software and documentation			 *
 * Copyright (c) 1999 by Myricom, Inc.					 *
 * All rights reserved.	 See the file `COPYING' for copyright notice.	 *
 *************************************************************************/

#include "gm_call_trace.h"
#include "gm_compiler.h"
#include "gm_internal.h"

GM_ENTRY_POINT
const char *
_gm_get_kernel_build_id (gm_port_t *port)
{
  gm_status_t status;
  int len;
  char *buffer;

  GM_CALLED_WITH_ARGS (("%p", port));
  
  /* If we got the kernel build ID from the kernel before, return the
     cached value. */

  if (port->kernel_build_id)
    GM_RETURN_PTR (port->kernel_build_id);

  /* determine the length of the kernel build ID string */

#if GM_KERNEL
  len = (int) (gm_strlen (_gm_build_id) + 1);
#else
  status = _gm_user_ioctl (port, GM_GET_KERNEL_BUILD_ID_LEN,
			   &len, sizeof (len));
  if (status != GM_SUCCESS)
    {
      gm_perror("get_kernel_build_id_len failed",status);
      goto abort_with_nothing;
    }
#endif
  if (len > GM_MAX_KERNEL_BUILD_ID_LEN) 
     len = GM_MAX_KERNEL_BUILD_ID_LEN;

  /* allocate space for the build ID string */

  buffer = gm_malloc (GM_MAX_KERNEL_BUILD_ID_LEN+8); /* freed in gm_close() */
  if (!buffer) 
    {
      status = GM_OUT_OF_MEMORY;
      gm_perror("gm_malloc failed in get_kernel_build_id", status);
      goto abort_with_nothing;
    }

  /* Copy the build ID string from the kernel */

#if GM_KERNEL
  gm_bcopy (_gm_build_id, buffer, len);
#else
  status = _gm_user_ioctl (port, GM_GET_KERNEL_BUILD_ID, buffer, len);
  if (status != GM_SUCCESS)
    {
      gm_perror("get_kernel_build_id ioctl failed",status);
      goto abort_with_buffer;
    }
#endif
  /* commit */

  port->kernel_build_id = buffer;
  GM_RETURN_PTR (buffer);
  
 abort_with_buffer:
  gm_free (buffer);
 abort_with_nothing:
  GM_RETURN_PTR (0);
}
