/******************************************************************-*-c-*-
 * Myricom GM networking software and documentation			 *
 * Copyright (c) 1999 by Myricom, Inc.					 *
 * All rights reserved.	 See the file `COPYING' for copyright notice.	 *
 *************************************************************************/

/* author: glenn@myri.com */

#include "gm_call_trace.h"
#include "gm_internal.h"

/****************************************************************
 * open port tracking
 ****************************************************************/

/* This is included here because it is one of the few things that
   needs to access gm_global_mutex. */

static gm_port_t *open_ports;	/* 0 at init */


int
_gm_open_ports_test(void)
{
  if (open_ports)
    return(1);
  else
    return(0);
}



/* add a port to the list of open ports */

void
_gm_open_ports_insert (gm_port_t * port)
{
  gm_mutex_enter (_gm_global_mutex);
  port->next_open_port = open_ports;
  open_ports = port;
  gm_mutex_exit (_gm_global_mutex);
}

/* splice a port out of the list of open ports */

gm_port_t *
_gm_open_ports_remove (gm_port_t * port)
{
  gm_port_t **p;

  GM_CALLED_WITH_ARGS (("%p", port));
  
  gm_mutex_enter (_gm_global_mutex);
  for (p = &open_ports; *p && *p != port; p = &(*p)->next_open_port)
    ;
  if (!*p)
    {
      port = 0;
      goto abort_with_mutex;
    }
  gm_assert (*p == port);
  *p = port->next_open_port;
  port->next_open_port = 0;
  /* fall through */
 abort_with_mutex:
  gm_mutex_exit (_gm_global_mutex);
  GM_RETURN_PTR (port);
}

/* close all open ports */

void
_gm_open_ports_close_all ()
{
  gm_port_t *p;

  /* Take care to be thread safe. */

  for (p = open_ports; p; p = open_ports)
    {
      p = _gm_open_ports_remove (p);
      if (!p)
	continue;
      gm_close (p);
    }
}


/****************************************************************
 * Clean up all state for systems like VxWorks that don't clean up
 * after exit.
 ****************************************************************/

void
_gm_finalize_all ()
{
  _gm_open_ports_close_all ();

  while (_gm_initialized)
    gm_finalize ();
}

/*
  This file uses GM standard indentation:

  Local Variables:
  c-file-style:"gnu"
  tab-width:8
  c-backslash-column:72
  End:
*/
