/******************************************************************-*-c-*-
 * Myricom GM networking software and documentation                      *
 * Copyright (c) 1996, 1997 by Myricom, Inc.                             *
 * All rights reserved.  See the file `COPYING' for copyright notice.    *
 *************************************************************************/

#include "gm_call_trace.h"
#include "gm_internal.h"

#if GM_KERNEL
#include "gm_pio.h"
#endif

#define GM_DEBUG_SET_HOST_NAME 0

#if GM_DEBUG_SET_HOST_NAME
#undef GM_LOCALLY_ENABLE_CALL_TRACE
#define GM_LOCALLY_ENABLE_CALL_TRACE 1
#endif

gm_status_t
_gm_set_host_name (gm_port_t * port, char (*name)[GM_MAX_HOST_NAME_LEN+1])
{
#if GM_KERNEL
  gm_port_state_t *ps;
  gm_instance_state_t *is;

  GM_CALLED_WITH_ARGS (("%p,\"%s\"", port, name ? (char *) name : "<NULL>"));
  
  gm_assert (port);
  ps = port->kernel_port_state;
  gm_assert (ps);
  is = ps->instance;
  gm_assert (is);

  /* Copy the host name into the LANai globals, taking care that the
     value cannot be used by the LANai processor while it is being
     updated.  However, we might be changing the name while it is
     being sent by the LANai packet engine, so the querying node might
     get a truncated host name if it is querying during the update. */

  gm_pause_lanai (is);
  gm_strncpy_to_lanai_globals (is, hostname, ""); /* clear name. */
  gm_strncpy_to_lanai_globals (is, hostname, (char *) name);
  gm_unpause_lanai (is);
  
  GM_RETURN_STATUS (GM_SUCCESS);
#else
  struct 
  {
    char host_name[GM_MAX_HOST_NAME_LEN+1];
  } in;
      
  gm_strncpy (in.host_name, &(*name)[0], sizeof (*name));
  return _gm_user_ioctl (port, GM_SET_HOST_NAME, &in, sizeof (in));
#endif
}

/*
  This file uses GM standard indentation.

  Local Variables:
  c-file-style:"gnu"
  tab-width:8
  End:
*/
