/******************************************************************-*-c-*-
 * Myricom GM networking software and documentation			 			 *
 * Copyright (c) 1999 by Myricom, Inc.					 				 *
 * All rights reserved.	 See the file `COPYING' for copyright notice.	 *
 *************************************************************************/

#include "gm_call_trace.h"
#include "gm_debug.h"
#include "gm_internal.h"
#include "gm_lanai_command.h"

#if GM_KERNEL
#include "gm_enable_security.h"
#include "gm_enable_software_loopback.h"
#include "gm_global_id.h"
#include "gm_pio.h"
#endif

#define GM_DEBUG_SET_ROUTE 0

#if GM_DEBUG_SET_ROUTE
#undef GM_LOCALLY_ENABLE_CALL_TRACE
#define GM_LOCALLY_ENABLE_CALL_TRACE 1
#endif

GM_ENTRY_POINT
gm_status_t
_gm_set_route (gm_port_t *p, const gm_u8_t unique_id[6], unsigned int length,
	       char *route)
{
#if GM_KERNEL
  gm_port_state_t *ps = p->kernel_port_state;
  gm_instance_state_t *is = ps->instance;
  gm_status_t status;
  unsigned int global_id;
  
  GM_CALLED_WITH_ARGS (("%p,%p,%d,%p", p, unique_id, length, route));

  /* Check privileges. */

  if (GM_ENABLE_SECURITY && !ps->privileged)
    {
      GM_RETURN_STATUS (GM_PERMISSION_DENIED);
    }

  /* Convert the unique ID to a GM global ID.  We use a global ID and
     not a normal GM node ID, because there might not be a local node
     ID for this remote node yet. */

  status = gm_mac_addr_to_global_id (unique_id, &global_id);
  if (status != GM_SUCCESS)
    {
      GM_RETURN_STATUS (status);
    }

  /* Print debugging info. */

  if (GM_DEBUG_SET_ROUTE)
    {
      unsigned int i;
      GM_PRINT (GM_DEBUG_SET_ROUTE, ("setting route to node 0x%x of length %d to:",
				     length, global_id));
      for (i=0; i<length; i++)
	{
	  GM_PRINT (GM_DEBUG_SET_ROUTE, (" %02x", route[i]));
	}
      GM_PRINT (GM_DEBUG_SET_ROUTE, ("\n"));
    }

  /* Have the LANai set the route, possibly allocating storage for a
     new connection in the process. */
  
  status = gm_lanai_command_set_route
    (is, global_id, (gm_u8_n_t *)route, length);
  GM_RETURN_STATUS (status);
#else
  gm_set_route_info_t ri;
  gm_status_t ret;
  unsigned int i;

  GM_CALLED_WITH_ARGS (("%p,%p,%d,%p", p, unique_id, length,route));

  ri.length = (gm_u8_t) length;
  gm_bcopy (unique_id, ri.mac_addr, 6);
  if (length && (length != GM_NO_ROUTE))
    {
      for (i=0; i<length; i++)
        ri.route[i] = route[i];
    }

  ret = _gm_user_ioctl (p, GM_SET_ROUTE, &ri, sizeof (ri));
  GM_RETURN_STATUS (ret);
#endif
}

/*
  This file uses GM standard indentation.

  Local Variables:
  c-file-style:"gnu"
  tab-width:8
  End:
*/
