/******************************************************************-*-c-*-
 * Myricom GM networking software and documentation                      *
 * Copyright (c) 1996, 1997 by Myricom, Inc.                             *
 * All rights reserved.  See the file `COPYING' for copyright notice.    *
 *************************************************************************/

/**
   @file gm_alloc_send_token.c
*/

#include "gm_call_trace.h"
#include "gm_internal.h"

/**
   gm_alloc_send_token() allocates a send token of priority \b priority
   previously freed with gm_free_send_token() and returns 0 if no
   token is available.  Clients may choose to maintain their own
   send token counts without using this utility function.
 
   @retval send_token_cnt
   @retval 0 Error occurred.
 
   @param p \b (IN) The GM port on the source/sender GM node from which
                    the communication will be sent.
 
   @param priority \b (IN) The priority of the message to be sent.
 
   @author Glenn Brown
 
   @version GM_API_VERSION (as defined in gm.h)
 
*/

GM_ENTRY_POINT
int
gm_alloc_send_token (gm_port_t * p, unsigned int priority)
{
  int ret;

  GM_CALLED ();
  if (p->send_token_cnt[priority])
    {
      ret = p->send_token_cnt[priority]--;
    }
  else
    {
      ret = 0;
    }
  GM_RETURN_INT (ret);
}
