/******************************************************************-*-c-*-
 * Myricom GM networking software and documentation                      *
 * Copyright (c) 1996, 1997 by Myricom, Inc.                             *
 * All rights reserved.  See the file `COPYING' for copyright notice.    *
 *************************************************************************/

/**
   @file gm_bzero.c
*/

#include "gm_config.h"

#if defined STDC_HEADERS && !GM_KERNEL
# include <string.h>
#ifndef memset			/* make sure memset is not a macro */
extern void *memset(void *s, int c, size_t n);
#endif
#elif defined HAVE_STRINGS_H && !GM_KERNEL
# include <strings.h>
#endif

#include "gm_internal.h"


#if GM_OS_FREEBSD && GM_KERNEL
/* HACK -- We need sys/systm.h for bzero prototype, but can't include
   sys/systm.h in gm_arch_types.h because of name collision with the
   system's crc32() prototype and our crc32() prototype. */
#include <sys/systm.h>
#endif


/* A reasonably fast bzero, that does not use partword I/O unless it must
   (for speed, especialy when doing PIO), and does not rely on the
   system bzero() functionality, which may not be safe for PIO mapped
   memory. */

void
gm_safer_bzero (void *ptr, int len)
{
  /* Clear 8 bytes at a time, if possible. */
  
  if (((gm_size_t) ptr & 7) == 0)
    {
      while (len >= 8)
	{
	  *(gm_u64_t *) ptr = (gm_u64_t) 0;
	  ptr = (void *) ((char *) ptr + 8);
	  len -= 8;
	}
    }
  
  /* Clear 4 bytes at a time, if possible. */
  
  if (((gm_size_t) ptr & 3) == 0)
    {
      while (len >= 4)
	{
	  *(gm_u32_t *) ptr = (gm_u32_t) 0;
	  ptr = (void *) ((char *) ptr + 4);
	  len -= 4;
	}
    }
  
  /* Clear remaining bytes, one at a time. */
  
  while (len >= 1)
    {
      *(gm_u8_t *) ptr = (gm_u32_t) 0;
      ptr = (void *) ((char *) ptr + 1);
      len -= 1;
    }
}

/**
    gm_bzero() clears the \b len bytes of memory starting at \b ptr.
    This function does not use partword I/O unless it must (for speed,
    especially when doing PIO), and does not rely on the
    system bzero() functionality, which may not be safe for PIO mapped
    memory.
 
    @param ptr \b (IN) The pointer to the memory location.
 
    @param len \b (IN) The number of bytes of memory to be bzero'ed.
 
    @author Glenn Brown
 
   @version GM_API_VERSION (as defined in gm.h)
 
*/

GM_ENTRY_POINT
void
gm_bzero (void *ptr, gm_size_t len)
{
#if defined (CSPI_MAP26xx) && (GM_NO_PARTWORD_PIO_WRITE || GM_NO_PARTWORD_PIO_READ)
#warning Using cspi lanai 5 bzero (no_partword_pio).
  extern void CK_bcopy(char *src, char *dest, int length);
  gm_u8_t *zero_ptr;
  int i;
  zero_ptr = (gm_u8_t *)gm_malloc(len);
  if (zero_ptr) 
    {
       for (i=0; i<len; i++)
         zero_ptr[i] = 0;
       CK_bcopy(zero_ptr,ptr,len);
       gm_free(zero_ptr);
    }
  return;
#elif GM_OS_VXWORKS
  bzero (ptr, len);
#elif GM_OS_LINUX || GM_CPU_powerpc
  gm_safer_bzero(ptr,len);
#elif defined WIN32 && GM_KERNEL
  RtlZeroMemory (ptr, len);
#elif defined WIN32 && !GM_KERNEL
  ZeroMemory (ptr, len);
#elif GM_OS_FREEBSD && GM_KERNEL
  bzero (ptr, len);
#elif GM_OS_OSF1
  bzero (ptr, len);
#elif defined STDC_HEADERS && !GM_KERNEL
  memset (ptr, 0, len);
#elif defined HAVE_STRINGS_H && !GM_KERNEL
  bzero (ptr, len);
#elif GM_KERNEL
  gm_safer_bzero (ptr, len);
#endif
}
