/******************************************************************-*-c-*-
 * Myricom GM networking software and documentation			 *
 * Copyright (c) 2000 by Myricom, Inc.					 *
 * All rights reserved.	 See the file `COPYING' for copyright notice.	 *
 *************************************************************************/

/* author: glenn@myri.com */

#include "gm_compiler.h"
#include "gm_internal.h"
#include "gm_call_trace.h"

/* state for the call stack */

GM_ENTRY_POINT char *gm_call_stack[GM_CALL_STACK_DEPTH];
GM_ENTRY_POINT void *gm_call_label[GM_CALL_STACK_DEPTH];
GM_ENTRY_POINT unsigned long gm_call_depth;
GM_ENTRY_POINT int gm_call_stack_overflowed;

/* state for call stack printing */

static gm_size_t depth;
static const char gm_call_trace_space[] = {' ',' ',' ',' ',' ',' ',' ',' ',
                                           ' ',' ',' ',' ',' ',' ',' ',' ',
                  /* array of 40 */        ' ',' ',' ',' ',' ',' ',' ',' ',
                                           ' ',' ',' ',' ',' ',' ',' ',' ',
                                           ' ',' ',' ',' ',' ',' ',' ',' ',
                                           '\0'};
#define GM_SIZEOF_CALL_TRACE_SPACE 41

/**/ GM_TOP_LEVEL_ASSERT (sizeof (gm_call_trace_space)
			  == GM_SIZEOF_CALL_TRACE_SPACE);
		     

/****************
 * serialization
 ****************/

/* This mutex is used below, and by gm_call_trace.h. */
struct gm_mutex *gm_call_trace_mutex;

/* Initialization of this package is done as early as possible in
   gm_init(). */

gm_status_t
_gm_call_trace_init (void)
{
  gm_call_trace_mutex = gm_create_mutex ();
  return gm_call_trace_mutex ? GM_SUCCESS : GM_FAILURE;
}

void
_gm_call_trace_fini (void)
{
  struct gm_mutex *mu;
  
  /* This should only be called in gm_finalize(), which is called in
     a single-threaded environment, but we take care just in case. */

  mu = gm_call_trace_mutex;
  if (mu)
    {
      gm_call_trace_mutex = 0;
      gm_mutex_enter (mu);
      gm_mutex_exit (mu);
      gm_destroy_mutex (mu);
    }
  
  _gm_report_memory_leaks ();
}

/* Functions to serialize call trace operations when possible. */

GM_ENTRY_POINT
void
gm_call_trace_mutex_enter (void)
{
  if (gm_in_intr)
    {
      return;
    }
  
  if (gm_call_trace_mutex)
    {
      gm_mutex_enter (gm_call_trace_mutex);
    }
}

GM_ENTRY_POINT
void
gm_call_trace_mutex_exit (void)
{
  if (gm_in_intr)
    {
      return;
    }
  
  if (gm_call_trace_mutex)
    {
      gm_mutex_exit (gm_call_trace_mutex);
    }
}

/****************
 * Indentation
 ****************/

static const char *
print_space_for_depth (gm_size_t the_depth)
{
  return &gm_call_trace_space[((sizeof (gm_call_trace_space) - 1)
			       - (the_depth
				  & (GM_SIZEOF_CALL_TRACE_SPACE - 2)))];
}

GM_ENTRY_POINT
const char *
__gm_print_called_space (void)
{
  const char *ret;
  
  depth++;
  ret = print_space_for_depth (depth);
  return ret;
}

GM_ENTRY_POINT
const char *
__gm_print_return_space (void)
{
  const char *ret;

  ret = print_space_for_depth (depth);
  depth--;
  return ret;
}

/****************
 * Call stack printing.
 ****************/

static
void
log_it (unsigned int ignored, const char *mesg, const char *ignored2)
{
  GM_PARAMETER_MAY_BE_UNUSED (ignored);
  GM_PARAMETER_MAY_BE_UNUSED (ignored2);
  _GM_PRINT (1, ("%s\n", mesg));
}

/* Append the call stack to the dispatch log.

   This is a bit of a hack because it relies on the fact that all the
   logged function names are string constants. */

void
gm_log_call_trace (void)
{
  unsigned int i = 0;
  char *caller;

  GM_CALLED ();
  for (caller=GM_CALLER(i); caller; i++, caller=GM_CALLER(i))
    {
      log_it (0, caller, caller);
    }
  GM_RETURN_NOTHING ();
}
