/******************************************************************-*-c-*-
 * Myricom GM networking software and documentation			 *
 * Copyright (c) 2002 by Myricom, Inc.					 *
 * All rights reserved.	 See the file `COPYING' for copyright notice.	 *
 *************************************************************************/

/* author: glenn@myri.com */

/* This function is used FOR PRODUCTION TESTING ONLY.  I allocates a
   single DMAable region to be associated with the port and returns a
   DMA pointer to the region.  There is no way for the client
   application to access this memory directly (it is not mmap()ed into
   user space). */

#include "gm_internal.h"
#include "gm_call_trace.h"
#include "gm_board_mapping.h"

GM_ENTRY_POINT
gm_status_t
gm_contiguous_dma_malloc (struct gm_port *port, gm_size_t len, gm_dp_t *ret)
{
  gm_status_t status;
  
  GM_CALLED_WITH_ARGS (("%p,"GM_U64_TMPL",%p",
			port, GM_U64_ARG (len), ret));
#if GM_KERNEL
  {
    gm_port_state_t *ps;
    gm_instance_state_t *is;

    ps = port->kernel_port_state;
    gm_assert (ps);
    is = ps->instance;
    gm_assert (is);
    
    /* Only allow one contiguous DMA allocation per port. */
	
    if (ps->contiguous_dma_region_allocated_len)
      GM_RETURN_STATUS (GM_PERMISSION_DENIED);

    /* Allocate the contiguous DMAable memory. */
    
    status
      = (gm_arch_dma_region_alloc
	 (is,
	  &ps->contiguous_dma_region,
	  len,
	  GM_ARCH_DMA_CONTIGUOUS
	  + GM_ARCH_DMA_STREAMING
	  + GM_ARCH_DMA_RDWR,
	  0,
	  0));
    if (status != GM_SUCCESS)
      {
	GM_RETURN_STATUS (status);
      }
    *ret = gm_arch_dma_region_dma_addr (&ps->contiguous_dma_region);
  }
#else
  {
    union 
    {
      gm_size_t in;
      gm_dp_t out;
    } in_out;

    in_out.in = len;
    status
      = (_gm_user_ioctl
	 (port, GM_CONTIGUOUS_DMA_MALLOC, &in_out, sizeof (in_out)));
    *ret = in_out.out;
  }
#endif
  GM_RETURN_STATUS (GM_SUCCESS);
}

GM_ENTRY_POINT
gm_status_t
gm_contiguous_dma_free (struct gm_port *port)
{
  GM_CALLED_WITH_ARGS (("%p", port));
#if GM_KERNEL
  {
    gm_port_state_t *ps;
    
    ps = port->kernel_port_state;
    gm_assert (ps);
    
    if (!ps->contiguous_dma_region_allocated_len)
      {
	GM_RETURN_STATUS (GM_INVALID_COMMAND);
      }
    gm_arch_dma_region_free (&ps->contiguous_dma_region);
    ps->contiguous_dma_region_allocated_len = 0;
    GM_RETURN_STATUS (GM_SUCCESS);
  }
#else
  {
    GM_RETURN_STATUS (_gm_user_ioctl (port, GM_CONTIGUOUS_DMA_FREE, 0, 0));
  }
#endif
}
