/******************************************************************-*-c-*-
 * Myricom GM networking software and documentation                      *
 * Copyright (c) 1996, 1997 by Myricom, Inc.                             *
 * All rights reserved.  See the file `COPYING' for copyright notice.    *
 *************************************************************************/

/**
   @file gm_directcopy.c
   This file contains the GM API function gm_directcopy_get().
*/

#include "gm_debug.h"
#include "gm_internal.h"

#define GM_DEBUG_DIRECTCOPY_GET 0

/**
   gm_directcopy_get() copies data of length \b length bytes, specified at
   the address \b source_addr of the local process using the
   port \b port_id of the board \b source_instance_id to the
   memory area specified at the address \b target_addr of the
   current process. This implementation bypasses all of the
   protection of the operating system to provide a memory copy
   from one process's memory space to another one.  The memory
   areas must have been registered by GM prior to calling this
   function in order to lock memory pages at their physical
   locations.  There are no alignment or length constraints
   but the maximum performance will be reached with aligned
   addresses on both sides.  This function is supported
   exclusively on Linux.
 
   @retval GM_SUCCESS Operation completed successfully.
   @retval GM_FAILURE Error occurred.
 
   @param p \b (IN) Handle to the GM port.
 
   @param source_addr \b (IN) Address of the data to be copied.
 
   @param target_addr \b (IN) Target address of the copied data.
 
   @param length \b (IN) The length (in bytes) of the area to be copied.
 
   @param source_instance_id \b (IN)  The id of the interface.
 
   @param source_port_id \b (IN) The port id of the interface.
 
   @author Patrick Geoffray
 
   @version GM_API_VERSION (as defined in gm.h)
 
*/

GM_ENTRY_POINT gm_status_t
gm_directcopy_get(struct gm_port *p, void *source_addr, void *target_addr, gm_size_t length, unsigned int source_instance_id, unsigned int source_port_id)
{
#if GM_OS_LINUX
  gm_directcopy_t args;
  
  GM_PRINT
    (GM_DEBUG_DIRECTCOPY_GET,
     ("directcopy: ioctl for directcopy from %p (%d,%d) to %p (length %lu)\n",
      source_addr, source_instance_id, source_port_id, target_addr, length));

  /* fill the structure for the args to pass to the kernel */
  args.source_addr = (gm_up_t)source_addr;
  args.target_addr = (gm_up_t)target_addr;
  args.length = length;
  args.source_instance_id = (gm_u16_t)source_instance_id;
  args.source_port_id = (gm_u16_t)source_port_id;
 
  /* pass the job to the kernel */ 
  return(_gm_user_ioctl(p, GM_DIRECTCOPY_GET, &args, sizeof(gm_directcopy_t))); 
#else
  GM_PRINT (1, ("directcopy: no support for directcopy in GM.\n"));
  return GM_FAILURE;
#endif
}
