/******************************************************************-*-c-*-
 * Myricom GM networking software and documentation                      *
 * Copyright (c) 1996, 1997 by Myricom, Inc.                             *
 * All rights reserved.  See the file `COPYING' for copyright notice.    *
 *************************************************************************/

/**
   @file gm_directed_send.c
   This file contains the GM API function
   gm_directed_send().
*/

#include "gm.h"
#include "gm_debug.h"
#include "gm_enable_put.h"
#include "gm_internal.h"
#include "gm_send_queue.h"

/* Handler to emulate the original-style GM behaviour of automatically
   freeing the send token upon the completion of a directed send. */

static
void
_gm_directed_sent (struct gm_port *p, void *context, gm_status_t stat)
{
  if (stat != GM_SUCCESS)
    {
      gm_perror ("directed send failed", stat);
#if !GM_KERNEL
      gm_exit (stat);
#endif /* GM_KERNEL */
    }
  

  __gm_free_send_tokens (p,
			 (context == (void *) GM_HIGH_PRIORITY
			  ? GM_HIGH_PRIORITY
			  : GM_LOW_PRIORITY),
			 1);
}

/**
 
   \defgroup deprecated Deprecated GM API functions

*/

/**

   \addtogroup deprecated

   gm_directed_send() is deprecated and included only for backward
   compatibility.  Use gm_directed_send_with_callback()
   instead.
 
   @see gm_directed_send_with_callback
 
 
*/
 
GM_ENTRY_POINT
void
gm_directed_send (gm_port_t * p, void *source_buffer,
		  gm_remote_ptr_t target_buffer,
		  gm_size_t len,
		  enum gm_priority priority,
		  unsigned int target_node_id,
		  unsigned int target_port_id)
{
  gm_directed_send_with_callback (p, source_buffer, target_buffer, len,
				  priority, target_node_id, target_port_id,
				  _gm_directed_sent, (void *) priority);
}
