/******************************************************************-*-c-*-
 * Myricom GM networking software and documentation			 *
 * Copyright (c) 2001 by Myricom, Inc.					 *
 * All rights reserved.	 See the file `COPYING' for copyright notice.	 *
 *************************************************************************/

/* author: glenn@myri.com */

#include "gm.h"
#include "gm_call_trace.h"
#include "gm_internal.h"
#include "gm_io.h"

#define GM_DEBUG_DMA_ADDR 0

#if GM_DEBUG_DMA_ADDR
#undef GM_LOCALLY_ENABLE_CALL_TRACE
#define GM_LOCALLY_ENABLE_CALL_TRACE 1
#endif

GM_ENTRY_POINT
gm_status_t
gm_dma_addr (struct gm_port *port, gm_up_t ptr, gm_dp_t *ret)
{
  gm_status_t status;
  
  GM_CALLED ();
#if GM_KERNEL
  {
    gm_port_state_t *ps;
    gm_instance_state_t *is;
    gm_page_hash_table_t *hash;
    
    gm_assert (port);
    ps = port->kernel_port_state;
    gm_assert (ps);
    is = ps->instance;
    gm_assert (is);
    hash = &is->page_hash;
    gm_assert (hash);

    gm_assert (ps->id < GM_PAGE_LEN);
    status = gm_dma_addr_for_port_addr (hash, ps->id, ptr, ret);
    
    GM_RETURN_STATUS (status);
  }
#else
  {
    union {
      gm_up_t in;
      gm_dp_t out;
    } buf;

    buf.in = ptr;
    status = _gm_user_ioctl (port, GM_DMA_ADDR, &buf, sizeof (buf));
    if (status != GM_SUCCESS)
      {
	GM_RETURN_STATUS (status);
      }
    gm_assert (ret);
    *ret = buf.out;
    GM_PRINT (GM_DEBUG_DMA_ADDR,
	      ("gm_dma_addr() sez: up %p -> dp "GM_U64_TMPL"\n",
	       ptr, GM_U64_ARG (buf.out)));
    GM_RETURN_STATUS (status);
  }
#endif
  /* never get here */
  gm_abort ();
}
