/******************************************************************-*-c-*-
 * Myricom GM networking software and documentation                      *
 * Copyright (c) 1996, 1997 by Myricom, Inc.                             *
 * All rights reserved.  See the file `COPYING' for copyright notice.    *
 *************************************************************************/

/**
   @file gm_dma_calloc.c
   This file contains the GM API function gm_dma_calloc().
*/

#include "gm_internal.h"
#include "gm_call_trace.h"

/**
   gm_dma_calloc() allocates and clears \b count * \b length bytes
   of DMAable memory aligned on a 4-byte boundary.  Memory should be
   freed using gm_dma_free().
 
   @param p \b (IN) Handle to the GM port.
 
   @param count \b (IN) The number of elements to be calloc'ed.
   @param length \b (IN) The size of each element to be calloc'ed.
 
   @see gm_dma_malloc gm_dma_free
 
   @author Glenn Brown
 
   @version GM_API_VERSION (as defined in gm.h)
 
*/

GM_ENTRY_POINT void *
gm_dma_calloc (struct gm_port *p, gm_size_t count, gm_size_t length)
{
  void *ret;

  GM_CALLED ();
  
  length *= count;
/* printf("gm_dma_calloc(%p, %ld, %ld)\n",p,count,length); */
  ret = gm_dma_malloc (p, length);
  if (ret)
    {
      gm_bzero (ret, length);
    }
  GM_RETURN_PTR (ret);
}
