/**
   @file gm_drop_sends.c
   This file contains the GM API function gm_drop_sends().
*/

#include "gm_internal.h"
#include "_gm_modsend.h"

/**
   gm_drop_sends() drops all enqueued sends for \b port of priority
   \b priority destined for target_port_id of \b target_node_id to be
   dropped, and reenable packet transmission on that connection. This
   function should only be called after an error is reported to a send
   completion callback routine. The first four parameters must match those
   of the failed send. It should be called only once per reported error. This
   function requires a send token, which will be returned to the client in
   the callback function. The dropped sends will then be returned to the
   client with a status of GM_SEND_DROPPED.
 
   gm_drop_sends() and gm_resume_sending(), as most gm requests, require a
   send token, and the callback you give to them is just meant to return
   this token. These gm_requests always succeed (if called in a valid
   manner), so the callback will \b always be called with \b GM_SUCCESS
   (which here does not mean at all that something was sent successfully,
   just that the request has been taken into account, and the token used
   for that request was recycled).
 
   @param port \b (IN) The handle to the GM port.
   @param priority \b (IN) The priority of the message.
   @param target_node_id \b (IN) The GM node to which the message is
                                 being sent.
   @param target_port_id \b (IN) The GM port on the destination GM node
                                 to which the message was sent.
   @param callback \b (IN) The function called when the send is complete.
   @param context \b (IN) Pointer to an integer or to a structure
                           that is passed to the callback function.
 
   @author Glenn Brown
 
   @version GM_API_VERSION (as defined in gm.h)
 
*/

GM_ENTRY_POINT
void
gm_drop_sends (struct gm_port *port,
	       unsigned int priority,
	       unsigned int target_node_id,
	       unsigned int target_port_id,
	       gm_send_completion_callback_t callback,
	       void *context)
{
  _gm_modsend (port, priority, target_node_id, target_port_id,
	       GM_DROP_SENDS_EVENT, callback, context);
}
