/******************************************************************-*-c-*-
 * Myricom GM networking software and documentation                      *
 * Copyright (c) 1998 by Myricom, Inc.                                   *
 * All rights reserved.  See the file `COPYING' for copyright notice.    *
 *************************************************************************/

/**
   @file gm_eprintf.c
*/

#include "gm_config.h"		/* must be first for firmware builds */

#include "gm.h"
#include "gm_debug.h"

#if GM_OS_VXWORKS && !GM_BUILDING_FIRMWARE
#include "vxWorks.h"
#endif /* GM_OS_VXWORKS && !GM_BUILDING_FIRMWARE*/

#if defined(STDC_HEADERS) && !GM_KERNEL
#include <stdarg.h>
#include <stdio.h>
#endif

/**
   gm_eprintf() emulates or invokes the ANSI standard vprintf() function.
 
   @retval 0 Operation completed successfully.
 
   @param format Specifies how the variable-length arguments are
                 converted for output.
 
   @author Glenn Brown
 
   @version GM_API_VERSION (as defined in gm.h)
 
*/

GM_ENTRY_POINT int
gm_eprintf (const char *format, ...)
{
  va_list ap;
  int ret = -1;

  va_start (ap, format);
#if GM_BUILDING_FIRMWARE
  ret = vfprintf (stdout, format, ap);
  fflush (stdout);
#elif GM_OS_VXWORKS && GM_KERNEL
  /* VxWorks does not have a stderr??? */
  ret = vprintf (format, ap);
#elif GM_KERNEL
  GM_NOTE_OCCASIONALLY
    (("Use of gm_eprintf() in the kernel is not supported\n"));
#else
  ret = vfprintf (stderr, format, ap);
#endif
  va_end (ap);

  return ret;
}


/*
  This file uses GM standard indentation:

  Local Variables:
  c-file-style:"gnu"
  tab-width:8
  c-backslash-column:72
  End:
*/
