/******************************************************************-*-c-*-
 * Myricom GM networking software and documentation			 *
 * Copyright (c) 2000 by Myricom, Inc.					 *
 * All rights reserved.	 See the file `COPYING' for copyright notice.	 *
 *************************************************************************/

/**
   @file gm_exit.c
*/

#include "gm_call_trace.h"
#include "gm_debug.h"
#include "gm_internal.h"
#include "gm_malloc_debug.h"

#if GM_OS_VXWORKS
#include <setjmp.h>
jmp_buf gm_vxworks_fake_exit_env;

GM_ENTRY_POINT void
vxworks_gm_exit (gm_status_t status)
{
  GM_PRINT (1,("faking gm_exit in VxWorks\n"));
  longjmp (gm_vxworks_fake_exit_env, status ? status : GM_FAILURE);
}
#endif /* GM_OS_VXWORKS */

#if GM_KERNEL && !GM_OS_VXWORKS
GM_ENTRY_POINT void
gm_exit (gm_status_t status)
{
  GM_CALLED ();
  gm_perror ("gm_exit called in kernel", status);
  GM_RETURN_NOTHING ();
}
#else

#ifdef STDC_HEADERS
#include <stdlib.h>
#else
extern void exit (int status);
#endif

/**
   gm_exit() causes the current process to exit with a status
   appropriate to the GM status code \b status.
 
   @param status \b (IN) The GM status code, as specified in gm.h.
 
   @see gm_init gm_open gm_close gm_finalize gm_abort
 
   @author Glenn Brown
 
   @version GM_API_VERSION (as defined in gm.h)
 
*/

GM_ENTRY_POINT void
gm_exit (gm_status_t status)
{
  /* clean up all GM global state. */

  GM_PRINT (GM_PRINT_LEVEL >= 2,("gm_exit entered in user mode\n"));

  _gm_perform_on_exit_callbacks (status);
  
  /* exit */

#if GM_OS_VXWORKS
  vxworks_gm_exit(status);
#else
  _gm_report_memory_leaks ();
  exit (status);
#endif
}
#endif
