/******************************************************************-*-c-*-
 * Myricom GM networking software and documentation                      *
 * Copyright (c) 1996, 1997 by Myricom, Inc.                             *
 * All rights reserved.  See the file `COPYING' for copyright notice.    *
 *************************************************************************/

/**
   @file gm_free.c
   This file contains the GM API function gm_free().
*/

#if !GM_KERNEL
#include <stdlib.h>
#endif

#include "gm_call_trace.h"
#include "gm_debug_malloc.h"
#include "gm_internal.h"
#include "gm_malloc_debug.h"

/* free a pointer without checking */

void
_gm_free (void *ptr)
{
  GM_CALLED_WITH_ARGS (("%p", ptr));
  
#if GM_KERNEL
  __gm_arch_kernel_free (ptr);
#else  /* !GM_KERNEL */
# undef free
  free (ptr);
# define free GM_DO_NOT_USE (free)
#endif /* !GM_KERNEL */
  
  GM_RETURN_NOTHING ();
}

/**
   gm_free() frees the memory buffer at \b ptr, which was previously
   allocated by gm_malloc(), or gm_calloc().
 
   @param ptr \b (IN) Address of the memory to be freed.
 
   @see gm_malloc gm_calloc
   @author Glenn Brown
 
   @version GM_API_VERSION (as defined in gm.h)
 
*/

GM_ENTRY_POINT
void
gm_free (void *ptr)
{
  GM_CALLED_WITH_ARGS (("%p", ptr));
  
  GM_MDEBUG_REMOVE_PTR (ptr, gm_free);
  _gm_free (ptr);

  GM_RETURN_NOTHING ();
}

