/******************************************************************-*-c-*-
 * Myricom GM networking software and documentation			 *
 * Copyright (c) 2003 by Myricom, Inc.					 *
 * All rights reserved.	 See the file `COPYING' for copyright notice.	 *
 *************************************************************************/

/* author: glenn@myri.com */

#include "gm_internal.h"

#if GM_KERNEL
#include "gm_pio.h"
#endif

/* Return any HW configuration strings associated with the card.  For
   the moment, we assume there are only 256 bytes worth, but since
   this is an undocumented internal function, we can easily change
   that in the future. */

GM_ENTRY_POINT
gm_status_t
_gm_get_board_config_strings (struct gm_port *port, char (*buf)[256])
{
#if GM_KERNEL
  {
    char stack_buf[sizeof (*buf)];
    gm_instance_state_t *is;
    gm_port_state_t *ps;

    gm_assert (port);
    ps = port->kernel_port_state;
    gm_assert (ps);
    is = ps->instance;
    gm_assert (is);
    
    /* Copy the strings to stack_buf, in case PIO is required, since
       copyout might not be able to handle PIO.  How we do this depends
       on the board revision. */
    
    if (GM_SUPPORT_PCI
	&& is->lanai.eeprom.bus_type == GM_MYRINET_BUS_PCI
	&& ((GM_SUPPORT_PCI_REV_4 && is->ifc.pci.config.Revision_ID == 4)
	    || (GM_SUPPORT_PCI_REV_5 && is->ifc.pci.config.Revision_ID == 5)))
      {
	gm_u32_t spec_offset;
	
	/* Strings are just below the LANaiX hardware's ROM configuration
	   space. */

	spec_offset = ((gm_read_lanai_special_reg_u32 (is, lX.DMA_CONFIG)
			& 0x0FFFFC00)
		       - 256);
	gm_bcopy ((char *) is->board_base + spec_offset, stack_buf, 256);
      }
    else
      {
	/* Empty string list for cards without a ROM config space. */
	
	stack_buf[0] = 0;
      }

    /* Copyout */

    return gm_copyout (ps, stack_buf, buf, sizeof (*buf));
  }
#else  /* !GM_KERNEL */
  {
    /* Pass the request to the driver. */
    return
      _gm_user_ioctl (port, GM_GET_BOARD_CONFIG_STRINGS, buf, sizeof (*buf));
  }
#endif /* !GM_KERNEL */
}

