/******************************************************************-*-c-*-
 * Myricom GM networking software and documentation                      *
 * Copyright (c) 1996, 1997 by Myricom, Inc.                             *
 * All rights reserved.  See the file `COPYING' for copyright notice.    *
 *************************************************************************/

/**
   @file gm_get_mapper_unique_id.c
   This file contains the GM API function gm_get_mapper_unique_id().
*/

#include "gm_internal.h"

#if GM_KERNEL
#include "gm_pio.h"
#include "gm_debug.h"
#endif

/**
   gm_get_mapper_unique_id() copies the 6-byte ethernet address of the
   network interface card associated with \b port to the buffer
   at \b id.
 
   @retval GM_SUCCESS Operation completed successfully.
 
 
   @param port \b (IN) The GM port associated with a specific
                       network interface card.
   @param id \b (OUT) Buffer to which the 6-byte ethernet address
                       (GM_GET_MAPPER_UNIQUE_ID) is copied.
 
   @author Glenn Brown
 
   @version GM_API_VERSION (as defined in gm.h)
 
*/

gm_status_t
gm_get_mapper_unique_id (struct gm_port *port, char *id)
{
#if GM_KERNEL
  gm_port_state_t *ps = port->kernel_port_state;
  gm_instance_state_t *is = ps->instance;
  gm_status_t status;
  gm_u8_t uid[6];

  /* In GM-2, each node runs its own mapper. */
  
  status = gm_node_id_to_unique_id (port, 1, uid);
  if (status != GM_SUCCESS)
    {
      return status;
    }
  
  GM_PRINT (GM_PRINT_LEVEL >= 4,
	    ("gm_get_mapper_unique_id returning %x:%x:%x:%x:%x:%x\n",
	     uid[0], uid[1], uid[2], uid[3], uid[4], uid[5]));
	    
  /* Copy the 6-byte ethernet address from scout reply message */

  return gm_copyout (ps, uid, id, 6);
#else
  /* indirectly call the kernel version of this function */
  return _gm_user_ioctl (port, GM_GET_MAPPER_UNIQUE_ID, id, 6);
#endif
}
