/******************************************************************-*-c-*-
 * Myricom GM networking software and documentation			 *
 * Copyright (c) 2001 by Myricom, Inc.					 *
 * All rights reserved.	 See the file `COPYING' for copyright notice.	 *
 *************************************************************************/

/* author: glenn@myri.com */

#include "gm.h"
#include "gm_board_mapping.h"
#include "gm_call_trace.h"
#include "gm_internal.h"
#include "gm_io.h"

GM_ENTRY_POINT
gm_status_t
gm_map_board (struct gm_port *port,
	      void **control_regs,
	      gm_size_t *control_regs_mapped_len,
	      void **special_regs,
	      gm_size_t *special_regs_mapped_len,
	      void **sram,
	      gm_size_t *sram_mapped_len)
{
  gm_status_t status;
  gm_status_t unmap_status;
  gm_mapping_specs_t *mappings;

  GM_CALLED_WITH_ARGS (("%p,%p,%p,%p", port, control_regs, special_regs,
			sram));
  
  gm_assert (port);
  mappings = port->mappings;
  gm_assert (mappings);

  /* map the control regs */

  status = _gm_mmap (port, mappings->control_regs.offset,
		     mappings->control_regs.len, GM_MAP_RDWR,
		     control_regs);
  if (status != GM_SUCCESS)
    {
      goto abort_with_nothing;
    }
  *control_regs_mapped_len = mappings->control_regs.len;
  
  /* map the special regs */

  status = _gm_mmap (port, mappings->special_regs.offset,
		     mappings->special_regs.len, GM_MAP_RDWR,
		     special_regs);
  if (status != GM_SUCCESS)
    {
      goto abort_with_control_regs;
    }
  *special_regs_mapped_len = mappings->special_regs.len;

  /* map the sram */

  status = _gm_mmap (port, mappings->sram.offset,
		     mappings->sram.len, GM_MAP_RDWR,
		     sram);
  if (status != GM_SUCCESS)
    {
      goto abort_with_special_regs;
    }
  *sram_mapped_len = mappings->sram.len;
  
  GM_RETURN_STATUS (GM_SUCCESS);
  
 abort_with_special_regs:
  unmap_status = _gm_munmap (port, special_regs, mappings->special_regs.len);
  if (unmap_status != GM_SUCCESS)
    {
      gm_perror ("could not unmap the LANai special registers", unmap_status);
    }
 abort_with_control_regs:
  unmap_status = _gm_munmap (port, control_regs, mappings->control_regs.len);
  if (unmap_status != GM_SUCCESS)
    {
      gm_perror ("could not unmap the LANai control registers", unmap_status);
    }
 abort_with_nothing:
  GM_RETURN_STATUS (status);
}

GM_ENTRY_POINT
void
gm_unmap_board (struct gm_port *port,
		void *control_regs,
		void *special_regs,
		void *sram)
{
  gm_status_t status;
  gm_mapping_specs_t *mappings;

  gm_assert (port);
  mappings = port->mappings;
  gm_assert (mappings);
  
  GM_CALLED_WITH_ARGS (("%p,%p,%p,%p", port, control_regs, special_regs,
			sram));

  status = _gm_munmap (port, sram, mappings->sram.len);
  if (status != GM_SUCCESS)
    {
      gm_perror ("could not unmap the LANai SRAM", status);
    }

  status = _gm_munmap (port, special_regs, mappings->special_regs.len);
  if (status != GM_SUCCESS)
    {
      gm_perror ("could not unmap the LANai special registers", status);
    }

  status = _gm_munmap (port, control_regs, mappings->control_regs.len);
  if (status != GM_SUCCESS)
    {
      gm_perror ("could not unmap the LANai control registers", status);
    }

  GM_RETURN_NOTHING ();
}
