/******************************************************************-*-c-*-
 * Myricom GM networking software and documentation                      *
 * Copyright (c) 1996, 1997 by Myricom, Inc.                             *
 * All rights reserved.  See the file `COPYING' for copyright notice.    *
 *************************************************************************/

/**
   @file gm_max_node_id.c
   This file contains the GM API function gm_max_node_id().
*/

#include "gm_call_trace.h"
#include "gm_internal.h"

/**
   gm_max_node_id() stores the maximum GM node ID supported by the
   network interface card corresponding to \b port at the address of \b n.
 
   @retval GM_SUCCESS Operation completed successfully.
 
   @param port \b (IN)
 
   @param n \b (OUT) Buffer to store the maximum GM node ID.
 
   @author Glenn Brown
 
   @version GM_API_VERSION (as defined in gm.h)
 
*/

GM_ENTRY_POINT
gm_status_t
gm_max_node_id (gm_port_t *port, unsigned int *n)
{
#if GM_KERNEL
  gm_port_state_t *ps;

  GM_CALLED_WITH_ARGS (("%p,%p", port, n));
  ps  = port->kernel_port_state;
  GM_RETURN_STATUS (gm_copyout (ps, &ps->instance->max_node_id, n,
				sizeof (*n)));
#else
  gm_status_t status = GM_SUCCESS;

  if (!port->max_node_id)
    {
      status = _gm_user_ioctl (port, GM_GET_MAX_NODE_ID, &port->max_node_id,
			       sizeof (port->max_node_id));
      if (status != GM_SUCCESS)
	port->max_node_id = 0;
    }
  *n = port->max_node_id;
  return status;
#endif
}

