/******************************************************************-*-c-*-
 * Myricom GM networking software and documentation                      *
 * Copyright (c) 1996, 1997 by Myricom, Inc.                             *
 * All rights reserved.  See the file `COPYING' for copyright notice.    *
 *************************************************************************/

/**
   @file gm_max_node_id_in_use.c
   This file contains the GM API function gm_max_node_id_in_use().
 
   This function exists to allow gm_board_info to have a guess
   for an upper limit to the actual number of nodes in the route table
   Without it, gm_board_info calls thousands of gm_get_route() calls.
   When debugging, this is really a mess.
*/

#include "gm_call_trace.h"
#include "gm_internal.h"

#if GM_KERNEL
#include "gm_pio.h"
#endif

#define GM_DEBUG_MAX_NODE_ID_INUSE 0

#if GM_DEBUG_MAX_NODE_ID_INUSE
#undef GM_LOCALLY_ENABLE_CALL_TRACE
#define GM_LOCALLY_ENABLE_CALL_TRACE 1
#endif

/**
   gm_max_node_id_in_use() returns the maximum GM node ID that is
   in use by the network attached to the port.
 
   @retval GM_SUCCESS Operation completed successfully.
   @retval GM_UNATTACHED
 
   @param port \b (IN) The GM port.
 
   @param n \b (OUT) Buffer containing the maximum GM node ID.
 
   @author Glenn Brown
 
   @version GM_API_VERSION (as defined in gm.h)
 
*/

GM_ENTRY_POINT
gm_status_t
gm_max_node_id_in_use (gm_port_t *port, unsigned int *n)
{
#if GM_KERNEL
  unsigned int max;

  gm_assert (port);
  gm_assert (port->kernel_port_state);
  gm_assert (port->kernel_port_state->instance);

  GM_CALLED_WITH_ARGS (("%p,%p", port, n));

  max = (gm_read_lanai_global_u32 (port->kernel_port_state->instance,
				 connection.first_free_id)
	- 1);
  GM_RETURN_STATUS (gm_copyout (port->kernel_port_state, &max, n, sizeof (*n)));
#else
  gm_status_t status = GM_SUCCESS;

  status = _gm_user_ioctl (port, GM_GET_MAX_NODE_ID_IN_USE, n,
			   sizeof (unsigned int));

  if (status != GM_SUCCESS)
    *n = 0;

  return status;
#endif
}

