/******************************************************************-*-c-*-
 * Myricom GM networking software and documentation                      *
 * Copyright (c) 1996, 1997 by Myricom, Inc.                             *
 * All rights reserved.  See the file `COPYING' for copyright notice.    *
 *************************************************************************/

/**
   @file gm_memcmp.c
   This file contains the GM API function gm_memcmp().
*/

#include "gm_internal.h"

/**
 
   gm_memcmp() emulates the ANSI memcmp() function.
 
   @retval int Returns an integer less than, equal to, or
               greater than zero if the first len bytes
               of a is found, respectively, to be less than,
               to match, or be greater than the first len bytes
               of b.
 
   @param a (IN) The first memory area for comparison.
   @param b (IN) The second memory area for comparison.
   @param len (IN) The number of bytes to compare.
 
   @author Glenn Brown
 
   @version GM_API_VERSION (as defined in gm.h)
 
*/
 
GM_ENTRY_POINT
int
gm_memcmp (const void *_a, const void *_b, gm_size_t len)
{
  char *a, *b;
  int result;

  a = (char *) _a;
  b = (char *) _b;
  while (len--)
    {
      result = *a++ - *b++;
      if (result)
	return result;
    }
  return 0;
}
