/******************************************************************-*-c-*-
 * Myricom GM networking software and documentation                      *
 * Copyright (c) 1996, 1997 by Myricom, Inc.                             *
 * All rights reserved.  See the file `COPYING' for copyright notice.    *
 *************************************************************************/

/**
   @file gm_memorize_message.c
   This file contains the GM API function gm_memorize_message().
*/

#include "gm_debug.h"
#include "gm_internal.h"

#define GM_DEBUG_MEMORIZE_MESSAGE 0

				/* for memcpy */
#if defined STDC_HEADERS && ( !GM_KERNEL || GM_OS_VXWORKS )
#include <string.h>
#endif

/**
   gm_memorize_message() is a wrapper around function memcpy().
   gm_memorize_message() copies a message into a buffer if needed.
   If \b message and \b buffer differ,
   gm_memorize_message(port,message,buffer) copies the message pointed
   to by \b message into the buffer pointed to by \b buffer.
   gm_memorize_message() returns \b buffer.  This function optionally
   optimizes the handling of FAST receive messages as described
   in "See Chapter 9 [Receiving Messages].
 
   @retval something
 
   @param message \b (IN) Address of the message to be copied.
   @param buffer \b (OUT) Address where the message is to be copied.
   @param len \b (IN) The length in bytes of the message to be copied.
 
   @author Glenn Brown
 
   @version GM_API_VERSION (as defined in gm.h)
 
*/

void *
gm_memorize_message (void *message, void *buffer, unsigned len)
{
  /* print input to copy */

  GM_PRINT (GM_DEBUG_MEMORIZE_MESSAGE, ("")); /* label location */
  if (GM_DEBUG_MEMORIZE_MESSAGE)
    {
      gm_hex_dump (message, len);
    }
  
  /* Copy message into buffer */

  /* FIXME: we may be able to do better than memcpy since
     we know some things about alignment and length */

  gm_bcopy (message, buffer, len);

  /* print copied output */

  if (GM_DEBUG_MEMORIZE_MESSAGE)
    {
      gm_hex_dump (buffer, len);
    }

  return message;
}

