/******************************************************************-*-c-*-
 * Myricom GM networking software and documentation                      *
 * Copyright (c) 1998 by Myricom, Inc.                                   *
 * All rights reserved.  See the file `COPYING' for copyright notice.    *
 *************************************************************************/

/**
   @file gm_mutex.c
   This file contains the GM API functions gm_create_mutex(),
   gm_destroy_mutex(), gm_mutex_enter(), gm_mutex_exit().
*/

#include "gm_call_trace.h"
#include "gm_compiler.h"
#include "gm_internal.h"

/**
 
   gm_create_mutex() creates a GM mutex.
 
   @retval gm_mutex \b (OUT) Handle to the GM mutex.
 
   @author Glenn Brown
 
   @version GM_API_VERSION (as defined in gm.h)
 
*/
 
GM_ENTRY_POINT struct gm_mutex *
gm_create_mutex ()
{
#if GM_KERNEL
  gm_arch_sync_t *sync;

  GM_CALLED ();
  sync = (gm_arch_sync_t *) gm_malloc (sizeof (gm_arch_sync_t));
  if (!sync)
    return 0;
  gm_arch_sync_init (sync, 0);
  GM_RETURN ((struct gm_mutex *) sync);
#else
  return (struct gm_mutex *) 1;
#endif
}

/**
  
   gm_destroy_mutex() destroys a GM mutex.
 
   @param mu \b (IN) The handle to the GM mutex.
 
   @author Glenn Brown
 
   @version GM_API_VERSION (as defined in gm.h)
 
*/
 
GM_ENTRY_POINT void
gm_destroy_mutex (struct gm_mutex *mu)
{
  GM_CALLED ();
#if GM_KERNEL
  gm_arch_sync_destroy ((gm_arch_sync_t *) mu);
  gm_free (mu);
#else
  GM_PARAMETER_MAY_BE_UNUSED (mu);
#endif
  GM_RETURN_NOTHING ();
}

/**
 
   gm_mutex_enter()
 
   @param mu \b (IN) The handle to the GM mutex.
 
   @author Glenn Brown
 
   @version GM_API_VERSION (as defined in gm.h)
 
*/
 
GM_ENTRY_POINT void
gm_mutex_enter (struct gm_mutex *mu)
{
  gm_assert (mu);
#if GM_KERNEL
  gm_arch_mutex_enter ((gm_arch_sync_t *) mu);
#else
#endif
}

/**
 
   gm_mutex_exit()
 
   @param mu \b (IN) The handle to the GM mutex.
 
   @author Glenn Brown
 
   @version GM_API_VERSION (as defined in gm.h)
 
*/
 
GM_ENTRY_POINT void
gm_mutex_exit (struct gm_mutex *mu)
{
  gm_assert (mu);
#if GM_KERNEL
  gm_arch_mutex_exit ((gm_arch_sync_t *) mu);
#else
#endif
}






/*
  This file uses GM standard indentation:

  Local Variables:
  c-file-style:"gnu"
  tab-width:8
  c-backslash-column:72
  End:
*/
