/******************************************************************-*-c-*-
 * Myricom GM networking software and documentation                      *
 * Copyright (c) 1996, 1997 by Myricom, Inc.                             *
 * All rights reserved.  See the file `COPYING' for copyright notice.    *
 *************************************************************************/

/**
   @file gm_next_event_peek.c
   This file contains the GM API function gm_next_event_peek().
*/

#include "gm_internal.h"

/**
   gm_next_event_peek() returns the nonzero event type if an event
   is pending.  If the event is a message receive event, then the
   sender parameter will be filled with the gmID of the message
   sender.  If an event is pending a call to any gm_receive*() function
   will return the event immediately, although gm_receive() is preferred
   in this case for efficiency.
 
   @retval gm_recv_event_t  Receive event type.
 
   @param p \b (IN) The GM port for which the communication is received.
 
   @param sender
 
   @see gm_receive_pending gm_receive
 
   @author Glenn Brown
 
   @version GM_API_VERSION (as defined in gm.h)
 
*/

GM_ENTRY_POINT
int
gm_next_event_peek(gm_port_t *p,  gm_u16_t *sender)
{
  int ret;

  ret = gm_ntoh_u8 (p->recv_queue_slot->event.recv.type);

  *sender =  gm_ntoh_u16 (p->recv_queue_slot->event.recv.sender_node_id);

  if ((unsigned int)ret >=128) 
      return (ret-128);
  else 
     return GM_NO_RECV_EVENT;
}

