/******************************************************************-*-c-*-
 * Myricom GM networking software and documentation                      *
 * Copyright (c) 1996, 1997 by Myricom, Inc.                             *
 * All rights reserved.  See the file `COPYING' for copyright notice.    *
 *************************************************************************/

/**
   @file gm_node_id_to_unique_id.c
   This file contains the GM API function gm_global_id_to_unique_id().
*/

#include "gm_call_trace.h"
#include "gm_internal.h"
#include "gm_global_id.h"

#define GM_DEBUG_NODE_ID_TO_UNIQUE_ID 0

#if GM_DEBUG_NODE_ID_TO_UNIQUE_ID
#undef GM_LOCALLY_ENABLE_CALL_TRACE
#define GM_LOCALLY_ENABLE_CALL_TRACE 1
#endif

/**
   gm_global_id_to_unique_id() stores the MAC address for the interface
   with GM ID \b n at \b unique.
 
   @retval GM_SUCCESS Operation completed successfully.
   @retval GM_INVALID_PARAMETER
 
   @param port \b (IN) A GM port.
   @param node_id \b (IN) The node ID to convert.
   @param uid \b (OUT) Where to store the unique ID corresponding to node_id.
 
   @author Glenn Brown
 
   @version GM_API_VERSION (as defined in gm.h)
 
*/

GM_ENTRY_POINT
gm_status_t
gm_node_id_to_unique_id (gm_port_t *port, unsigned int node_id, char *uid)
{
  gm_status_t status;
  unsigned int global_id;

  GM_CALLED_WITH_ARGS (("%p,%d,%p", port, node_id, uid));
  
  /* Convert the node ID to a global ID. */
  
  status = gm_node_id_to_global_id (port, node_id, &global_id);
  if (status != GM_SUCCESS)
    {
      GM_RETURN_STATUS (status);
    }
  GM_PRINT (GM_DEBUG_NODE_ID_TO_UNIQUE_ID, ("global_id = 0x%x\n", global_id));
  
  /* Convert the global id to a MAC addr. */

  status = gm_global_id_to_mac_addr (global_id, uid);
  if (status != GM_SUCCESS)
    {
      GM_RETURN_STATUS (status);
    }
  _GM_PRINT
    (GM_DEBUG_NODE_ID_TO_UNIQUE_ID,
     ("uid = %02x:%02x:%02x:%02x:%02x:%02x\n",
      uid[0], uid[1], uid[2], uid[3], uid[4], uid[5]));
  
  GM_RETURN_STATUS (GM_SUCCESS);
}
