/******************************************************************-*-c-*-
 * Myricom GM networking software and documentation                      *
 * Copyright (c) 1996, 1997 by Myricom, Inc.                             *
 * All rights reserved.  See the file `COPYING' for copyright notice.    *
 *************************************************************************/

#ifdef WIN32
#include <stdio.h>
#endif

#include "gm_call_trace.h"
#include "gm_debug.h"
#include "gm_enable_fast_small_send.h"
#include "gm_enable_trace.h"
#include "gm_internal.h"
#include "gm_send_queue.h"

#define GM_DEBUG_OPIOM_SEND 0

#if GM_DEBUG_OPIOM_SEND
#undef GM_LOCALLY_ENABLE_CALL_TRACE
#define GM_LOCALLY_ENABLE_CALL_TRACE 1
#endif

GM_ENTRY_POINT
void
gm_opiom_send_with_callback (gm_port_t *p, unsigned int offset,
			     unsigned int size,
			     unsigned long len, unsigned int priority,
			     unsigned int target_node_id,
			     unsigned int target_port_id,
			     gm_send_completion_callback_t callback,
			     void *context)
{
  GM_CALLED ();

  GM_LOG_EVT(GM_SEND_EVENT);

  gm_assert (p);
  gm_assert (offset+len <= GM_OPIOM_MAX);
  gm_assert (gm_min_size_for_length (len) <= size);
  gm_assert (size >= GM_MIN_MESSAGE_SIZE && size <= 32);
  gm_assert (priority == GM_LOW_PRIORITY ||
	       priority == GM_HIGH_PRIORITY);
  
#if GM_DEBUG_OPIOM_SEND
  gm_printf ("gm_opiom_send_with_callback "
	     "(%p, %p, %d, 0x%lx, %s, %d, %d, %p, %p) called.\n",
	     p, message, size, len,
	     priority ? "GM_HIGH_PRIORITY" : "GM_LOW_PRIORITY",
	     target_node_id, target_port_id, callback, context);
#endif

  GM_SIMULATE_SIMULATE ( );

  /* Pass send token to LANai.  Care must be taken to flush the message to
     memory before flushing the token, and the token should be flushed to
     memory immediately. */

  {
    struct gm_send_send_event volatile *se;
    struct gm_send_send_event batch_write;

    se = GM_SEND_QUEUE_SLOT (p, callback, context, send);

    if (GM_STRUCT_WRITE_COMBINING)
      {
	GM_PRINT (GM_PRINT_LEVEL >= 8,
		  ("gm_opiom_send (%p, %d, %d, 0x%lx, %s, %d, %d) called.\n",
		   p, offset, size, len,
		   priority ? "GM_HIGH_PRIORITY" : "GM_LOW_PRIORITY",
		   target_node_id, target_port_id));
	GM_PRINT (GM_PRINT_LEVEL >= 8,
		  ("gm_opiom_send: port = %p  send = %p\n", p, se));
	
	batch_write.message = gm_hton_up ((gm_up_t) offset);
	batch_write.length = gm_hton_u32 ((gm_u32_t) len);
	batch_write.target_node_id = gm_hton_u16 ((gm_u16_t) target_node_id);
	batch_write.target_subport_id
	  = gm_hton_u8 ((gm_u8_t) GM_SUBPORT (priority, target_port_id));
	batch_write.type = gm_hton_u8 ((gm_u8_t) (GM_OPIOM_SEND_EVENT_0 + size));

	GM_COPY_TO_IO_SPACE (*se, batch_write);
      }
    else
      {
	se->message = gm_hton_up ((gm_up_t) offset);
	se->target_node_id = gm_hton_u16 ((gm_u16_t) target_node_id);
	se->length = gm_hton_u32 ((gm_u32_t) len);
	se->target_subport_id =
	  gm_hton_u8 ((gm_u8_t) GM_SUBPORT (priority, target_port_id));
	GM_STBAR ();
	se->type = gm_hton_u8 ((gm_u8_t) (GM_OPIOM_SEND_EVENT_0 + size));
      }
    GM_FLUSH_SEND_EVENT (p, se);
  }
  GM_RETURN_NOTHING ();
}
