/******************************************************************-*-c-*-
 * Myricom GM networking software and documentation                      *
 * Copyright (c) 1999,2000 by Myricom, Inc.                              *
 * All rights reserved.  See the file `COPYING' for copyright notice.    *
 *************************************************************************/

/**
   @file gm_perror.c
   This file contains the GM API function gm_perror().
*/

#include "gm_debug.h"
#include "gm_internal.h"

#if !GM_KERNEL
#if defined(STDC_HEADERS) || GM_OS_AIX
#include <stdio.h>
#endif
#endif /* !GM_KERNEL */

/**
   gm_perror() is similar to ANSI perror(), but takes the error code
   as a parameter to allow thread safety in future implementations,
   and only supports GM error numbers.  Prints \b message followed
   by a description of \b errno.
 
   @param message \b (OUT) Textual description of the GM error.
 
   @param error \b (IN) GM Error code.
 
   @author Glenn Brown
 
   @version GM_API_VERSION (as defined in gm.h)
 
*/

GM_ENTRY_POINT void
gm_perror (const char *message, gm_status_t error)
{
  const char *description;
  char *tmpl;
  
  description = gm_strerror (error);
  
  if ((unsigned int) error >= (unsigned int) GM_NUM_STATUS_CODES)
    {
      tmpl = "%s: %s %u\n";
    }
  else
    {
      tmpl = "%s: %s\n";
    }
  
#if GM_KERNEL
  GM_INFO ((tmpl, message, description, (unsigned int) error));
#else
  fprintf (stderr, tmpl, message, description, (unsigned int) error);
  fflush (stderr);
#endif
}

/*
  This file uses GM standard indentation:

  Local Variables:
  c-file-style:"gnu"
  tab-width:8
  End:
*/
