/******************************************************************-*-c-*-
 * Myricom GM networking software and documentation                      *
 * Copyright (c) 1998 by Myricom, Inc.                                   *
 * All rights reserved.  See the file `COPYING' for copyright notice.    *
 *************************************************************************/

/**
   @file gm_printf.c
*/

#include "gm_config.h"		/* must be first for firmware builds */

#if (GM_KERNEL && (GM_OS_SOLARIS || GM_OS_OSF1))
/*
"/usr/include/iso/stdarg_iso.h", line 117: syntax error before or at: void
"/ufs/gm/feldy/main/libgm/gm_printf.c", line 47: cannot recover from previous errors
cc: acomp failed for /ufs/gm/feldy/main/libgm/gm_printf.c
*/
static int gm_printf_compile_unused;
#else

#include "gm.h"
#if GM_KERNEL
#include "gm_debug.h"		/* Needed for GM_PANIC */
#endif

#if GM_OS_VXWORKS && !GM_BUILDING_FIRMWARE
#include "vxWorks.h"
#endif /* GM_OS_VXWORKS && !GM_BUILDING_FIRMWARE*/

#if defined(__STDC__) || (defined(STDC_HEADERS) && (!GM_KERNEL || GM_OS_IRIX || GM_OS_FREEBSD || GM_OS_VXWORKS || GM_OS_HPUX))
#include <stdarg.h>
#endif

#if !GM_KERNEL || GM_OS_VXWORKS
#include <stdio.h>
#endif

/**

   gm_printf() emulates or invokes the ANSI standard printf() function.
 
   @retval 0 Operation completed successfully.
 
   @param format Specifies how the arguments are
                 converted for output.
 
   @author Glenn Brown
 
   @version GM_API_VERSION (as defined in gm.h)
 
*/

GM_ENTRY_POINT int
gm_printf (const char *format, ...)
{
  va_list ap;
  int ret = 0;

  va_start (ap, format);
#if GM_BUILDING_FIRMWARE
  ret = vfprintf (stdout, format, ap);
  fflush (stdout);
#elif GM_OS_VXWORKS && GM_KERNEL
  ret = vprintf (format, ap);
#elif GM_KERNEL
  ret = 0;
  GM_PANIC (("gm_printf not supported in kernel\n"));
#else
  ret = vprintf (format, ap); 
#endif
  va_end (ap);

  return ret;
}

#endif /* !SOLARIS */

/*
  This file uses GM standard indentation:

  Local Variables:
  c-file-style:"gnu"
  tab-width:8
  c-backslash-column:72
  End:
*/

