/*******************************************************************-*-c-*-
 * Myricom GM networking software and documentation                      *
 * Copyright (c) 1996, 1997 by Myricom, Inc.                             *
 * All rights reserved.  See the file `COPYING' for copyright notice.    *
 *************************************************************************/

/**
   @file gm_rand.c

   This file contains the GM API functions gm_rand() and gm_srand().
*/
 

#include "gm_internal.h"
#include "gm_crc32.h"

static gm_u32_t _gm_rand_seed;

/**

   gm_rand() returns a pseudo-random integer, using a poor but fast
   random number generator.
 
   @retval RANDOM_NUMBER The random number that was generated.
 
   @author Glenn Brown
 
   @version GM_API_VERSION (as defined in gm.h)
 
*/

GM_ENTRY_POINT int
gm_rand ()
{
  gm_u32_t in = 0;

  GM_CRC32_COMPUTE_1_BIT (in, _gm_rand_seed);
  return (int) _gm_rand_seed;
}

/**

   gm_srand() returns a pseudo-random integer, and requires a seed
   for the random number generator.
 
   @param seed \b (IN) Seed for the random number generator.
 
   @author Glenn Brown
 
   @version GM_API_VERSION (as defined in gm.h)
 
*/

GM_ENTRY_POINT void
gm_srand (int seed)
{
  _gm_rand_seed = seed ? seed : 1;
}


/*
  This file uses GM standard indentation:

  Local Variables:
  c-file-style:"gnu"
  tab-width:8
  c-backslash-column:72
  End:
*/
