/******************************************************************-*-c-*-
 * Myricom GM networking software and documentation                      *
 * Copyright (c) 1996, 1997 by Myricom, Inc.                             *
 * All rights reserved.  See the file `COPYING' for copyright notice.    *
 *************************************************************************/

/**
   @file gm_resume_sending.c
   The file contains the GM API function gm_resume_sending().
*/

#include "gm_internal.h"
#include "_gm_modsend.h"

/**
   gm_resume_sending(0 reenables packet transmission of messages from \b port
   of priority \b priority destined for target_port_id of \b target_node_id.
   This function should only be called after an error is reported to a send
   completion callback routine. The message that generated the error is not
   resent. The first four parameters must match those of the failed send. It
   should be called only once per reported error. This function requires a
   send token, which will be returned to the client in the callback function.
 
   gm_resume_sending() and gm_drop_sends(), as most gm requests, require a
   send token, and the callback you give to them is just meant to return
   this token. These gm_requests always succeed (if called in a valid
   manner), so the callback will \b always be called with \b GM_SUCCESS
   (which here does not mean at all that something was sent successfully,
   just that the request has been taken into account, and the token used
   for that request was recycled).
 
   @param p \b (IN) The handle to the GM port.
 
   @param priority \b (IN) The priority of the message being sent.
 
   @param target_node_id \b (IN) The GM node to which the message is
                     being sent.
 
   @param target_port_id \b (IN) The GM port on the destination GM node
                     to which the message is being sent.
 
   @param callback \b (IN) The function called when the send is complete.
 
   @param context \b (IN) Pointer to an integer or to a structure
                           that is passed to the callback function.
 
   @author Glenn Brown
 
   @version GM_API_VERSION (as defined in gm.h)
 
*/

GM_ENTRY_POINT
void
gm_resume_sending (struct gm_port *p,
		   unsigned int priority,
		   unsigned int target_node_id,
		   unsigned int target_port_id,
		   gm_send_completion_callback_t callback,
		   void *context)
{
  _gm_modsend (p, priority, target_node_id, target_port_id,
	       GM_RESUME_SENDING_EVENT, callback, context);
}
