/******************************************************************-*-c-*-
 * Myricom GM networking software and documentation                      *
 * Copyright (c) 1996, 1997 by Myricom, Inc.                             *
 * All rights reserved.  See the file `COPYING' for copyright notice.    *
 *************************************************************************/

#include "gm_debug.h"
#include "gm_internal.h"

#if GM_KERNEL
#include "gm_pio.h"
#endif


GM_ENTRY_POINT gm_status_t
gm_set_image_properties(struct gm_port *port,
			gm_up_t window_offset,
			gm_u16_t window_width,
			gm_u16_t window_height,
			gm_u16_t image_width,
			gm_u16_t screen_width,
			gm_u8_t depth_in_bits)
{
#if GM_ENABLE_OPIOM
  gm_u16_t depth = depth_in_bits >> 3;
#if GM_KERNEL
  gm_instance_state_t *is;
  gm_port_state_t *ps;
  
  gm_assert (port);
  ps = port->kernel_port_state;
  gm_assert (ps);
  is = ps->instance;
  gm_assert (is);
  gm_assert (is->lanai.globals);

  gm_write_lanai_global_up (is, window_offset, window_offset);
  gm_write_lanai_global_u16 (is, window_width, window_width * depth);
  gm_write_lanai_global_u16 (is, window_height, window_height);
  gm_write_lanai_global_u16 (is, image_width, image_width * depth);
  gm_write_lanai_global_u16 (is, screen_width, screen_width * depth);

  GM_PRINT (GM_PRINT_LEVEL >= 4,
            ("gm_set_image_properties: set values off %d Ww %d Wh %d Sw %d Iw%d.\n",
	     window_offset, window_width, window_height, screen_width, image_width));
  return GM_SUCCESS;
#else
  gm_image_properties_t vals;
  gm_status_t status;

  vals.window_offset = window_offset;
  vals.window_width = window_width * depth;
  vals.window_height = window_height;
  vals.screen_width = screen_width * depth;
  vals.image_width = image_width * depth;

  status = _gm_user_ioctl (port, GM_SET_IMAGE_PROPERTIES,
                           &vals, sizeof (gm_image_properties_t));
  return status;
#endif

#else /* GM_ENABLE_OPIOM */

  return GM_FAILURE;

#endif /* GM_ENABLE_OPIOM */  
}
