/**
   @file gm_sleep.c
   This file contains the GM API function gm_sleep().
*/

#include "gm_internal.h"

#if GM_KERNEL && !GM_OS_VXWORKS
#else

#ifdef WIN32

#else
extern int sleep (unsigned int seconds);
#endif

/**
   gm_sleep() emulates the ANSI standard sleep(), sleeping the
   entire process for \b seconds seconds.
 
   @retval SLEEP
 
   @param seconds \b (IN) The number of seconds for which the
             process should sleep.
 
   @author Glenn Brown
 
   @version GM_API_VERSION (as defined in gm.h)
 
*/

GM_ENTRY_POINT int
gm_sleep (unsigned int seconds)
{
#ifdef WIN32
  Sleep (seconds * 1000);

  return 0; /* sleep() does not return a useful value */


#elif GM_OS_VXWORKS
#include <time.h>

  taskDelay(CLOCKS_PER_SEC * seconds);
  return (0);
#else
  return sleep (seconds);
#endif

}
#endif
