/******************************************************************-*-c-*-
 * Myricom GM networking software and documentation			 *
 * Copyright (c) 2000 by Myricom, Inc.					 *
 * All rights reserved.	 See the file `COPYING' for copyright notice.	 *
 *************************************************************************/

/**
   @file gm_strerror.c
   This file contains the GM API function gm_strerror().
*/

#include "gm.h"

/**
   gm_strerror() is an error function for GM.  The error is
   only valid until next call to this function.
 
   @retval char
 
   @param error \b (IN) GM status code.
 
   @author Glenn Brown
 
   @version GM_API_VERSION (as defined in gm.h)
 
*/

GM_ENTRY_POINT
char *gm_strerror (gm_status_t error)
{
  switch (error)
    /* NOTE: Each status code defined in gm.h should have an entry here */
    {
    case GM_SUCCESS:
      return  "success";
    case GM_FAILURE:
      return  "unqualified error";
    case GM_INPUT_BUFFER_TOO_SMALL:
      return  "input buffer is too small";
    case GM_OUTPUT_BUFFER_TOO_SMALL:
      return  "output buffer is too small";
    case GM_TRY_AGAIN:
      return  "try again";
    case GM_BUSY:
      return  "busy";
    case GM_MEMORY_FAULT:
      return  "memory fault";
    case GM_INTERRUPTED:
      return  "interrupted";
    case GM_INVALID_PARAMETER:
      return  "invalid parameter";
    case GM_OUT_OF_MEMORY:
      return  "out of memory";
    case GM_INVALID_COMMAND:
      return  "invalid command";
    case GM_PERMISSION_DENIED:
      return  "permission denied";
    case GM_INTERNAL_ERROR:
      return  "internal error";
    case GM_UNATTACHED:
      return  "unattached";
    case GM_UNSUPPORTED_DEVICE:
      return  "unsupported device";
    case GM_SEND_TIMED_OUT:
      return  "send timed out";
    case GM_SEND_REJECTED:
      return  "send was rejected by receiver";
    case GM_SEND_TARGET_PORT_CLOSED:
      return  "target port was closed";
    case GM_SEND_TARGET_NODE_UNREACHABLE:
      return  "target node was unreachable";
    case GM_SEND_DROPPED:
      return  "send dropped as requested by the client";
    case GM_SEND_PORT_CLOSED:
      return  "send port was closed";
    case GM_NODE_ID_NOT_YET_SET:
      return "node ID not set (node has not been configured by the mapper)";
    case GM_STILL_SHUTTING_DOWN:
      return "port is still shutting down from previous open";
    case GM_CLONE_BUSY:
      return  "clone device is busy";
    case GM_NO_SUCH_DEVICE:
      return  "no such device";
    case GM_ABORTED:
      return  "GM aborted";
    case GM_INCOMPATIBLE_LIB_AND_DRIVER:
      return  "libgm and GM driver are incompatible";
    case GM_UNTRANSLATED_SYSTEM_ERROR:
      return  "untranslated system error";
    case GM_ACCESS_DENIED:
      return "access denied";
#if GM_API_VERSION >= GM_API_VERSION_2_0
    case  GM_NO_DRIVER_SUPPORT:
      return "no driver support";
    case GM_PTE_REF_CNT_OVERFLOW:
      return "page table entry reference count overflow";
    case GM_NOT_SUPPORTED_IN_KERNEL:
      return "not supported in the kernel";
    case GM_NOT_SUPPORTED_ON_ARCH:
      return "not supported for this architecture";
    case GM_NO_MATCH:
      return "no match";
    case GM_USER_ERROR:
      return "user error";
    case GM_TIMED_OUT:
      return "timed out";
    case GM_DATA_CORRUPTED:
      return "data has been corrupted";
    case GM_HARDWARE_FAULT:
      return "hardware fault";
    case GM_SEND_ORPHANED:
      return "send orphaned";
    case GM_MINOR_OVERFLOW:
      return "minor number overflow";
    case GM_PAGE_TABLE_FULL:
      return "page table is full";
    case GM_UC_ERROR:
      return "UC error";
    case GM_INVALID_PORT_NUMBER:
      return "invalid port number";
    case GM_DEV_NOT_FOUND:
      return "no device files found";
    case GM_FIRMWARE_NOT_RUNNING:
      return "firmware is not running";
    case GM_YP_NO_MATCH:
      return "No match for yellow pages query";
#endif
    default:
      ;
    }

  return "unknown error";
}

