/******************************************************************-*-c-*-
 * Myricom GM networking software and documentation                      *
 * Copyright (c) 1996, 1997 by Myricom, Inc.                             *
 * All rights reserved.  See the file `COPYING' for copyright notice.    *
 *************************************************************************/

/**
   @file gm_strncasecmp.c
   This file contains the GM API function gm_strncasecmp().
*/

#include "gm_internal.h"
#include "gm_debug.h"

static char
gm_tolower(char c)
{
  if ((c>='A') && (c<='Z')) 
    {
      return c-'A'+'a';
    }
  return c;
}

/**
 
   gm_strncasecmp() reimplements strncasecmp().

   @retval int Returns an integer less than, equal to, or greater
        than  zero  if the first len bytes of a is found, respectively, to
        be less than, to match, or be greater than b. 

   @param a (IN) The first string to be compared.
   
   @param b (IN) The second string to be compared.

   @param len (IN) The number of bytes.

   @author Glenn Brown
 
   @version GM_API_VERSION (as defined in gm.h)
 
*/
 
GM_ENTRY_POINT 
int
gm_strncasecmp (const char *a, const char *b, int len)
{ 
  int result;

  gm_assert (a);
  gm_assert (b);
  while (len--)
    {   
      if (!*a || !*b)
        return gm_tolower(*a) - gm_tolower(*b);
      result = gm_tolower(*a++) - gm_tolower(*b++);
      if (result)
        return result;
    }
  return 0;
}
