/******************************************************************-*-c-*-
 * Myricom GM networking software and documentation                      *
 * Copyright (c) 1996, 1997 by Myricom, Inc.                             *
 * All rights reserved.  See the file `COPYING' for copyright notice.    *
 *************************************************************************/

/**
   @file gm_unique_id_to_node_id.c
   This file contains the GM API function gm_unique_id_to_node_id().
*/

#include "gm_call_trace.h"
#include "gm_debug.h"
#include "gm_internal.h"
#include "gm_global_id.h"

#define GM_DEBUG_UNIQUE_ID_TO_NODE_ID 0

#if GM_DEBUG_UNIQUE_ID_TO_NODE_ID
#undef GM_LOCALLY_ENABLE_CALL_TRACE
#define GM_LOCALLY_ENABLE_CALL_TRACE 1
#endif

/**
   gm_unique_id_to_node_id() takes the MAC address and returns the GM node
                             id for a specific port.
 
   @retval GM_SUCCESS Operation completed successfully.
   @retval GM_INVALID_PARAMETER
 
   @param port \b (IN) The GM port.
 
   @param unique \b (IN) The unique ID to translate.
 
   @param node_id (OUT) Where to store the node ID corresponding to unique.
 
   @author Glenn Brown
 
   @version GM_API_VERSION (as defined in gm.h)
 
*/

gm_status_t
gm_unique_id_to_node_id (gm_port_t *port, char *unique,
			 unsigned int *node_id)
{
  gm_status_t status;
  unsigned int global_id;

  GM_CALLED_WITH_ARGS
    (("%p,%02x:%02x:%02x:%02x:%02x:%02x,%p",
      port,
      unique[0]&0xff, unique[1]&0xff, unique[2]&0xff,
      unique[3]&0xff, unique[4]&0xff, unique[5]&0xff,
      node_id));

  /* Convert the MAC addr to a global ID. */
  
  status = gm_mac_addr_to_global_id (unique, &global_id);
  if (status != GM_SUCCESS)
    {
      GM_RETURN_STATUS (status);
    }

  /* Convert the global ID into a node ID. */
  
  GM_PRINT
    (GM_DEBUG_UNIQUE_ID_TO_NODE_ID,
     ("%02x:%02x:%02x:%02x:%02x:%02x -> 0x%x\n",
      unique[0]&0xff, unique[1]&0xff, unique[2]&0xff,
      unique[3]&0xff, unique[4]&0xff, unique[5]&0xff,
      global_id));
  status = gm_global_id_to_node_id (port, global_id, node_id);
  if (status != GM_SUCCESS)
    {
      GM_RETURN_STATUS (status);
    }

  /* Return */
  
  _GM_PRINT (GM_DEBUG_UNIQUE_ID_TO_NODE_ID, ("-> 0x%x\n", node_id));
  GM_RETURN_STATUS (GM_SUCCESS);
}
