#include "lx.h"

lx_node_t*lx_peek (lx_queue_t*q)
{
  insist (q);
  return q->head;
  except: return 0;
}

lx_node_t*lx_get (lx_queue_t*q)
{
  insist (q);
  if (q->head)
    return lx_remove (q, q->head);
  except: return 0;
}

int lx_queued (lx_queue_t*q, lx_node_t*n)
{
  insist (q && n);
  return n->q == q;
  except: return 0;
}

int lx_put (lx_queue_t*q, lx_node_t*n)
{
  insist (q && n);
  insist (!n->next && !n->prev);
  insist (!n->q);

  if (q->tail)
  {
    q->tail->next = mi_node2index (n);
    n->prev = mi_node2index (q->tail);
  }
  else
  {
    insist (!q->head);
    q->head = n;
    n->prev = 0;
  }
  q->tail = n;
  n->next = 0;
  n->q = q;
  q->count++;
  
  return 1;
  except: return 0;
}

int lx_put_at_front (lx_queue_t*q, lx_node_t*n)
{
  insist (q && n);
  insist (!n->next && !n->prev);
  insist (!n->q);

  if (q->head)
  {
    n->next = mi_node2index (q->head);
    q->head->prev = mi_node2index (n);
  }
  else
  {
    insist (!q->tail);
    q->tail = n;
    n->prev = 0;
  }
  q->head = n;
  n->prev = 0;
  n->q = q;
  q->count++;
  
  return 1;
  except: return 0;
}

int lx_insert (lx_queue_t*q, lx_node_t*before, lx_node_t*n)
{
  insist (q && n);
  insist (!n->next && !n->prev);
  insist (!n->q);
  
  if (q->tail)
  {
    insist (q->head);
    insist (q->head != n);
    
    if (!before)
    {
      n->next = mi_node2index (q->head);
      q->head->prev = mi_node2index (n);
      q->head = n;
      n->prev = 0;
    }
    else if (before == q->tail)
      return lx_put (q, n);
    else
    {
      mi_index2node (before->next)->prev = mi_node2index (n);
      n->next = before->next;
      before->next = mi_node2index (n);
      n->prev = mi_node2index (before);
    }
  }
  else
  {
    insist (!q->head);
    insist (!before);
    q->head = q->tail = n;
    n->prev = n->next = 0;
  }

  q->count++;
  n->q = q;

  return 1;
  except: return 0;
}

lx_node_t*lx_remove (lx_queue_t*q, lx_node_t*n)
{
  lx_node_t*nn,*pp;
  
  insist (n);
  insist (n->q == q);
  n->q = 0;
  
  nn = mi_index2node (n->next);
  pp = mi_index2node (n->prev);

  insist (q->head && q->tail);
  
  if (nn)
    nn->prev = mi_node2index (pp);
  if (pp)
    pp->next = mi_node2index (nn);
  
  if (n == q->tail)
    q->tail = pp;
  if (n == q->head)
    q->head = nn;
  
  n->prev = n->next = 0;
  q->count--;
  
  return n;
  
  except: return 0;
}

#if 0
lx_node_t*lx_nth (lx_queue_t*q, int n)
{
  int i;
  lx_node_t*nn = 0;
  insist (q && n >= 0);
  
  for (i = 0, nn = q->tail; nn && i < n; i++, nn = mi_index2node (nn->prev));

  return nn;
  except: return 0;
}
#endif

lx_node_t*lx_remove_from_anywhere (lx_node_t*n)
{
  insist (n);
  return n->q ? lx_remove (n->q, n) : n;
  except: return 0;
}


#ifdef MI_SIMULATION
#define LX_TOTAL_HOSTS 200
#define LX_TOTAL_XBARS 100
#else
#define LX_TOTAL_HOSTS LX_MAX_HOSTS
#define LX_TOTAL_XBARS LX_MAX_XBARS
#endif

lx_node_t*lx_new_xbar (lx_map_t*m, lx_route_t*route)
{
  lx_node_t*n;
  
  insist (m->lx->free_xbars.count + m->lx->staged_xbars.count + m->frontier.count + m->xbars.count + m->merge_queue.count == LX_TOTAL_XBARS);
  
  n = (lx_node_t*) lx_get (&m->lx->free_xbars);

  insist (m && n);
  lx_bzero ((char*)n, sizeof (lx_xbar_t));

  m->num_xbars++;

  n->xbar = 1;
  n->first = 0;
  n->index = m->xbars_created++;
  n->original.index = 0;
  n->original.port = 0;
  
  n->route.length = 0;
  if (route)
  {
    n->route.length = route->length;
    lx_copy_route (n->route.hops, route->hops, n->route.length);
  }

  mi_c (("created " lx_node_format, lx_node_args (n)));
  
  insist (mi_index2node (mi_node2index (n)) == n);

  return n;
  except: return 0;  
}

lx_node_t*lx_new_host (lx_map_t*m, unsigned char mac_address [6], int host_type, int level, lx_route_t*route)
{
  lx_node_t*n;

  insist (m->lx->free_hosts.count + m->lx->staged_hosts.count + m->hosts.count == LX_TOTAL_HOSTS - 1);

  n = (lx_node_t*) lx_get (&m->lx->free_hosts);

  insist (m && n);
  lx_bzero ((char*)n, sizeof (lx_host_t));
  n->first = 0;
  n->index = m->num_hosts++;
  n->original.index = 0;
  n->original.port = 0;
  
  n->route.length = 0;
  lx_host_c (n)->level = level;
  lx_host_c (n)->host_type = host_type;

  if (mac_address)
    mi_memcpy (lx_host_c (n)->mac_address, mac_address, 6);
  
  if (route)
  {
    n->route.length = route->length;
    lx_copy_route (n->route.hops, route->hops, n->route.length);
  }
  mi_c (("created " lx_node_format, lx_node_args (n)));

  insist (mi_index2node (mi_node2index (n)) == n);

  return n;
  except: return 0;
}

void lx_delete_hosts (lx_map_t*m, lx_node_t*n)
{
  int i;
  lx_node_t*nn;
  
  insist (m && n && n->xbar);

  for (i = 0; i < LX_XBAR_SIZE; i++)
  {
    if ((nn = lx_get_node (n, i + n->first)) && !nn->xbar)
      lx_delete_node (m, lx_remove_from_anywhere (nn));
  }
  except:;
}

int lx_has_hosts (lx_node_t*n)
{
  int i;
  lx_node_t*o;
  insist (n->xbar);
  
  for (i = 1 - LX_XBAR_SIZE; i < LX_XBAR_SIZE; i++)
  {
    if ((o = lx_get_node (n, i)) && !o->xbar)
      return 1;
  }
  except: return 0;
}

void lx_delete_node (lx_map_t*m, lx_node_t*n)
{
  int i;
  lx_node_t*o;
  
  insist (m && n);

  mi_c (("deleted " lx_node_format, lx_node_args (n)));
  
  for (i = 1 - LX_XBAR_SIZE; i < LX_XBAR_SIZE; i++)
  {
    if ((o = lx_get_node (n, i)))
    {
      if (lx_get_node (o, lx_get_port (n, i)) == n)
	lx_disconnect (o, lx_get_port (n, i));
    }
  }  
  
  if (n->xbar)
    m->num_xbars--;
  else
    m->num_hosts--;

  insist (m->num_hosts >= 0 && m->num_xbars >= 0);

  lx_put (n->xbar ? &m->lx->free_xbars : &m->lx->free_hosts, n);

  except:;
}

static void lx_shift (lx_node_t*n, unsigned amount)
{
  int i;
  
  insist (n && amount < LX_XBAR_SIZE);
  for (i = LX_XBAR_SIZE - 1; i >= (int) amount; i--)
    n->links [i] = n->links [i - amount];
  for (; i >= 0; i--)
    n->links [i].index = 0;
  except:;
}

void lx_disconnect (lx_node_t*n, int p)
{
  insist (n);
  insist (p > -LX_XBAR_SIZE && p < LX_XBAR_SIZE);
  insist (p - n->first >= 0 && p - n->first < LX_XBAR_SIZE);
  n->links [p - n->first].index = 0;
  except:;
}


int lx_connect (lx_node_t*na, int pa, lx_node_t*nb, int pb)
{
  insist (na && nb);
  insist (pa > -LX_XBAR_SIZE && pa < LX_XBAR_SIZE);
  insist (pb >= -LX_XBAR_SIZE && pb < LX_XBAR_SIZE);
  insist (na->xbar || !pa);
  insist (nb->xbar || !pb);
  
  mi_c (("connecting " lx_node_format " port %d to " lx_node_format " port %d", lx_node_args (na), pa, lx_node_args (nb), pb));

  if (pa < na->first)
  {
    lx_shift (na, na->first - pa);
    na->first = pa;
  }
  if (pb < nb->first)
  {
    lx_shift (nb, nb->first - pb);
    nb->first = pb;
  }
  
  na->links [pa - na->first].port = pb;  
  na->links [pa - na->first].index = mi_node2index (nb);
  
  nb->links [pb - nb->first].port = pa;  
  nb->links [pb - nb->first].index = mi_node2index (na);

  return 1;
  except: return 0;
}

lx_node_t*lx_get_node (lx_node_t*n, int port)
{
  insist (n);
  insist (n->xbar || !n->first);
  
  if ((!n->xbar && port) || port <= -LX_XBAR_SIZE || port > LX_XBAR_SIZE || port - n->first < 0 || port - n->first >= LX_XBAR_SIZE)
    return 0;
  
  return mi_index2node (n->links [port - n->first].index);
  
  except: return 0;
}

void lx_copy_route (signed char*dst, signed char*src, int length)
{
  int i;
  insist (dst && src && length <= LX_ROUTE_SIZE);
  
  for (i = 0; i < length; i++)
  {
    insist (src [i] >= 1 - LX_XBAR_SIZE && src [i] < LX_XBAR_SIZE);
    dst [i] = src [i];
  }
  except:;
}

void lx_reverse_route (signed char*dst, signed char*src, int length)
{
  int i;
  insist (dst && src && length <= LX_ROUTE_SIZE);

  for (i = 0; i < length; i++)
  {
    insist (src [length - 1 - i] >= 1 - LX_XBAR_SIZE && src [length - 1 - i] < LX_XBAR_SIZE);
    dst [i] = - src [length - 1 - i];
  }

  except:;
}

int lx_there_and_back_again (signed char*dst, signed char*src, signed char port, int length)
{
  insist (dst && src && length * 2 + 3 <= LX_ROUTE_SIZE);
 
  lx_copy_route (dst, src, length);
  dst [length] = port;
  dst [length + 1] = 0;
  dst [length + 2] = -port;
  
  lx_reverse_route (dst + length + 3, src, length);
  
  return length * 2 + 3;
  except: return 0;
}

void lx_physical_route (signed char*p, int length)
{
  int i;
  insist (p && length <= LX_ROUTE_SIZE);
  
  for (i = 0; i < length; i++)
    p[i] = (p[i] & '\x3f') | '\x80';

  except:;
}

void lx_logical_route (signed char*p, int length)
{
  int i;
  
  insist (p && length <= LX_ROUTE_SIZE);

  for (i = 0; i < length; i++)
    if ((p [i] = (p[i] & 0x3f)) & 0x20)
      p [i] |= 0xf0;

  except:;
}

lx_route_t*lx_append_route (lx_route_t*dst, lx_route_t*route, signed char port)
{
  insist (dst && route && route->length < LX_ROUTE_SIZE);
  insist (port >= 1 - LX_XBAR_SIZE && port < LX_XBAR_SIZE);

  lx_copy_route (dst->hops, route->hops, route->length);
  dst->hops [route->length] = port;
  dst->length = route->length + 1;

  return dst;
  except: return 0;;
}


char*lx_print_route (signed char*p, int length)
{
  static char s [LX_ROUTE_SIZE * 4];
  char*t = s;
  int i;

  insist (s && p && length <= LX_ROUTE_SIZE);

  *s = 0;

  for (i = 0; i < length; i++)
    t += sprintf (t, "%d,", p [i]);

  if (t > s)
    t [-1] = 0;
  return s;
  except: return 0;
}

static int lx_print_node (lx_t*lx, lx_node_t*n)
{
  int i, c = 1;
  lx_node_t*nn;
  
  insist (lx && n);
  
  if (n->xbar)
  {
    for (c = 0, i = 1 - LX_XBAR_SIZE; i < LX_XBAR_SIZE; i++)
      c += lx_connected (n, i) ? 1 : 0;
  }
  
  mi_println_map (lx, "%s - " lx_node_format, n->xbar ? "s" : "h", lx_node_args (n));
  mi_println_map (lx, "%d", c);

  for (i = 1 - LX_XBAR_SIZE; i < LX_XBAR_SIZE; i++)
  {
    if ((nn = lx_get_node (n, i)))
    {
      insist (i >= n->first);
      insist (lx_get_port (n, i) >= nn->first);
      mi_println_map (lx, "%d %s - " lx_node_format " %d", i - n->first, nn->xbar ? "s" : "h", lx_node_args (nn), lx_get_port (n, i) - nn->first);
    }
  }
  
  if (!n->xbar)
  {
    mi_println_map (lx, "address " lx_mac_format, lx_mac_args (lx_host_c (n)->mac_address));
    mi_println_map (lx, "hostType %d", lx_host_c (n)->host_type);
  }
  mi_println_map (lx, "");
  
  return 1;  
  except: return 0;
}

static int lx_print_nodes (lx_t*lx, lx_queue_t*queue)
{
  lx_node_t*n;
  
  insist (lx && queue);

  for (n = queue->head; n; n = mi_index2node (n->next))
    if (!lx_print_node (lx, n))
      return 0;
  
  return 1;
  except: return 0;
  
}

int lx_map_print (lx_t*lx, lx_map_t*m)
{
  insist (lx && m);
  mi_println_map (lx, "; map_address " lx_mac_format " map_version %u", lx_mac_args (lx->map_address), lx->map_version);
  
  return lx_print_nodes (lx, &m->hosts) && lx_print_nodes (lx, &m->xbars);
  except: return 0;
}

void lx_bzero (char*p, unsigned length)
{
  unsigned i;
  for (i = 0; i < length; i++)
    p [i] = 0;
}

int lx_memcmp (unsigned char*a, unsigned char*b, unsigned length)
{
  unsigned i;
  for (i = 0; i < length; i++)
  {
    if (a [i] < b [i])
      return -1;
    else if (a [i] > b [i])
      return 1;
  }
  return 0;
}

int lx_zero (unsigned char*p, unsigned length)
{
  unsigned i;
  for (i = 0; i < length; i++)
    if (p [i]) return 0;
  return 1;
}

unsigned lx_dice (unsigned throws, unsigned sides)
{
  return throws < sides ? (unsigned) mi_rand () <  (unsigned) MI_RAND_MAX / sides * throws : 1;
}

void lx_reset_map (lx_t*lx, lx_map_t*m)
{
  lx_node_t*n;
  
  insist (lx && m);
  insist (!lx->map_valid);

  m->lx = lx;
  
  while ((n = (lx_node_t*) lx_get (&m->hosts)))
  {
    lx_put (&lx->free_hosts, n);
    mi_punt (lx);
  }
  
  while ((n = (lx_node_t*) lx_get (&m->xbars)))
  {
    lx_put (&lx->free_xbars, n);
    mi_punt (lx);
  }

  while ((n = (lx_node_t*) lx_get (&m->frontier)))
  {
    lx_put (&lx->free_xbars, n);
    mi_punt (lx);
  }
  
  insist (!lx_get (&m->merge_queue));
  
  m->root = 0;
  m->num_hosts = m->num_xbars = 0;
  m->xbars_created = 0;
  
  except:;
}


void lx_init (lx_t*lx)
{
  int i;
  
  insist (lx);  

  lx_bzero ((char*)&lx->free_hosts, sizeof (lx_queue_t));
  lx_bzero ((char*)&lx->free_xbars, sizeof (lx_queue_t));
  
  lx_bzero ((char*)&lx->host_block, sizeof (lx_host_t) * LX_TOTAL_HOSTS);
  lx_bzero ((char*)&lx->xbar_block, sizeof (lx_xbar_t) * LX_TOTAL_XBARS);

  lx_bzero ((char*)&lx->staged_hosts, sizeof (lx_queue_t));
  lx_bzero ((char*)&lx->staged_xbars, sizeof (lx_queue_t));

  lx_bzero ((char*)&lx->map.xbars, sizeof (lx_queue_t));
  lx_bzero ((char*)&lx->map.hosts, sizeof (lx_queue_t));
  lx_bzero ((char*)&lx->map.frontier, sizeof (lx_queue_t));
  lx_bzero ((char*)&lx->map.merge_queue, sizeof (lx_queue_t));
    
  lx_bzero ((char*)&lx->other.xbars, sizeof (lx_queue_t));
  lx_bzero ((char*)&lx->other.hosts, sizeof (lx_queue_t));
  lx_bzero ((char*)&lx->other.frontier, sizeof (lx_queue_t));
  lx_bzero ((char*)&lx->other.merge_queue, sizeof (lx_queue_t));

  lx->bigs_sent = lx->smalls_sent = lx->bigs_pending = lx->smalls_pending = lx->replies_pending = lx->replies_sent = 0;
  lx->printeurs_pending = lx->printeurs_sent = 0;
  
  lx->duration = 0;
  lx->die = 0;
  lx->in_receive = 0;
  lx->map_valid = 0;
  
  lx_bzero ((char*) lx->map_address, 6);
  lx_bzero ((char*) lx->winner_address, 6);

  for (i = 1; i < LX_TOTAL_HOSTS; i++)
  {
    lx_node_t*n = (lx_node_t*) &lx->host_block [i];
    n->xbar = 0;
    lx_put (&lx->free_hosts, n);
  }
  for (i = 0; i < LX_TOTAL_XBARS; i++)
  {
    lx_node_t*n = (lx_node_t*) &lx->xbar_block [i];
    n->xbar = 1;
    lx_put (&lx->free_xbars, n);
  }
  lx_reset_map (lx, &lx->map);
  lx_reset_map (lx, &lx->other);

  lx->printeur_route.length = 0;
  
  except:;
}

int lx_receive (lx_t*lx, char*p, unsigned length, int early)
{
  int reason;
  unsigned i;
  void*contexts [LX_NUM_SMALLS + LX_NUM_REPLIES + LX_NUM_BIGS + LX_NUM_PRINTEURS];
  unsigned message_length;
  
  insist (lx->bigs_pending || lx->smalls_pending || lx->replies_pending || lx->printeurs_pending || lx->duration);
  insist (!lx->in_receive);

  lx->in_receive = 1;  
  
  if (lx->set_alarm)
  {
    insist (lx->duration);
    lx->set_alarm = 0;
    mi_set_alarm (lx, lx->duration);
  }  

  reason = mi_wait (lx, (char*) lx->receive_buffer, &message_length, contexts, early);  
  
  switch (reason)
  {
    case LX_RECEIVE:
      lx_map_receive_callback (lx, (lx_packet_t*) lx->receive_buffer, message_length);

      if (message_length >= length && p)
      {
	if (p != lx->receive_buffer)
	  mi_memcpy (p, (char*) lx->receive_buffer, length);
	lx->in_receive = 0;
	return message_length;
      }

      break;
    case LX_ALARM:
      insist (lx->duration);
      lx->duration = 0;
      /*mi_c (("alarm time %u", mi_rtc - rtc));*/
      
      if (lx->timeout > LX_TIMEOUT_DECREASE)
      {
	lx->timeout -= LX_TIMEOUT_DECREASE;
	if (lx->timeout < lx->start_timeout)
	  lx->timeout = lx->start_timeout;
      }
      lx->in_receive = 0;
      return -1;
    case LX_SEND_DONE:
      for (i = 0; i < message_length; i++)
      {  
	if (contexts [i] == LX_SMALL_DONE)
	{	  
	  insist (lx->smalls_pending && lx->smalls_pending <= LX_NUM_SMALLS);
	  lx->smalls_pending--;
	}
	else if (contexts [i] == LX_REPLY_DONE)
	{  
	  insist (lx->replies_pending && lx->replies_pending <= LX_NUM_REPLIES);
	  lx->replies_pending--;
	}
	else if (contexts [i] == LX_BIG_DONE)
	{
	  insist (lx->bigs_pending && lx->bigs_pending <= LX_NUM_BIGS);
	  lx->bigs_pending--;
	}
	else 
	{
	  insist (contexts [i] == LX_PRINTEUR_DONE);
	  insist (lx->printeurs_pending && lx->printeurs_pending <= LX_NUM_PRINTEURS);
	  lx->printeurs_pending--;
	}
      }
      break;
  
    case LX_PUNT:
      break;
    case LX_ABORT:
      /*mi_c (("aborted with %d", length));*/
      insist (0);
      break;
    default:
      lx->in_receive = 0;
      insist (0);
  }
  lx->in_receive = 0;
  except: return 0;
}

int lx_send_scout (lx_t*lx, lx_route_t*route, unsigned timestamp, int reason, int*port, int*host_type, unsigned char mac_address [6], unsigned*map_version, unsigned char map_address [6], int*level, int*flags, int*context)
{
  lx_scout_message_t*m;
  lx_scout_reply_message_t t;
  lx_small_t*p;
  int r;  

  insist (route);
  insist (port && host_type && mac_address && map_version && map_address && level && flags && context);
  insist (sizeof (m) <= LX_SMALL_MTU);

  while (lx->smalls_pending == LX_NUM_SMALLS)
  {
    if ((r = lx_receive (lx, (char*) &t, sizeof (t), 0)) < 0)
      return 0;
    else if (r >= (int) sizeof (t))
    {
      if (t.header.subtype == mi_htons (LX_SCOUT_REPLY) &&
	  !lx_memcmp (t.header.echo_address, lx->mac_address, 6))
      {
	if (lx_compare_time (lx, t.header.timestamp, mi_htonl (timestamp)))
	{
	  *port = t.header.port;
	  *host_type = t.host_type;
	  *flags = mi_htonl (t.flags);
	  *level = t.header.level;
	  *context = mi_htonl (t.header.context);
	  mi_memcpy (mac_address, t.header.sender_address, 6);
	  *map_version = mi_htonl (t.map_version);
	  mi_memcpy (map_address, t.map_address, 6);

	  return -1;
	}
      }
    }
  }

  /*mi_c (("sending scout %d to route %s", timestamp, lx_print_route (route->hops, route->length)));*/

  if (++lx->smalls_sent == LX_NUM_SMALLS)
    lx->smalls_sent = 0;
  
  p = &lx->smalls [lx->smalls_sent];
  m = (lx_scout_message_t*) p->p;
  
  m->header.type = mi_htons (LX_TYPE);
  m->header.subtype = mi_htons (LX_SCOUT);
  m->header.timestamp = mi_htonl (timestamp);
  mi_memcpy (m->header.sender_address, lx->mac_address, 6);
  m->header.port = *port;
  m->header.level = lx->level;
  m->header.reason = reason;
  m->header.context = mi_htonl (*context);

  mi_memcpy (m->map_address, lx->map_address, 6);
  m->map_version = mi_htonl (lx->map_version);

  m->flags = 0;
  
  if (lx->map_valid)
    m->flags |= LX_FLAG_MAP_VALID;
  if (lx->kids_match)
    m->flags |= LX_FLAG_KIDS_MATCH;
  if (lx->done_everywhere)
    m->flags |= LX_FLAG_DONE_EVERYWHERE;
  if (lx->printeur)
    m->flags |= LX_FLAG_PRINTEUR;
  if (lx->pause)
    m->flags |= LX_FLAG_PAUSE;

  m->flags = mi_htons (m->flags);
  m->route_length = route->length;
  lx_reverse_route (m->route, route->hops, route->length);
  p->r.length = route->length;

  lx_copy_route (p->r.hops, route->hops, p->r.length);
  lx_physical_route (p->r.hops, p->r.length);
  
  lx->smalls_pending++;
  mi_send (lx, p->r.hops, p->r.length, (char*) m, sizeof (lx_scout_message_t), LX_SMALL_DONE);

  return 1;
  except: return 0;  
}

int lx_send_scout_reply (lx_t*lx, lx_scout_message_t*scout)
{
  lx_scout_reply_message_t*m;
  lx_small_t*p;  
  
  insist (sizeof (m) <= LX_SMALL_MTU);
  insist (lx->replies_pending != LX_NUM_REPLIES);
  
  if (++lx->replies_sent == LX_NUM_REPLIES)
    lx->replies_sent = 0;
  
  p = &lx->replies [lx->replies_sent];
  m = (lx_scout_reply_message_t*) p->p;
  
  /*mi_c (("sending scout reply %d to " lx_mac_format, mi_htonl (scout->header.timestamp), lx_mac_args (scout->header.sender_address)));*/
  
  m->header.type = mi_htons (LX_TYPE);
  m->header.subtype = mi_htons (LX_SCOUT_REPLY);
  m->header.timestamp = scout->header.timestamp;
  m->header.port = scout->header.port;
  m->header.context = scout->header.context;
  m->header.level = lx->level;
  mi_memcpy (m->header.sender_address, lx->mac_address, 6);
  mi_memcpy (m->header.echo_address, scout->header.sender_address, 6);
  mi_memcpy (m->map_address, lx->map_address, 6);
  m->map_version = mi_htonl (lx->map_version);
  m->host_type = lx->host_type;

  m->flags = 0;  
  if (lx->map_valid)
    m->flags |= LX_FLAG_MAP_VALID;
  if (lx->kids_match)
    m->flags |= LX_FLAG_KIDS_MATCH;
  if (lx->done_everywhere)
    m->flags |= LX_FLAG_DONE_EVERYWHERE;
  if (lx->printeur)
    m->flags |= LX_FLAG_PRINTEUR;
  if (lx->pause)
    m->flags |= LX_FLAG_PAUSE;
  
  m->flags = mi_htonl (m->flags);
  
  p->r.length = scout->route_length;  
  lx_copy_route (p->r.hops, scout->route, scout->route_length);
  lx_physical_route (p->r.hops, p->r.length);

  /*lx_dump ((char*)m, sizeof (lx_scout_reply_message_t));*/

  lx->replies_pending++;
  mi_send (lx, p->r.hops, p->r.length, (char*) m, sizeof (lx_scout_reply_message_t), LX_REPLY_DONE);

  return 1;
  except: return 0;
}

int lx_send_map_request (lx_t*lx, lx_route_t*route, unsigned timestamp, int type, int first)
{
  lx_map_request_message_t*m;
  lx_small_t*p;
  
  insist (route);
  insist (sizeof (m) <= LX_SMALL_MTU);
  
  mi_c (("sending map request %d to " lx_mac_format, timestamp, lx_mac_args (lx->winner_address)));
  
  while (lx->smalls_pending == LX_NUM_SMALLS)
    lx_receive (lx, 0, 0, 0);

  if (++lx->smalls_sent == LX_NUM_SMALLS)
    lx->smalls_sent = 0;
  
  p = &lx->smalls [lx->smalls_sent];
  m = (lx_map_request_message_t*) p->p;
  
  m->header.type = mi_htons (LX_TYPE);
  m->header.subtype = mi_htons (LX_MAP_REQUEST);
  m->header.timestamp = mi_htonl (timestamp);
  mi_memcpy (m->header.sender_address, lx->mac_address, 6);
  m->type = type;
  m->first = mi_htons (first);
  
  m->route_length = route->length;
  lx_reverse_route (m->route, route->hops, route->length);

  p->r.length = route->length;
  lx_copy_route (p->r.hops, route->hops, route->length);
  lx_physical_route (p->r.hops, p->r.length);
  
  lx->smalls_pending++;
  mi_send (lx, p->r.hops, p->r.length, (char*) m, sizeof (lx_scout_message_t), LX_SMALL_DONE);

  return 1;
  except: return 0;  
}

int lx_send_xbar (lx_t*lx, lx_route_t*route, int timestamp, int*port, int*context)
{
  lx_xbar_message_t t, *m;
  int r;
  lx_small_t*p;

  insist (route && port && context);
  insist (sizeof (m) <= LX_SMALL_MTU);
  
  while (lx->smalls_pending == LX_NUM_SMALLS)
  {
    if ((r = lx_receive (lx, (char*) &t, sizeof (t), 0)) < 0)
      return 0;
    else if (r >= (int) sizeof (t))
    {
      if (t.header.subtype == mi_htons (LX_XBAR) && 
	  !lx_memcmp (t.header.sender_address, lx->mac_address, 6))
      {
	if (lx_compare_time (lx, t.header.timestamp, mi_htonl (timestamp)))
	{
	  *port = t.header.port;
	  *context = mi_htonl (t.header.context);
	  return -1;
	}
      }
    }
  }

  /*mi_c (("sending xbar to route %s", lx_print_route (route->hops, route->length)));*/

  if (++lx->smalls_sent == LX_NUM_SMALLS)
    lx->smalls_sent = 0;
  
  p = &lx->smalls [lx->smalls_sent];
  m = (lx_xbar_message_t*) p->p;
  
  m->header.type = mi_htons (LX_TYPE);
  m->header.subtype = mi_htons (LX_XBAR);
  m->header.timestamp = mi_htonl (timestamp);
  mi_memcpy (m->header.sender_address, lx->mac_address, 6);
  m->header.port = *port;
  m->header.context = mi_htonl (*context);
  
  p->r.length = route->length;
  lx_copy_route (p->r.hops, route->hops, p->r.length);
  lx_physical_route (p->r.hops, p->r.length);

  lx->smalls_pending++;
  mi_send (lx, p->r.hops, p->r.length, (char*) m, sizeof (lx_xbar_message_t), LX_SMALL_DONE);
  
  return 1;
  except: return 0;
}

int lx_wait_xbar (lx_t*lx, unsigned duration, int timestamp, int*port, int*context, int early)
{
  lx_xbar_message_t m;
  int r;

  insist (port && context);

  if (duration)
    lx_set_alarm (lx, duration);

  while ((r = lx_receive (lx, (char*) &m, sizeof (m), early)) >= 0)
  {
    if (r < (int) sizeof (m) || m.header.subtype != mi_htons (LX_XBAR) || lx_memcmp (m.header.sender_address, lx->mac_address, 6))
      continue;
    
    if (lx_compare_time (lx, mi_htonl (m.header.timestamp), timestamp))
      break;
  }
  
  if (r < 0)
    return 0;

  if (duration)
    lx_clear_alarm (lx);

  *port = m.header.port;
  *context = mi_htonl (m.header.context);

  return 1;
  except: return 0;
}


int lx_wait_map_reply (lx_t*lx, unsigned duration, int timestamp, int type, lx_map_reply_message_t*reply)
{
  int r;
  lx_map_reply_message_t*m;
  
  insist (reply);

  m = (lx_map_reply_message_t*) lx->receive_buffer;

  lx_set_alarm (lx, duration);
  
  while ((r = lx_receive (lx, (char*) m, sizeof (lx_map_reply_message_t), 0)) >= 0)
  {
    if  (r < (int) sizeof (lx_map_reply_message_t) ||
	 m->header.subtype != mi_htons (LX_MAP_REPLY) ||
	 lx_memcmp (m->header.echo_address, lx->mac_address, 6))
      continue;
    
    if (lx_compare_time (lx, mi_htonl (m->header.timestamp), timestamp))
      break; 
  }
  
  if (r < 0)
    return 0;
  lx_clear_alarm (lx);

  mi_memcpy ((char*) reply, (char*) m, r);
  
  return r;
  except: return 0;
}

int lx_wait_scout_reply (lx_t*lx, unsigned duration, int timestamp, int*port, int*host_type, unsigned char mac_address [6], unsigned*map_version, unsigned char map_address [6], int*level, int*flags, int*context)
{
  lx_scout_reply_message_t m;
  int r;
  
  insist (port && host_type && mac_address && map_version && map_address && flags && context);
  insist (level);
  
  if (duration)
    lx_set_alarm (lx, duration);

  while ((r = lx_receive (lx, (char*) &m, sizeof (m), 0)) >= 0)
  {
    if (r < (int) sizeof (m) || m.header.subtype != mi_htons (LX_SCOUT_REPLY) || lx_memcmp (m.header.echo_address, lx->mac_address, 6))
      continue;

    if (lx_compare_time (lx, mi_htonl (m.header.timestamp), timestamp))
      break;
  }
  
  if (r < 0)
    return 0;
	  
  if (duration)
    lx_clear_alarm (lx);

  /*lx_dump ((char*)&m, r);*/

  *port = m.header.port;
  *host_type = m.host_type;
  mi_memcpy (mac_address, m.header.sender_address, 6);
  *map_version = mi_htonl (m.map_version);
  *level = m.header.level;
  *flags = mi_htonl (m.flags);
  *context = mi_htonl (m.header.context);
  
  mi_memcpy (map_address, m.map_address, 6);

  return 1;
  except: return 0;
}

int lx_wait (lx_t*lx, unsigned duration, int*variable, int early)
{
  int r = 0;  
  insist (lx);
  
  if (variable && *variable) return 1;
  
  lx_set_alarm (lx, duration);
  
  while ((!variable || !*variable) && (r = lx_receive (lx, 0, 0, early)) >= 0)
    if (lx->die)
      return 0;

  if (r < 0)
    return 0;
  
  lx_clear_alarm (lx);
  
  return 1;
  except: return 0;
}

lx_node_t*lx_seen (lx_map_t*m, unsigned char mac_address [6])
{
  lx_node_t*n;
  insist (m && mac_address);
  
  for (n = m->hosts.head; n; n = mi_index2node (n->next))
  { 
    if (!lx_memcmp (lx_host_c (n)->mac_address, mac_address, 6))
      return n;
    mi_punt (m->lx);
  }
  
  except: return 0;
}

void lx_set_alarm (lx_t*lx, unsigned duration)
{
  insist (!lx->duration && duration);
  lx->set_alarm = 1;
  lx->duration = duration;  
  except:;
}

void lx_clear_alarm (lx_t*lx)
{
  insist (lx->duration);
  insist (!lx->set_alarm);
  
  mi_set_alarm (lx, lx->duration = 0);
  except:;
}


void lx_xbar2net (lx_map_t*m, lx_xbar_t*from, lx_map_xbar_t*to)
{
  int i;
  lx_node_t*n;
  
  insist (m && from && to);
  
  to->level = from->level;
  
  for (i = 0; i < LX_XBAR_SIZE; i++)
  {
    if ((n = mi_index2node (lx_node_c (from)->links [i].index)))
    {
      to->links [i].index = 1 + (n->index + (n->xbar ? m->num_hosts : 0));
      insist (to->links [i].index > 0 && to->links [i].index <= m->num_hosts + m->num_xbars);
      to->links [i].port = lx_node_c (from)->links [i].port -  n->first;
      to->links [i].index = mi_htons (to->links [i].index);
    }
    else
      to->links [i].index = to->links [i].port = 0;
  }
  except:;
}

void lx_net2xbar (lx_map_t*m, lx_map_xbar_t*from, lx_xbar_t*to, int num_hosts, int num_xbars)
{
  int i;
  lx_node_t*n;
  lx_t*lx = m->lx;
  
  insist (!lx->map_valid);  
  insist (m && from && to);
  
  lx_node_c (to)->first = 0;
  to->level = from->level;
  
  for (i = 0; i < LX_XBAR_SIZE; i++)
  {
    int index = mi_htons (from->links [i].index);
    
    insist (index >= 0 && index <= num_hosts + num_xbars);

    if (index)
    {
      index--;
      n = mi_index2node ((index < num_hosts ? lx->hosts [index] : lx->xbars [index - num_hosts]));
      insist (n);
      
      lx_node_c (to)->links [i].index = mi_node2index (n);
      lx_node_c (to)->links [i].port = from->links [i].port;
    }
    else
      lx_node_c (to)->links [i].index = 0;
  }
  except:;
}

void lx_host2net (lx_map_t*m, lx_host_t*from, lx_map_host_t*to)
{
  lx_node_t*n;
  
  insist (m && from && to);

  insist (!lx_zero (from->mac_address, 6));
  
  mi_memcpy (to->mac_address, from->mac_address, 6);
  to->host_type = from->host_type;
  to->level = from->level;
  
  if ((n = mi_index2node (lx_node_c (from)->links [0].index)))
  {  
    to->link.index = 1 + (n->index + (n->xbar ? m->num_hosts : 0));
    insist (to->link.index > 0 && to->link.index <= m->num_hosts + m->num_xbars);
    to->link.port = lx_node_c (from)->links [0].port - n->first;
    to->link.index = mi_htons (to->link.index);
  }
  else
    to->link.index = to->link.port = 0;

  except:;
}

void lx_net2host (lx_map_t*m, lx_map_host_t*from, lx_host_t*to, int num_hosts, int num_xbars)
{ 
  int index;
  lx_node_t*n;
  lx_t*lx = m->lx;
  
  insist (m && from && to);
  insist (num_hosts <= LX_MAX_HOSTS);
  insist (num_xbars <= LX_MAX_XBARS);
  insist (!lx->map_valid);
  
  lx_node_c (to)->first = 0;
  insist (!lx_zero (from->mac_address, 6));
  
  mi_memcpy (to->mac_address, from->mac_address, 6);
  to->host_type = from->host_type;
  to->level = from->level;
  
  index = mi_htons (from->link.index);
  insist (index >= 0 && index <= num_hosts + num_xbars);

  if (index)
  {
    index--;
    n = mi_index2node ((index < num_hosts ? lx->hosts [index] : lx->xbars [index - num_hosts]));
    insist (n);
      
    lx_node_c (to)->links [0].index = mi_node2index (n);
    lx_node_c (to)->links [0].port = from->link.port;
  }
  else
    lx_node_c (to)->links [0].index = 0;

  except:;
}


int lx_number_xbars_bfs (lx_map_t*m)
{
  lx_queue_t q;
  lx_node_t*n, *nn;
  int xbars_found, i;
  
  insist (m && m->root);
  
  xbars_found = 0;
  
  for (i = 0, n = m->xbars.head; n; n = mi_index2node (n->next), i++)
    lx_xbar_c (n)->level = (unsigned char) -1;

  insist (i == m->num_xbars);
  
  n = lx_get_node (m->root, 0);
  insist (n);
  
  if (!n->xbar)
    return 1;

  lx_bzero ((char*) &q, sizeof (q));
  
  lx_xbar_c (n)->level = 0;

  lx_put (&q, lx_remove (&m->xbars, n));
  
  while ((n = lx_get (&q)))
  {
    lx_put (&m->xbars, n);
  
    for (i = 0; i < LX_XBAR_SIZE; i++)
    { 
      mi_punt (m->lx);
      if ((nn = lx_get_node (n, i + n->first)) && nn->xbar && lx_xbar_c (nn)->level == (unsigned char) -1)
      {
	lx_xbar_c (nn)->level = lx_xbar_c (n)->level + 1;
	lx_put (&q, lx_remove (&m->xbars, nn));
      }
    }
  }

  insist (!q.count);
  return 1;
  except: return 0;
}



int lx_number_xbars_clos (lx_map_t*m, int max_level)
{
  lx_queue_t a;
  lx_queue_t b;

  lx_node_t*n, *nn;
  int num_xbars, i;
  
  insist (m && m->root); 

  lx_bzero ((char*) &a, sizeof (a)); 
  lx_bzero ((char*) &b, sizeof (b)); 

  num_xbars = m->xbars.count;
  
  while ((n = lx_get (&m->xbars)))
  {
    insist (n->xbar);
    mi_punt (m->lx);

    if (n->route.length == max_level)
    {
      lx_xbar_c (n)->level = max_level;
      lx_put (&b, n);
    }
    else
    {
      lx_xbar_c (n)->level = (unsigned char) -1;
      lx_put (&a, n);
    }
  }
  
  while ((n = lx_get (&b)))
  {
    insist (lx_xbar_c (n)->level != (unsigned char) -1);

    for (i = 0; i < LX_MAX_XBARS; i++)
    {
      mi_punt (m->lx);
      
      if ((nn = lx_get_node (n, i + n->first)) && nn->xbar && lx_xbar_c (nn)->level == (unsigned char) -1)
      {
	lx_xbar_c (nn)->level = lx_xbar_c (n)->level - 1;
	/*mi_c ((lx_node_format " level %d", lx_node_args (nn), lx_xbar_c (nn)->level));*/
	
	if (lx_xbar_c (nn)->level > lx_xbar_c (n)->level)
	{
	  lx_xbar_c (nn)->level = 0; /*not a clos?*/
	}
	
	insist (lx_queued (&a, nn));
	lx_put (&b, lx_remove_from_anywhere (nn));
      }
    }
    lx_put (&m->xbars, n);
  }

  insist (m->xbars.count == num_xbars);
  insist (!a.count && !b.count);
  
  except: return 0;
}

int lx_before_sending (lx_map_t*m)
{
  int i;
  lx_node_t*n;
  lx_t*lx = m->lx;
  
  insist (m);
  insist (!lx->map_valid);

  n = (lx_node_t*) lx_peek (&m->hosts);
  for (i = 0; n; i++, n = mi_index2node (n->next))
  {
    insist (!n->next || lx_hostcmp (lx_host_c (n)->mac_address, lx_host_c (n)->level, lx_host_c (mi_index2node (n->next))->mac_address, lx_host_c (mi_index2node (n->next))->level) > 0);
    
    lx->hosts [i] = mi_node2index (n);
    n->index = i;
    mi_punt (m->lx);
  }

  insist (i == m->num_hosts);

  n = (lx_node_t*) lx_peek (&m->xbars);
  for (i = 0; n; i++, n = mi_index2node (n->next))
  {
    lx->xbars [i] = mi_node2index (n);
    n->index = i;
    mi_punt (m->lx);
  }
  insist (i == m->num_xbars);

  return 1;
  except: return 0;
}

void lx_before_receiving (lx_map_t*m, int num_hosts, int num_xbars)
{
  int i;
  lx_node_t*n;
  lx_t*lx = m->lx;
  
  insist (m);
  insist (m->num_hosts == 0 && m->num_xbars == 0);
  insist (!lx->map_valid);
  
  for (i = 0; i < num_hosts; i++)
  {
    n = lx_new_host (m, 0, 0, 0, 0);
    insist (n);
    lx->hosts [i] = mi_node2index (n);
    n->index = i;
    lx_put (&m->lx->staged_hosts, n);
    mi_punt (m->lx);
  } 
  insist (i == m->num_hosts);
  
  for (i = 0; i < num_xbars; i++)
  {
    n = lx_new_xbar (m, 0);
    insist (n);

    lx->xbars [i] = mi_node2index (n);
    n->index = i;
    lx_put (&m->lx->staged_xbars, n);
    mi_punt (m->lx);
  }
  insist (i == m->num_xbars);
  except:;
}

int lx_send_map_reply (lx_t*lx, lx_map_t*m, lx_map_request_message_t*request)
{
  int i, first;
  lx_map_reply_message_t*mr;
  lx_big_t*p;
  int piece_size;
  
  insist (m && request);
  insist (lx->map_version);
  insist (m == &lx->map);
  insist (lx->map_valid);
  insist (lx->bigs_pending != LX_NUM_BIGS);
  
  if (++lx->bigs_sent == LX_NUM_BIGS)
    lx->bigs_sent = 0;
  
  p = (lx_big_t*) &lx->bigs [lx->bigs_sent];
  
  mr = (lx_map_reply_message_t*) p->p;

#ifndef MI_XM
  {
    lx_node_t*higher = 0;

    if (!lx_find_host (lx, &lx->map.hosts, request->header.sender_address, 0, &higher))
    {
      mi_c ((lx_mac_format " is not in my map", lx_mac_args (request->header.sender_address)));
      mi_c (("Therefore my map is invalid."));
      lx->map_valid = 0;
      lx->aborted = LX_ABORT_ACTIVE;
    }
  }
  
#endif
  mi_c (("sending map reply to " lx_mac_format, lx_mac_args (request->header.sender_address)));
  
  mr->header.type = mi_htons (LX_TYPE);
  mr->header.subtype = mi_htons (LX_MAP_REPLY);
  mr->header.timestamp = request->header.timestamp;
  mr->header.port = request->header.port;
  mr->header.context = request->header.context;
  mi_memcpy (mr->header.sender_address, lx->mac_address, 6);
  mi_memcpy (mr->header.echo_address, request->header.sender_address, 6);
  mi_memcpy (mr->map_address, lx->map_address, 6);

  mr->type = request->type;
  mr->num_hosts = mi_htons (m->num_hosts);
  mr->num_xbars = mi_htons (m->num_xbars);
  mr->num_nodes = 0;
  mr->first = request->first;
  lx_bzero ((char*) mr->map_address, 6);
  mr->map_version = mi_htonl (lx->map_version);

  first = mi_htons (request->first);

  mi_memcpy (mr->map_address, lx->map_address, 6);
  
  switch (request->type)
  {
    case LX_MAP_HOST:
      piece_size = sizeof (lx_map_host_t);
      for (i = first; i < m->num_hosts && i - first < lx_piece_count (lx_map_host_t); i++)
      {
	insist (i >= 0 && i < m->num_hosts);
	lx_host2net (m, (lx_host_t*) mi_index2node (lx->hosts [i]), &mr->nodes.host [i - first]);
	mr->num_nodes++;
      }
      break;
    case LX_MAP_XBAR:
      piece_size = sizeof (lx_map_xbar_t);
      for (i = first; i < m->num_xbars && i - first < lx_piece_count (lx_map_xbar_t); i++)
      {
	insist (i >= 0 && i < m->num_xbars);
	lx_xbar2net (m, (lx_xbar_t*) mi_index2node (lx->xbars [i]), &mr->nodes.xbar [i - first]);
	mr->num_nodes++;
      }
      break;
    default:
      insist (0);
      return 0;
  }

  p->r.length = request->route_length;
  lx_copy_route (p->r.hops, request->route, p->r.length);
  lx_physical_route (p->r.hops, p->r.length);

  lx->bigs_pending++;
  piece_size = sizeof (lx_map_reply_message_t) + piece_size * ((mr->num_nodes - (mr->num_nodes ? 1 : 0)));
  insist (piece_size <= LX_BIG_MTU);
  
  mr->num_nodes = mi_htons (mr->num_nodes);
  mi_send (lx, p->r.hops, p->r.length, (char*) mr, piece_size, LX_BIG_DONE);
  return 1;
  except: return 0;
}

int lx_compare_time (lx_t*lx, unsigned received, unsigned expected)
{
  insist (lx);

  if (received == expected)
  {
    lx->last_seen_timestamp = received;
    return 1;
  }
  
  if (received < expected && received != lx->last_seen_timestamp)
  {
    mi_c (("late message timestamp %u, expected %u", received, expected));
    if ((lx->timeout += lx->timeout / LX_TIMEOUT_INCREASE_FACTOR) > lx->start_timeout * LX_MAX_TIMEOUT_FACTOR)
      lx->timeout = lx->start_timeout * LX_MAX_TIMEOUT_FACTOR;
    
    mi_c (("timeout is now %u", lx->timeout));
  }
  except: return 0;
}

unsigned lx_advance_time (lx_t*lx)
{
  insist (lx);
  mi_punt (lx);
  return ++lx->timestamp;
  except: return 0;
}


lx_node_t*lx_follow (lx_node_t*n, lx_route_t*r)
{ 
  int i, p, in;
  insist (n && r);
  
  in = lx_get_port (n, 0);
  n = lx_get_node (n, 0);
  
  for (i = 0; i < r->length; i++)
  {
    p = lx_get_port (n, in + r->hops [i]);
    n = lx_get_node (n, in + r->hops [i]);
    insist (n);
    in = p;
  }
  return n;
  except: return 0;
}

int lx_hostcmp (unsigned char address1 [6], int level1, unsigned char address2 [6], int level2)
{
  insist (address1 && address2);
  return level1 == level2 ? lx_memcmp (address1, address2, 6) : (level1 > level2 ? 1 : -1);
  except: return 0;
}

int lx_higher_than_me (lx_t*lx, unsigned char mac_address [6], int level)
{
  insist (lx && mac_address);
  return lx_hostcmp (mac_address, level, lx->mac_address, lx->level) > 0;
  except: return 0;
}

void lx_set_timeout (lx_t*lx, unsigned timeout, unsigned tries)
{
  insist (lx);
  lx->timeout = lx->start_timeout = timeout;
  lx->tries = tries;
  
  except:;
}

lx_node_t*lx_find_host (lx_t*lx, lx_queue_t*queue, unsigned char mac_address [6], int level, lx_node_t**higher)
{
  int r;
  lx_node_t*n;
  insist (lx && queue && mac_address);
  insist (higher);
 
  *higher = 0;
  
  for (n = queue->head; n; n = mi_index2node (n->next))
  {
    if ((r = lx_hostcmp (lx_host_c (n)->mac_address, lx_host_c (n)->level, mac_address, level)) > 0)
      *higher = n;
    
    if (!r || !lx_memcmp (lx_host_c (n)->mac_address, mac_address, 6))
      return n;
    mi_punt (lx);
  }

  except: return 0;
}

void lx_reclaim_staged (lx_t*lx)
{
  lx_node_t*n;

  insist (lx);

  while ((n = (lx_node_t*) lx_get (&lx->staged_hosts)))
  {
    mi_punt (lx);
    lx_put (&lx->free_hosts, n);
  }

  while ((n = (lx_node_t*) lx_get (&lx->staged_xbars)))
  {
    mi_punt (lx);
    lx_put (&lx->free_xbars, n);
  }

  except:;
}	  

int lxd_cur;
int lxd_lines [LXD_MAX_LINES];
void lxd_line (int line)
{
  if (++lxd_cur == LXD_MAX_LINES)
    lxd_cur = 0;
    
  lxd_lines [lxd_cur] = line;

}

void lx_dump (char*p, int length)
{
#define MAX_DUMP 200
  char buffer [MAX_DUMP * 3];
  char*s = buffer;
  int i;
  
  if (length > MAX_DUMP ) length = MAX_DUMP;
  
  for (i = 0; i < length; i++)
    s += sprintf
      (s, "%2.2x%s", ((unsigned char*)p) [i],
       i  && (i + 1) % 16 == 0 ? "\n" : i && i % 2 == 1  ? " " : "");
  mi_c (("%d:\n%s\n\n", length, buffer));
}

void lx_send_printeur_message (const char*format, ...)
{
  va_list args;
  static int in = 0;
  lx_t*lx = mi_printeur_lx;
  lx_printeur_t*p;
  lx_printeur_message_t*m;

  if (in || !lx || !lx->printeur_route.length || lx->duration || lx->in_receive)
    return;

  in = 1;

  while (lx->printeurs_pending == LX_NUM_PRINTEURS)
    lx_receive (lx, 0, 0, 0);
  
  if (++lx->printeurs_sent == LX_NUM_PRINTEURS)
    lx->printeurs_sent = 0;

  p = &lx->printeurs [lx->printeurs_sent];
  m = (lx_printeur_message_t*) p->p;

  m->header.type = mi_htons (LX_TYPE);
  m->header.subtype = mi_htons (LX_PRINTEUR);
  mi_memcpy (m->header.sender_address, lx->mac_address, 6);
  p->r.length = lx->printeur_route.length;
  lx_copy_route (p->r.hops, lx->printeur_route.hops, p->r.length);
  lx_physical_route (p->r.hops, p->r.length);

  va_start (args, format);
  vsprintf ((char*) m->s, format, args);

  lx->printeurs_pending++;
  mi_send (lx, p->r.hops, p->r.length, (char*) m, sizeof (lx_printeur_message_t), LX_PRINTEUR_DONE);

  in = 0;
}


int lx_inflate_map (lx_map_t*m, int xbar_fan, int extra_hosts)
{
  int p, i, j, num_xbars, num_host_kids, num_xbar_kids;
  unsigned char mac_address [6];
  int hosts_made = 0;
  lx_node_t*n, *nn;
  lx_route_t route;
  lx_t*lx = m->lx;
  
  insist (m && lx);
  insist (xbar_fan < LX_XBAR_SIZE);
  insist (!lx->staged_xbars.count);
  
  num_xbar_kids = xbar_fan - 1;
  num_host_kids = LX_XBAR_SIZE - num_xbar_kids - 1;
  insist (num_xbar_kids > 0 && num_host_kids > 0);
  
  num_xbars = extra_hosts / num_host_kids;
  insist (num_xbars > 0);
  
  i = 0;
  for (n = (lx_node_t*) lx_peek (&m->xbars); n; n = mi_index2node (n->next))
  {
    for (i = 0; i < LX_XBAR_SIZE && lx_get_node (n, n->first + i); i++);
  
    if (i < LX_XBAR_SIZE)
      break;
  }

  if (!n)
    return 0;
  p = n->first + i;
  insist (p > -LX_XBAR_SIZE && p < LX_XBAR_SIZE);

  nn = lx_new_xbar (m, lx_append_route (&route, &n->route, p));
  insist (nn);
  lx_connect (n, p, nn, 0);
  lx_put (&lx->staged_xbars, nn);

  mac_address [0] = 0;
  mac_address [1] = 0x60;
  mac_address [2] = 0xdd;
  
  while ((n = lx_get (&lx->staged_xbars)))
  {
    lx_put (&m->xbars, n);
    
    for (i = 0, p = 1; i < num_xbar_kids && num_xbars; i++, num_xbars--, p++)
    {
      nn = lx_new_xbar (m, lx_append_route (&route, &n->route, p));
      insist (nn);
      lx_connect (n, p, nn, 0);
      lx_put (&lx->staged_xbars, nn);
    }
    for (i = 0; i < num_host_kids && hosts_made < extra_hosts; i++, p++)
    {
      insist (p < LX_XBAR_SIZE);

      hosts_made++;
      
      for (j = 3; j < 6; j++)
	mac_address [j] = (extra_hosts + 1 - hosts_made) >> (8 * (5 - j));
      
      nn = lx_new_host (m, mac_address, LX_DEBUG_TYPE, 0, lx_append_route (&route, &n->route, p));
      insist (nn);
      lx_connect (n, p, nn, 0);
      insist (m->hosts.tail);
      insist (lx_hostcmp (lx_host_c (m->hosts.tail)->mac_address, lx_host_c (m->hosts.tail)->level, mac_address, 0) > 0);
      lx_put (&m->hosts, nn);
    }
  }
  return 1;
  except: return 0;
}

int lx_get_highest_mac_address (lx_t*lx, unsigned char mac_address [6])
{
  int i;
  lx_map_t*m = &lx->map;
  lx_node_t*n;
  
  insist (lx && mac_address);
  
  if (!lx->map_valid)
    return 0;

  n = (lx_node_t*) lx_peek (&m->hosts);
  for (i = 0; n; i++, n = mi_index2node (n->next))
  {
    if (lx_host_c (n)->host_type == lx->host_type)
    {
      mi_memcpy (mac_address, lx_host_c (n)->mac_address, 6);
      return 1;
    }
  }

  except: return 0;
}
