#ifndef lx_h
#define lx_h

#define LX_MAX_HOSTS 2400
#define LX_MAX_XBARS 1200
#define LX_SMALL_MTU 100
#define LX_PRINTEUR_MTU (sizeof (lx_printeur_message_t))
#define LX_BIG_MTU 1000
#define LX_ROUTE_SIZE 23
#define LX_XBAR_SIZE 16
#define LX_MAX_TIMEOUT_FACTOR 10
#define LX_TIMEOUT_INCREASE_FACTOR 2
#define LX_PASSIVE_TIMEOUT (1 * 1000 * 1000)
#define LX_ACTIVE_TIMEOUT (4 * 1000)
#define LX_VERIFY_DISCONNECTED_TIMEOUT (1 * 1000 * 100)
#define LX_VERIFY_DISCONNECTED_TRIES 3
#define LX_VERIFY_CONNECTED_TIMEOUT (1 * 1000 * 1000)
#define LX_VERIFY_CONNECTED_TRIES 30
#define LX_BRANCH 2
#define LX_MAX_DEPTH 20

#define LX_PASSIVE_TRIES 3
#define LX_ACTIVE_TRIES 3
#define LX_COLLECT_TRIES 30
#define LX_COLLECT_TIMEOUT (1 * 1000 * 1000)
#define LX_TIMEOUT_DECREASE 500
#define LX_SMALL_DONE ((void*)1)
#define LX_REPLY_DONE ((void*) 2)
#define LX_BIG_DONE ((void*) 3)
#define LX_PRINTEUR_DONE ((void*) 4)
#define LX_NO_PORT -LX_XBAR_SIZE
#define LX_PRINTEUR_MAX_STRING 255

enum
{
  LX_GM_HOST_TYPE = 1,
  LX_XM_HOST_TYPE = 2,
  LX_MX_HOST_TYPE = 3,
  LX_DEBUG_TYPE = 4
};

enum
{
  LX_UP = 1,
  LX_DOWN = 2
};

enum
{
  LX_CHANGED_HOST = 1,
  LX_MISSING_HOST,
  LX_INVALID_MAP
};

enum
{
  LX_FLAG_MAP_VALID = 1,
  LX_FLAG_KIDS_MATCH = 2,
  LX_FLAG_DONE_EVERYWHERE = 4,
  LX_FLAG_PRINTEUR = 8,
  LX_FLAG_PAUSE = 16
};


enum
{
  LX_NONE,
  LX_RECEIVE,
  LX_ALARM,
  LX_SEND_DONE,
  LX_PUNT,
  LX_ABORT
};

#define LX_TYPE 0x700f
#define LX_SUBTYPE 0xa

enum
{
  LX_SCOUT = LX_SUBTYPE,
  LX_SCOUT_REPLY,
  LX_XBAR,
  LX_MAP_REQUEST,
  LX_MAP_REPLY,
  LX_PRINTEUR
};

enum
{
  LX_SCOUTING,
  LX_QUERYING_MAP_VERSION,
  LX_VERIFYING,
  LX_TREE
};

enum
{
  LX_ABORT_PASSIVE = 1,
  LX_ABORT_ACTIVE
};

#define lx_piece_count(t) (int)(((LX_BIG_MTU - sizeof (lx_map_reply_message_t)) / sizeof (t)))

enum
{
  LX_MAP_HOST = 0,
  LX_MAP_XBAR = 1
};

#define LX_NUM_SMALLS (LX_XBAR_SIZE * 2)
#define LX_NUM_REPLIES (LX_XBAR_SIZE)
#define LX_NUM_BIGS 4
#define LX_NUM_PRINTEURS 1

#define LX_MAX_LINE 100

#define lx_mac_format "%2.2x:%2.2x:%2.2x:%2.2x:%2.2x:%2.2x"
#define lx_mac_args(m) m[0] & 0xff , m[1] & 0xff , m[2] & 0xff , m[3] & 0xff , m[4] & 0xff , m[5] & 0xff
#define lx_node_args(n) ((n)->xbar ? "x" : "h"), (n)->index
#define lx_node_format "%s%d"

#define lx_host_c(n)((lx_host_t*)n)
#define lx_xbar_c(n)((lx_xbar_t*)n)
#define lx_node_c(n)((lx_node_t*)n)

struct lx_node_t;
struct lx_port_t;
struct lx_queue_t;

#ifdef	__cplusplus
extern "C"
{
#endif

#include "mi.h"

#ifdef	__cplusplus
}
#endif

typedef struct lx_route_t
{
  unsigned char length;
  signed char hops [(LX_ROUTE_SIZE - 1) / 2];
}lx_route_t;

typedef struct lx_extended_t
{
  lx_route_t _r;
  signed char hops [LX_ROUTE_SIZE / 2 + 1];
}lx_extended_t;

struct lx_t;
struct lx_node_t;

typedef struct lx_event_t
{
  int type;
  char buffer [LX_SMALL_MTU];
}lx_event_t;

typedef struct lx_queue_t
{
  int count;
  struct lx_node_t*head;
  struct lx_node_t*tail;
}lx_queue_t;


typedef struct mi_link_t
{
  mi_index_t index;
  signed char port;
  unsigned char weight;
}mi_link_t;

typedef struct lx_node_t
{
  lx_queue_t*q;
  mi_index_t next;
  mi_index_t prev;
  char xbar;
  signed char first;
  lx_route_t route;
  
  mi_link_t original;
  unsigned short index;
  
  mi_link_t links [1];
}lx_node_t;

typedef struct lx_host_t
{
  lx_node_t _n;
  unsigned char mac_address [6];
  unsigned char host_type;
  unsigned char level;
  unsigned char pad [3];
}lx_host_t;

typedef struct lx_xbar_t
{
  lx_node_t _n;
  mi_link_t _l [LX_XBAR_SIZE - 1];
  unsigned char level;
  unsigned char last;
}lx_xbar_t;

typedef struct lx_header_t
{
  mi_u16_t type;
  mi_u16_t subtype;
  mi_u32_t timestamp;
  mi_u8_t sender_address [6];
  mi_u8_t echo_address [6];
  mi_s8_t port;
  mi_u8_t reason;
  mi_u8_t level;
  mi_u8_t pad;
  mi_u32_t context;
}lx_header_t;


typedef struct lx_scout_message_t
{
  lx_header_t header;
  mi_u32_t map_version;
  mi_u8_t map_address [6];
  mi_u16_t flags;
  mi_u8_t route_length;
  signed char route [LX_ROUTE_SIZE];

}lx_scout_message_t;

typedef struct lx_scout_reply_message_t
{
  lx_header_t header;
  mi_u32_t map_version;
  mi_u8_t map_address [6];
  mi_u8_t host_type;
  mi_u8_t pad;
  mi_u32_t flags;
}lx_scout_reply_message_t;

typedef struct lx_xbar_message_t
{
  lx_header_t header;
}lx_xbar_message_t;

typedef struct lx_map_link_t
{
  mi_u16_t index;
  mi_s8_t port;
  mi_u8_t pad;
}lx_map_link_t;

typedef struct lx_map_host_t
{
  unsigned char mac_address [6];
  mi_u8_t host_type;
  mi_u8_t level;
  lx_map_link_t link;
}lx_map_host_t;

typedef struct lx_map_xbar_t
{
  mi_u8_t level;
  mi_u8_t pad [3];
  lx_map_link_t links [LX_XBAR_SIZE];
}lx_map_xbar_t;

typedef struct lx_map_request_message_t
{
  lx_header_t header;
  mi_u8_t type;
  mi_u8_t pad;
  mi_u16_t first;
  mi_u8_t route_length;
  signed char route [LX_ROUTE_SIZE];
}lx_map_request_message_t;

typedef struct lx_map_reply_message_t
{
  lx_header_t header;
  mi_u8_t type;
  mi_u8_t pad;
  mi_u16_t num_hosts;
  mi_u16_t num_xbars;
  mi_u16_t num_nodes;
  mi_u16_t first;
  unsigned char map_address [6];
  mi_u32_t map_version;

  union node
  {
    lx_map_host_t host [1];
    lx_map_xbar_t xbar [1];
  } nodes;
  
}lx_map_reply_message_t;

typedef struct lx_printeur_message_t
{
  lx_header_t header;
  mi_u8_t s [LX_PRINTEUR_MAX_STRING];
}lx_printeur_message_t;
  
typedef union lx_packet_t
{
  lx_scout_message_t scout;
  lx_scout_reply_message_t scout_reply;
  lx_xbar_message_t xbar;
  lx_map_request_message_t map_request;
  lx_map_reply_message_t map_reply;
  lx_printeur_message_t printeur;
}lx_packet_t;


enum
{
  MAPPING,
  WAITING_FOR_MAP,
  COLLECTING_MAP,
  ACTIVE_WITH_MAP,
  PASSIVE_WITH_MAP
};

typedef struct lx_map_t
{
  struct lx_t*lx;
  lx_queue_t xbars;
  lx_queue_t hosts;
  lx_queue_t frontier;
  lx_queue_t merge_queue;
  lx_node_t*root;
  int num_xbars;
  int num_hosts;
  int xbars_created;
}lx_map_t;

#ifdef __GNUC__
#define lx_aligned __attribute__ ((aligned(8)))
#else
#define lx_aligned
#endif

typedef struct lx_small_t
{
  lx_route_t r;
  lx_route_t _r;
  char p [LX_SMALL_MTU];
}lx_small_t lx_aligned;

typedef struct lx_printeur_t
{
  lx_route_t r;
  lx_route_t _r;
  char p [LX_PRINTEUR_MTU];
}lx_printeur_t lx_aligned;

typedef struct lx_big_t
{
  lx_route_t r;
  lx_route_t _r;
  char p [LX_BIG_MTU];
}lx_big_t lx_aligned;

typedef struct lx_t
{
  mi_t mi;
  
  unsigned int aborted;
  unsigned int die;
  lx_map_t map;
  lx_map_t other;

  lx_host_t host_block [LX_MAX_HOSTS];
  lx_xbar_t xbar_block [LX_MAX_XBARS];
  lx_queue_t free_hosts;
  lx_queue_t free_xbars;

  lx_queue_t staged_hosts;
  lx_queue_t staged_xbars;
  
  mi_index_t hosts [LX_MAX_HOSTS];
  mi_index_t xbars [LX_MAX_XBARS];

  lx_small_t replies [LX_NUM_REPLIES];
  lx_small_t smalls [LX_NUM_SMALLS];
  lx_big_t bigs [LX_NUM_BIGS];
  lx_printeur_t printeurs [LX_NUM_PRINTEURS];
  
  char receive_buffer [LX_BIG_MTU];
  char map_reply_buffer [LX_BIG_MTU];
  
  unsigned replies_sent;
  unsigned smalls_sent;
  unsigned bigs_sent;
  unsigned printeurs_sent;
  
  unsigned duration;
  unsigned char set_alarm;

  unsigned char replies_pending;
  unsigned char smalls_pending;
  unsigned char bigs_pending;
  unsigned char printeurs_pending;
  
  unsigned char map_valid;
  unsigned char active;
  
  unsigned char mac_address [6];
  unsigned char map_address [6];
  unsigned char winner_address [6];
  lx_route_t winner_route;  
  unsigned char pause_address [6];
  
  unsigned timestamp;
  unsigned last_seen_timestamp;
  unsigned start_timeout;
  unsigned timeout;
  unsigned longest_route;
  unsigned max_level;
  int tries;
  int extra_hosts;
  int num_passes;
  int thrash;
  unsigned map_version;
  unsigned char clos;
  unsigned char level;
  unsigned char host_type;
  unsigned char num_routes;
  unsigned char map_once;
  unsigned char printeur;
  unsigned char check_xbars;
  unsigned char map_incrementally;
  unsigned char kids_match;
  unsigned char pause;
  unsigned char paused;
  unsigned char done_everywhere;
  unsigned char in_receive;
  unsigned char parent_address [6];
  lx_route_t parent_route;
  lx_route_t printeur_route;
}lx_t;

void lx_init (lx_t*lx);
void lx_reset_map (lx_t*lx, lx_map_t*m);

lx_node_t*lx_new_xbar (lx_map_t*m, lx_route_t*route);
lx_node_t*lx_new_host (lx_map_t*m, unsigned char mac_address [6], int host_type, int level, lx_route_t*route);

void lx_delete_node (lx_map_t*m, lx_node_t*node);
void lx_delete_hosts (lx_map_t*m, lx_node_t*n);
int lx_connect (lx_node_t*na, int pa, lx_node_t*nb, int pb);
int lx_map_connect (lx_node_t*na, int pa, int index, int pb);
void lx_disconnect (lx_node_t*n, int p);
lx_node_t*lx_get_node (lx_node_t*n, int port);
#define lx_get_port(n, i)((n)->links[i - n->first].port)
#define lx_connected(n, p)(lx_get_node (n, p) ? 1 : 0)

int lx_there_and_back_again (signed char*dst, signed char*src, signed char port, int length);
void lx_copy_route (signed char*dst, signed char*src, int length);
void lx_reverse_route (signed char*dst, signed char*src, int length);
void lx_physical_route (signed char*p, int length);
void lx_logical_route (signed char*p, int length);
lx_route_t*lx_append_route (lx_route_t*dst, lx_route_t*route, signed char port);
char*lx_print_route (signed char*src, int length);

int lx_put (lx_queue_t*q, lx_node_t*n);
int lx_put_at_front (lx_queue_t*q, lx_node_t*n);
int lx_insert (lx_queue_t*q, lx_node_t*before, lx_node_t*n);
lx_node_t*lx_get (lx_queue_t*q);
lx_node_t*lx_peek (lx_queue_t*q);
int lx_queued (lx_queue_t*q, lx_node_t*n);
lx_node_t*lx_remove (lx_queue_t*q, lx_node_t*n);
lx_node_t*lx_remove_from_anywhere (lx_node_t*n);
lx_node_t*lx_seen (lx_map_t*m, unsigned char mac_address [6]);

int lx_map_print (lx_t*lx, lx_map_t*m);
int lx_map_route (lx_map_t*m, int num_passes, int num_routes);
int lx_firsts (lx_queue_t*queue);
unsigned lx_dice (unsigned throws, unsigned sides);
int lx_memcmp (unsigned char*a, unsigned char*b, unsigned length);
int lx_zero (unsigned char*p, unsigned length);
int lx_has_hosts (lx_node_t*n);

int lx_receive (lx_t*lx, char*p, unsigned length, int early);

int lx_send_xbar (lx_t*lx, lx_route_t*route, int timestamp, int*port, int*context);
int lx_send_scout (lx_t*lx, lx_route_t*route, unsigned timestamp, int reason, int*port, int*host_type, unsigned char mac_address [6], unsigned*map_version, unsigned char map_address [6], int*level, int*flags, int*context);
int lx_send_scout_reply (lx_t*lx, lx_scout_message_t*scout);
int lx_send_map_request (lx_t*lx, lx_route_t*route, unsigned timestamp, int type, int first);
int lx_send_map_reply (lx_t*lx, lx_map_t*m, lx_map_request_message_t*request);

int lx_wait_xbar (lx_t*lx, unsigned duration, int timestamp, int*port, int*context, int early);
int lx_wait_scout_reply (lx_t*lx, unsigned duration, int timestamp, int*port, int*host_type, unsigned char mac_address [6], 
			 unsigned*map_version, unsigned char map_address [6], int*level, int*flags, int*context);
int lx_wait_map_reply (lx_t*lx, unsigned duration, int timestamp, int type, lx_map_reply_message_t*reply);
int lx_wait (lx_t*lx, unsigned duration, int*variable, int early);

void lx_xbar2net (lx_map_t*m, lx_xbar_t*from, lx_map_xbar_t*to);
void lx_net2xbar (lx_map_t*m, lx_map_xbar_t*from, lx_xbar_t*to, int num_hosts, int num_xbars);
void lx_host2net (lx_map_t*m, lx_host_t*from, lx_map_host_t*to);
void lx_net2host (lx_map_t*m, lx_map_host_t*from, lx_host_t*to, int num_hosts, int num_xbars);

void lx_set_alarm (lx_t*lx, unsigned duration);
void lx_clear_alarm (lx_t*lx);
int lx_compare_time (lx_t*lx, unsigned received, unsigned expected);
unsigned lx_advance_time (lx_t*lx);
void lx_mapper_main (lx_t*lx);
void lx_swap_maps (void);
int lx_number_xbars_bfs (lx_map_t*m);
int lx_number_xbars_clos (lx_map_t*m, int max_level);
int lx_before_sending (lx_map_t*m);
void lx_before_receiving (lx_map_t*m, int num_hosts, int num_xbars);
void lx_bzero (char*p, unsigned length);
void lx_map_receive_callback (lx_t*lx, lx_packet_t*p, int length);
lx_node_t*lx_follow (lx_node_t*n, lx_route_t*r);
int lx_higher_than_me (lx_t*lx, unsigned char mac_address [6], int level);
int lx_hostcmp (unsigned char address1 [6], int level1, unsigned char address2 [6], int level2);
void lx_set_timeout (lx_t*lx, unsigned timeout, unsigned tries);
void lx_reclaim_staged (lx_t*lx);
lx_node_t*lx_find_host (lx_t*lx, lx_queue_t*queue, unsigned char mac_address [6], int level, lx_node_t**higher);
void lx_send_printeur_message (const char*format, ...);
int lx_inflate_map (lx_map_t*m, int xbar_fan, int extra_hosts);
int lx_get_highest_mac_address (lx_t*lx, unsigned char mac_address [6]);


int lx_route_valid (struct lx_t*lx, unsigned char mac_address [6], signed char*hops, unsigned length, int physical);


#define LXD_MAX_LINES 64
extern int lxd_cur;
extern int lxd_lines [LXD_MAX_LINES];
void lxd_line (int line);
void lx_dump (char*p, int length);

#endif /*lx_h*/
