#include "lx.h"

void lx_map_receive_callback (lx_t*lx, lx_packet_t*p, int length)
{
  int higher_than_me = 0;
  int r;
  
  insist (p);
  
  if (length < 4 || p->scout.header.type != mi_htons (LX_TYPE))
    return;
    
  higher_than_me = lx_higher_than_me (lx, p->scout.header.sender_address, p->scout.header.level);

  if (lx->printeur)
  {
    if (p->scout.header.subtype == mi_htons (LX_PRINTEUR))
      mi_c ((lx_mac_format " %s", lx_mac_args (p->scout.header.sender_address), ((lx_printeur_message_t*) p)->s));
    return;
    
  }
  if (p->scout.header.subtype == mi_htons (LX_SCOUT))
  {
    insist ((unsigned) length >= sizeof (lx_scout_message_t));

    mi_c (("scouted by " lx_mac_format " (level %d, reason %d)", lx_mac_args (p->scout.header.sender_address), p->scout.header.level, p->scout.header.reason));

    if (mi_htons (p->scout.flags) & LX_FLAG_PRINTEUR)
    {
      lx->printeur_route.length = p->scout.route_length;
      lx_copy_route (lx->printeur_route.hops, p->scout.route, lx->printeur_route.length);
      r = lx->replies_pending == LX_NUM_REPLIES || lx_send_scout_reply (lx, &p->scout);
      insist (r);
      return;
    }    
    if (p->scout.header.reason == LX_SCOUTING && lx->active && higher_than_me)
    {
      mi_memcpy (lx->winner_address, p->scout.header.sender_address, 6);
      lx->winner_route.length = p->scout.route_length;
      lx_copy_route (lx->winner_route.hops, p->scout.route, lx->winner_route.length);
	
      mi_c (("my winner is " lx_mac_format " (level %d)", lx_mac_args (p->scout.header.sender_address), p->scout.header.level));
      lx->map_valid = 0;
      lx->aborted = LX_ABORT_PASSIVE;
    }
    else if ((p->scout.header.reason == LX_TREE) && !lx->active)
    {
      int done_everywhere;
      
      mi_memcpy (lx->parent_address, p->scout.header.sender_address, 6);
      lx->parent_route.length = p->scout.route_length;
      lx_copy_route (lx->parent_route.hops, p->scout.route, lx->parent_route.length);

      mi_c (("my parent is " lx_mac_format " (level %d)", lx_mac_args (p->scout.header.sender_address), p->scout.header.level));
      if (lx_hostcmp (p->scout.header.sender_address, p->scout.header.level, lx->mac_address, lx->level) <= 0)
      {
	mi_c (("my parent is evil. remap"));
	lx->aborted = LX_ABORT_ACTIVE;
      }
      done_everywhere = (lx->map_valid && (mi_htons (p->scout.flags) & LX_FLAG_DONE_EVERYWHERE)) != 0;

      if (done_everywhere != lx->done_everywhere)
      {
	lx->done_everywhere = done_everywhere;
	mi_set_map_version (lx, lx->map_address, lx->map_version, lx->map.num_hosts, lx->done_everywhere, lx->map_valid);
      }      
    }
  }
  if (p->scout.header.subtype == mi_htons (LX_SCOUT))
  {
    if (higher_than_me)
    {
      int pause = (mi_htons (p->scout.flags) & LX_FLAG_PAUSE) != 0;
      if (pause || lx_zero (lx->pause_address, 6) || !lx_memcmp (p->scout.header.sender_address, lx->pause_address, 6))
      {
	if ((lx->paused = pause))
	{
	  mi_c ((lx_mac_format " is pausing me", lx_mac_args (p->scout.header.sender_address)));
	  mi_memcpy (lx->pause_address, p->scout.header.sender_address, 6);
	}
      }
    }
    else
    {
      if (lx->map_version == mi_htonl (p->scout.map_version) && !lx_memcmp (p->scout.map_address, lx->map_address, 6) && lx->map_valid && !(mi_htons (p->scout.flags) & LX_FLAG_MAP_VALID))
      {
	mi_c (("my child is telling me my map is invalid."));
	lx->map_valid = 0;
      }
    }
    
    if (lx->replies_pending == LX_NUM_REPLIES)
    {
      mi_c (("dropped reply"));
    }
    r = lx->replies_pending == LX_NUM_REPLIES || lx_send_scout_reply (lx, &p->scout);
    insist (r);
  }
  else if (p->scout_reply.header.subtype == mi_htons (LX_SCOUT_REPLY))
  {
    /*mi_c (("got reply message %d from " lx_mac_format, mi_htonl (p->scout.header.timestamp), lx_mac_args (p->scout.header.sender_address)));*/
  }
  else if (p->map_request.header.subtype == mi_htons (LX_MAP_REQUEST))
  {
    mi_c (("got map request %d from " lx_mac_format, mi_htonl (p->scout.header.timestamp), lx_mac_args (p->map_request.header.sender_address)));

    if (lx->aborted || !lx->map_valid)
    {
      mi_c (("I don't have a map to give."));
    }
    else
    {
      if (lx->bigs_pending == LX_NUM_BIGS)
      {
	mi_c (("no map buffers, he'll have to ask again"));
      }
      else
      {
	r = lx_send_map_reply (lx, &lx->map, &p->map_request);
	insist (r);
      }
    }
  }
  except:;
}

static lx_node_t*lx_map_add_host (lx_map_t*m, lx_node_t*n, signed char port, int host_type, int level, unsigned char mac_address [6], lx_node_t*higher)
{
  lx_node_t*nn;
  lx_route_t route;
  
  insist (m && n && mac_address);
  insist (!lx_zero (mac_address, 6));
  
  mi_c (("found host " lx_mac_format " (type %d) on port %d", lx_mac_args (mac_address), host_type, port));

  insist (port > -LX_XBAR_SIZE && port < LX_XBAR_SIZE);
  
  if ((nn = lx_get_node (n, port)))
  {
    insist (!nn->xbar && !lx_memcmp (lx_host_c (nn)->mac_address, mac_address, 6));
    mi_c (("already found"));
    return 0;
  }

  nn = lx_new_host (m, mac_address, host_type, level, lx_append_route (&route, &n->route, port));
  insist (nn);
  lx_insert (&m->hosts, higher, nn);
  lx_connect (n, port, nn, 0);
  insist (m->hosts.head == m->root);
  
  return nn;

  except: return 0;
}

static lx_node_t*lx_map_add_xbar (lx_map_t*m, lx_node_t*n, signed char port)
{
  lx_node_t*nn;
  lx_route_t route;
  
  insist (m && n);
  
  mi_c (("found xbar on port %d", port));

  insist (port > -LX_XBAR_SIZE && port < LX_XBAR_SIZE);

  if ((nn = lx_get_node (n, port)))
  {
    insist (nn->xbar);
    mi_c (("already found"));
    return nn;
  }

  nn = lx_new_xbar (m, lx_append_route (&route, &n->route, port));
  insist (nn);
  return lx_connect (n, port, nn, 0) ? nn : 0;

  except: return 0;
}

static int lx_map_merge (lx_map_t*m, lx_node_t*copy, lx_node_t*original, int offset)
{
  lx_node_t*n;
  int i, p;
  
  insist (m && copy && original);
  insist (!m->merge_queue.head);
  insist (!copy->original.index);
  insist (copy != original);
  insist (copy->xbar && original->xbar);

  mi_c (("merging duplicate " lx_node_format " with original " lx_node_format,
	 lx_node_args (copy), lx_node_args (original)));
  
  insist (mi_match (&copy->route, &original->route));

  lx_remove_from_anywhere (copy);
  lx_put (&m->merge_queue, copy);
  copy->original.index = mi_node2index (original);
  copy->original.port = offset;
  insist (mi_index2node (copy->original.index) == original);

  while ((n = (lx_node_t*) lx_get (&m->merge_queue)))
  {
    lx_node_t*nn, *oo;
    lx_node_t*no = mi_index2node (n->original.index);
    insist (no);
    
    for (i = 1 - LX_XBAR_SIZE; i < LX_XBAR_SIZE; i++)
    {
      mi_punt (m->lx);

      p = i + n->original.port;
      if ((nn = lx_get_node (n, i)))
      {
	if (!(oo = lx_get_node (no, p)))
	{
	  if (!lx_connect (no, p, nn, lx_get_port (n, i)))
	    return 0;
	}
	else
	{
	  if (nn != oo && !oo->original.index && nn->xbar == oo->xbar)
	  {
	    mi_c ((lx_node_format " matches " lx_node_format, 
		   lx_node_args (nn),
		   lx_node_args (oo)));

	    if (!nn->xbar && lx_memcmp (lx_host_c (nn)->mac_address, lx_host_c (oo)->mac_address, 6))
	      return -1;

	    insist (mi_match (&nn->route, &oo->route));
	    
	    lx_remove_from_anywhere (nn);
	    
	    if (nn->xbar)
	    {
	      nn->original.index = mi_node2index (oo);
	      insist (mi_index2node (nn->original.index));
	      nn->original.port = lx_get_port (no, p)  - lx_get_port (n, i);
	      insist (nn->original.port > -LX_XBAR_SIZE && nn->original.port < LX_XBAR_SIZE);

	      if (!lx_queued (&m->merge_queue, nn))
	      {
		mi_c (("adding " lx_node_format " to merge queue", lx_node_args (nn)));
		
		lx_remove_from_anywhere (nn);
		lx_put (&m->merge_queue, nn);
		insist (lx_queued (&m->merge_queue, nn));
	      }
	    }
	    else
	    {
	      lx_remove_from_anywhere (nn);
	      lx_delete_node (m, nn);
	    }
	    
	  }
	}
      }
    }
    lx_delete_node (m, n);
  }
  
  return 1;
  except: return 0;
}


static int lx_map_find_hosts (lx_map_t*m, lx_node_t*n)
{
  unsigned char mac_address [6];
  unsigned char map_address [6];
  unsigned map_version;
  int i, j,  rr, port, host_type, level, flags, context, hop;
  lx_node_t*h, *higher;
  lx_route_t r;
  lx_t*lx = m->lx;
  int sends;
  
  insist (m && n && n->xbar);
  insist (n->route.length < LX_ROUTE_SIZE);

  mi_c (("looking for hosts on " lx_node_format, lx_node_args (n)));

  for (j = 0; j < (int) lx->tries && !lx->aborted; j++)
  {
    sends = 0;
    lx_advance_time (lx);

    while (lx->smalls_pending)
      lx_receive (lx, 0, 0, 0);

    for (i = 1 - LX_XBAR_SIZE; i < LX_XBAR_SIZE && !lx->aborted; i++)
    {
      if (lx_connected (n, i))
	continue;
    
      insist (!n->original.port);
      hop = i - n->original.port;

      if (hop < 1 - LX_XBAR_SIZE || hop >= LX_XBAR_SIZE)
	continue;

      lx_copy_route (r.hops, n->route.hops, n->route.length);
      r.length = n->route.length + 1;  
      insist (r.length > 0);
     
      r.hops [r.length - 1] = hop;
      port = i;

      rr = lx_send_scout (lx, &r, lx->timestamp, LX_SCOUTING, &port, &host_type, mac_address, &map_version, map_address, &level, &flags, &context);
      insist (rr > 0);
      sends++;
    }
    if (!sends)
      break;
    insist (!lx->duration);
    insist (lx->timeout);
    lx_set_alarm (lx, lx->timeout);
    
    while (lx_wait_scout_reply (lx, 0, lx->timestamp, &port, &host_type, mac_address, &map_version, map_address, &level, &flags, &context))
    {
      insist (!lx_connected (n, port));
	
      if ((h = lx_find_host (lx, &m->hosts, mac_address, level, &higher)))
      {
	lx_node_t*o = mi_index2node (h->links[0].index);
	insist (o);
	  
	mi_c (("we have seen " lx_mac_format " before on " lx_node_format " port %d", 
	       lx_mac_args (mac_address), lx_node_args (o), h->links [0].port));
	if (lx_map_merge (m, n, o, h->links [0].port - port) < 0)
	{
	  mi_c (("merge failed. probably a duplicate board ID " lx_mac_format, lx_mac_args (mac_address)));
	  lx->die = lx->aborted = LX_ABORT_ACTIVE;
	}
	lx_clear_alarm (lx);
	return 0;
      }
      else
      {
	r.hops [r.length - 1] = port;
	if (flags & LX_FLAG_PRINTEUR)
	{
	  lx->printeur_route.length = r.length;
	  lx_reverse_route (lx->printeur_route.hops, r.hops, r.length);
	}

	if (lx_higher_than_me (lx, mac_address, level))
	{
	  mi_c ((lx_mac_format " (level %d) is higher than me",  lx_mac_args (mac_address), level));
	    
	  mi_memcpy (lx->winner_address, mac_address, 6);
	  lx->winner_route.length = r.length;
	  lx_copy_route (lx->winner_route.hops, r.hops, r.length);
	  /*mi_c (("winner route is %s", lx_print_route (lx->winner_route.hops, lx->winner_route.length)));*/
	  lx->aborted = LX_ABORT_PASSIVE;
	  lx_clear_alarm (lx);
	  return 1;
	}
	if (!(h = lx_map_add_host (m, n, port, host_type, level, mac_address, higher)))
	{
	  lx_clear_alarm (lx);
	  return 0;
	}
      }
    }
    
    if (lx->aborted)
      return 1;
  }

  return 1;
  except: return 0;
}


static int lx_map_can_match (lx_node_t*copy, lx_node_t*original, int offset)
{  
  insist (copy && original);
  insist (offset > -LX_XBAR_SIZE && offset < LX_XBAR_SIZE);
  
  return offset && !lx_get_node (original, offset);

  except: return 0;
}

#define lx_xbar2offset(lx, xbar) ((lx_xbar_t*) xbar - lx->xbar_block)

static lx_node_t*lx_offset2xbar (lx_map_t*m, int offset)
{
  lx_node_t*n;
  insist (m && m->lx);
  insist (offset >= 0 && offset < LX_MAX_XBARS);
  
  n = (lx_node_t*) (m->lx->xbar_block + offset);
  insist (lx_queued (&m->frontier, n));
  return n;
  except: return 0;
}
  
static int lx_map_match (lx_map_t*m, lx_node_t*n)
{
  int i, j, t;
  lx_node_t*o;
  lx_extended_t e;
  lx_route_t*r = (lx_route_t*) &e;
  int port, context;

  lx_t*lx = m->lx;
  
  insist (m && n && n->xbar);
  insist (!lx_has_hosts (n));
  
  mi_c (("trying to match "  lx_node_format, lx_node_args (n)));
  
  lx_advance_time (lx);

  for (o = m->frontier.tail; o; o = mi_index2node (o->prev))
  {    
    lx_copy_route (r->hops, n->route.hops, n->route.length);

    for (i = 1 - LX_XBAR_SIZE; i < LX_XBAR_SIZE && !lx->aborted; i++)
    {
      for (j = 0; j < (int) lx->tries && i && !lx->aborted; j++)
      {
	if (lx_map_can_match (n, o, -i))
	{
	  r->length = n->route.length + o->route.length + 1;
	  r->hops [n->route.length] = i;
	  lx_reverse_route (r->hops + n->route.length + 1, o->route.hops, o->route.length);
	  port = i;
	  context = lx_xbar2offset (lx, o);
	  
	  /*mi_punt (lx);*/
	  
	  if ((t = lx_send_xbar (lx, r, lx->timestamp, &port, &context)) < 0)
	    return lx_map_merge (m, n, lx_offset2xbar (m, context), -port);
	  else if (!t)
	    return 0;
#ifdef LX_XM
	  if (lx_wait_xbar (lx, lx->timeout, 100, &port, &context, 1))
	    return lx_map_merge (m, n, lx_offset2xbar (m, context), -port);
#endif
	}
      }
    }
  }

  if (lx_wait_xbar (lx, lx->timeout, lx->timestamp, &port, &context, 0))
    return lx_map_merge (m, n, lx_offset2xbar (m, context), -port);

  mi_c ((lx_node_format " is original", lx_node_args (n)));
  except: return 0;  
}

static int lx_map_find_xbars (lx_map_t*m, lx_node_t*n, int ports [2 * LX_XBAR_SIZE])
{
  int i, j, rr;
  lx_extended_t e;
  lx_route_t*r = (lx_route_t*) &e;
  int port, context, hop;
  lx_t*lx = m->lx;
  int sends = 0;
  
  insist (m && n && n->xbar);
  insist (!lx->staged_xbars.count);
  insist (ports);

  insist (n->route.length < LX_ROUTE_SIZE / 2 - 1);  

  mi_c (("looking for xbars on " lx_node_format, lx_node_args (n)));

  for (i = 0; i < 2 * LX_XBAR_SIZE; i++)
    ports [i] = 0;
  
  for (j = 0; j < (int) lx->tries && !lx->aborted; j++)
  {
    lx_advance_time (lx);
    sends = 0;
 
    while (lx->smalls_pending)
      lx_receive (lx, 0, 0, 0); 
  
    for (i = 1 - LX_XBAR_SIZE; i < LX_XBAR_SIZE && !lx->aborted; i++)
    {
      if (lx_connected (n, i) || ports [i - (1 - LX_XBAR_SIZE)])
	continue;

      insist (!n->original.port);
      hop = i - n->original.port;

      if (hop < 1 - LX_XBAR_SIZE || hop >= LX_XBAR_SIZE)
	continue;
      
      r->length = lx_there_and_back_again (r->hops, n->route.hops, hop, n->route.length);
      insist (r->length);
      port = i;

      rr = lx_send_xbar (lx, r, lx->timestamp, &port, &context);
      insist (rr > 0);
      sends++;
    }
    
    if (!sends)
      break;
 
    insist (!lx->duration);
    insist (lx->timeout);
    lx_set_alarm (lx, lx->timeout);
    
    while (lx_wait_xbar (lx, 0, lx->timestamp, &port, &context, 0))
    {
      insist (port >= 1 - LX_XBAR_SIZE && port < LX_XBAR_SIZE);
      insist (!lx_connected (n, port) && !ports [port - (1 - LX_XBAR_SIZE)]);
      ports [port - (1 - LX_XBAR_SIZE)] = 1;
    }
  }

  return 1;
  except: return 0;
}

static int lx_map_explore (lx_map_t*m)
{
  int ports [2 * LX_XBAR_SIZE];
  unsigned char mac_address [6];
  unsigned char map_address [6];
  lx_node_t*n, *nn;
  lx_route_t r;
  int j, i, rr, port, context, host_type, level, flags;
  unsigned map_version;
  lx_t*lx = m->lx;
  
  insist (m);
  insist (!m->root);

  r.length = 0;

  m->root = lx_new_host (m, lx->mac_address, lx->host_type, lx->level, &r);
  insist (m->root);
 
  lx_put (&m->hosts, m->root);

  mi_c (("looking for directly connected host"));

  port = 0;

  for (j = 0; j < (int) lx->tries && !lx->aborted; j++)
  {  
    rr = lx_send_scout (lx, &r, lx_advance_time (lx), LX_SCOUTING, &port, &host_type, mac_address, &map_version, map_address, &level, &flags, &context);
    insist (rr);
    
    if (rr < 0 || lx_wait_scout_reply (lx, lx->timeout, lx->timestamp, &port, &host_type, mac_address, &map_version, map_address, &level, &flags, &context))
    {
      mi_c (("found host " lx_mac_format, lx_mac_args (mac_address)));
    
      if (lx_memcmp (lx_host_c (m->root)->mac_address, mac_address, 6))
      {
	n = lx_new_host (m, mac_address, host_type, level, &r);
	
	if (lx_higher_than_me (lx, mac_address, level))
	  lx_put_at_front (&m->hosts, n);
	else
	  lx_put (&m->hosts, n);
	
	return lx_connect (m->root, 0, n, 0);
      }
      else return lx_connect (m->root, 0, m->root, 0);
    }
    if (lx->aborted)
      return 0;
  }
  
  mi_c (("looking for first xbar"));

  r.length = 1;
  r.hops [0] = 0;
  m->root->route.length = r.length;
  lx_copy_route (m->root->route.hops, r.hops, r.length);
  
  for (j = 0; j < (int) lx->tries && !lx->aborted; j++)
  {  
    port = 0;
    
    rr = lx_send_xbar (lx, &r, lx_advance_time (lx), &port, &context);
    insist (rr);
    
    if (rr < 0 || lx_wait_xbar (lx, lx->timeout, lx->timestamp, &port, &context, 0))
    {
      mi_c (("found xbar"));
      r.length = 0;
      lx_connect (m->root, 0, (n = lx_new_xbar (m, &r)), 0);
    
      insist (!lx_get (&m->frontier));
      rr = lx_map_find_hosts (m, n);
      insist (rr);
      
      lx_put (&m->frontier, n);
    
      while (!lx->aborted && (n = m->frontier.head))
      {
	if (!lx_map_find_xbars (m, n, ports))
	  return 0;
	
	for (i = 1 - LX_XBAR_SIZE; i < LX_XBAR_SIZE; i++)
	{
	  if (lx_connected (n, i) || !ports [i - (1 - LX_XBAR_SIZE)])
	    continue;
	  
	  if (!(nn = lx_map_add_xbar (m, n, i)))
	    return 0;

	  if (lx_map_find_hosts (m, nn) && (lx_has_hosts (nn) || !lx_map_match (m, nn)))
	  {
	    mi_c (("adding " lx_node_format " to frontier", lx_node_args (nn)));
	    lx_put (&m->frontier, nn);
	    insist (m->frontier.tail == nn);
	  }

	  if (m->frontier.head != n)
	  {
	    mi_c (("xbar I was exploring was deleted"));
	    break;
	  }
	}
	if (m->frontier.head == n && !lx_put (&m->xbars, lx_get (&m->frontier)))
	  return 0;
      }
      
      insist (!m->merge_queue.head);

      return !lx->aborted;
    }
  }
  return -1;
  except: return 0;
}

static int lx_map_query_new_version (lx_t*lx, unsigned*map_version, unsigned char map_address [6])
{
  int i, r, port, host_type, level, flags, context;
  unsigned char mac_address [6];
  insist (map_version && map_address);
  
  lx_advance_time (lx);
  lx_set_timeout (lx, LX_VERIFY_CONNECTED_TIMEOUT, LX_VERIFY_CONNECTED_TRIES);

  for (i = 0; i < lx->tries; i++)
  {
    /*mi_c (("time is %u", mi_time (lx)));*/
    
    r = lx_send_scout (lx, &lx->winner_route, lx->timestamp, LX_QUERYING_MAP_VERSION, &port, &host_type, mac_address, map_version, map_address, &level, &flags, &context);
    insist (r);
    if (r < 0 || lx_wait_scout_reply (lx, lx->timeout, lx->timestamp, &port, &host_type, mac_address, map_version, map_address, &level, &flags, &context))
    {      
      if (*map_version && !lx_zero (map_address, 6) && (*map_version != lx->map_version || lx_memcmp (map_address, lx->map_address, 6)))
      {
	mi_c ((lx_mac_format " has a new map (" lx_mac_format " version %d)",  lx_mac_args (lx->winner_address), lx_mac_args (map_address), *map_version));
	
	if (flags & LX_FLAG_MAP_VALID)
	  return 1;

	mi_c (("but it is invalid. updating my version and marking it invalid too"));

	mi_memcpy (lx->map_address, map_address, 6);
	lx->map_version = *map_version;
	lx->map_valid = 0;
	return -1;
      }
      if (!lx_higher_than_me (lx, mac_address, level))
      {
	mi_c (("winner changed to a loser"));
	return 0;
      }
      
      return -1;
    }
  }
  mi_c (("no reply from " lx_mac_format,  lx_mac_args (lx->winner_address)));
  except: return 0;
}


static int lx_map_check_for_xbar (lx_t*lx, lx_node_t*n, int p, int connected, int fast)
{
  int port;
  int j, rr, context, hop;
  lx_extended_t e;
  lx_route_t*r = (lx_route_t*) &e;
  
  insist (n);
  insist (lx);

  insist (p > -LX_XBAR_SIZE && p < LX_XBAR_SIZE);
  insist (n->original.port > -LX_XBAR_SIZE && n->original.port < LX_XBAR_SIZE);

  hop = p - n->original.port;
  insist (hop > -LX_XBAR_SIZE && hop < LX_XBAR_SIZE);
  
  if (!fast && connected)
    lx_set_timeout (lx, LX_VERIFY_CONNECTED_TIMEOUT, LX_VERIFY_CONNECTED_TRIES);
  else
    lx_set_timeout (lx, LX_VERIFY_DISCONNECTED_TIMEOUT, LX_VERIFY_DISCONNECTED_TRIES);

  lx_advance_time (lx);

  r->length = lx_there_and_back_again (r->hops, n->route.hops, hop, n->route.length);
  insist (r->length);

  for (j = 0; j < (int) lx->tries; j++)
  {
    port = p;
    
    rr = lx_send_xbar (lx, r, lx->timestamp, &port, &context);
    insist (rr);
    
    if (rr < 0 || (lx_wait_xbar (lx, lx->timeout, lx->timestamp, &port, &context, 0) && port == p))
      return 1;    
  }
  except: return 0;
}

static int lx_map_check_xbar_new_hosts (lx_t*lx, lx_node_t*n)
{
  int i, j, context, hop;
  lx_extended_t e;
  int port, host_type, level, flags;
  unsigned char mac_address [6], map_address [6];
  unsigned map_version;
  
  lx_route_t*r = (lx_route_t*) &e;
  
  insist (n);
  insist (lx);

  mi_c (("checking for new hosts on " lx_node_format, lx_node_args (n)));

  lx_set_timeout (lx, LX_VERIFY_DISCONNECTED_TIMEOUT, LX_VERIFY_DISCONNECTED_TRIES);
  lx_advance_time (lx);
  
  for (i = 0; i < (int) lx->tries; i++)
  {
    for (j = 1 - LX_XBAR_SIZE; j < LX_XBAR_SIZE; j++)
    {
      hop = j - n->original.port;

      if (hop > -LX_XBAR_SIZE && hop < LX_XBAR_SIZE && !lx_get_node (n, j))
      {   
	r->length = 0;
	lx_append_route (r, &n->route, hop);
	insist (r->length);
	port = j;
	
	if (lx_send_scout (lx, r, lx->timestamp, LX_VERIFYING, &port, &host_type, mac_address, &map_version, map_address, &level, &flags, &context) < 0)
	{
	  mi_c (("new host " lx_mac_format " on port %d",  lx_mac_args (lx->mac_address), port));
	  lx->aborted = lx_higher_than_me (lx, mac_address, level) ? LX_ABORT_PASSIVE : LX_ABORT_ACTIVE;
	  mi_c (("aborting %s", lx->aborted == LX_ABORT_PASSIVE ? "passive" : "active"));
	  return 0;
	}
      }
    }
  }
  if (lx_wait_scout_reply (lx, lx->timeout, lx->timestamp, &port, &host_type, mac_address, &map_version, map_address, &level, &flags, &context))
  {
    mi_c (("new host " lx_mac_format " on port %d",  lx_mac_args (lx->mac_address), port));
    return 0;
  }
  return 1;
  except: return 0;
}

static int lx_map_check_xbar_for_xbars (lx_t*lx, lx_node_t*n, int fast)
{
  int i, hop, there;
  lx_node_t*nn;
  
  insist (lx && n);
  
  for (i = 1 - LX_XBAR_SIZE; i < LX_XBAR_SIZE; i++)
  {
    hop = i - n->original.port;
    
    if (hop <= -LX_XBAR_SIZE || hop >= LX_XBAR_SIZE)
      continue;
    
    nn = lx_get_node (n, i);
    
    if (!nn || nn->xbar)
    {
      mi_c (("checking %s link %d on " lx_node_format, nn ? "connected" : "disconnected", i, lx_node_args (n)));
      there = lx_map_check_for_xbar (lx, n, i, nn ? 1 : 0, fast);

      if (!nn && there)
      {
	mi_c (("new link"));
	return 0;
      }
      if (nn && !there)
      {
	mi_c (("missing link"));
	return 0;
      }
    }
  }
  return 1;
  except: return 0;
}

static int lx_map_check_xbars (lx_t*lx, int max_level)
{
  int p, hop, there;
  lx_node_t*n,*before;
  lx_map_t *m = &lx->map;
  
  insist (lx);

  if (!lx->map.num_xbars || !lx->check_xbars)
    return 1;
  
  before = lx_get_node (m->root, 0);
  insist (before && before->xbar);
  
  for (;;)
  {
    insist (before && before->xbar);
    do
    {
      p = lx_xbar_c (before)->last++ % (LX_XBAR_SIZE * 2 - 2) + 1 - LX_XBAR_SIZE;
      hop = p - before->original.port;
    } while (hop <= -LX_XBAR_SIZE || hop >= LX_XBAR_SIZE);

    if ((n = lx_get_node (before, p)) && n->xbar)
    {
      if ((lx->clos && lx_xbar_c (before)->level > lx_xbar_c (n)->level) || (!lx->clos && lx_xbar_c (before)->level < lx_xbar_c (n)->level))
	return 1;
    }
    
    if (!n || n->xbar)
    {
      mi_c (("checking %s link %d on " lx_node_format, n ? "connected" : "disconnected", p, lx_node_args (before)));
      there = lx_map_check_for_xbar (lx, before, p, n ? 1 : 0, 0);

      if (!n && there)
      {
	mi_c (("new link"));
	return 0;
      }
      if (n && !there)
      {
	mi_c (("missing link"));
	return 0;
      }
    }
    if (!n || !n->xbar)
      return 1;
    if (lx->clos && lx_xbar_c (n)->level == max_level)
      return 1;

    before = n;
  }
  except: return 0;
}


static int lx_map_check_xbars_new_hosts (lx_t*lx)
{
  int i;
  insist (lx);
  
  for (i = lx->map.root->index; i < lx->map.num_xbars; i += lx->map.num_hosts)
  {
    lx_node_t*n = mi_index2node (lx->xbars [i]);
    insist (n && n->xbar && n->index == i);
    if (!lx_map_check_xbar_new_hosts (lx, n))
      return 0;
  }
  
  return 1;
  except: return 0;
}

static int lx_map_check_host (lx_t*lx, lx_node_t*n, int reason, int*kids_match, int*matches_me, int*change)
{
  int r, j;
  unsigned char mac_address [6];
  unsigned char map_address [6];
  int port, host_type, level, flags, context;
  unsigned map_version;

  if (lx_host_c (n)->host_type == LX_DEBUG_TYPE /*|| n == lx->map.root*/)
    return 1;
  
  insist (lx && n && !n->xbar);
  insist (reason == LX_VERIFYING || reason == LX_TREE);
  
  mi_c (("checking host " lx_node_format, lx_node_args (n)));
  insist (lx_follow (lx->map.root, &n->route) == n);
  insist (kids_match && matches_me && change);

  /*mi_c (("root is " lx_node_format, lx_node_args (lx->map.root)));*/

  lx_advance_time (lx);
  
  for (j = 0; j < (int) lx->tries; j++)
  {
    port = 0;
    r = lx_send_scout (lx, &n->route, lx->timestamp, reason, &port, &host_type, mac_address, &map_version, map_address, &level, &flags, &context);
    insist (r);    
    
    if (r < 0 || lx_wait_scout_reply (lx, lx->timeout, lx->timestamp, &port, &host_type, mac_address, &map_version, map_address, &level, &flags, &context))
    {
      if (host_type != lx_host_c (n)->host_type || level != lx_host_c (n)->level || lx_memcmp (mac_address, lx_host_c (n)->mac_address, 6))
      {
	mi_c (("host changed " lx_mac_format, lx_mac_args (mac_address)));
	mi_c ((lx_mac_format " map_address " lx_mac_format " map_version %d flags %d" , lx_mac_args (mac_address), lx_mac_args (map_address), map_version, flags));

	*change = LX_CHANGED_HOST;
	return 0;
      }
      *matches_me = map_version == lx->map_version && !lx_memcmp (map_address, lx->map_address, 6);
      
      if (!(flags & LX_FLAG_MAP_VALID) && map_version == lx->map_version && !lx_memcmp (map_address, lx->map_address, 6))
      {
	lx->aborted = lx_higher_than_me (lx, mac_address, level) ? LX_ABORT_PASSIVE : LX_ABORT_ACTIVE;
	mi_c (("host map invalid"));
	*change = LX_INVALID_MAP;
	return 0;
      }
      *kids_match = (flags & LX_FLAG_KIDS_MATCH) != 0;
      return 1;
    }
  }
  *change = LX_MISSING_HOST;
  mi_c (("missing host " lx_node_format, lx_node_args (n)));
  
  except: return 0;
}

static int lx_map_check_xbar_old_hosts (lx_t*lx, lx_node_t*n)
{
  int kids_match, matches_me, change, i;
  lx_node_t*nn;
  
  insist (lx && n);

  mi_c (("checking old hosts on " lx_node_format, lx_node_args (n)));

  for (i = 0; i < LX_XBAR_SIZE; i++)
  {
    if ((nn = lx_get_node (n, i + n->first)) && !nn->xbar && !lx_map_check_host (lx, nn, LX_VERIFYING, &kids_match, &matches_me, &change) && change != LX_INVALID_MAP)
      return 0;  
  }
  return 1;
  except: return 0;
}


static int lx_map_check_hosts (lx_t*lx)
{
  int i;
  int kids_match, change;
  int matches_me;
  int num_kids = 0;
  int num_matches = 0;
  
  insist (lx);
  insist (lx->map.num_hosts);
  
  mi_c ((lx_node_format " checking hosts", lx_node_args (lx->map.root)));

  lx_set_timeout (lx, LX_VERIFY_CONNECTED_TIMEOUT, LX_VERIFY_CONNECTED_TRIES);
  
  if (lx->map.root->index)
  {
    lx_node_t*kid = mi_index2node (lx->hosts [(lx->map.root->index - 1) / LX_BRANCH ]);
    insist (kid);
      
    if (!lx_map_check_host (lx, kid, LX_VERIFYING, &kids_match, &matches_me, &change))
      return 0;
  }

  for (i = 0; i < LX_BRANCH; i++)
  {
    int n;
    if ((n = lx->map.root->index * LX_BRANCH + 1 + i) < lx->map.num_hosts)
    {
      lx_node_t*kid = mi_index2node (lx->hosts [n]);
      insist (kid);
      
      num_kids++;
      insist (lx_hostcmp (lx->mac_address, lx->level, lx_host_c (kid)->mac_address, lx_host_c (kid)->level) > 0);

      if (!lx_map_check_host (lx, kid, LX_TREE, &kids_match, &matches_me, &change))
	return 0;
      if (kids_match && matches_me)
	num_matches++;
    }
  }
  lx->kids_match = num_matches == num_kids;
  return 1;
  except: return 0;
}

static int lx_map_collect_map (lx_map_t*m, lx_route_t*map_route, unsigned char map_address [6], int map_version)
{
  int first, num_nodes, i, r = 0;
  int num_hosts, num_xbars;
  lx_node_t*higher;
  lx_map_reply_message_t*rm = (lx_map_reply_message_t*)&m->lx->map_reply_buffer;
  lx_t*lx = m->lx;
  
  insist (m && map_address);
  insist (map_route);
  insist (!lx->map_valid);
  
  num_nodes = 0;
  
  lx_set_timeout (lx, LX_COLLECT_TIMEOUT, LX_COLLECT_TRIES);

  num_xbars = LX_MAX_XBARS;
  num_hosts = LX_MAX_HOSTS;
  
  mi_c (("collecting map from " lx_mac_format,  lx_mac_args (lx->winner_address)));

  for (first = 0; first < num_hosts;)
  { 
    lx_advance_time (lx);

    for (i = 0; i < lx->tries; i++)
    {
      if (lx_send_map_request (lx, map_route, lx->timestamp, LX_MAP_HOST, first) &&
	  (r = lx_wait_map_reply (lx, lx->timeout, lx->timestamp, LX_MAP_HOST, rm)))
	break;
    }
    
    if (i == lx->tries)
    {
      mi_c (("no reply from " lx_mac_format,  lx_mac_args (lx->winner_address)));
      return 0;
    }
    
    insist (m);
    insist (r >= (int) sizeof (lx_map_reply_message_t));
    
    mi_c (("got map reply %d message", mi_htonl (rm->header.timestamp)));

    if (lx_memcmp (rm->map_address, map_address, 6) || rm->map_version != (unsigned) mi_htonl (map_version))
    {
      mi_c (("map address / map version changed to " lx_mac_format " %u",  lx_mac_args (rm->map_address), mi_htonl (rm->map_version)));
      return 0;
    }
    
    if (first == 0)
    {
      num_hosts = mi_htons (rm->num_hosts);
      num_xbars = mi_htons (rm->num_xbars);
      
      insist (num_hosts && num_hosts < LX_MAX_HOSTS);
      insist (rm->num_nodes);
      
      insist (m->num_xbars == 0 && m->num_hosts == 0);
      
      lx_before_receiving (m, num_hosts, num_xbars);
    }
    else
    {
      insist (num_hosts == mi_htons (rm->num_hosts));
      insist (num_xbars = mi_htons (rm->num_xbars));
    }
    
    num_nodes = mi_htons (rm->num_nodes);
    
    for (i = 0; i < num_nodes; i++, first++)
    {     
      lx_node_t*n = mi_index2node (lx->hosts [first]);
      insist (n);
      insist (first < num_hosts);

      mi_c (("reading " lx_node_format, lx_node_args (n)));
      lx_net2host (m, &rm->nodes.host [i], lx_host_c (n), num_hosts, num_xbars);
      
      insist (!m->hosts.tail || lx_hostcmp (lx_host_c (m->hosts.tail)->mac_address, lx_host_c (m->hosts.tail)->level, lx_host_c (n)->mac_address, lx_host_c (n)->level) > 0);
      lx_put (&m->hosts, lx_remove_from_anywhere (n));
      mi_punt (m->lx);
    }
  }

  if (!(m->root = lx_find_host (lx, &m->hosts, lx->mac_address, 0, &higher)) || lx->level != lx_host_c (m->root)->level)
  { 
    mi_c (("my %s has changed", m->root ? "level" : "existence"));
    mi_c (("Therefore this map version is invalid"));
    mi_memcpy (lx->map_address, map_address, 6);
    lx->map_version = map_version;
    lx->map_valid = 0;
    return 0;
  }
  
  for (first = 0; first < num_xbars;)
  {
    lx_advance_time (lx);

    for (i = 0; i < lx->tries; i++)
    {
      if (lx_send_map_request (lx, map_route, lx->timestamp, LX_MAP_XBAR, first) &&
	  (r = lx_wait_map_reply (lx, lx->timeout, lx->timestamp, LX_MAP_XBAR, rm)))
	break;
    }
    
    if (i == lx->tries)
    {
      mi_c (("no reply from " lx_mac_format,  lx_mac_args (lx->winner_address)));
      return 0;
    }
    
    insist (m);
    insist (r >= (int) sizeof (lx_map_reply_message_t));

    /*mi_c (("got map reply %d message", mi_htonl (rm->header.timestamp)));*/
    
    if (lx_memcmp (rm->map_address, map_address, 6) || rm->map_version != (unsigned) mi_htonl (map_version))
    {
      mi_c (("map address / map version changed to " lx_mac_format " %u",  lx_mac_args (rm->map_address), mi_htonl (rm->map_version)));
      return 0;
    }

    insist (num_hosts == mi_htons (rm->num_hosts));
    insist (num_xbars == mi_htons (rm->num_xbars));
    
    num_nodes = mi_htons (rm->num_nodes);
    insist (num_nodes);
    
    for (i = 0; i < num_nodes; i++, first++)
    {
      lx_node_t*n = mi_index2node (lx->xbars [first]);
      insist (n);
      insist (first < num_xbars);

      mi_c (("reading " lx_node_format, lx_node_args (n)));
      lx_net2xbar (m, &rm->nodes.xbar [i], lx_xbar_c (n), num_hosts, num_xbars);
      lx_put (&m->xbars, lx_remove_from_anywhere (n));
      mi_punt (m->lx);
    }
  }
  return 1;
  except: return 0;
}



static int lx_map_init_frontier (lx_t*lx, lx_map_t*m)
{
  lx_queue_t q;
  lx_node_t*n, *nn;
  int kids_match, matches_me, change, i;
  
  insist (lx && m);
  insist (!lx->staged_xbars.count);
  insist (!m->frontier.count);
  
  if (!m->root)
    return 0;
  
  if (!(n = lx_get_node (m->root, 0)) || !n->xbar)
    return 0;
 
  while ((n = lx_get (&m->xbars)))
  {
    lx_put (&lx->staged_xbars, n);
    mi_punt (lx);
  }
  
  lx_bzero ((char*) &q, sizeof (q));
  lx_put (&q, lx_remove (&lx->staged_xbars, lx_get_node (m->root, 0)));

  while ((n = lx_peek (&q)))
  {
    insist (n->xbar);

    mi_c (("seeing if we have to remap " lx_node_format, lx_node_args (n)));

    for (i = 0; i < LX_XBAR_SIZE; i++)
    {	
      mi_punt (lx);
      if ((nn = lx_get_node (n, i + n->first)) && nn->xbar && lx_queued (&lx->staged_xbars, nn))
      {
	lx_set_timeout (lx, LX_VERIFY_DISCONNECTED_TIMEOUT, LX_VERIFY_DISCONNECTED_TRIES);

	if (!lx_map_check_xbar_new_hosts (lx, nn) || !lx_map_check_xbar_old_hosts (lx, nn) || !lx_map_check_xbar_for_xbars (lx, nn, 1))
	  lx_put (&m->frontier, lx_remove (&lx->staged_xbars, nn));
	else
	  lx_put (&q, lx_remove (&lx->staged_xbars, nn));
      }
    }
    lx_put (&m->frontier, lx_remove_from_anywhere (n));
  }
  
  while ((n = lx_get (&lx->staged_xbars)))
  {
    lx_delete_hosts (m, n);
    lx_delete_node (m, n);
  }

  lx_set_timeout (lx, LX_VERIFY_DISCONNECTED_TIMEOUT, LX_VERIFY_DISCONNECTED_TRIES);

  for (i = 0, n = m->frontier.head; n; n = mi_index2node (n->next), i++)
  {
    for (i = 0; i < LX_XBAR_SIZE; i++)
    {
      if ((nn = lx_get_node (n, i + n->first)) && !nn->xbar && (!m->root || (!lx_map_check_host (lx, nn, LX_VERIFYING, &kids_match, &matches_me, &change) && change != LX_INVALID_MAP)))
      {
	if (nn == m->root)
	  m->root = 0;
	lx_remove_from_anywhere (nn);
	lx_delete_node (m, nn);
      }
    }
  }

  return m->root && m->num_hosts && m->num_xbars;
  except: return 0;
}

static int lx_map_incrementally (lx_t*lx, lx_map_t*m)
{
  lx_node_t*n,*nn;
  int ports [2 * LX_XBAR_SIZE];
  int i;
  
  insist (lx && m);
  insist (!lx->map_valid);
  
  if (!lx_map_init_frontier (lx, m))
    return 0;

  lx->aborted = 0;
  
  insist (m->num_hosts && m->num_xbars);
  insist (m->num_xbars == m->xbars.count + m->frontier.count);
  
  lx_set_timeout (lx, LX_ACTIVE_TIMEOUT, LX_ACTIVE_TRIES);

  mi_c (("mapping incrementally starting with %d host%s and %d xbar%s (%d on frontier)", m->num_hosts, m->num_hosts == 1 ? "" : "s", m->num_xbars, m->num_xbars == 1 ? "" : "s", m->frontier.count));
  
  while (!lx->aborted && (n = m->frontier.head))
  {
    if (!lx_map_find_xbars (m, n, ports))
      return 0;
    
    for (i = 1 - LX_XBAR_SIZE; i < LX_XBAR_SIZE; i++)
    {
      if (!ports [i - (1 - LX_XBAR_SIZE)])
	continue;
      
      insist (!lx_connected (n, i));
      
      if (!(nn = lx_map_add_xbar (m, n, i)))
	return 0;
      
      if (lx_map_find_hosts (m, nn) && (lx_has_hosts (nn) || !lx_map_match (m, nn)))
      {
	mi_c (("adding " lx_node_format " to frontier", lx_node_args (nn)));
	lx_put (&m->frontier, nn);
	insist (m->frontier.tail == nn);
      }
      if (m->frontier.head != n)
      {
	mi_c (("xbar I was exploring was deleted"));
	break;
      }
    }
    if (m->frontier.head == n && !lx_put (&m->xbars, lx_get (&m->frontier)))
      return 0;
  }
  return 1;
  except: return 0;
}

static int lx_map_simple_routes (lx_map_t*m, int index, int pass, int num_passes)
{
  lx_queue_t q;
  lx_node_t*n, *nn;
  int hosts_found, xbars_found, i, j, a, num_lowests, the_lowest, lowests, iter, lowest;
  int longest_route = 0;
  
  lx_t*lx = m->lx;
  
  insist (m && m->root);
  insist (LX_UP && LX_DOWN);
  insist (num_passes);
  
  hosts_found = xbars_found = 0;
  
  for (i = 0, n = m->hosts.head; n; n = mi_index2node (n->next), i++)
  {
    n->original.port = LX_NO_PORT;
    mi_punt (lx);
  }
  insist (i == m->num_hosts);
  
  for (i = 0, n = m->xbars.head; n; n = mi_index2node (n->next), i++)
  {
    n->original.port = LX_NO_PORT;
    n->route.length = 0;
    mi_punt (lx);
  }
  insist (i == m->num_xbars);
  
  n = lx_get_node (m->root, 0);
  insist (n);
  
  if (!n->xbar)
  {
    n->route.length = 0;
    /*mi_c ((lx_node_format " %s", lx_node_args (n), lx_print_route (n->route.hops, n->route.length)));*/
    mi_set_route (lx, index, lx_host_c (n)->mac_address, lx_host_c (n)->host_type, n->index, n->route.hops, n->route.length);
    
    hosts_found++;

    if (n != m->root)
    {
      m->root->route.length = 0;
      /*mi_c ((lx_node_format " %s", lx_node_args (m->root), lx_print_route (m->root->route.hops, m->root->route.length)));*/
      mi_set_route (lx, index, lx_host_c (m->root)->mac_address, lx_host_c (m->root)->host_type, m->root->index, m->root->route.hops, m->root->route.length);
    }
    return 1;
  }
  
  lx_bzero ((char*) &q, sizeof (q));
  
  n->original.port = lx_get_port (m->root, 0);
  n->xbar = LX_DOWN;
  lx_put (&q, lx_remove (&m->xbars, n));

  while ((n = lx_get (&q)))
  {
    insist (n->xbar && n->original.port > -LX_XBAR_SIZE && n->original.port < LX_XBAR_SIZE);
    lx_put (&m->xbars, n);
    xbars_found++;    
    
    for (;;)
    {
      a = m->root->index;
      num_lowests = lowests = the_lowest = 0;
      for (lowest = -1, iter = 0; iter < 3 && (!iter || lowest != -1); iter++)
      {
	for (i = 0; i < LX_XBAR_SIZE; i++)
	{	
	  mi_punt (lx);
	  j = (i + a) % LX_XBAR_SIZE;
	  
	  nn = 0;
	  if ((nn = lx_get_node (n, j + n->first)) && nn->original.port == LX_NO_PORT && (lx->clos || !nn->xbar || (n->xbar == LX_DOWN || lx_xbar_c (nn)->level > lx_xbar_c (n)->level)))
	  { 
	    if (iter == 0)
	    {
	      if (lowest == -1 || lowest > n->links [j].weight)
		lowest = n->links [j].weight;
	    }
	    else if (iter == 1)
	    {
	      insist (lowest != -1);
	      if (n->links [j].weight == lowest)
		num_lowests++;
	    }
	    else 
	    {
	      insist (num_lowests);
	      if (n->links [j].weight == lowest)
	      {
		if (lowests == the_lowest)
		  break;
		
		lowests++;
	      }
	    }
	  }
	  /*insist (!nn || !nn->xbar || lx_xbar_c (nn)->level != lx_xbar_c (n)->level);*/
	}
	if (iter == 1)
	{
	  insist (num_lowests);
	  the_lowest = mi_rand () % num_lowests;
	  lowests = 0;

	  /*mi_c (("num_lowests is %d the_lowest is %d", num_lowests, the_lowest));*/
	}
      }
      if (lowest == -1)
	break;

      insist (nn);
      insist (nn->original.port == LX_NO_PORT);

      n->links [j].weight++;

      lx_copy_route (nn->route.hops, n->route.hops, n->route.length);
      nn->route.hops [n->route.length] = j + n->first - n->original.port;
      nn->route.length = n->route.length + 1;

      insist (lx_follow (m->root, &nn->route) == nn);

      nn->original.port = lx_get_port (n, j + n->first);

      insist (nn->original.port > -LX_XBAR_SIZE && nn->original.port < LX_XBAR_SIZE);

      mi_punt (m->lx);

      if (nn->xbar)
      {
	nn->xbar = lx->clos ? LX_DOWN : ((n->xbar == LX_DOWN ? lx_xbar_c (nn)->level < lx_xbar_c (n)->level : 0) ? LX_DOWN : LX_UP);
	lx_put (&q, lx_remove (&m->xbars, nn));
      }
      else
      {
	if (nn->route.length > longest_route)
	  longest_route = nn->route.length;

	if ((nn->index + m->root->index) % num_passes == pass)
	{
	  mi_c ((lx_node_format " %s", lx_node_args (nn), lx_print_route (nn->route.hops, nn->route.length)));
	  mi_set_route (lx, index, lx_host_c (nn)->mac_address, lx_host_c (nn)->host_type, nn->index, nn->route.hops, nn->route.length);
	}
	
	hosts_found++;
      }
    }
  }
  
  insist (hosts_found == m->num_hosts);
  insist (xbars_found == m->num_xbars);
  
  return longest_route;
  except: return 0;
}


int lx_map_route (lx_map_t*m, int num_passes, int num_routes)
{
  int i, j;
  int longest_route = 0;
  lx_node_t*n;
  insist (m);
  
  mi_c (("computing routes"));
  insist (num_passes > 0 && num_routes > 0);
  
  for (n = m->hosts.head; n; n = mi_index2node (n->next))
  {
    n->links [0].weight  = 0;
    mi_punt (m->lx);

  }

  for (n = m->xbars.head; n; n = mi_index2node (n->next))
  {
    mi_punt (m->lx);

    for (i = 0; i < LX_XBAR_SIZE; i++)
      n->links [i].weight  = 0;
  }
  
  for (i = 0; i < num_routes; i++)
  {
    for (j = 0; j < num_passes; j++)
    {
      if (!(longest_route = lx_map_simple_routes (m, i, j, num_passes)))
	return 0;
    }
  }

  return longest_route;
  except: return 0;
}

static int lx_map_compute_in_ports (lx_t*lx, lx_node_t*n)
{
  int i, j, in, num_xbars;
  lx_node_t*o;
  lx_route_t*r;
  
  insist (n && !n->xbar);

  num_xbars = lx->map.num_xbars;

  if (!num_xbars)
    return 1;
  
  for (i = 0; i < num_xbars; i++)
  { 
    o = mi_index2node (lx->xbars [i]);
    insist (o);
    r = &o->route;
    
    insist (lx_follow (n, r) == o);

    in = lx_get_port (n, 0);
    o = lx_get_node (n, 0);
    
    insist (o && o->xbar);

    o->original.port = in;
    mi_punt (lx);
    
    for (j = 0; j < r->length; j++)
    {
      insist (in + r->hops [j] - o->first >= 0 && in + r->hops [j] - o->first < LX_XBAR_SIZE);
      
      in = lx_get_port (o, in + r->hops [j]);
      o = lx_get_node (o, o->original.port + r->hops [j]);

      insist (o);
      o->original.port = in;
    }
    insist (o == mi_index2node (lx->xbars [i]));
    mi_punt (lx);
  }
  return 1;  
  except: return 0;
}

void lx_mapper_main (lx_t*lx)
{ 
  int r, t;
  unsigned char map_address [6];
  unsigned map_version;
  int longest_route, done_everywhere, max_level = 0;
  int thrash;
  
  lxd_cur = 0;
  
  lx_init (lx);
  mi_get_address (lx, lx->mac_address);
  insist (!lx_zero (lx->mac_address, 6));
  
  mi_c (("I am host " lx_mac_format " (type %d)", lx_mac_args (lx->mac_address), lx->host_type));

  mi_c (("%d hosts, %d xbars per map message piece",lx_piece_count (lx_map_host_t), lx_piece_count (lx_map_xbar_t)));
  
  lx_advance_time (lx);

  lx->map_version = mi_rand ();

  for (;;)
  {
    mi_c (("going active"));
    lx->active = 1;
    lx->aborted = !lx->level;

    while (!lx->aborted)
    {
      lx->map_valid = 0;
      mi_set_map_version (lx, lx->map_address, 0, 0, 0, 0);
      
      if (lx->die)
	return;
      
      t = mi_time (lx);
      r = 0;
      
      if (!lx->map_incrementally || !(r = lx_map_incrementally (lx, &lx->map)))
      {
	lx_reset_map (lx, &lx->map);
	lx_set_timeout (lx, LX_ACTIVE_TIMEOUT, LX_ACTIVE_TRIES);
	r = lx_map_explore (&lx->map);
      }
      
      if (r && !lx->aborted)
      {
	if (r < 0)
	{
	  mi_c (("mapper is disconnected"));
	  lx_wait (lx, LX_PASSIVE_TIMEOUT, (int*)&lx->aborted, 0);
	  continue;
	}
	
	t = mi_time (lx) - t;
	mi_c (("mapping took %d usecs", t));

	if (lx->extra_hosts)
	  lx_inflate_map (&lx->map, 8, lx->extra_hosts);
	
	mi_c (("%d host%s and %d xbar%s", lx->map.num_hosts, lx->map.num_hosts == 1 ? "" : "s", lx->map.num_xbars, lx->map.num_xbars == 1 ? "" : "s"));

	lx_before_sending (&lx->map);

	r = lx->clos || lx_number_xbars_bfs (&lx->map);
	insist (r);  

	mi_set_route_begin (lx);
	longest_route = lx_map_route (&lx->map, lx->num_passes, lx->num_routes);
	mi_set_route_end (lx);

	max_level = (longest_route - 1) / 2;

	r = lx_map_compute_in_ports (lx, lx->map.root);
	insist (r);

	r = !lx->clos || lx_number_xbars_clos (&lx->map, max_level);

	while (!++lx->map_version)
	  lx->map_version = mi_rand ();
	
	mi_c (("map version is now %d", lx->map_version));

	mi_memcpy (lx->map_address, lx->mac_address, 6);

	lx->kids_match = 0;
	lx->map_valid = 1;

	lx_set_timeout (lx, LX_PASSIVE_TIMEOUT, LX_PASSIVE_TRIES);
	
	mi_dump_map (lx, &lx->map);
	mi_set_map_version (lx, lx->map_address, lx->map_version, lx->map.num_hosts, 0, lx->map_valid);

	while (lx->printeur)
	  lx_wait (lx, LX_PASSIVE_TIMEOUT, 0, 0);    
	thrash = 0;
	
	while (!lx->aborted && lx_map_check_hosts (lx) && lx_map_check_xbars (lx, max_level) && lx_map_check_xbars_new_hosts (lx))
	{
	  mi_c (("map version " lx_mac_format " %u", lx_mac_args (lx->map_address), lx->map_version));
	  mi_c (("%d host%s and %d xbar%s", lx->map.num_hosts, lx->map.num_hosts == 1 ? "" : "s", lx->map.num_xbars, lx->map.num_xbars == 1 ? "" : "s"));

	  done_everywhere = lx->kids_match;
	  
	  if (done_everywhere != lx->done_everywhere)
	  {
	    lx->done_everywhere = done_everywhere;
	    mi_set_map_version (lx, lx->map_address, lx->map_version, lx->map.num_hosts, lx->done_everywhere, lx->map_valid);
	  }

	  if (lx->done_everywhere)
	  {
	    mi_c (("mapping is done everywhere"));
	    if (lx->map_once)
	      return;
	  }
	  
	  if (lx->die)
	    return;

	  lx_wait (lx, LX_PASSIVE_TIMEOUT, (int*)&lx->aborted, 0);
	  mi_c (("I am " lx_node_format, lx_node_args (lx->map.root)));
	  mi_c (("verifying again"));

	  if (lx->thrash && lx->done_everywhere && ++thrash == 10)
	    lx->aborted = LX_ABORT_ACTIVE;	  
	}
      }
      if (lx->aborted == LX_ABORT_ACTIVE)
	lx->aborted = 0;
    }

    mi_c (("going passive"));

    lx->aborted = 0;    
    lx->active = 0;
    lx->map_valid = 0;
    lx_reset_map (lx, &lx->map);
    lx->kids_match = 0;

    mi_set_map_version (lx, lx->map_address, lx->map_version, lx->map.num_hosts, 0, lx->map_valid);

    if (!lx->level)
    {
      lx_bzero ((char*) lx->parent_address, 6);

      while (lx_zero (lx->parent_address, 6))
      {
	if (lx->die)
	  return;
      
	lx_wait (lx, LX_PASSIVE_TIMEOUT, 0, 0);
      }
    }
    
    while (!lx->aborted)
    { 
      lx_set_timeout (lx, LX_PASSIVE_TIMEOUT, LX_PASSIVE_TRIES);

      if (!lx_zero (lx->parent_address, 6))
      {	
	mi_memcpy (lx->winner_address, lx->parent_address, 6);
	lx->winner_route.length = lx->parent_route.length;
	lx_copy_route (lx->winner_route.hops, lx->parent_route.hops, lx->parent_route.length);
      }

      if (!(r = lx_map_query_new_version (lx, &map_version, map_address)))
      {
	mi_c (("query new_version failed"));
	lx->aborted = 0;
	break;
      }
      
      if (r > 0)
      {
	if (lx->map_valid)
	{
	  lx_bzero ((char*) lx->map_address, 6);
	  lx->map_version = 0;
	}

	lx->map_valid = 0;
	lx_reset_map (lx, &lx->map);
	
	if (!lx_map_collect_map (&lx->map, &lx->winner_route, map_address, map_version))
	{
	  lx_reclaim_staged (lx);
	  mi_c (("collect map failed. I'll try again."));
	  lx_wait (lx, LX_PASSIVE_TIMEOUT, (int*)&lx->aborted, 0);
	  continue;
	}

	lx_before_sending (&lx->map);

	mi_set_route_begin (lx);
	longest_route = lx_map_route (&lx->map, lx->num_passes, lx->num_routes);
	mi_set_route_end (lx);
	max_level = (longest_route - 1) / 2;

	mi_memcpy (lx->map_address, map_address, 6);
	lx->map_version = map_version;
	lx->map_valid = 1;
	mi_dump_map (lx, &lx->map);
	mi_set_map_version (lx, lx->map_address, lx->map_version, lx->map.num_hosts, 0, lx->map_valid);

	r = lx_map_compute_in_ports (lx, lx->map.root);
	insist (r);
      }
      lx_set_timeout (lx, LX_PASSIVE_TIMEOUT, LX_PASSIVE_TRIES);
      
      if (lx->map_valid && !lx->paused && map_version == lx->map_version)
      {
	mi_c (("verifying"));

	if (!lx_map_check_hosts (lx) || !lx_map_check_xbars (lx, max_level) || !lx_map_check_xbars_new_hosts (lx))
	{
	  mi_c (("network verify failed"));
	  lx->map_valid = 0;
	  break;
	}      
	mi_c (("I am " lx_node_format, lx_node_args (lx->map.root)));
	mi_c (("map version " lx_mac_format " %u", lx_mac_args (lx->map_address), lx->map_version));
	mi_c (("%d host%s and %d xbar%s", lx->map.num_hosts, lx->map.num_hosts == 1 ? "" : "s", lx->map.num_xbars, lx->map.num_xbars == 1 ? "" : "s"));
	
	if (lx->done_everywhere)
	{
	  mi_c (("mapping is done everywhere"));
	}
      }
      do lx_wait (lx, LX_PASSIVE_TIMEOUT, (int*)&lx->aborted, 0);
      while (lx->paused);      
      
      if (lx->die)
	return;
    }
    except:;
  }
}

int lx_route_valid (struct lx_t*lx, unsigned char mac_address [6], signed char*hops, unsigned length, int physical)
{
  lx_node_t*n, *higher;
  lx_route_t route;
  
  insist (lx && mac_address && hops && length >= 0 && length < LX_ROUTE_SIZE);

  if (!lx->map_valid)
    return 0;

  mi_memcpy (route.hops, hops, route.length = length);

  if (physical)
    lx_logical_route (route.hops, length);
  
  insist (lx->map.root);
  return (n = lx_find_host (lx, &lx->map.hosts, mac_address, 0, &higher)) && lx_follow (lx->map.root, &route) == n;
  except: return 0;
}

