/*
  sm_Port.c
  map tools
  finucane@myri.com (David Finucane)
*/


#include "insist.h"
#include "sm_Port.h"
#include "sm_Packet.h"

sm_Port::sm_Port ()
{
}

sm_Port::sm_Port (sm_Node*node, int port, int size, int latency, int timeForByte)
{
  this->port = port;
  this->node = node;
  head = tail = 0;
  packet = 0;
  next = 0;
  numPackets = 0;
  disabled = 0;
  
  timeUtilized = 0;
  
  define (size, latency, timeForByte);
}

void sm_Port::setDisabled (int disabled)
{
  this->disabled = disabled;
}

void sm_Port::define (int size, int latency, int timeForByte)
{
  insist (this);
  insist (size > 0 && latency >= 0 &&  timeForByte > 0);
  this->size = size;
  this->latency = latency;
  this->timeForByte = timeForByte;
  exception: return;
}

void sm_Port::call (sm_EventList* /*list*/, int /*type*/, void*data)
{
  insist (this);

  timeUtilized += (int) data * timeForByte;
  numPackets++;
  
  exception: return;
}

int sm_Port::getAddedLatency ()
{
  insist (this);
  return latency;
  exception: return 0;
}

void sm_Port::setNext (sm_Port*next)
{
  insist (this);
  this->next = next;
  exception: return;
}
sm_Port*sm_Port::getNext ()
{
  insist (this);
  return next;
  exception: return 0;
}
void sm_Port::setPacket (sm_Packet*packet)
{
  insist (this);
  perr ("nport %x setpacket %x \n", this, packet);
  this->packet = packet;
  exception: return;
}
sm_Packet*sm_Port::getPacket ()
{
  insist (this);
  return packet;
  exception: return 0;
}
int sm_Port::getSize ()
{
  insist (this);
  return size;
  exception: return 0;
}
int sm_Port::getPort ()
{
  insist (this);
  return port;
  exception: return 0;
}
sm_Node*sm_Port::getNode ()
{
  insist (this);
  return node;
  exception: return 0;
}
sm_Time sm_Port::getTimeUtilized()
{
  insist (this);
  return timeUtilized;
  exception: return 0;
}

int sm_Port::getNumPackets ()
{
  insist (this);
  return numPackets;
  exception: return 0;
}

sm_Packet*sm_Port::getBlocked ()
{
  sm_Packet*k;
  k = getPacket ();
  if (k && (k->getTailPort () != this || k->getTailPosition () == 0))
    return k;
  return head;
}

void sm_Port::unblock (sm_EventList*list)
{
  insist (list); 
  sm_Packet*p;
  if ((p = getBlocked ()))
  {
    perr ("unblocking message %d", p->id);
    p->blocked = 0;
    p->lastTime = list->getCurrentTime () - 1;
    p->addEvent (list, 0, sm_EventList::MOVE, 0);
  }
  exception:;
}

void sm_Port::removeBlocked ()
{
  insist (head);
  perr ("removing blocked message %d from list on port %x", head->id, this);
  sm_Packet*t;
  t = head;
  
  if (!(head = head->nextBlocked))
    tail = 0;
  
  t->nextBlocked = 0;
  t->blockedPort = 0;
  
  sm_Packet*k;
  for (k = head; k && k != tail; k = k->nextBlocked)
    k->check ();
  
  insist (k == tail);
  
  exception:;
}
