/*
  add_host.c
  add a host to a map file
  finucane@myri.com (David Finucane)
*/


#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>

#include "mt_Component.h"
#include "mt_MapFile.h"
#include "mt_Graph.h"
#include "mt_Args.h"
#include "mt_Switch.h"

#define die(s){mt_Component::printFormat s; exit (1);}
#define insist(e) do { if (!(e)) {mt_Component::printFormat ("Assertion failed in file %s, line %d",__FILE__,__LINE__); goto exception;}} while (0)
#define insistp(e, s) do { if (!(e)) {mt_Component::printFormat s; goto exception;}} while (0)


static void printLine (char*s)
{
  printf ("%s\n", s);
}
int main (int argc, char*argv[])
{  
  mt_Component::initialize (printLine);
  mt_Address address;
  mt_Node*s,*h;
  
  if (argc != 6)
    die (("usage: %s <mapfile> <switch> <port> <hostname> <mac_address>", argv[0]));
 
  char*mapFile = argv [1];
  char*switchName = argv [2];
  int port = atoi (argv [3]);
  
  if (port < 0 || port > 16)
    die (("bad port %d", argv [3]));
  
  char*hostname = argv [4];

  if (!address.fromString (argv[5]))
    die (("malformed address %s", argv [5]));

  mt_MapFile mf (mapFile, mt_File::_READ);
  mt_Graph g;
  
  if (!mf.read (&g))
    die (("couldn't parse map file %s", mapFile));

  if (!(s = g.getNode (switchName)))
    die (("couldn't find switch %s in %s", switchName, mapFile));
  
  if ((h = s->getNode (port)))
  {
    printf ("replacing host");
    s->disconnect (port);
    delete h;
  }
  
  if (g.getNode (hostname))
    die (("node %s already exists not on %s:%d", hostname, switchName, port));

  if (!(h = g.newNode (mt_Node::HOST, hostname, "-")))
    die (("malloc failed."));

  g.add (h);
  
  h->setAddress (&address);
  s->connect (port, 0, h);
  h->connect (0, port, s);
  
  mt_MapFile m (stdout);
  m.write (&g);
  
  return 0;
}
