#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "mt_RouteFile.h"
#include "mt_MapFile.h"

#define MAX_NAME 50
#define MAX_LINE 500
#define DELIMITERS " "
#define MAC_FORMAT "%2.2x:%2.2x:%2.2x:%2.2x:%2.2x:%2.2x"
#define MAC_ARGS(m) m[0] & 0xff , m[1] & 0xff , m[2] & 0xff , m[3] & 0xff , m[4] & 0xff , m[5] & 0xff

#define die(s){printf s; printf ("\n");exit (1);}

void usage (char*s)
{
  printf ("usage: %s <map-file> <gm_board_info-file>\n", s);
  exit (1);
}
int match (FILE*fp, char*s)
{
  char*t;
  char line [MAX_LINE];

  while ((t = fgets (line, MAX_LINE, fp)))
    if (strstr (t, s))
      return 1;
  return 0;
}

int scan_mac_address (unsigned char mac_address [6], char*s)
{
  int i;
  int t [6];
  
  if (sscanf (s, "%x:%x:%x:%x:%x:%x", t,t+1,t+2,t+3,t+4,t+5) != 6)
    return 0;
  for (i = 0; i < 6; i++)
    mac_address [i] = t [i];
  return 1;
}


int convert (mt_Graph*g, FILE*fp)
{
  unsigned char mac_address [6];
  char*t;
  char line [MAX_LINE];


  if (!match (fp, "----"))
    die (("couldn't find route table"));  
  
  while ((t = fgets (line, MAX_LINE, fp)) && *t != '\n')
  {
    strtok (t, DELIMITERS);
    
    if (!(t = strtok (0, DELIMITERS)) || !scan_mac_address (mac_address, t))
      break;
    
    if (!(t = strtok (0, DELIMITERS)))
      die (("expected something in the gmName column"));

    mt_Address a ((char*) mac_address);
    mt_Node*h = g->getHost (&a);
    if (h)
    {
      h->setName (t);
    }
  }
  return 1;
}

static void printLine (char*s)
{
  printf ("%s\n", s);
}


int main (int argc, char*argv [])
{
  FILE*fp;  

  mt_Component::initialize (printLine);

  if (argc != 3)
    usage (argv [0]);

  mt_MapFile mf (argv [1], mt_File::_READ);
  mt_Graph g;
    
  if (!mf.read (&g))
    die (("main: couldn't parse map file %s", argv [1]));

  if (!(fp = fopen (argv [2], "r")))
    die (("couldn't open %s", argv [2]));
  
  convert (&g, fp);
  
  mt_MapFile of (stdout);
  of.write (&g);
}
