#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "mt_RouteFile.h"
#include "mt_MapFile.h"

#define MAX_NAME 50
#define MAX_LINE 500
#define MAX_HOPS 16
#define MIN_HOP -32
#define MAX_HOP 32
#define DELIMITERS " "
#define MAC_FORMAT "%2.2x:%2.2x:%2.2x:%2.2x:%2.2x:%2.2x"
#define MAC_ARGS(m) m[0] & 0xff , m[1] & 0xff , m[2] & 0xff , m[3] & 0xff , m[4] & 0xff , m[5] & 0xff


typedef struct entry_t
{
  struct entry_t*next;
  unsigned char mac_address [6];
  int num_hops;
  char hops [MAX_HOPS];
}entry_t;

#define die(s){printf s; printf ("\n");exit (1);}


void usage (char*s)
{
  printf ("usage: %s <map-file> <gm_board_info-file>\n", s);
  exit (1);
}
int match (FILE*fp, char*s)
{
  char*t;
  char line [MAX_LINE];

  while ((t = fgets (line, MAX_LINE, fp)))
    if (strstr (t, s))
      return 1;
  return 0;
}

int scan_mac_address (unsigned char mac_address [6], char*s)
{
  int i;
  int t [6];
  
  if (sscanf (s, "%x:%x:%x:%x:%x:%x", t,t+1,t+2,t+3,t+4,t+5) != 6)
    return 0;
  for (i = 0; i < 6; i++)
    mac_address [i] = t [i];
  return 1;
}

char*get_hostname (mt_Graph*g, unsigned char mac_address [6])
{
  mt_Node*n;
  mt_Address a;
  a.fromBytes ((char*) mac_address);
  
  if (!(n = g->getHost (&a)))
    die (("couldn't find " MAC_FORMAT " in map file\n", MAC_ARGS (mac_address)));
  
  return n->getName ();
}

int convert (mt_Graph*g, FILE*fp)
{
  unsigned char mac_address [6];
  unsigned char a [6];
  int i, num_entries, num_bytes;
  char*t;
  char line [MAX_LINE];  
  entry_t*head = 0;
  entry_t*tail = 0;
  static int preamble = 1;
  char match_string[MAX_LINE];
  long offset;


  offset = ftell (fp);
  if (match (fp, "LANai10"))
    strcpy (match_string, "ROM");
  else
    strcpy (match_string, "lanai_cpu_version");
  fseek (fp, offset, SEEK_SET);

  if (!match (fp, match_string))
    return 0;

  if (!(t = fgets (line, MAX_LINE, fp)))
    die (("no mac address"));
  if ((strstr (t, "SN=")))
    t = fgets (line, MAX_LINE, fp);
  if (!(t = strstr (t, "=")))
    die (("no mac address"));
  t++;

  if (!scan_mac_address (mac_address, t))
    die (("couldn't read mac address"));
 
  if (!match (fp, "----"))
    die (("couldn't find route table"));  
  
  num_entries = num_bytes = 0;
  
  while ((t = fgets (line, MAX_LINE, fp)) && *t != '\n')
  {
    strtok (t, DELIMITERS);
    
    if (!(t = strtok (0, DELIMITERS)) || !scan_mac_address (a, t))
      break;
   
    num_entries++;
    
    if (!tail)
      head = tail = (entry_t*) calloc (sizeof (entry_t), 1);
    else
    {
      tail->next = (entry_t*) calloc (sizeof (entry_t), 1);
      tail = tail->next;
    }
    
    if (!tail)
      die (("malloc failed"));
    
    memcpy (tail->mac_address, a, 6);
    
    if (!(t = strtok (0, DELIMITERS)))
      die (("expected something in the gmName column"));

    if (*t == '[')
      while ((t = strtok (0, DELIMITERS)) && strcmp (t, "YP]"));
    t = strtok (0, DELIMITERS);
    
    while (t && *t != '(' && *t != '\n')
    {
      int n;
      char c;
      
      if (tail->num_hops > MAX_HOPS)
	die (("route too long"));
      
      if (sscanf (t, "%x", &n) != 1)
	die (("bad hop"));
      c = n;
      
      if ((c = (c & 0x3f)) & 0x20)
	c |= 0xf0;

      if (c < MIN_HOP || c > MAX_HOP)
	die (("bad hop"));

      tail->hops [tail->num_hops++] = c;
      t = strtok (0, DELIMITERS);
    }
    if (!tail->num_hops)       /*work around gm2's lies*/
      tail->num_hops = 1;

    num_bytes += tail->num_hops;
  }

  if (!num_entries || !tail || !head)
    die (("empty routing table"));
  
  if (preamble)
  {
    printf ("%d 1\n", num_entries);
    preamble = 0;
  }

  printf ("%s (%d %d)\n", get_hostname (g, mac_address), num_entries, num_bytes);

  while (head)
  {
    printf ("%s ", get_hostname (g, head->mac_address)); 
  
    for (i = 0; i < head->num_hops; i++)
      printf ("%d%s", head->hops [i], i != head->num_hops - 1 ? ", ":"");
    printf ("\n");
    
    tail = head->next;
    free (head);
    head = tail;
  }
  return 1;
}

static void printLine (char*s)
{
  printf ("%s\n", s);
}


int main (int argc, char*argv [])
{
  FILE*fp;  

  mt_Component::initialize (printLine);

  if (argc != 3)
    usage (argv [0]);

  mt_MapFile mf (argv [1], mt_File::_READ);
  mt_Graph g;
    
  if (!mf.read (&g))
    die (("main: couldn't parse map file %s", argv [1]));

  if (!(fp = fopen (argv [2], "r")))
    die (("couldn't open %s", argv [2]));
  
  while (convert (&g, fp));
}
