/*
  load_routes.c
  finucane@myri.com
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "mt_RouteFile.h"
#include "mt_Gm.h"
#include "mt_Host.h"
#include "mt_MapFile.h"

#define die(s){mt_Component::printFormat s; exit (1);}
#define insist(e) do { if (!(e)) {mt_Component::printFormat ("Assertion failed in file %s, line %d",__FILE__,__LINE__); exit (1);}} while (0)
#define insistp(e, s) do { if (!(e)) {mt_Component::printFormat s; exit (1);}} while (0)


static void printLine (char*s)
{
  printf ("%s\n", s);
}

void usage (char*s)
{
  printf ("usage: %s <map-file> <routes-file> <unit>\n", s);
  exit (1);
}

class mt_Fake : public mt_Job
{
  public:
  mt_Fake (mt_Node*node, mt_Network*network) : mt_Job (node, network)
  {
  }
  virtual void dump (FILE*fp)
  {
  }
  virtual int willWait ()
  {
    return 1;
  }
  virtual int start ()
  {
    return 0;
  }
  virtual void receive (int event, char*p, int length)
  {
  }
};

int main (int argc, char*argv[])
{ 
  mt_Component::initialize (printLine);

  mt_Host host ("fake", "-");
  mt_Fake fake (&host, mt_getNetwork ());
  mt_Address address;
  char name [mt_Network::HOSTNAME_LENGTH + 1];
  int type, unit;
  mt_Network*net = mt_getNetwork ();
  if (!net)
    die (("something evil"));
  
  int numHosts, maxRoutes;
  char from [mt_Node::NAME_LENGTH + 1];
  char to [mt_Node::NAME_LENGTH + 1];
  char lastFrom [mt_Node::NAME_LENGTH + 1];
  
  mt_Route route, r2;

  if (argc == 2)
  {
    if (!net->open (&fake, &address, name, &type, atoi (argv[1])))
      die (("couldn't open gm mapper port"));
    printf ("%s", address.toNewString ());
    exit (0);
  }
  
  if (argc != 4)
    usage (argv [0]);
  
  mt_MapFile mf (argv[1], mt_File::_READ);
  mt_Graph g;

  if (!mf.read (&g))
    die (("main: couldn't parse map file %s", argv [1]));

  mt_RouteFile rf (argv [2], mt_File::_READ);  
  unit = atoi (argv [3]);
  
  if (!net->open (&fake, &address, name, &type, unit))
    die (("couldn't open gm mapper port"));
  
  if (!rf.readHeader (&numHosts, &maxRoutes))
    die (("error reading header from %s", argv [1]));

  *lastFrom = 0;
  
  int numRoutes, maxBytes, numBytes;
  numRoutes = maxBytes = numBytes = 0;

  if (!rf.readFromLine (from, &numRoutes, &maxBytes) || numRoutes < 0 && maxBytes < 0)
    die (("couldn't read from line"));

  if (strcmp (from, address.toNewString ()))
    die (("route file mismatch, expecting from mac address %s and not %s", address.toNewString (), from));
    
  insist (strcmp (from, lastFrom));
  strcpy (lastFrom, from);

  for (int j = 0; j < numRoutes; j++)
  { 
    if (!rf.readToLine (to, &route))
      die (("couldn't read to line"));

    mt_Node*n = g.getHost (to);

    if (!n)
      die (("%s not found in map file %s", to, argv [1]));

    ((mt_Gm*)net)->setRoute (&fake, n->getAddress (), &route);
  }  
}
