/*
  safe_routes.c
  replace deadlockable routes until deadlocks go away.
  finucane@myri.com
*/

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>

#include "de_Deadlock.h"
#include "mt_RouteTable.h"
#include "mt_MapFile.h"

#define die(s){mt_Component::printFormat s; exit (1);}

static void printLine (char*s)
{
  printf ("%s\n", s);
}

int main (int argc, char*argv [])
{
  mt_Component::initialize (printLine);
  mt_RouteTable*routes = 0;
  mt_RouteTable safeRoutes;
  
  if (argc < 5)
    die (("usage: %s <mapfile> <deadlock-routefile> <safe-routefile> <outfile> [-verbose]", argv [0]));
  
  char*mapFile = argv[1];
  char*deadlockFile = argv[2];
  char*safeFile = argv [3];
  char*outfile = argv [4];
  int verbose = argc > 5 && !strcmp (argv [5], "-verbose");
  
  mt_MapFile mf (mapFile, mt_File::_READ);
  de_Deadlock deadlock;

  if (!mf.read (&deadlock))
    die (("main: couldn't parse map file %s", mapFile));

  if (!safeRoutes.fromFile (safeFile, &deadlock))
    die (("couldn't parse routes file %s", safeFile));

  routes = new mt_RouteTable;
  if (!routes)
    die (("couldn't alloc route table"));

  if (!routes->fromFile (deadlockFile, &deadlock))
    die (("couldn't parse routes file %s", deadlockFile));

  routes = deadlock.removeDeadlocks (&safeRoutes, routes, verbose);
  if (!routes)
    die (("coudln't remove deadlocks"));
  
  {
    mt_RouteFile rf (outfile, mt_File::_WRITE);
    rf.write (&deadlock, routes);
  }
}
