#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "mt_RouteFile.h"
#include "mt_MapFile.h"

#define die(s){mt_Component::printFormat s; exit (1);}
#define insist(e) do { if (!(e)) {mt_Component::printFormat ("Assertion failed in file %s, line %d",__FILE__,__LINE__); exit (1);}} while (0)
#define insistp(e, s) do { if (!(e)) {mt_Component::printFormat s; exit (1);}} while (0)


static void printLine (char*s)
{
  printf ("%s\n", s);
}

void usage (char*s)
{
  printf ("usage: %s <map-file> <route-file>\n", s);
  exit (1);
}

int main (int argc, char*argv[])
{ 
  mt_Component::initialize (printLine);

  int numHosts, maxRoutes;
  char from [mt_Node::NAME_LENGTH + 1];
  char to [mt_Node::NAME_LENGTH + 1];
  char lastFrom [mt_Node::NAME_LENGTH + 1];
  char filename [mt_Node::NAME_LENGTH * 2];

  mt_Route route, r2;
  FILE*fp;
  
  if (argc != 3)
    usage (argv [0]);

  mt_RouteFile rf (argv [2], mt_File::_READ);
  
  if (!rf.readHeader (&numHosts, &maxRoutes))
    die (("error reading header from %s", argv [2]));

  mt_MapFile mf (argv [1], mt_File::_READ);
  mt_Graph g;
    
  if (!mf.read (&g))
    die (("main: couldn't parse map file %s", argv [1]));

  *lastFrom = 0;
  
  for (int i = 0; i < numHosts; i++)
  {
    int numRoutes, maxBytes, numBytes;
    numRoutes = maxBytes = numBytes = 0;

    if (!rf.readFromLine (from, &numRoutes, &maxBytes) || numRoutes < 0 && maxBytes < 0)
      die (("couldn't read from line"));

    mt_Node*n = g.getNode (from);
    if (!n)
      die (("%s not found in map file %s", from, argv [1]));

    sprintf (filename, "%s.routes", n->getAddress ()->toNewString ());

    if (!(fp = fopen (filename, "w")))
      die (("couldn't create %s", filename));
    
    if (!rf.writeLimits (fp, numHosts, maxRoutes))
      die (("couldn't write header"));
    
    insist (strcmp (from, lastFrom));
    strcpy (lastFrom, from);

    rf.writeFromLine (fp, n->getAddress ()->toNewString (), numRoutes, maxBytes);
    
    for (int j = 0; j < numRoutes; j++)
    {
      if (!rf.readToLine (to, &route))
	die (("couldn't read to line"));

      mt_Node*n = g.getNode (to);
      if (!n)
	die (("%s not found in map file %s", to, argv [1]));
      
      rf.writeToLine (fp, n->getAddress ()->toNewString (), &route);
    }
    fclose (fp);
  }
}
