#! /bin/sh

# This is the core command line processing for a /etc/init.d/* script,
# designed to be included by a daemon script.

gm_do test x"$gm_service" != x || gm_die Did not define gm_service above.

# Syntax:  .../*.d/<gm_service> [--verbose] {start|stop|restart}
#
# Where:   <gm_service>  names the service to be started or stopped -- eg, gm
#          --verbose     is optional, for debugging
#          Exactly one of 'start' or 'stop' must be specified

# Start a service

gm_start ()
{
    # Arrange to automatically cleanup failed starts using uninstall logs.
    #
    # We don't use these logs to stop the service, since missing uninstall
    # logs might make it impossible to stop the service.  Stopping is
    # essential, where uninstalls are not, since new software can be installed
    # on top of old.
    gm_unstart_log="/tmp/gm_start.$$"
    gm_create_uninstall_file "${gm_unstart_log}"
    gm_log_uninstall_command "rm -f \"${gm_unstart_log}\""

    # Start the service
    service="$1"
    gm_printf "Starting $service... "
    gm_start_${service}
    if test $? != 0
    then
        gm_echo FAILED.
        return 1
    fi
    gm_echo done.

    # Succeeded, so remove the cleanup log.
    gm_close_uninstall_log
    rm -f "${gm_unstart_log}"
    return 0
}

# Stop a service

gm_stop ()
{
    service="$1"
    gm_printf "Stopping $service... "
    gm_stop_${service}
    if test $? != 0
    then
        gm_echo FAILED.
        return 1
    fi
    gm_echo done.
    return 0
}

# Parse command line option and execute command

if test x"$1" = x"--verbose"
then
    gm_verbose=1
    shift
fi
	
case "$1" in
    start)
	gm_start $gm_service
	exit $?
	;;
    stop)
	gm_stop $gm_service
	exit $?
	;;
    restart | force-reload)
	gm_stop $gm_service
	gm_start $gm_service
	exit $?
	;;
    *)
	gm_die Unknown option \""$1"\".
esac

# Never get here

exit 1
