; melp.el
; finucane@myri.com
;
; emacs code to extract uuencoded help tar.gz file from a RMAIL buffer
; into a directory for ruth.

;
; this thing depends on the uuencode filename starting with
; a number and a dash. This number is supposed to be a help number and
; it is used as the default subdirectory name.
; 
; e.g.
;
; begin 644 0-010815_1350_glenn@asl.myri.com.tar.gz
;
;
; might extract into /tmp/0
;
; unless the user overrides the number when prompted.
;
; the script will delete the temporary .tar.gz file it makes.
; 
; Also this pretty much depends on gnu tar (-C option?)
; and if the programs mkdir, uudecode, tar, and rm
; aren't on your path when emacs uses them you'll
; have to set the full pathnames to these in the call-process
; calls below. I didn't bother to make them variables.
;
; to use this, save this file to ~/melp.el and add
; to your .emacs file
;
; (load-library "~/melp.el")
; (setq melp-dir "/tmp/")
;
; and in any buffer with the uuencoded stuff type M-x melp
;
; You can bind melp to a key by adding something like this to
; you .emacs file
;
; (global-set-key [f1] 'melp)
;
;
;
(defvar melp-dir "/tmp"
  "*pathname for toplevel directory to untar everything to")

(defun melp-get-filename ()
  (beginning-of-buffer)
  (if (search-forward-regexp "^begin [0-7][0-7][0-7] " nil  t)
      (let ((start (point)))
        (end-of-line)
        (buffer-substring start (point)))))

(defun melp ()
  (interactive)
  (setq filename (melp-get-filename))
  (setq fullname (concat melp-dir "/" filename))
  (setq default (substring filename 0 (string-match "-" filename)))
  (setq subdir (substring filename
			  (+ 1 (string-match "-" filename))
			  (string-match ".tar.gz" filename)))
  (setq number (read-string "enter help number: " default 'my-history))
  (setq newdir (concat melp-dir "/" number))
  (beginning-of-line)
  (setq start (point))
  (re-search-forward "^end$")
  (forward-line)
  (call-process "mkdir" nil "melp-buffer" t newdir)
  (call-process-region start (point) "uudecode" nil "melp-buffer" t "-o"
		       fullname)
  (call-process "tar" nil "melp-buffer" t "-C"
   (concat melp-dir "/" number) "-xzv" "-f" fullname)
  (delete-file fullname)
  ; (switch-to-buffer "melp-buffer")
  (find-file (concat newdir "/" subdir))
  (beginning-of-buffer))
