/******************************************************************-*-c-*-
 * Myricom GM networking software and documentation			 *
 * Copyright (c) 2001 by Myricom, Inc.					 *
 * All rights reserved.	 See the file `COPYING' for copyright notice.	 *
 *************************************************************************/

/* author: glenn@myri.com */

#include "gm_config.h"
#include "gm.h"
#include "gm_lanai.h"
#include "gm_stbar.h"
#include <stdio.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/mman.h>

#define GM_LX(x) (31 - (x ^ 0x7))

extern unsigned int sleep (unsigned int);

static void
my_sleep (unsigned int seconds)
{
  do
    {
      seconds = sleep (seconds);
    }
  while (seconds);
}

static
int
gm_board_mapping (int argc, char *argv[])
{
  const size_t board_span = 16 * 1024 * 1024;
  gm_size_t control_regs_len;
  gm_size_t dma_region_len;
  gm_size_t special_regs_len;
  gm_size_t sram_len;
  gm_status_t status;
  gm_u32_n_t *control_regs;
  gm_lanai_special_registers_t *special_regs;
  gm_u64_t *dma_region;
  gm_u32_n_t *sram;
  int fd;
  void *board = 0;

  status = gm_init ();
  if (status != GM_SUCCESS)
    {
      goto abort_with_nothing;
    }
  
  /* Open the board */
  
  fd = open ("/dev/gmp0", O_RDWR);
  if (fd == -1)
    {
      perror ("could not open file");
      goto abort_with_init;
    }

  /* Memory map the entire board */
  
  dma_region_len = 64 * 1024;
  board = mmap (0, board_span + dma_region_len,
		PROT_READ | PROT_WRITE, MAP_SHARED, fd, 0);
  if (board == MAP_FAILED)
    {
      perror ("could not mmap()");
      status = GM_FAILURE;
      goto abort_with_open_port;
    }

  /* Generate pointers to useful parts of the board */
  
  sram = (gm_u32_n_t *)board;
  sram_len = 0x100000;		/* fake 1MB SRAM */
  control_regs = (gm_u32_n_t *)((char *)board + 0x800000);
  control_regs_len = 0x100;
  special_regs = (gm_lanai_special_registers_t *)((char *)board + 0xc00000);
  special_regs_len = sizeof (gm_lanai_special_registers_t);
  dma_region = (gm_u64_t *) ((char *) board + board_span);
  
  printf ("Mapped 0x%lx bytes of control registers\n"
	  "Mapped 0x%lx bytes of special registers\n"
	  "Mapped 0x%lx bytes of SRAM\n",
	  (unsigned long) control_regs_len,
	  (unsigned long) special_regs_len,
	  (unsigned long) sram_len);
  if (dma_region_len)
    {
      printf ("DMA region UVMA=%p, DMAA=0x%qx\n", dma_region, dma_region[0]);
    }
  
  /* Make sure the board is in reset. */

  printf ("Resetting board.\n");
  printf ("control_regs[16] = 0x%x\n", gm_ntoh_u32 (control_regs[16]));
  /* lanai reset on */
  control_regs[0x10] = gm_hton_u32 (1 << GM_LX (31));
  GM_STBAR ();
  printf ("control_regs[16] = 0x%x\n", gm_ntoh_u32 (control_regs[16]));
  /* mem reset on */
  control_regs[0x10] = gm_hton_u32 (1 << GM_LX (29));
  GM_STBAR ();
  printf ("control_regs[16] = 0x%x\n", gm_ntoh_u32 (control_regs[16]));
  my_sleep (1);

  /* Take the board out of mem reset so we can play with the specials */

  printf ("Taking board out of reset, except for LANai.\n");
  /* mem reset off */
  control_regs[16] = gm_hton_u32 (1 << GM_LX (28));
  GM_STBAR ();
  my_sleep (1);

  /* Flash the LED for a while by writing to the the LED via PIO.  */

  printf ("Flashing LED with PIO\n");
  {
    int i;
    for (i=1; i<=30; i++)
      {
	special_regs->write.lX.LED = gm_hton_u8 (i);
	GM_STBAR ();
	printf ("%d 0x%x\n", i, gm_ntoh_s32 (special_regs->read.lX.RTC));
	my_sleep (1);
      }
  }

  /* Load a simple control program to flash the LED faster and to
     copy the RTC to the bottom word of SRAM. */

  printf ("Loading program\n");

  sram[ 0] = gm_hton_u32 (0x4704fd30); /* nop */
  sram[ 1] = gm_hton_u32 (0x4684fd50); /* nop */
  sram[ 2] = gm_hton_u32 (0x87ba0000); /* ld [ fffffe68 <*ABS*+0xfffffe68> ] , %r15 */
  sram[ 3] = gm_hton_u32 (0x00000001); /* nop */
  sram[ 4] = gm_hton_u32 (0x00000001); /* nop */
  sram[ 5] = gm_hton_u32 (0xf7810000); /* st %r15 , [ 0 <*ABS*> ]	! 0 <*ABS*> */
  sram[ 6] = gm_hton_u32 (0x77bcffee); /* sh %r15 , ffffffee <*ABS*+0xffffffee> , %r15 */
  sram[ 7] = gm_hton_u32 (0xf7810004); /* st %r15 , [ 4 <*ABS*+0x4> ]	! 4 <*ABS*+0x4> */
  sram[ 8] = gm_hton_u32 (0xf7b76800); /* st %r15 , [ ffffff40 <*ABS*+0xffffff40> ] */
  sram[ 9] = gm_hton_u32 (0xe0000008); /* bt 8 <*ABS*+0x8>	! 8 <*ABS*+0x8> */
  sram[10] = gm_hton_u32 (0x00000001); /* nop */
  sram[11] = gm_hton_u32 (0x00000001); /* nop */
  GM_STBAR ();

  /* Dump the portion of SRAM holding the program. */
  
  gm_hex_dump (sram, 11*4);

  /* Take the LANai out of reset. */

  printf ("Running program\n");
  printf ("control_regs[16] = 0x%x\n", gm_ntoh_u32 (control_regs[16]));
  /* lanai reset off */
  control_regs[16] = gm_hton_u32 (1 << GM_LX (30));
  GM_STBAR ();
  printf ("control_regs[16] = 0x%x\n", gm_ntoh_u32 (control_regs[16]));

  /* Let the program run a while */

  {
    int i;
    
    for (i=0; i<15; i++)
      {
	my_sleep (1);
	printf ("LANai SRAM copy of RTC = 0x%08x\n", gm_ntoh_u32 (sram[0]));
      }
  }

  /* Put the board back in reset */

  printf ("Resetting board\n");
  /* lanai reset on */
  control_regs[16] = gm_hton_u32 (1 << GM_LX (31));
  GM_STBAR ();
  /* mem reset on */
  control_regs[16] = gm_hton_u32 (1 << GM_LX (29));
  GM_STBAR ();

  /* cleanup and exit */

  munmap (board, board_span);
 abort_with_open_port:
  close (fd);
 abort_with_init:
  gm_finalize ();
 abort_with_nothing:
  gm_exit (status);
 return 0;
}

/****************************************************************
 * Entry point
 ****************************************************************/

int
main (int argc, char *argv[])
{
  return gm_board_mapping (argc, argv);
}


/*
  This file uses GM standard indentation:

  Local Variables:
  c-file-style:"gnu"
  tab-width:8
  c-backslash-column:72
  End:
*/
