#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/*
 * Simple Send/Recv for GM
 * Author: Cam Macdonell - University of Alberta 
 * */


#include "gm.h"
#include <gm_example.h>

void my_send_callback (struct gm_port *port, void *context, gm_status_t status);

int min_buf_size;
int todo;

int
main(int argc, char **argv)
{
	static struct gm_port *p = NULL;
	static void           *in_buffer = NULL,
		                  *out_buffer = NULL,
		                  *remote_buffer;
	static int            recv_sz;
	char                  dest_node_name[64];
	int                   dest_node_id;
	gm_recv_event_t       *e;

	todo=2;
	
	if (argc!=2){
		printf("USAGE: simple_recv <dest mach name>\n");
		exit(-1);
	}
	
	if (gm_open(&p, GM_EXAMPLE_BOARD_RECEIVER, 
				GM_EXAMPLE_PEER_PORT, "gm_basic_p",
				(enum gm_api_version) GM_API_VERSION)
		!= GM_SUCCESS)
	{
		printf("[recv] Couldn't open board %d port %d\n",
			   GM_EXAMPLE_BOARD_RECEIVER, GM_EXAMPLE_PEER_PORT);
		exit(-1);
	}
	else
		printf("[recv] Opened board %d port %d\n",
			   GM_EXAMPLE_BOARD_RECEIVER, GM_EXAMPLE_PEER_PORT);

	strcpy(dest_node_name,argv[1]);	
	dest_node_id=gm_host_name_to_node_id(p,dest_node_name);
	printf("[recv] Destination node name %s, node id %d\n",
		   dest_node_name, dest_node_id);

	if ((out_buffer = gm_dma_calloc(p, 1, 64)) != 0) {
		printf("[recv] out_buffer allocated at %p\n\n", out_buffer);
	}
	else {
		printf("[recv] Couldn't allocate out_buffer\n");
    		gm_close(p);
		gm_exit(GM_FAILURE);
	}
	if ((in_buffer = gm_dma_calloc(p, 1, 64)) != 0) {
		printf("[recv] in_buffer allocated at %p\n\n", in_buffer);
	}
	if ((remote_buffer = gm_dma_calloc(p, 1, 64)) != 0) {
		printf("[recv] remote_buffer allocated at %p\n\n", remote_buffer);
	}
	
	gm_allow_remote_memory_access(p);

	sprintf(out_buffer,"%p",remote_buffer);

	printf("[recv] out_buffer contains \"%s\"\n",out_buffer);

	/* send the remote address buffer */
	gm_send_to_peer_with_callback(p, out_buffer, 7,
		(unsigned long) strlen(out_buffer)+1,
		GM_EXAMPLE_PRIORITY, dest_node_id, my_send_callback,NULL);
	
	gm_provide_receive_buffer(p, in_buffer, 7, GM_EXAMPLE_PRIORITY);
	gm_provide_receive_buffer(p, remote_buffer, 7, GM_EXAMPLE_PRIORITY);

	sprintf(remote_buffer,"lots a filler, don't need this");

	printf("[recv] remote_buffer contains \"%s\"\n",remote_buffer);
	
	while (todo) {
		e = gm_receive(p);
		
		switch (gm_ntoh_u8 (e->recv.type)) {
		 case GM_RECV_EVENT:
		 case GM_PEER_RECV_EVENT:	 
		 case GM_FAST_PEER_RECV_EVENT:   
			printf("[recv] Receive Event\n");
			printf("[recv] Received: \"%s\"\n",
				(char *) gm_ntohp (e->recv.message));
			recv_sz=(int)gm_ntoh_u8(e->recv.size);
			printf("[recv] Receive Size is %d\n",recv_sz);
			todo--; 
			break;

		 case GM_NO_RECV_EVENT:
			 break;

		 default:
			 printf("[recv] got GM unknown\n");
			 gm_unknown(p, e);
		}

	}

	printf("[recv] internal buffer = %s\n",remote_buffer);
	
	gm_close(p);

    	return(0);
}

void my_send_callback (struct gm_port *port, void *context, gm_status_t status)
{

  printf("[recv] Send callback entry = %d\n",todo--);

  if (status != GM_SUCCESS)
    {
      if (status != GM_SEND_DROPPED)
	{
	  gm_perror ("send completed with error", status);
	}
    }
}

/*
  This file uses feldy indentation:

  Local Variables:
  tab-width:4
  c-file-style:"bsd"
  End:
*/
