#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/*
 * Simple Send/Recv for GM
 * Author: Cam Macdonell - University of Alberta
 * Slightly modified by Myricom
 * */


#include "gm.h"
#include <gm_example.h>

void my_send_callback (struct gm_port *port, void *context, gm_status_t status);

int todo;
int min_buf_size;

int
main(int argc, char **argv)
{
	static struct gm_port *p             = NULL;
	static void           *out_buffer    = NULL,
		                  *in_buffer     = NULL,
		                  *remote_buffer = NULL;
	char                  dest_node_name[64],
		                  temp[64];
	int                   dest_node_id,
		                  bufsize=65536;
	gm_remote_ptr_t       address;
	unsigned int          my_node;
	gm_recv_event_t       *e;
	
	
	if (argc!=2){
		printf("USAGE: simple_send <dest mach name>\n");
		exit(-1);
	}
	
	
	if (gm_open(&p, GM_EXAMPLE_BOARD_SENDER, 
				GM_EXAMPLE_PEER_PORT, "gm_basic_p1",
				(enum gm_api_version) GM_API_VERSION)
		!= GM_SUCCESS)
	{
		printf("[send] Couldn't open board %d port %d\n",
			   GM_EXAMPLE_BOARD_SENDER, GM_EXAMPLE_PEER_PORT);
		exit(-1);
	}
	else
		printf("[recv] Opened board %d port %d\n",
			   GM_EXAMPLE_BOARD_SENDER, GM_EXAMPLE_PEER_PORT);

	todo=1;

	strcpy(dest_node_name,argv[1]);	
	dest_node_id=gm_host_name_to_node_id(p,dest_node_name);
	printf("[send] Destination node name %s, node id %d\n",
		   dest_node_name, dest_node_id);
	gm_get_node_id(p,&my_node);
	printf("[send] My node id %d\n", my_node);

	printf("[send] Buf Size = %d, Min Length = %d\n",bufsize,min_buf_size);
	if ((out_buffer = gm_dma_calloc(p, 1, 64)) == 0) {
		printf("[send] Couldn't allocate out_buffer\n");
	}
	if ((in_buffer = gm_dma_calloc(p, 1, 64)) == 0) {
		printf("[send] Can't allocate in_buffer\n");
	}
	if ((remote_buffer = gm_dma_calloc(p, 1, 64)) == 0) {
		printf("[send] Can't allocate remote_buffer\n");
	}
	gm_free_send_tokens(p, GM_EXAMPLE_PRIORITY, gm_num_send_tokens(p));

	gm_provide_receive_buffer(p,in_buffer,7,GM_EXAMPLE_PRIORITY);
	
	while (todo) {
		e = gm_receive(p);

		switch (gm_ntoh_u8 (e->recv.type)) {
		 case GM_RECV_EVENT:
		 case GM_PEER_RECV_EVENT:
		 case GM_FAST_PEER_RECV_EVENT:   
			 printf("[send] Receive Event\n");
			 strcpy(temp,(char *) gm_ntohp (e->recv.message));
			 printf("[send] Received: \"%s\"\n",temp);
			 printf("[send] Address is 0x%lx\n",address);
			 todo=0; 
			 break;

		 case GM_NO_RECV_EVENT:
			 break;

		 default:
			 gm_unknown(p, e);	/* gm_unknown calls the callback function, I think*/
		}
	}

	sprintf(remote_buffer,"this was sent directly");

	printf("[send] remote_buffer contains \"%s\"\n",remote_buffer);

	gm_directed_send_with_callback(p, remote_buffer, 
		(gm_remote_ptr_t)address,
		(unsigned long) strlen(remote_buffer)+1,
		GM_EXAMPLE_PRIORITY, dest_node_id, 2, my_send_callback,NULL);
	
	sprintf(out_buffer,"this is the sender!!");

	printf("[send] out_buffer contains \"%s\"\n",out_buffer);
	gm_send_to_peer_with_callback(p, out_buffer, 7,
		(unsigned long) strlen(out_buffer)+1,
		GM_EXAMPLE_PRIORITY, dest_node_id, my_send_callback,NULL);

	todo=1;
	while (todo) {
		e = gm_receive(p);

		switch (gm_ntoh_u8 (e->recv.type)) {
		 case GM_RECV_EVENT:
		 case GM_PEER_RECV_EVENT:
		 case GM_FAST_PEER_RECV_EVENT:   
			 printf("[send] Receive Event\n");
			 break;
		 case GM_NO_RECV_EVENT:
			 break;

		 default:
			 gm_unknown(p, e);	/* gm_unknown calls the callback function, I think*/
		}
	}
		
    gm_close(p);
    gm_exit(GM_SUCCESS);
    return(0);
}

void my_send_callback (struct gm_port *port, void *context, gm_status_t status)
{

  printf("[send] Send callback entry = %d\n",todo--);

  if (status != GM_SUCCESS)
  {
      if (status != GM_SEND_DROPPED)
	  {
		  gm_perror ("send completed with error", status);
	  }
  }
}

/*
  This file uses feldy indentation:

  Local Variables:
  tab-width:4
  c-file-style:"bsd"
  End:
*/
