/******************************************************************-*-c-*-
 * Myricom GM networking software and documentation                      *
 * Copyright (c) 1996, 1997 by Myricom, Inc.                             *
 * All rights reserved.  See the file `COPYING' for copyright notice.    *
 *************************************************************************/

/* NOTE: If you change the operating characteristics or the flags
 *       used by this program, please notify Max so he can change the
 *       Myrinet Handbook document.
 */

#if defined __STDC__ || defined WIN32
#if GM_OS_VXWORKS
#include "vxWorks.h"
#endif  /* GM_OS_VXWOKRS */
#include <stdarg.h>
#else
#include <varargs.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>

#include "gm.h"

#include <time.h>
#if (!defined WIN32) && (defined HAVE_SYS_TIME_H)
#include <sys/time.h>
#endif

				/* sleep() */
#ifdef HAVE_UNISTD_H
#  include <unistd.h>
#endif

/*
 * quiet is: 0 for verbose, 1 for quiet, and 2 for silent (except for errors)
 */
#define TEST_PRINT(level, args)  do { \
  if (level>=quiet) printf args; \
} while (0)

/****************************************************************
 * typedefs
 ****************************************************************/

enum test_message_types
{
  test_bogus_type = 0,
  test_ping_type,
  test_forward_type
};

typedef struct test_message
{
  gm_u32_n_t seq_no;
  gm_u32_n_t type;

  gm_u32_n_t size;
  gm_u32_n_t sender;

  gm_u8_n_t payload[1];

}
test_message_t;

/****************************************************************
 * Macros, test parameter defaults, etc.
 ****************************************************************/

#define DEBUG_NWAY 0
#define ABORT_NWAY 0

#define NWAY_ERR_TOLERANCE 30

#define MAX_NODES 2048

#define GM_NWAY_DELAY 5

#define MIN_TEST_SIZE	5
#define MAX_TEST_SIZE	14

#define ROUNDUP(n,m) (((gm_u32_t)(n)+(gm_u32_t)(m)-1)&~((gm_u32_t)(m)-1))
#define DMA_ROUNDUP(p) ROUNDUP (p, GM_DMA_GRANULARITY)

#define TIME_CALIBRATION 1e6


/****************************************************************
 * Global variables
 ****************************************************************/

/* Global variables set by command-line options  */

/* Device Control and Selection:                 */
int      board_num = 0;		/* Our board number                         */
unsigned port_id_g = 2;		/* Port number for all participants         */
char     *node_name[MAX_NODES];	/* node1, node2, ...                        */
int      num_nodes = -1;

/* Test Parameters:                              */
int           cycle = 0;	/* True ==> cycle through nodes in sequence */
int           neighbor= 0;	/* True ==> send only to next neighbor in the list */
int           fast = 0;		/* True ==> "fast mode"                     */
unsigned int  max_test_size = MAX_TEST_SIZE,
              min_test_size = MIN_TEST_SIZE;
unsigned int  min_test_len = 0;	/* 0 ==> don't apply a minimum              */
int           verify = 0;	/* True ==> verify contents of messages     */

/* Looping and Time-stamping:                    */
long  iterate_limt = 0;		/* 0 ==> loop forever                       */
int   iterate_time = GM_NWAY_DELAY; /* Interval between alarms              */

/* Information Display:                          */

int   nodomain = 0;     	/* True ==> Strip domain from nodename      */
int   quiet = 0;		/* Print level for TEST_PRINT macro         */
int   timestamps = 0;		/* True ==> Timestamp the output            */


/* Other control variables */

float PROB_FORWARD = 0.5;
float PROB_SLEEP = (float) 0.000005;


int   node_id[MAX_NODES];
int   in_seq[MAX_NODES],	/* Sequence checking, only if (verify) */
      out_seq[MAX_NODES];
int   in_fw_seq[MAX_NODES],	/* For message-filling */
      out_fw_seq[MAX_NODES];
int   errs_by_node[MAX_NODES];

struct gm_port    *port_g = NULL;
static gm_alarm_t my_alarm;

int           test_underway = 0;

unsigned int  target_index = 0;	/* Array index for current send target      */
unsigned int  this_node_id;	/* Our GM Node ID                           */
unsigned int  this_node_index;	/* Our index in the node table              */
int           slp = 0;		/* True ==> Sleep before next send          */
int           caught_sigint = 0; /* True ==> User hit ^C, probably          */
static int    got_alarm;	/* True ==> We fielded the gm_alarm         */

unsigned long iterations = 0;	/* Actual interations tally                 */
unsigned long gm_send_tally = 0, /* For making sure we consume all the      */
              send_cb_tally = 0, /*   send callbacks                        */
              forward_tally = 0,
              fwrd_cb_tally = 0;
unsigned long gm_unkn_tally = 0; /* Times we called gm_unknown()            */
#if DEBUG_NWAY
unsigned int  max_len_sent = 0,
              min_len_sent = 0,
              max_len_recd = 0,
              min_len_recd = 0;
#endif


/* Statistics              */

gm_s64_t total_sent      = 0,
         total_recvd     = 0;
gm_s64_t bytes_sent      = 0,
         bytes_recvd     = 0;
gm_s64_t old_bytes_sent  = 0,
         old_bytes_recvd = 0;
gm_s64_t low_recvd       = 0,
         high_recvd      = 0;
gm_s64_t sleeps          = 0;

gm_s64_t total_senta[MAX_NODES];
gm_s64_t total_recvda[MAX_NODES + 1];	/* last location is for error */

#ifdef PRINT_LOTS
gm_s64_t bytes_senta[MAX_NODES];
gm_s64_t bytes_recvda[MAX_NODES + 1];
gm_s64_t old_bytes_senta[MAX_NODES];
gm_s64_t old_bytes_recvda[MAX_NODES];
#endif



/****************************************************************
 * Command line processing
 ****************************************************************/

#define TP(a) fprintf (stderr, a)

static void
print_synopsis(void)
{
  TP ("\n");
  TP ("Myricom 'gm_nway' facilitates setting up message passing tests\n");
  TP ("among arbitrary numbers of GM nodes.  Multiple test variations are\n");
  TP ("provided via the command-line options.\n");
}


static void
print_options (void)
{
  TP ("\nusage: gm_nway [-p port] [-B board] [-N #] <other flags> "
      "node1 node2 ...\n\n");
  TP ("       The gm_nway command should be run with the same set of nodes"
      " on\n");
  TP ("       each participating host for each participating node."
      "  For example,\n");
  TP ("       to use nodes host1, host1:1, and host2, run the following"
      " commands:\n");
  TP ("         On host host1:\n");
  TP ("            gm_nway host1 host1:1 host2\n");
  TP ("            gm_nway --board=1 host1 host1:1 host2\n");
  TP ("         On host host2:\n");
  TP ("            gm_nway host1 host1:1 host2\n\n");

  TP ("\nDevice Control and Selection:\n");
  TP ("-B, --board=NUM  [0]  board number for this process to use\n");
  TP ("-p, --port=NUM   [2]  port number to use for all participants\n");
  TP ("node1, node2, ...     the complete list of participating nodes\n");

  TP ("\nTest Parameters:\n");
  TP ("    --cycle          [off]  choose message destinations in sequence\n"
      "                            rather than randomly\n");
  TP ("-f, --fast           [off]  use faster algorithms\n");
  fprintf (stderr,
      "    --max-size=NUM   [%d]   maximum size to test\n", MAX_TEST_SIZE);
  fprintf (stderr,
      "    --min-length=NUM [%d]   minimum length to test\n", DMA_ROUNDUP(16));
  fprintf (stderr,
      "    --min-size=NUM   [%d]    minimum size to test\n", MIN_TEST_SIZE);
  TP ("    --neighbor       [off]  send only to adjacent higher node in sequence\n"
      "                            rather than randomly\n");
  TP ("-s, --size=NUM       [none] alias for --min-size=NUM --max-size=NUM\n");
  TP ("-v, --verify         [off]  use message verification algorithms (slow)\n");

  TP ("\nLooping and Time-stamping:\n");
  TP ("-N, --iterations=NUM     [off] stop after NUM interations\n");
  fprintf (stderr,
      "    --iteration-time=NUM [%d]   duration of each iteration in seconds\n"
	   , GM_NWAY_DELAY);

  TP ("\nInformation Display:\n");
  TP ("    --no-domain-names [off] suppress printing of node domain-name suffixes\n");
  TP ("    --quiet           [off] omit some information from output display\n");
  TP ("    --silent          [off] omit most information from output display\n");
  TP ("    --timestamp       [off] display time of day with each iteration's data\n");
  
  TP ("\nInformative Options:\n");
  TP ("    --help        output usage and options information, then exit\n");
  TP ("\n");
  fprintf (stderr, 
      "  GM sizes must be from %d to 31.\n",
	   gm_min_message_size ((struct gm_port *)0));
  TP ("\n");

#undef TP
}


static void
#ifdef va_dcl
usage (fmt, va_alist)
     const char *fmt;
     va_dcl
#else
usage (const char *fmt, ...)
#endif
{
  va_list ap;

  print_options();

/* Now print the caller's diagnostic and exit */
#if defined __STDC__ || defined WIN32
  va_start (ap, fmt);
#else
  va_start (ap);
#endif
  vfprintf (stderr, fmt, ap);
  va_end (ap);

  if (port_g)
    {
      gm_close (port_g);
      port_g = NULL;
    }
  gm_exit (GM_INVALID_PARAMETER);

} /* usage() */


static void
parse_args (int argc, char *argv[])
{
/* NOTE: If you change the operating characteristics or the flags
 *       used by this program, please notify Max so he can change the
 *       Myrinet Handbook document.
 */
  int i;

  /* The 'F' macro is for all flags with no argument (parameter)     */
#define F(name,action) if (strcmp (argv[i], name) == 0) { action; continue; }

  for (i = 1; i < argc; i++)
  {
    char dummy_string[256];

    strcpy (dummy_string, "");

    /* Dispose of the --help flag first                                */
    if ((strcmp (argv[i], "--help") == 0) ||
	(strcmp (argv[i], "-h")     == 0)   )             /* "-h" Deprecated */
    {
      print_synopsis();
      print_options();
      gm_exit (GM_SUCCESS);
    }

    /* Here are the flags with no argument.                                  */
    /* Please keep short and/or deprecated forms, if any, after their long-  */
    /* form equivalent, but keep the long form options in alpha sequence.    */

    F ("--cycle"          , cycle      = 1);
    F ("--neighbor"       , neighbor   = 1);
    F ("--no-domain-names", nodomain   = 1);
    F ("--nodomain"       , nodomain   = 1);                   /* Deprecated */
    F ("--fast"           , fast       = 1);
    F ("-f"               , fast       = 1);
    F ("--quiet"          , quiet      = 1);
    F ("--silent"         , quiet      = 2);
    if (strcmp (argv[i], "-q") == 0)                           /* Deprecated */
    {
      if (++quiet > 2)
	quiet = 2;
      continue;
    }
    F ("--timestamp"      , timestamps = 1);
    F ("-t"               , timestamps = 1);                   /* Deprecated */
    F ("--verify"         , verify     = 1);
    F ("-v"               , verify     = 1);
#undef F

    /* The check_opt macro is for long-form options with numeric        */
    /* arguments and simple validation tests.                           */
#define check_opt(prefix, var, fmt, test)			\
    if (sscanf (argv[i], "--" prefix "=%" fmt, &var) == 1)	\
    {								\
      if (!(test))						\
      {								\
	usage ("bad value for option \"%s\"\n", argv[i]);	\
      }								\
      continue;							\
    }

    /* Here are the flags with arguments, sequenced as above */

    check_opt ("board", board_num, "i", board_num >= 0);
    if (strcmp (argv[i], "-B") == 0)
      {
	if (i + 1 >= argc)
	  usage ("Board number expected after '-B'\n");

	board_num = atoi (argv[++i]);
	if (board_num < 0)
	  usage ("bad board number: %d\n", board_num);

	continue;
      }

    check_opt ("iteration-time", iterate_time, "i", iterate_time >= 0);

    check_opt ("iterations", iterate_limt, "li", iterate_limt >= 0);
    if (strcmp (argv[i], "-N") == 0)
      {
	if (i + 1 >= argc)
	  usage ("Iteration count expected after '-N'\n");

	iterate_limt = atoi (argv[++i]);
	if (iterate_limt < 0)
	  usage ("bad iteration count: %d\n", iterate_limt);

	continue;
      }

    check_opt ("max-size", max_test_size, "i", max_test_size < 32);
    if (strcmp (argv[i], "--max-size") == 0)                   /* Deprecated */
      {
	if (i + 1 >= argc)
	  usage ("Size value expected after '--max-size'\n");

	max_test_size = atoi (argv[++i]);
	if (max_test_size > 31)
	  usage ("Bad max-size value: %d\n", max_test_size);

	continue;
      }

    check_opt ("min-length", min_test_len, "i", min_test_len > 0);
    if (strcmp (argv[i], "--min-length") == 0)                 /* Deprecated */
      {
	if (i + 1 >= argc)
	  usage ("Size value expected after '--min-length'\n");

	min_test_len = atoi (argv[++i]);
	if (min_test_size < 1)
	  usage ("Bad min-size value: %d\n", min_test_size);

	continue;
      }

    /* Note that the parameter to gm_min_message_size() is a dummy */
    check_opt ("min-size", min_test_size, "i",
	       min_test_size >= gm_min_message_size ((struct gm_port *)0));
    if (strcmp (argv[i], "--min-size") == 0)                   /* Deprecated */
      {
	if (i + 1 >= argc)
	  usage ("Size value expected after '--min-size'\n");

	min_test_size = atoi (argv[++i]);
	if ((min_test_size < gm_min_message_size ((struct gm_port *)0)) ||
	    (min_test_size > 31))
	  usage ("Bad min-size value: %d\n", min_test_size);

	continue;
      }

    /* Note that the parameter to gm_num_ports() is a dummy */
    /* Note that port_id_g is unsigned ... no compare to 0  */
    check_opt ("port", port_id_g, "i",
	       (port_id_g < gm_num_ports ((struct gm_port *)0)));
    if (strcmp (argv[i], "-p") == 0)
      {
	if (i + 1 >= argc)
	  usage ("Port number expected after '-p'.\n");

	port_id_g = atoi (argv[++i]);
	if (port_id_g >= gm_num_ports ((struct gm_port *)0))
	  usage ("Bad port number: %d\n", port_id_g);

	continue;
      }

    if (sscanf (argv[i], "--size=%i", &min_test_size) == 1)
      {
	max_test_size = min_test_size;
	/* Note that the parameter to gm_min_message_size() is a dummy */
	if (!(min_test_size >= gm_min_message_size ((struct gm_port *)0)
	      && max_test_size < 32))
	  usage ("Bad value for option \"%s\"\n", argv[i]);

	continue;
      }
    if ((strcmp (argv[i], "-s") == 0) ||
        (strcmp (argv[i], "--size") == 0))                     /* Deprecated */
      {
	if (i + 1 >= argc)
	  usage ("size expected after '%s'.\n", argv[i]);

	min_test_size = max_test_size = atoi (argv[++i]);
	/* Note that the parameter to gm_min_message_size() is a dummy */
	if ((min_test_size < gm_min_message_size ((struct gm_port *)0) ||
	     max_test_size > 31))
	  usage ("bad value for size: %d", min_test_size);

	continue;
      }
#undef check_opt

    /* skip over any arg beginning with '-' */
    if (sscanf (argv[i], "-%s", dummy_string) == 1)
      usage ("unknown option \"%s\"\n", argv[i]);

    break;			/* Unknown arg not beginning with '-' */
  }

  if (i >= argc)
    usage ("No node names specified\n");

  /* Assume the remaining args are node names */

  num_nodes = 0;
  for (; i < argc; i++)
    {
      if (num_nodes >= MAX_NODES)
	{
	  TEST_PRINT (2, ("too many nodes; only first %d will be used\n",
			  MAX_NODES));
	  break;
	}

      node_name[num_nodes] = argv[i];

      num_nodes++;
    }

} /* parse_args() */



static void
cross_check_args (void)
{

  if (gm_max_length_for_size(min_test_size) < 16)
    usage ("Min size %ld is too small for gm_nway on this system; "
	   "try a larger --min-size\n", min_test_size);

  if (max_test_size > 31)
    usage ("Max size %ld is greater than GM limit of 31\n", max_test_size);

  if (min_test_size > max_test_size)
    usage ("Min size is greater than max size (%ld > %ld)\n",
	   min_test_size, max_test_size);

  if (min_test_len != 0)
    {
      if (min_test_len < DMA_ROUNDUP(16))
	usage ("Min length is less than nway minimum (%d < %d)\n",
	       min_test_len, DMA_ROUNDUP(16));

      if (min_test_len > gm_max_length_for_size(max_test_size))
	usage ("Min length is too great for max size (%d > %ld)\n",
	       min_test_len, gm_max_length_for_size(max_test_size));

      if (min_test_len > gm_max_length_for_size(min_test_size))
	  usage ("Min length is too great for min size (%d > %ld)\n"
		 "... some messages will be smaller than %d\n",
		 min_test_len, gm_max_length_for_size(min_test_size),
		 min_test_len);
    }

  if (fast)
    {
      /* don't forward, don't sleep */
      PROB_FORWARD = 0.0;
      PROB_SLEEP = 0.0;
    }

} /* cross_check_args() */



static void
cross_check_nodes(void)
{
  int i;


  /* Try to convert the node names to GM IDs. */

  for (i=0; i < num_nodes; i++)
    {
      unsigned int id;

      id = gm_host_name_to_node_id (port_g, node_name[i]);
      if (id == GM_NO_SUCH_NODE_ID)
	usage ("Could not translate node name \"%s\" to a GM ID\n", 
	       node_name[i]);

      if (nodomain)
	{
	  /* if nodomain, then null terminate at the 1st '.' */
	  char *ptr = 0;

	  ptr = strchr(node_name[i], '.');
	  if (ptr)
	    ptr[0] = 0;
	}
      node_id[i] = id;
    }
}



static void
announce_run_parameters()
{
  int i;

  /*
   * let the user know what the parameters are
   */
  TEST_PRINT (2, ("Interacting with nodes:\n"));
  for (i = 0; i < num_nodes; i++)
    TEST_PRINT (2, ("\t\"%s\" (GM ID %d)\n", node_name[i], node_id[i]));

  TEST_PRINT (1, ("Using port %d.\n", port_id_g));
  if (fast)
    TEST_PRINT (1, ("'fast' mode - sending as fast as possible\n"));

  if (verify)
    TEST_PRINT (1, ("Verififying sequence and contents of messages.\n"));
  else
    TEST_PRINT (1, ("Not touching contents of messages.\n"));

  if (iterate_limt)
    TEST_PRINT (1, ("Test will terminate after %ld iterations.\n",
                    iterate_limt));
}



/****************************************************************
 * Service functions
 ****************************************************************/

static gm_inline int
payload_len (int len)
{
  return len - GM_OFFSETOF (test_message_t, payload);
}


static void
die (void)
{
  if (port_g)
    {
      gm_close (port_g);
      port_g = NULL;
    }
  if (ABORT_NWAY)
    {
      TEST_PRINT (2, ("Calling gm_abort()\n"));
      gm_abort();
    }
  else
    {
      TEST_PRINT (2, ("Calling gm_exit(GM_FAILURE)\n"));
      gm_exit (GM_FAILURE);
    }
}


static void
tolerate_errors(int the_err_cnt, int the_node)
{
  if (the_err_cnt > NWAY_ERR_TOLERANCE)
    {
      TEST_PRINT (2, ("%d verify errors from node %d\n",
		      the_err_cnt, the_node));
      die ();
    }
}


static int
get_local_ix (int gm_id)
{
  int j;
  /* if no match is found, the count will go into a safe place */
  int val = MAX_NODES;

  for (j = 0; j < num_nodes; j++)
    {
      if (node_id[j] == gm_id)
	{
	  val = j;
	  break;
	}
    }
  return (val);
}


static void
set_got_alarm (void *ignored)
{
  GM_PARAMETER_MAY_BE_UNUSED (ignored);
  
  got_alarm = 1;
}

/****************************************************************
 * DMAable Buffer allocation
 ****************************************************************/

/* The type of message we will
   be sending and receiving. */

typedef struct chainable_test_message
{
  /* Use the 8 bytes at the head of the
     message for user data that will not
     be sent over the network. */
  struct chainable_test_message *next;
  gm_u32_t size;
  test_message_t message;

}
chainable_test_message_t;

chainable_test_message_t *free_send_messages[32];

static gm_inline chainable_test_message_t *
ctm_parent (test_message_t * p)
{
  return ((chainable_test_message_t *)
	  ((char *) p - GM_OFFSETOF (chainable_test_message_t, message)));

}

static void
verify_send_buffer_size (void *buffer, unsigned int size)
{
  gm_assert (ctm_parent (buffer)->size == size);
}

/*
 * add a buffer to the (user managed) list of free send buffers
 */
static gm_inline void
free_a_send_message (test_message_t * p)
{
  chainable_test_message_t *ctm = ctm_parent (p);
  gm_u32_t p__size;

  p__size = gm_ntoh_u32 (p->size);
  gm_assert (ctm->size == p__size);

  ctm->next = free_send_messages[p__size];
  free_send_messages[p__size] = ctm;
}

static gm_inline test_message_t *
alloc_send_message (unsigned int size)
{
  chainable_test_message_t *ctm;
  ctm = free_send_messages[size];
  if (!ctm)
    return 0;
  free_send_messages[size] = ctm->next;
  return &ctm->message;
}

static gm_status_t
init_send_buffer_allocator (struct gm_port *port,
			    int            min_size,
			    int            max_size,
			    int            tokens_in_use)
{
  unsigned int send_buffers_per_size;
  unsigned int ui,
               size,
               send_buffer_tally = 0;


  send_buffers_per_size = tokens_in_use / (max_size - min_size + 1);
  if (send_buffers_per_size > 8)
    send_buffers_per_size = 8;

  for (size = min_size; size <= max_size; size++)
    {
#if DEBUG_NWAY
      TEST_PRINT(1, ("Send buffers for size %d:\n", size));
#endif
      /* allocate send buffers */
      for (ui = 0; ui < send_buffers_per_size; ui++)
	{
	  chainable_test_message_t *ctm;

	  ctm = ((chainable_test_message_t *)
		 gm_dma_calloc (port, 1,
				GM_OFFSETOF (chainable_test_message_t,
					     message) +
				gm_max_length_for_size (size)));
	  if (!ctm)
	    return GM_OUT_OF_MEMORY;

	  gm_always_assert (GM_DMA_ALIGNED (ctm));

	  ctm->next = 0;
	  ctm->size = size;
	  ctm->message.size = gm_hton_u32 (size);
	  ctm->message.type = gm_hton_u32 (-1);

#if GM_DEBUG_BUFFERS
	  gm_register_buf (&ctm->message, size);
#endif
#if DEBUG_NWAY
	  TEST_PRINT(1, ("  Buffer %d at %p\n", ui, &ctm->message));
#endif
	  free_a_send_message (&ctm->message);
	  send_buffer_tally++;
	}
    }
  TEST_PRINT (1, ("Allocated %d total send buffers\n", send_buffer_tally));
  return GM_SUCCESS;
}


static void
recycle_receive_buffer (struct gm_port *the_port,
			test_message_t *the_buffer,
			unsigned int   the_size,
			unsigned int   the_prio)
{
  gm_always_assert (the_size > 0);
  if (DEBUG_NWAY)
    memset (the_buffer, 0xee, gm_max_length_for_size(the_size));
  gm_provide_receive_buffer (the_port, the_buffer, the_size, the_prio);
}



/****************************************************************
 * 
 ****************************************************************/

#if 0
#error not updated
/*
 * simple buffer fill and verify routines
 */
static gm_inline void
fill_buffer (char *buf, int len, char seed)
{
  int i;

  buf[0] = seed;

  for (i = 1; i < len; i++)
    {
      buf[i] = seed;
    }
}

static gm_inline int
verify_buffer (char *buf, int len)
{
  int i;
  char seed = buf[0];

  for (i = 1; i < len; i++)
    {
      if (buf[i] != seed)
	{
	  TEST_PRINT (2,
		      ("data error; at pos %d expected 0x%x and got 0x%x\n",
		       i, val, buf[i]));
	  return 1;
	}
    }

  return 0;
}

#else

/*
 * better (but slow) buffer fill and verify routines; these limit
 * throughput to a few MB/s
 */

static gm_inline int
verify_buffer (char *buf, int len)
{
  int i;
  char val;
  char seed = buf[0];
  gm_srand (seed);

  for (i = 1; i < len; i++)
    {
      val = gm_rand_mod (256);
      if (buf[i] != val)
	{
	  TEST_PRINT (2,
		      ("data error; at pos %d expected 0x%x and got 0x%x\n",
		       i, val, buf[i]));
	  return 1;
	}
    }

  return 0;
}

static gm_inline void
fill_buffer (char *buf, int len, char seed)
{
  int i;

  buf[0] = seed;
  gm_srand (seed);

  for (i = 1; i < len; i++)
    {
      buf[i] = gm_rand_mod (256);
    }
}

#endif

/****************************************************************
 * error handling
 ****************************************************************/

static void
err_exit (const char *str, gm_status_t status)
{
  gm_perror (str, status);
  if (port_g)
    {
      gm_close (port_g);
      port_g = NULL;
    }
  gm_exit (status);
}

/****************************************************************
 * sending
 ****************************************************************/

/* structure for recording information about sends for debugging
   purposes. */

struct nway_send_record
{
  struct gm_port *port;
  void *message;
  unsigned int size;
  unsigned long len;
  unsigned int priority;
  unsigned int target_node_id;
};

static struct gm_lookaside *send_record_lookaside;
static int outstanding_send_cnt[GM_NUM_PRIORITIES];


/* Initialize debugging support.  This function is called only if
   DEBUG_NWAY is nonzero */
/* The above comment is somewhat bogus, but I haven't figured out the
 * correction yet.                                    - Max 2001/04/25
 */

static void
nway_debug_init (struct gm_port *port)
{
  if (!send_record_lookaside)
    {
      send_record_lookaside
	= gm_create_lookaside (sizeof (struct nway_send_record),
			       gm_num_send_tokens (port));
      if (!send_record_lookaside)
	err_exit ("could not create send record lookaside list",
		  GM_OUT_OF_MEMORY);
    }
}

#if DEBUG_NWAY
void dump_recv_event(gm_recv_event_t *the_event, char *event_type_string)
{
  gm_u8_t        event_type;


  event_type = GM_RECV_EVENT_TYPE(the_event);

  gm_printf("  event type = %d (%s)\n", event_type, event_type_string);

  gm_printf("  sender node ID = %d, port id = %d\n",
	    gm_ntoh_u16(the_event->recv.sender_node_id),
	    gm_ntoh_u8(the_event->recv.sender_port_id));
  gm_printf("  message = %p\n", gm_ntohp(the_event->recv.message));
  gm_printf("  buffer  = %p\n", gm_ntohp(the_event->recv.buffer));
  gm_printf("  size    = %d\n", gm_ntoh_u8(the_event->recv.size));
  gm_printf("  length  = %d\n", gm_ntoh_u32(the_event->recv.length));
  gm_printf("  tag     = %d\n", gm_ntoh_u8(the_event->recv.tag));
}



void dump_message(test_message_t *the_message, int the_length)
{
  gm_printf("  Message at %p, length %d\n", the_message, the_length);
  gm_printf("  seq_no = %d\n", gm_ntoh_u32(the_message->seq_no));
  gm_printf("  type   = %d\n", gm_ntoh_u32(the_message->type  ));
  gm_printf("  size   = %d\n", gm_ntoh_u32(the_message->size  ));
  gm_printf("  sender = %d\n", gm_ntoh_u32(the_message->sender));
  if (payload_len(the_length) > 0)
    {
      gm_printf("  hex dump of payload follows:\n");
      /* gm_hex_dump() produces output only if GM_DEBUG */
      gm_hex_dump(the_message->payload, payload_len(the_length));
    }
  else
    gm_printf("  Message of this length has no payload.\n");
}



void dump_event_and_message(gm_recv_event_t *the_event,
			    test_message_t  *the_message,
			    int             the_loop_flag)
{
  int            tally_it = 0;	/* Just to give the loop something to do */
  gm_u8_t        event_type;
  test_message_t *evt_message;


  event_type = GM_RECV_EVENT_TYPE(the_event);

  gm_printf("Event display; event at %p:\n", the_event);

  /* We shouldn't get here unless the event is one of the following types */
  switch (event_type)
    {
    case GM_PEER_RECV_EVENT:
      evt_message = (test_message_t *) gm_ntohp (the_event->recv.buffer);
      dump_recv_event(the_event, "GM_PEER_RECV_EVENT");
      break;

    case GM_HIGH_PEER_RECV_EVENT:
      evt_message = (test_message_t *) gm_ntohp (the_event->recv.buffer);
      dump_recv_event(the_event, "GM_HIGH_PEER_RECV_EVENT");
      break;

    case GM_FAST_PEER_RECV_EVENT:
      evt_message = (test_message_t *) gm_ntohp (the_event->recv.message);
      dump_recv_event(the_event, "GM_FAST_PEER_RECV_EVENT");
      break;

    case GM_FAST_HIGH_PEER_RECV_EVENT:
      evt_message = (test_message_t *) gm_ntohp (the_event->recv.message);
      dump_recv_event(the_event, "GM_FAST_HIGH_PEER_RECV_EVENT");
      break;

    default:
      gm_printf("Something is VERY WRONG here -- unexpected event type\n");
      /* gm_hex_dump() produces output only if GM_DEBUG */
      gm_hex_dump(the_event, sizeof(*the_event));
    }
  
  gm_printf("Message display:\n");
  dump_message(the_message, gm_ntoh_u32(the_event->recv.length));

  if (the_message != evt_message)
    {
      gm_printf("Something is VERY WRONG here -- message pointer mismatch\n");
      gm_printf ("%p != %p\n", the_message, evt_message);
      dump_message(evt_message, gm_ntoh_u32(the_event->recv.length));
    }

  /* You could put a breakpoint here */
  while (the_loop_flag)
    {
      tally_it++;
      gm_sleep(2);			/* Go to sleep for 2 seconds */
    }
}
#endif

/****************
 * Send token management
 ****************/

static int _lst, _hst;

static gm_inline void
free_low_send_tokens (int cnt)
{
  _lst += cnt;
}
static gm_inline void
free_high_send_tokens (int cnt)
{
  _hst += cnt;
}
static gm_inline void
free_low_send_token (void)
{
  _lst++;
}
static gm_inline void
free_high_send_token (void)
{
  _hst++;
}
static gm_inline int
alloc_low_send_token (void)
{
  return _lst ? _lst-- : 0;
}
static gm_inline int
alloc_high_send_token (void)
{
  return _hst ? _hst-- : 0;
}

/****************
 * Statistics
 ****************/

static void
send_accounting(the_target, the_bytes)
{
  total_sent++;
  total_senta[the_target]++;
  bytes_sent += the_bytes;
#if DEBUG_NWAY
  if (the_bytes > max_len_sent)
    max_len_sent = the_bytes;
  if (the_bytes < min_len_sent)
    min_len_sent = the_bytes;
#endif
#ifdef PRINT_LOTS
  bytes_senta[the_target] += the_bytes;
#endif
}



static void
receive_accounting(the_sender, the_bytes)
{
  total_recvd++;
  bytes_recvd += the_bytes;
  total_recvda[the_sender]++;
#if DEBUG_NWAY
  if (the_bytes > max_len_recd)
    max_len_recd = the_bytes;
  if (the_bytes < min_len_recd)
    min_len_recd = the_bytes;
#endif
#ifdef PRINT_LOTS
  bytes_recvda[the_sender] += the_bytes;
#endif
}



/****************
 * Send Completion Handlers
 ****************/

static void check_sent (struct gm_port *port, void *message,
			enum gm_priority priority);


static void
dropped_or_resumed (struct gm_port *port, void *context, gm_status_t status)
{
  enum gm_priority priority;
  GM_PARAMETER_MAY_BE_UNUSED (port);
  
  gm_assert (status == GM_SUCCESS);

  priority = (enum gm_priority) ((unsigned long) context);
  gm_assert (priority == GM_LOW_PRIORITY || priority == GM_HIGH_PRIORITY);

  if (priority == GM_LOW_PRIORITY)
    free_low_send_token ();
  else
    free_high_send_token ();
}

/* either drop or resume sends on a connection that has experienced an error */

static void
drop_or_resume (struct gm_port *port, struct nway_send_record *r)
{
  if ((gm_rand () & 0xf) < 8)
    {
      TEST_PRINT (1, ("dropping %s priority sends\n",
		      r->priority == GM_LOW_PRIORITY ? "low" : "high"));
      gm_drop_sends (port, r->priority, r->target_node_id, port_id_g,
		     dropped_or_resumed, (void *) ((gm_size_t) r->priority));
    }
  else
    {
      TEST_PRINT (1, ("resuming %s priority sends\n",
		      r->priority == GM_LOW_PRIORITY ? "low" : "high"));
      gm_resume_sending (port, r->priority, r->target_node_id, port_id_g,
			 dropped_or_resumed, (void *) ((gm_size_t) r->priority));
    }
}

/* Handler called when simple sends complete. */

static void
my_send_callback (struct gm_port *port, void *context, gm_status_t status)
{
  struct nway_send_record *r;
  test_message_t *m;

  send_cb_tally++;
  if (DEBUG_NWAY)
    TEST_PRINT (0, ("Send callback entry %ld\n", send_cb_tally));

  r = (struct nway_send_record *) context;

  if (status == GM_SUCCESS || status == GM_SEND_DROPPED)
    {
      free_low_send_token ();
    }
  else
    {
      /* drop or resume will reuse this token */
      gm_perror ("send completed with error", status);
      TEST_PRINT (1,
		  ("(size=%u, len=%lu, %s priority, target_node_id=%u)\n",
		   r->size, r->len, r->priority ? "high" : "low",
		   r->target_node_id));
      drop_or_resume (port, r);
    }

  m = r->message;
  check_sent (port, m, GM_LOW_PRIORITY);
  gm_always_assert (gm_ntoh_u32 (m->type) == test_ping_type);
  free_a_send_message (m);
  gm_lookaside_free (r);
}

/* Handler called when forwarded sends complete. */

static void
my_forward_callback (struct gm_port *port, void *context, gm_status_t status)
{
  struct nway_send_record *r;
  test_message_t *m;

  fwrd_cb_tally++;

  if (DEBUG_NWAY)
    TEST_PRINT (0, ("Forward callback entry %ld\n", fwrd_cb_tally));

  r = (struct nway_send_record *) context;

  if (status == GM_SUCCESS || status == GM_SEND_DROPPED)
    {
      free_high_send_token ();
    }
  else
    {
      gm_perror ("send completed with error", status);
      drop_or_resume (port, r);
    }

  m = r->message;
  check_sent (port, m, GM_HIGH_PRIORITY);
  gm_always_assert (gm_ntoh_u32 (m->type) == test_forward_type);
  recycle_receive_buffer (port, m, r->size, GM_LOW_PRIORITY);
  gm_lookaside_free (r);
}

/****************
 * send functions
 ****************/

/* cover functions for gm_send_to_peer() to add debugging hooks. */

static void
send_to_peer (struct gm_port *port, void *message, unsigned int size,
	      unsigned long len, unsigned int priority,
	      unsigned int target_node_id)
{
  struct nway_send_record *r;

  nway_debug_init (port);

  gm_always_assert (target_node_id);

  /* record the details about the send */

  r = ((struct nway_send_record *)
       gm_lookaside_alloc (send_record_lookaside));
  if (!r)
    err_exit ("could not record send details", GM_OUT_OF_MEMORY);
  r->port = port;
  r->message = message;
  r->size = size;
  r->len = len;
  r->priority = priority;
  r->target_node_id = target_node_id;

  /* count the number of sends */

  if (DEBUG_NWAY)
    {
      ++outstanding_send_cnt[priority];
      TEST_PRINT (0, ("Outstanding sends before send: %d(low) %d(high)\n",
		      outstanding_send_cnt[GM_LOW_PRIORITY],
		      outstanding_send_cnt[GM_HIGH_PRIORITY]));
      TEST_PRINT (0, ("Sending size %d to Node %d\n", size, target_node_id));
    }
  if (priority == GM_HIGH_PRIORITY)
    {
      gm_send_to_peer_with_callback (port, message, size, len, priority,
				     target_node_id,
				     my_forward_callback,
				     (void *) r);
      forward_tally++;
    }
  else
    {
      gm_send_to_peer_with_callback (port, message, size, len, priority,
				     target_node_id,
				     my_send_callback,
				     (void *) r);
      gm_send_tally++;
    }
}

/* debug function to verify that a buffer was actually used for an
   earlier send and remove the record for that send.  Also, this
   updates the outstanding send counters. */

static void
check_sent (struct gm_port *port, void *message, enum gm_priority priority)
{
  GM_PARAMETER_MAY_BE_UNUSED (message);
  
  if (DEBUG_NWAY)
    {
      nway_debug_init (port);

      /* count the number of sends */

      --outstanding_send_cnt[priority];
      TEST_PRINT (0, ("outstanding sends after sent: %d(low) %d(high)\n",
		      outstanding_send_cnt[GM_LOW_PRIORITY],
		      outstanding_send_cnt[GM_HIGH_PRIORITY]));
    }
}

static void
try_send (void)
{
  unsigned int max_len_for_size,
               send_len;
  unsigned int send_size;
  unsigned int target;

  if (DEBUG_NWAY)
    TEST_PRINT (0, ("try_send() entry\n"));

  while (!caught_sigint)
    {
      test_message_t *message;

      if ((fast) || (max_test_size == min_test_size))
	{
	  send_size = max_test_size;
	}
      else
	{
	  send_size = (((unsigned int) gm_rand ()
			% (max_test_size - min_test_size + 1))
		       + min_test_size);
	}

      /* send_size = 12; */

      max_len_for_size = gm_max_length_for_size (send_size);
      if (fast)
	send_len = max_len_for_size;
      else if (max_len_for_size <= 16)
	send_len = DMA_ROUNDUP (16);
      else
	send_len = (DMA_ROUNDUP (gm_rand () % (max_len_for_size - 16)) + 16);
      if (send_len < min_test_len)
        {
	  /* May increase, but never reduce */
	  if (min_test_len > max_len_for_size)
	    {
              send_len = max_len_for_size;
            }
	  else
            {
	      send_len = min_test_len;
            }
        }

      gm_always_assert (send_len <= max_len_for_size);
      gm_always_assert (send_size >= min_test_size
			&& send_size <= max_test_size);

      if (!alloc_low_send_token ())
	{
	  if (DEBUG_NWAY)
	    TEST_PRINT (0, ("alloc_low_send_token() failure\n"));
	  break;
	}

      message = alloc_send_message (send_size);
      if (!message)
        {
	  free_low_send_token();
	  if (DEBUG_NWAY)
	    TEST_PRINT (0, ("alloc_send_message() failure\n"));
	  break;
        }

      if ((gm_rand () & 0xffffff) < PROB_SLEEP * 0x1000000)
	slp = 1;

      if (neighbor)
	{
      	  target = (unsigned int) (this_node_index+1) % (unsigned int) num_nodes;
	}
      else if (cycle)
	{
      	  target = (unsigned int) (target_index++) % (unsigned int) num_nodes;
	}
      else 
	{
      	  target = (unsigned int) gm_rand () % (unsigned int) num_nodes;
	}

      gm_assert (send_len >= 8 && (send_len <= max_len_for_size));

      /*
       * fill message with known values
       */
      if (payload_len (send_len) > 0)
	{
	  if (verify)
	    {
	      fill_buffer ((char *) message->payload, payload_len (send_len),
			   (char) out_seq[target]);
	    }
	  else
	    {
	      if (DEBUG_NWAY)
		{
		  memset (message->payload, 0xbb, payload_len (send_len));
		}
	    }
	}
      if (verify)
	{
	  message->seq_no = gm_hton_u32 (out_seq[target]);
	  out_seq[target]++;
	}
      
      message->type = gm_hton_u32 (test_ping_type);

      verify_send_buffer_size (message, send_size);
      gm_always_assert (target < (unsigned int) num_nodes);

      message->sender = gm_hton_u32 ((gm_u32_t) this_node_id);

      gm_assert (gm_ntoh_u32 (message->type) == test_ping_type);
      send_to_peer (port_g, message, send_size,
		    send_len, GM_LOW_PRIORITY, node_id[target]);

      send_accounting(target, send_len);

    }
} /* try_send() */



void progress(void)
{
  gm_recv_event_t *event;
  int             loop_limit = 20;


  if (DEBUG_NWAY)
    TEST_PRINT (0, ("progress() entry\n"));

  while (loop_limit--)
    {
      
      event = gm_receive (port_g);

      if (!fast)
	{
	  if (slp)
	    {
	      gm_printf ("sleeping\n");
	      gm_sleep (1);
	      sleeps++;
	      gm_printf ("awake\n");
	      slp = 0;
	    }
	}

      switch (GM_RECV_EVENT_TYPE (event))
	{
	case GM_PEER_RECV_EVENT:
	case GM_FAST_PEER_RECV_EVENT:
	  {
	    test_message_t *buffer;
	    test_message_t *p;
	    unsigned int size = gm_ntoh_u8 (event->recv.size);
	    unsigned int len = gm_ntoh_u32 (event->recv.length);
	    int sender = gm_ntoh_u16 (event->recv.sender_node_id);
	    int local_ix;
	    int ptype;

	    local_ix = get_local_ix(sender);
	    low_recvd++;
	    receive_accounting(local_ix, len);

	    buffer = (test_message_t *) gm_ntohp (event->recv.buffer);

	    if (GM_RECV_EVENT_TYPE (event) == GM_FAST_PEER_RECV_EVENT)
	      {
		p = (test_message_t *) gm_ntohp (event->recv.message);
	      }
	    else
	      {
		p = buffer;
	      }

	    ptype = gm_ntoh_u32 (p->type);
#if DEBUG_NWAY
	    if (low_recvd == 97)
	      {
		gm_printf("Printing 97th LOW received message\n");
		dump_event_and_message(event, p, 0);
	      }
	    if (ptype != test_ping_type)
	      {
		gm_printf("Failed assertion (ptype == test_ping_type)"
			  " %d == %d\n",
			  ptype, test_ping_type);
		dump_event_and_message(event, p, 1);
	      }
#else
	    gm_always_assert (ptype == test_ping_type);
#endif

	    /*
	     * verify message contents
	     */
	    if (verify)
	      {
		int seq         = gm_ntoh_u32 (p->seq_no);
		int found_error = 0;

		if (in_seq[local_ix] != seq)
		  {
		    TEST_PRINT (2, ("low  seqerr; size %d, len %d, sender %d, "
				    "buf %p; exp 0x%x, got 0x%x\n",
				    size, len, sender,
				    p, in_seq[local_ix], seq));

		    /*
		      for (i = 0; i < len; i++)
		      TEST_PRINT
		      (2, ("%02x ", ((int) ((char *) p)[i]) & 0xff));
		      TEST_PRINT (2, ("\n"));
		    */
		    /* print_all_buffers_size (6); */
		    /* die(); */
		    errs_by_node[local_ix]++;
		    found_error = 1;
		  }

		if (verify_buffer ((char *) p->payload,
				   payload_len (len)) != 0)
		  {
		    TEST_PRINT (2,
				("data error; size = %d, len = %d, p = %p\n",
				 size, len, p));
		    errs_by_node[local_ix]++;
		    found_error = 1;
		    /* print_all_buffers_size (6); */
		    /* die(); */
		  }

		in_seq[local_ix]++;

		if (found_error)
		  {
#if DEBUG_NWAY
		    dump_event_and_message(event, p, 0);
#endif
		    tolerate_errors(errs_by_node[local_ix], sender);
		  }
	      } /* (verify) */

	    if ((gm_rand () & 0xffff) / 16384.0 > PROB_FORWARD)
	      {
		/*
		 * Not forwarding, so just recycle receive buffer
		 */
		recycle_receive_buffer (port_g, buffer,
					gm_ntoh_u8 (event->recv.size),
					GM_LOW_PRIORITY);
	      }
	    else
	      {
		/*
		 * forward the message to a random target
		 */
		int target;

		if (!alloc_high_send_token ())
		  {
		    /*
		     * Unable to forward, so recycle receive buffer
		     */
		    recycle_receive_buffer (port_g, buffer,
					    gm_ntoh_u8 (event->recv.size),
					    GM_LOW_PRIORITY);
		    break;
		  }

		target = (unsigned int) gm_rand () % (unsigned int) num_nodes;

		buffer->type = gm_hton_u32 (test_forward_type);

		/*
		 * fill message with known values
		 */
		if (verify)
		  {
		    fill_buffer ((char *) buffer->payload,
				 payload_len (len),
				 (char) out_fw_seq[target]);
		    buffer->seq_no = gm_hton_u32 (out_fw_seq[target]);
		    out_fw_seq[target]++;
		  }

		gm_always_assert (target >= 0 && target < num_nodes);
		gm_assert (len <= gm_max_length_for_size (size));

		buffer->sender = gm_hton_u32 ((gm_u32_t) this_node_id);

		gm_assert (gm_ntoh_u32 (buffer->type) == test_forward_type);
		send_to_peer (port_g, buffer, size, len,
			      GM_HIGH_PRIORITY, node_id[target]);

		send_accounting(target, len);
	      }
	  } /* case GM_[FAST_]PEER_RECV_EVENT */

	  return;

	case GM_HIGH_PEER_RECV_EVENT:
	case GM_FAST_HIGH_PEER_RECV_EVENT:

	  /* Handle the reception of forwarded packets */

	  {
	    test_message_t *p;
	    int size = gm_ntoh_u8 (event->recv.size);
	    int len = gm_ntoh_u32 (event->recv.length);
	    int sender = gm_ntoh_u16 (event->recv.sender_node_id);
	    int local_ix;
	    int ptype;

	    local_ix = get_local_ix(sender);
	    high_recvd++;
	    receive_accounting(local_ix, len);

	    if (GM_RECV_EVENT_TYPE (event) == GM_FAST_HIGH_PEER_RECV_EVENT)
	      {
		p = (test_message_t *) gm_ntohp (event->recv.message);
	      }
	    else
	      {
		p = (test_message_t *) gm_ntohp (event->recv.buffer);
	      }

	    ptype = gm_ntoh_u32 (p->type);
#if DEBUG_NWAY
	    if (high_recvd == 97)
	      {
		gm_printf("Printing 97th HIGH received message\n");
		dump_event_and_message(event, p, 0);
	      }
 
	    if (ptype != test_forward_type)
	      {
		gm_printf("Failed assertion (ptype == test_forward_type)"
			  " %d == %d\n",
			  ptype, test_forward_type);
		dump_event_and_message(event, p, 1);
	      }
#else
	    gm_always_assert (ptype == test_forward_type);
#endif

	    /*
	     * verify message contents
	     */
	    if (verify)
	      {
		int seq         = gm_ntoh_u32 (p->seq_no);
		int found_error = 0;

		if (in_fw_seq[local_ix] != seq)
		  {
		    TEST_PRINT (2, ("high seqerr; size %d, len %d, "
				    "sender %d, buf %p;"
				    " exp 0x%x, got 0x%x\n",
				    size, len, sender, p,
				    in_fw_seq[local_ix], seq));

		    /*
		      for (i = 0; i < len; i++)
		      TEST_PRINT (2, ("%02x ",
		      ((int) ((char *) p)[i]) & 0xff));
		      TEST_PRINT (2, ("\n"));
		    */
		    errs_by_node[local_ix]++;
		    found_error = 1;
		  }

		if (verify_buffer ((char *) p->payload, payload_len (len))
		    != 0)
		  {
		    TEST_PRINT (2, ("data error; size = %d, len = %d, "
				    "p = %p\n", size, len, p));
		    errs_by_node[local_ix]++;
		    found_error = 1;
		  }

		in_fw_seq[local_ix]++;

		if (found_error)
		  {
#if DEBUG_NWAY
		    dump_event_and_message(event, p, 0);
#endif
		    tolerate_errors(errs_by_node[local_ix], sender);
		  }
	      } /* (verify) */

	    recycle_receive_buffer (port_g, gm_ntohp (event->recv.buffer),
				    size,   GM_HIGH_PRIORITY);
	  }

	  return;

	case GM_NO_RECV_EVENT:

	  return;

	case GM_ALARM_EVENT:
	  gm_unknown (port_g, event);
	  gm_unkn_tally++;
	  break;

	  /*
	   * should never happen; gives GM a chance to invoke our callbacks
	   */
	default:
	  gm_unknown (port_g, event);
	  gm_unkn_tally++;
	  break;

	} /* switch */

    } /* while */

} /* progress() */


/****************************************************************
 * test proper
 ****************************************************************/

static void
test_init (struct gm_port *port,    int port_id,
	   int            min_size, int max_size)
{
  int         size;
  int         number_of_sizes = max_size - min_size + 1;
  unsigned    port_send_tokens,
              port_recv_tokens,
              low_send_tokens,
              high_send_tokens,
              recv_buffers_per_size,
              recv_buffer_tally = 0;
  gm_status_t status;
  gm_size_t   acceptable_sizes_mask;
  

  GM_PARAMETER_MAY_BE_UNUSED (port_id);
  
  /*
   * initialize GM
   */
  gm_assert (port);
  TEST_PRINT (1, ("GM is initialized.\n"));

  gm_initialize_alarm (&my_alarm);

  if (max_size > 30)
    acceptable_sizes_mask = (gm_size_t)(0xffffffff ^
					((1<<min_size) - 1));
  else
    acceptable_sizes_mask = (gm_size_t)(((1<<(max_size+1)) - 1) ^
					((1<<min_size) - 1));

  TEST_PRINT (1, ("Calling gm_set_acceptable_sizes "
		  "(%p, GM_LOW_PRIORITY, 0x%lx)\n",
		  port, acceptable_sizes_mask));

  status = gm_set_acceptable_sizes (port, GM_LOW_PRIORITY,
                                    acceptable_sizes_mask);
  if (status == GM_SUCCESS)
    status = gm_set_acceptable_sizes (port, GM_HIGH_PRIORITY,
                                    acceptable_sizes_mask);

  if (status != GM_SUCCESS)
    {
      gm_perror ("could not set GM acceptable sizes", status);
      if (port)
	{
	  gm_close (port);
	  port = NULL;
	}

      gm_exit (status);
    }

  /* note the initial number of send tokens */

  port_send_tokens = gm_num_send_tokens(port);
  port_recv_tokens = gm_num_receive_tokens(port);

  low_send_tokens  = port_send_tokens / 2;
  high_send_tokens = port_send_tokens / 2;
  free_low_send_tokens (low_send_tokens);
  free_high_send_tokens (high_send_tokens);
  recv_buffers_per_size = port_recv_tokens / (number_of_sizes * 2);
  /* total over all sizes must be <= GM_NUM_RECV_TOKENS */

  if (recv_buffers_per_size > 5)
    {
      recv_buffers_per_size = 5;
    }

  TEST_PRINT(1, ("Min. size = %d, max. size = %d\n",
		 min_test_size, max_test_size));
  if (min_test_len != 0)
    {
      if (fast)
        {
	  TEST_PRINT (2, ("Both -f and --min-test-length specified; "
	  		  "--min-test-length has no effect;\n"));
	  TEST_PRINT (2, ("... -f means always use max length for size\n"));
        }
      else
        {
	  TEST_PRINT (1, ("Min. length = %d\n", min_test_len));
	  if (min_test_len != DMA_ROUNDUP(min_test_len))
            {
	      TEST_PRINT (1, ("NOTE: will NOT be rounded up to %d\n",
			      DMA_ROUNDUP(min_test_len)));
            }            
        }
    }

  TEST_PRINT(1, ("Port has %d send and %d receive tokens\n",
		 port_send_tokens, port_recv_tokens));
  TEST_PRINT(1, ("Using %d send tokens for LOW and %d for HIGH priority\n",
		 low_send_tokens, high_send_tokens));
  TEST_PRINT(1, ("Using %d receive buffers for each size\n",
		 recv_buffers_per_size * 2));

  /* allocate receive buffers */

  for (size = min_test_size; size <= max_test_size; size++)
    {
      int i;
      gm_size_t max_length_for_size = gm_max_length_for_size(size);
      
#if DEBUG_NWAY
      TEST_PRINT(1, ("Receive buffers for size %d:\n", size));
#endif
      for (i = 0; i < recv_buffers_per_size; i++)
	{
	  int j;

	  /* allocate low and high priority receive buffers */
	  for (j = 0; j < 2; j++)
	    {
	      void *message;

	      /* allocate a receive buffer */
	      message = gm_dma_calloc (port, 1, max_length_for_size);
	      if (!message)
		err_exit ("could not allocate receive buffer\n",
			  GM_OUT_OF_MEMORY);

	      /* Make sure message is aligned. */
	      gm_always_assert (GM_DMA_ALIGNED (message));

#if GM_DEBUG_BUFFERS
	      gm_register_buf (message, size);
#endif
#if DEBUG_NWAY
	      TEST_PRINT(1, ("  Buffer %d at %p for GM_%s_PRIORITY\n",
			     i, message, (j == 0) ? "LOW" : "HIGH"));
	      memset (message, 0xee, max_length_for_size);
#else
	      memset (message, 0, max_length_for_size);
#endif
	      gm_provide_receive_buffer (port, message, size, 
					 (j == 0) ? GM_LOW_PRIORITY:
					 GM_HIGH_PRIORITY);
	      recv_buffer_tally++;
	    }
	}
    }
  TEST_PRINT (1, ("Allocated %d total receive buffers\n", recv_buffer_tally));

  /* allocate send buffers */

  status = init_send_buffer_allocator (port, min_test_size, max_test_size,
				       low_send_tokens + high_send_tokens);
  if (status != GM_SUCCESS)
    err_exit ("could not initialize send buffer allocator", status);

  status = gm_get_node_id (port, &this_node_id);
  if (status != GM_SUCCESS)
    {
      fprintf (stderr, "could not get node id\n");
      if (port)
	{
	  gm_close (port);
	  port = NULL;
	}
      gm_exit (status);
    }
#if DEBUG_NWAY
  min_len_sent = min_len_recd = gm_max_length_for_size(31);
#endif

  this_node_index = get_local_ix (this_node_id);

} /* test_init() */


static gm_inline void
handle_sigint (int i)
{
  GM_PARAMETER_MAY_BE_UNUSED (i);
  
  caught_sigint = 1;
}

static void
print_stats(void)
{
  int i;
  int k;
  double gm_stime = 0;
  double oldstime = 0;


  GM_VAR_MAY_BE_UNUSED (gm_stime);
  GM_VAR_MAY_BE_UNUSED (oldstime);

  if (timestamps)
    {
      time_t uct;
      char   *time_string;

      uct = time(NULL);
      time_string = asctime(localtime(&uct));
      TEST_PRINT (1, ("As of %s", time_string));
    }

  for (i = 1; i <= num_nodes; i++)
    {
      if (errs_by_node[i] != 0)
	{
	  TEST_PRINT (1, ("%d: in = %d  out = %d  in fwd = %d  "
			  "out fwd = %d  errs_by_node = %d %s\n",
			  i, in_seq[i], out_seq[i], in_fw_seq[i],
			  out_fw_seq[i], errs_by_node[i],
			  (i == (int) this_node_id) ? "(*)" : ""));
	}
    }

#if GM_OS_VXWORKS
  TEST_PRINT (1, ("Msgs sent = %ld  Msgs received = %ld"
		  " (HIGH %ld, LOW %ld).\n",
		  (unsigned long) total_sent, (unsigned long) total_recvd,
		  (unsigned long) high_recvd, (unsigned long) low_recvd));
  TEST_PRINT (1, ("MBytes sent = %ld.  MBytes received = %ld.\n",
		  ((unsigned long)bytes_sent)  / (1024 * 1024),
		  ((unsigned long)bytes_recvd) / (1024 * 1024)));
  TEST_PRINT (1,
	      ("MB/s sent = %ld.  MB/s received = %ld.\n",
	       ((unsigned long)(bytes_sent - old_bytes_sent)) 
	       / (iterate_time * 1024 * 1024),
	       ((unsigned long)(bytes_recvd - old_bytes_recvd))
	       / (iterate_time * 1024 * 1024)));
#else /* !GM_OS_VXWORKS */
  TEST_PRINT (1, ("Msgs sent = %.0f  Msgs received = %.0f"
		  " (HIGH %.0f, LOW %.0f).\n",
		  (float) total_sent, (float) total_recvd,
		  (float) high_recvd, (float) low_recvd));
  TEST_PRINT (1, ("MBytes sent = %.0f.  MBytes received = %.0f.\n",
		  (float) (bytes_sent / (1024 * 1024)),
		  (float) (bytes_recvd / (1024 * 1024))));
  TEST_PRINT (1,
	      ("MB/s sent = %3.3f.  MB/s received = %3.3f.\n",
	       ((float) ((int) bytes_sent - (int) old_bytes_sent) /
		(float) (iterate_time * 1024 * 1024)),
	       ((float) ((int) bytes_recvd - (int) old_bytes_recvd) /
		(float) (iterate_time * 1024 * 1024))));
#endif /* !GM_OSVXWORKS */

  /* PRINT Messages sent and received */
  TEST_PRINT (1, ("Msgs sent, to nodes:\n"));

  for (k = 0; k < num_nodes; k++)
    {
      TEST_PRINT (1, (" %12s", node_name[k]));
    }
  TEST_PRINT (1, ("\n"));
  for (k = 0; k < num_nodes; k++)
    {
#if GM_OS_VXWORKS
      TEST_PRINT(1, ("%013ld", (unsigned long) total_senta[k]));
#else
      TEST_PRINT (1, ("%13.0f", (float) total_senta[k]));
#endif
    }
  TEST_PRINT (1, ("\n\n"));

  TEST_PRINT (1, ("Msgs received, from nodes:\n"));
  for (k = 0; k < num_nodes; k++)
    {
      TEST_PRINT (1, (" %12s", node_name[k]));
    }
  TEST_PRINT (1, ("\n"));
  for (k = 0; k < num_nodes; k++)
    {
#if GM_OS_VXWORKS
      TEST_PRINT(1, ("%13ld", (unsigned long) total_recvda[k]));
#else
      TEST_PRINT (1, ("%13.0f", (float) total_recvda[k]));
#endif
    }
  TEST_PRINT (1, ("\n\n"));

#ifdef PRINT_LOTS
  /* PRINT MBytes sent to and received from each node */
  TEST_PRINT (1, ("MBytes sent, to nodes:\n"));

  for (k = 0; k < num_nodes; k++)
    {
      TEST_PRINT (1, (" %12s", node_name[k]));
    }
  TEST_PRINT (1, ("\n"));
  for (k = 0; k < num_nodes; k++)
    {
      TEST_PRINT (1, ("%13.0f", (float) (bytes_senta[k] / (1024 * 1024))));
    }
  TEST_PRINT (1, ("\n\n"));

  TEST_PRINT (1, ("Msgs received, from nodes:\n"));
  for (k = 0; k < num_nodes; k++)
    {
      TEST_PRINT (1, (" %12s", node_name[k]));
    }
  TEST_PRINT (1, ("\n"));
  for (k = 0; k < num_nodes; k++)
    {
      TEST_PRINT (1, ("%13.0f", (float) (bytes_recvda[k] / (1024 * 1024))));
    }
  TEST_PRINT (1, ("\n\n"));

  /* PRINT MB/s sent to and received from each node */
  TEST_PRINT (1, ("MB/s sent, to nodes:\n"));

  for (k = 0; k < num_nodes; k++)
    {
      TEST_PRINT (1, (" %12s", node_name[k]));
    }
  TEST_PRINT (1, ("\n"));
  for (k = 0; k < num_nodes; k++)
    {
      TEST_PRINT (1, ("%13.3f",
		      ((float)
		       ((int) bytes_senta[k] - (int) old_bytes_senta[k]) /
		       (float) (DELAY * 1024 * 1024))));
      old_bytes_senta[k] = bytes_senta[k];
    }
  TEST_PRINT (1, ("\n\n"));

  TEST_PRINT (1, ("MB/s received, from nodes:\n"));
  for (k = 0; k < num_nodes; k++)
    {
      TEST_PRINT (1, (" %12s", node_name[k]));
    }
  TEST_PRINT (1, ("\n"));
  for (k = 0; k < num_nodes; k++)
    {
      TEST_PRINT (1, ("%13.3f",
		      ((float)
		       ((int) bytes_recvda[k] - (int) old_bytes_recvda[k]) /
		       (float) (DELAY * 1024 * 1024))));
      old_bytes_recvda[k] = bytes_recvda[k];
    }
  TEST_PRINT (1, ("\n\n"));

  gm_stime = ((gm_s64_t) gm_ticks (port_g)) / 2e6; /* 2 million ticks per sec. */
  TEST_PRINT (1, ("Slept %.0f time(s).  Last printed status %3.2f "
		  "seconds ago.\n\n",
		  (float) sleeps, gm_stime - oldstime));
  oldstime = gm_stime;
#endif

  old_bytes_sent = bytes_sent;
  old_bytes_recvd = bytes_recvd;
}



static int
gm_nway (int argc, char *argv[])
{
  int         i;
  gm_status_t status;


  signal (SIGINT, handle_sigint);

  parse_args (argc, argv);

  cross_check_args();

  /* Open the user-specified port */

  port_g = NULL;

  status = gm_open (&port_g, board_num, port_id_g, "n-way test",
		    (enum gm_api_version) GM_API_VERSION);
  if (status != GM_SUCCESS)
    {
      gm_perror ("could not open GM port", status);
      if (port_g)
	{
	  gm_close (port_g);
	  port_g = NULL;
	}
      gm_exit (status);
    }

  cross_check_nodes();

  announce_run_parameters();

  test_init (port_g, port_id_g, min_test_size, max_test_size);

  for (i = 0; i < MAX_NODES; i++)
    {
      total_senta[i] = 0;
      total_recvda[i] = 0;
#ifdef PRINT_LOTS
      bytes_senta[i] = 0;
      bytes_recvda[i] = 0;
      old_bytes_senta[i] = 0;
      old_bytes_recvda[i] = 0;
#endif
      in_seq[i] = out_seq[i] = 0;
      in_fw_seq[i] = out_fw_seq[i] = 1000000;
      errs_by_node[i] = 0;
    }

  gm_set_alarm (port_g, &my_alarm,
		(unsigned int) (iterate_time * TIME_CALIBRATION),
		set_got_alarm, 0);

  if (caught_sigint)
    die ();
  test_underway = 1;

  while (!caught_sigint)
    {
      try_send ();

      progress();

      if (got_alarm)
	{
	  iterations++;
	  got_alarm = 0;
	  print_stats();
	  gm_set_alarm (port_g, &my_alarm,
			(unsigned int) (iterate_time * TIME_CALIBRATION),
			set_got_alarm, 0);
	}

      if ((iterate_limt) && (iterations >= iterate_limt))
	{
	  TEST_PRINT (2, ("Iteration limit of %ld reached\n", iterate_limt));
	  break;
	}

    } /* while */

  test_underway = 0;

  if ((send_cb_tally < gm_send_tally) || (fwrd_cb_tally < forward_tally))
    {
#define WAIT_CONVERGED 20

      unsigned int sleep_tally     = 0,
                   no_change_tally = 0;
      unsigned int prev_scb_tally  = send_cb_tally,
	           prev_fcb_tally  = fwrd_cb_tally,
	           prev_send_tally = gm_send_tally,
	           prev_frwd_tally = forward_tally;


      TEST_PRINT (2, ("Waiting for send callbacks to complete\n"));
      TEST_PRINT (1, ("  sends    = %ld\tsend callbacks     = %ld\n",
		      gm_send_tally, send_cb_tally));
      TEST_PRINT (1, ("  forwards = %ld\tforward callbacks  = %ld\n",
		      forward_tally, fwrd_cb_tally));

      /* We will wait until all send callbacks are complete, or until
       * we have slept WAIT_CONVERGED times in a row for 1 second, with
       * no change in any of the four tallies.
       */

      do
	{
          if (caught_sigint)
            die();
	  progress();

	  if ((send_cb_tally == prev_scb_tally)  &&
	      (fwrd_cb_tally == prev_fcb_tally)  &&
	      (gm_send_tally == prev_send_tally) &&
	      (forward_tally == prev_frwd_tally)    )
	    {
	      no_change_tally++;
	      gm_sleep(1);
	      sleep_tally++;
	    }
	  else
	    {
	      TEST_PRINT (1, ("  sends    = %ld\tsend callbacks     = %ld\n",
			      gm_send_tally, send_cb_tally));
	      TEST_PRINT (1, ("  forwards = %ld\tforward callbacks  = %ld\n",
			      forward_tally, fwrd_cb_tally));
	      prev_scb_tally  = send_cb_tally;
	      prev_fcb_tally  = fwrd_cb_tally;
	      prev_send_tally = gm_send_tally;
	      prev_frwd_tally = forward_tally;
	      no_change_tally = 0;
	    }
	}
      while (((send_cb_tally < gm_send_tally) ||
	      (fwrd_cb_tally < forward_tally)    ) &&
	     (no_change_tally < WAIT_CONVERGED)       );

      TEST_PRINT (1, ("Slept %d times waiting for convergence\n",
		      sleep_tally));

      if ((send_cb_tally < gm_send_tally) || (fwrd_cb_tally < forward_tally))
	{
	  TEST_PRINT (2, ("Never received all callbacks...\n"));
	  TEST_PRINT (2, ("  sends    = %ld\tsend callbacks     = %ld\n",
			  gm_send_tally, send_cb_tally));
	  TEST_PRINT (2, ("  forwards = %ld\tforward callbacks  = %ld\n",
			  forward_tally, fwrd_cb_tally));
	}
    }

  TEST_PRINT (2, ("Test completed after %ld iterations.\n", iterations));
  TEST_PRINT (1, ("Called gm_unknown() %ld times\n", gm_unkn_tally));
#if DEBUG_NWAY
  TEST_PRINT (2, ("Min length sent = %d (0x%x); max = %d (0x%x)\n",
		  min_len_sent, min_len_sent, max_len_sent, max_len_sent));
  TEST_PRINT (2, ("Min length recd = %d (0x%x); max = %d (0x%x)\n",
		  min_len_recd, min_len_recd, max_len_recd, max_len_recd));
#endif

  print_stats();

  gm_close (port_g);
  port_g = NULL;

  TEST_PRINT (1, ("GM deinitialized.\n"));

#if 0
  die ();			/* Why would we want to do this? */
#endif

  return 0;

} /* gm_nway() */


/****************************************************************
 * Entry point
 ****************************************************************/

#define GM_MAIN gm_nway
#include "gm_main.h"

/*
  This file uses GM standard indentation.

  Local Variables:
  c-file-style:"gnu"
  c-backslash-column:72
  tab-width:8
  End:
*/
