/******************************************************************-*-c-*-
 * Myricom GM networking software and documentation			 *
 * Copyright (c) 2001 by Myricom, Inc.					 *
 * All rights reserved.	 See the file `COPYING' for copyright notice.	 *
 *************************************************************************/

/* author: feldy@myri.com */

#if defined __STDC__ || defined WIN32
#if GM_OS_VXWORKS
#include "vxWorks.h"
#endif  /* GM_OS_VXWOKRS */
#include <stdarg.h>
#else
#include <varargs.h>
#endif

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "gm_internal.h"


/****************************************************************
 * global data
 ****************************************************************/

int portid    = 1;
int board_num = 0;

gm_port_t *g_port = NULL;

char new_host_name[GM_MAX_HOST_NAME_LEN + 1];


/****************************************************************
 * Command line processing
 ****************************************************************/

#define TP(a) fprintf (stderr, a)

static void
print_synopsis (void)
{
  TP ("\n");
  TP ("Myricom 'gm_set_name' sets the GM name (host name)\n"
      " for the specified board.\n");
}


static void
print_options (void)
{
  TP ("\nusage: gm_set_name --host-name=name [--board=board]\n\n");

  TP ("-B, --board=NUM     [0] board for which the name is to be set\n");
  TP ("-h, --host-name=NAME []  GM name to set -- MANDATORY PARAMETER\n");
  TP ("    --help              output usage information, then exit\n");
  fprintf (stderr,
	   "\nThe new name must be no more than %d characters long.\n\n",
	   GM_MAX_HOST_NAME_LEN);
#undef TP
}


static void
#ifdef va_dcl
usage (fmt, va_alist)
     const char *fmt;
     va_dcl
#else
     usage (const char *fmt, ...)
#endif
{
  va_list ap;

  print_options ();

  /* Now print the caller's diagnostic */
#if defined __STDC__ || defined WIN32
  va_start (ap, fmt);
#else
  va_start (ap);
#endif
  vfprintf (stderr, fmt, ap);
  va_end (ap);

  if (g_port)
    {
      gm_close (g_port);
    }
  
  gm_exit (GM_INVALID_PARAMETER);

} /* usage() */


static void
parse_args (int argc, char *argv[])
{
  /* NOTE: If you change the operating characteristics or the flags
   *       used by this program, please notify Max so he can change the
   *       Myrinet Handbook document.
   */
  int i;

  /* The 'F' macro is for all flags with no argument (parameter) */
#define F(name,action) if (strcmp (argv[i], name) == 0) { action; continue; }

  for (i = 1; i < argc; i++)
    {
      /* Dispose of the --help flag first */
      if (strcmp (argv[i], "--help") == 0)
	{
	  print_synopsis ();
	  print_options ();
	  gm_exit (GM_SUCCESS);
	}

      /* Here are the flags with no argument.  Please keep short form,
	 if any, after its long form equivalent, but keep the long
	 form options in alpha sequence.  */

      /* None in gm_set_name() at present */

#undef F

      /* The check_opt macro is for long-form options with numeric
	 arguments and simple validation tests. */
#define check_opt(prefix, var, test)					\
      if (sscanf (argv[i], "--" prefix "=%i", &var) == 1)		\
	{								\
	  if (!(test))							\
	    {								\
	      usage ("bad value for option \"%s\"\n", argv[i]);		\
	      gm_exit (GM_INVALID_PARAMETER);				\
	    }								\
	  continue;							\
	}

      /* Here are the flags with arguments, sequenced as above */

      check_opt ("board", board_num, board_num >= 0);
      if ((strcmp (argv[i], "-B") == 0) || (strcmp (argv[i], "--unit") == 0))
	{
	  if (i + 1 >= argc)
	    {
	      usage ("Board number expected after '%s'.\n", argv[i]);
	    }
	  
	  board_num = atoi (argv[++i]);
	  if (board_num < 0)
	    {
	      usage ("bad board number: %d\n", board_num);
	    }
	  
	  continue;
	}

      if (sscanf (argv[i], "--host-name=%s", new_host_name) == 1)
	{
	  new_host_name[GM_MAX_HOST_NAME_LEN] = '\0';
	  continue;
	}
      if (strcmp (argv[i], "-h") == 0)
	{
	  if (i + 1 >= argc)
	    {
	      usage ("host_name expected after '-h'.\n");
	    }
	  
	  strncpy (new_host_name, argv[++i], GM_MAX_HOST_NAME_LEN);
	  new_host_name[GM_MAX_HOST_NAME_LEN] = '\0';
	  continue;
	}

#undef check_opt

      usage ("unknown arg: '%s'\n", argv[i]);
      gm_exit (GM_FAILURE);
    }
}


static int
gm_set_name (int argc, char *argv[])
{
  gm_status_t status;

  gm_bzero (new_host_name, sizeof(new_host_name));

  parse_args (argc, argv);

  if (strlen (new_host_name) == 0)
    {
      usage ("Must specify name; use \'--host-name=<host_name>\'\n");
    }
  /* Function usage() does not return */

  status = gm_open (&g_port, board_num, portid, argv[0],
		    (enum gm_api_version) GM_API_VERSION);
  if (status != GM_SUCCESS)
    {
      printf ("Could not open board %d, port %d : %s\n", board_num, portid,
	      gm_strerror (status));
      goto abort_with_nothing;
    }

  status = _gm_set_host_name (g_port, &new_host_name);
  if (status != GM_SUCCESS)
    {
      printf ("Couldn't set the GM host name\n");
    }
  else
    {
      printf ("Set host_name of board %d to '%s'\n", 
	      board_num, new_host_name);
      strcpy (new_host_name, "");
      gm_get_host_name (g_port,new_host_name);
      printf ("Read    name of board %d as '%s'\n",
	      board_num,new_host_name);
    }

  gm_close (g_port);

 abort_with_nothing:
  return (0);
}

#define GM_MAIN gm_set_name
#include "gm_main.h"
