/******************************************************************-*-c-*-
 * Myricom GM networking software and documentation                      *
 * Copyright (c) 1996, 1997, 1998, 1999 by Myricom, Inc.                 *
 * All rights reserved.  See the file `COPYING' for copyright notice.    *
 *************************************************************************/

/* author: finucane@myri.com */

#if defined __STDC__ || defined WIN32
#if GM_OS_VXWORKS
#include "vxWorks.h"
#endif  /* GM_OS_VXWOKRS */
#include <stdarg.h>
#else
#include <varargs.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#ifdef HAVE_UNISTD_H
#include "unistd.h"
#endif

#include "gm_internal.h"
#include "gm_enable_crc32.h"

#define num_elements(e)(sizeof(e)/sizeof(e[0]))

#include "gm.h"
#include "lanai9_def.h"


typedef struct speed_t
{
  int clock;
  int speed;
  int mbyte_per_sec;
} speed_t;


/****************************************************************
 * global data
 ****************************************************************/

int            board_num = 0;
int            speed     = 0;
int            ix_speed;

struct gm_port *port = NULL;


static speed_t speeds [] =
{
  {
    0x606c81,  1280, 160
  },

  {
    0x61e80e,  2000, 250
  },
  
  {
    0x60a003,  2200, 275
  },
  
  {
    0x612006,  2400, 300
  },
 
  {
    0x606c01,  2560, 320
  },
 
};

/****************************************************************
 * Command line processing
 ****************************************************************/

#define TP(a) fprintf (stderr, a)

static void
print_synopsis(void)
{
  TP ("\n");
  TP ("Myricom 'gm_set_speed' sets the interface link speed"
      " for SAN interfaces.\n");
}


static void
print_options (void)
{
     int i;


     TP ("\nusage: gm_set_speed --speed=speed [--board=board]\n\n");

     TP ("-B, --board=NUM  [0] board on which to set the speed\n");
     TP ("    --speed=NUM  []  speed to set -- MANDATORY PARAMETER\n");
     TP ("    --help           output usage and options information, then exit\n");
     printf ("\nspeed must be one of:\n");

     for (i = 0; i < 2 /*num_elements (speeds)*/; i++)
	  printf ("  %d\n", speeds [i].speed);
     printf ("\n");

#undef TP
}


static void
#ifdef va_dcl
usage (fmt, va_alist)
     const char *fmt;
     va_dcl
#else
usage (const char *fmt, ...)
#endif
{
  va_list ap;

  print_options();

/* Now print the caller's diagnostic */
#if defined __STDC__ || defined WIN32
  va_start (ap, fmt);
#else
  va_start (ap);
#endif
  vfprintf (stderr, fmt, ap);
  va_end (ap);

  if (port)
       gm_close(port);

  gm_exit (GM_INVALID_PARAMETER);

} /* usage() */


static void
parse_args (int argc, char *argv[])
{
/* NOTE: If you change the operating characteristics or the flags
 *       used by this program, please notify Max so he can change the
 *       Myrinet Handbook document.
 */
  int i;

  /* The 'F' macro is for all flags with no argument (parameter)     */
#define F(name,action) if (strcmp (argv[i], name) == 0) { action; continue; }

  for (i = 1; i < argc; i++)
  {
       /* Dispose of the --help flag first                                */
       if (strcmp (argv[i], "--help") == 0)
       {
	    print_synopsis();
	    print_options();
	    gm_exit (GM_SUCCESS);
       }

       /* Here are the flags with no argument.                            */
       /* Please keep short form, if any, after its long form equivalent, */
       /* but keep the long form options in alpha sequence.               */

       /* None in gm_set_speed at present */

#undef F

       /* The check_opt macro is for long-form options with numeric        */
       /* arguments and simple validation tests.                           */
#define check_opt(prefix, var, test)				\
       if (sscanf (argv[i], "--" prefix "=%i", &var) == 1)	\
       {							\
	  if (!(test))						\
	  {							\
	     usage ("bad value for option \"%s\"\n", argv[i]);	\
	     gm_exit (GM_INVALID_PARAMETER);			\
	  }							\
	  continue;						\
       }

       /* Here are the flags with arguments, sequenced as above */

       check_opt ("board", board_num, board_num >= 0);
       if ((strcmp (argv[i], "-B") == 0) || (strcmp (argv[i], "--unit") == 0))
       {
	    if (i + 1 >= argc)
		 usage ("Board number expected after '%s'.\n", argv[i]);

	    board_num = atoi (argv[++i]);
	    if (board_num < 0)
		 usage ("bad board number: %d\n", board_num);

	    continue;
       }

       check_opt ("speed", speed, speed > 0);
       if (strcmp (argv[i], "--speed") == 0)
       {
	    if (i + 1 >= argc)
		 usage ("Speed expected after '%s'.\n", argv[i]);

	    speed = atoi (argv[++i]);
	    if (speed <= 0)
		 usage ("bad speed: %d\n", board_num);

	    continue;
       }

#undef check_opt

       usage ("unknown arg: '%s'\n", argv[i]);
       gm_exit (GM_FAILURE);
  }
}



static void
check_speed_option(void)
{
     if (speed == 0)
	  usage ("A speed must be specified\n");  

     for (ix_speed = 0; ix_speed < num_elements (speeds); ix_speed++)
	  if (speeds [ix_speed].speed == speed)
	       break;

     if (ix_speed == num_elements (speeds))
	  usage ("%d is a bad speed.\n", speed);
}



/****************************************************************
 * perform the actual test
 ****************************************************************/


static int dl_clock (struct gm_port *port, long int c)
{
  return
    !_gm_cop_wakeup  (port) &&
    !_gm_cop_send (port, 0x21) &&
    !_gm_cop_send (port, (c >> 16) & 0xFF) &&
    !_gm_cop_send (port, (c >> 8) & 0xFF) &&
    !_gm_cop_send (port, c & 0xFF) &&
    !_gm_cop_end (port);
}

#if 0
static int do_bitwr (struct gm_port *port, int addr, int bitinfo)
{
  return
    !_gm_cop_wakeup  (port) &&
    !_gm_cop_send (port, 0x2F) &&
    !_gm_cop_send (port, addr) &&
    !_gm_cop_send (port, bitinfo & 0xF7) &&
    !_gm_cop_end (port);
}
#endif


static int gm_set_speed (int argc, char *argv[])
{
	 gm_status_t status;
	 gm_u32_t offset;
	 gm_u32_t myrinet_bits
			 = (TIMEOUT0
				| TIMEOUT1
				| WINDOW0
				| WINDOW1
				| RX_CRC8_ENABLE_BIT
				| TX_CRC8_ENABLE_BIT);

     parse_args(argc, argv);

     check_speed_option();

     if (_gm_mapper_open (&port, board_num,
			  (enum gm_api_version) GM_API_VERSION)
	 != GM_SUCCESS)
     {
	  printf ("_gm_mapper_open failed\n");
	  exit (1);
     }

	 if (GM_ENABLE_CRC32)
	 {		 
	  myrinet_bits |= CRC32_ENABLE_BIT;
	 }

	 if (speed == 1280)
	 {
	  printf("Disabling BEAT on unit %d\n", board_num);
	 } 
	 else
	 {
	  printf("Enabling BEAT on unit %d\n", board_num);
	  myrinet_bits |= BEAT_ENABLE;
	 }

	 offset = (char *)&(MYRINET) - (char *)&MM_REG_FULLWORD (0);
	 status = _gm_user_write_lanai_register (port, offset, myrinet_bits);
	 if (status != GM_SUCCESS)
	 {
	  printf("_gm_user_write_lanai_register() failed\n");
	  exit (1);
	 }

     printf("Setting link speed on unit %d to %d\n", board_num, speed);
     if (!dl_clock (port, 0x1e0000) ||
	 !dl_clock (port, speeds [ix_speed].clock))
	  printf ("dl_clock failed\n");

     gm_close(port);

     return(0);
}

#define GM_MAIN gm_set_speed
#include "gm_main.h"
