#include "gm.h"
#include <stdio.h>
#include <stdlib.h>

/*

  This program allow to do some testing that the MCP support for Sparc
  streaming does what is expected and that we never read stale data,
  (and allow to test that without the MCP changes, we are actually
  SDMAING stale data sometimes).
  For instance you can use it with:
  gm_sparcstreaming 4 0 4
  gm_sparcstreaming 64 0 64
  gm_sparcstreaming 64 0 0
  gm_sparcstreaming 64 8128 8128
  to see a couple of intersting cases

 */

#define BUFSIZE 16384
#define GMSIZE 20

#define NOFAIL == GM_SUCCESS || die(__FILE__,__LINE__)

#define NBITER 4

int die (char *file, int line)
{
  fprintf(stderr, "gm called failed in %s, line %d\n",file, line);
  exit (2);
  return 0;
}

void callback(struct gm_port *p, void *ctxt, gm_status_t status)
{
  status NOFAIL;
  *(int*)ctxt += 1;
}

int main(int argc, char*argv[])
{
  
  unsigned char *sbuf;
  unsigned char *rbuf;
  struct gm_port *port;
  int cc;
  int flag;
  unsigned int id;
  int len, offset, offset2;
  int done, i;
  
  if (argc != 4)
    {
      fprintf (stderr, "usage: gm_sparcstream_test <len> <offset sbuf(0)> <offset sbuf(1)>\n");
      exit (1);
    }
  len = atoi (argv[1]);
  offset = atoi (argv[2]);
  offset2 = atoi (argv[3]);

  gm_open(&port, 0, 2, "sparcstr-test", GM_API_VERSION_1_0) NOFAIL;

  gm_allow_remote_memory_access (port) NOFAIL;

  gm_get_node_id (port, &id) NOFAIL;

  rbuf = gm_dma_malloc (port, BUFSIZE);
  memset (rbuf, 0xff, BUFSIZE);

  sbuf = gm_dma_malloc (port, BUFSIZE);
  memset (sbuf, 0xff, BUFSIZE);
  printf ("rbuf=0x%pbuf=0x%p\n",rbuf,sbuf);

  for (i=0;i<NBITER;i++)
    {
      memset (sbuf, 2 * i, BUFSIZE);
      done = 0;
      gm_directed_send_with_callback (port, sbuf + offset, (gm_remote_ptr_t)rbuf + 2 * i * len, len, 
				      GM_LOW_PRIORITY, id, 2,
				      callback, &done);
      while (!done)
	{
	  gm_recv_event_t *e = gm_receive (port);
	  gm_unknown (port, e);
	}
      memset (sbuf, 2 * i + 1, BUFSIZE);
      
      done = 0;
      gm_directed_send_with_callback (port, sbuf + offset2, (gm_remote_ptr_t)rbuf + (2 * i + 1)* len, len, 
				      GM_LOW_PRIORITY, id, 2,
				      callback, &done);
      while (!done)
	{
	  gm_recv_event_t *e = gm_receive (port);
	  gm_unknown (port, e);
	}
    }

  if (offset + len > offset2)
    {
      printf("sbuf(0) and sbuf(1) overlap, do not interpret results blindly\n");
    }
  for (i=0;i<NBITER;i++)
    {
      int snum;
      for (snum=0;snum < 2;snum++)
	{
	  int j, error;
	  unsigned expect = 2 * i + snum;
	  printf("iter %d-%d: data sent 0x%02x\n", i, snum, expect);
	  for (error = 0,j=0;j<len;j++)
	    if (rbuf[(2*i + snum)*len+j] != expect)
	      error += 1;
	  if (error)
	    {
	      for (j=0;j<len;j++)
		printf ("%02x", rbuf[(2*i+snum)*len+j]);
	      printf ("\n");
	    }
	  else
	    {
	      printf ("SUCCESS\n");
	    }
	}
    }
  gm_exit (0);
}
