/******************************************************************-*-c-*-
 * Myricom GM networking software and documentation			 *
 * Copyright (c) 2002 by Myricom, Inc.					 *
 * All rights reserved.	 See the file `COPYING' for copyright notice.	 *
 *************************************************************************/

/* author: glenn@myri.com */

/* This program generates a structure definition for the LANaiX
   special registers memory region that matches the definitions in
   lanaiX_def.h. */

#include "lanaiX_def.h"
#include <stdio.h>
#include <stdlib.h>

/* read/write flags */

#define GM_R 1
#define GM_W 2
#define GM_RW (GM_R + GM_W)

#define REG(s, n, rw) do {						\
  if (GM_ ## rw & GM_R)							\
    {									\
      r->name = #n;							\
      r->offset = (unsigned long) ((void *) &n);			\
      r->size = sizeof (n);						\
      r->signedness = s;						\
      ++r;								\
      rcnt++;								\
    }									\
  if (GM_ ## rw & GM_W)							\
    {									\
      w->name = #n;							\
      w->offset = (unsigned long) ((void *) &n);			\
      w->size = sizeof (n);						\
      w->signedness = s;						\
      ++w;								\
      wcnt++;								\
    }									\
} while (0)

struct reg_rec {
  const char *name;
  unsigned long offset;
  unsigned long size;
  char signedness;
  int flags;
};

int
compare_offsets (const void *_a, const void *_b)
{
  const struct reg_rec *a, *b;
  
  a = (struct reg_rec *)_a;
  b = (struct reg_rec *)_b;

  /* HACK: assume difference in offsets fits in an int. */
  return (int) ((long) a->offset - (long) b->offset);
}

void print_list (struct reg_rec *p, const char *type);

int
main (int argc, char *argv[])
{
  struct reg_rec _r[100], *r = &_r[0];
  struct reg_rec _w[100], *w = &_w[0];
  unsigned int i;
  unsigned int rcnt = 0;
  unsigned int wcnt = 0;
  
  /****************
   * Misc
   ****************/

  /* Interface Status */
  
  REG ('u', ISR, RW);		/*  */
  REG ('u', AISR, RW);		/*  */
  REG ('u', AISR_ON, RW);	/*  */
  REG ('u', AISR_OFF, RW);	/*  */

  /* Event Dispatch */

  REG ('u', DISPATCH_STATE, W); /*  */
  REG ('u', DISPATCH_INDEX, R);	/*  */
  REG ('u', DISPATCH_CONFIG, W); /*  */
  REG ('u', DISPATCH_ISR_ON, W); /*  */
  REG ('u', DISPATCH_ISR_OFF, W); /*  */
  REG ('u', DISPATCH_STATE_ON, W); /*  */
  REG ('u', DISPATCH_STATE_OFF, W); /*  */

  /* Counters, Timers */

  REG ('u', CPUC, RW);		/*  */
  REG ('s', RTC, RW);		/*  */
  REG ('u', IT0, RW);		/*  */
  REG ('u', IT1, RW);		/*  */
  REG ('u', IT2, RW);		/*  */

  /* Programmable Output Pins */
  
  REG ('u', LED, RW);		/*  */
  REG ('u', MDI, RW);		/*  */

  /* Memory Control */
  
  REG ('u', ARBITER_SLOT, RW);	/*  */
  REG ('u', ARBITER_CODE, RW);	/*  */
  REG ('u', MP_LOWER, RW);	/*  */
  REG ('u', MP_UPPER, RW);	/*  */

  /* Port-Register Access */
  
  REG ('u', PORT_ADDR, RW);	/*  */
  REG ('u', PORT_DATA, RW);	/*  */

  /* Hardware Configuration */
  
  REG ('u', PLL, RW);		/*  */
  REG ('u', SYNC, RW);		/*  */

  /****************
   * P0
   ****************/

  /* Send Special registers */
  
  REG ('u', P0_SEND, W);	/*  */
  REG ('u', P0_SEND_FREE_COUNT, R); /*  */
  REG ('u', P0_SEND_FREE_LIMIT, RW); /*  */
  REG ('u', P0_SEND_COUNT, RW);	/*  */
  REG ('u', P0_PAUSE_COUNT, RW); /*  */
  REG ('u', P0_SEND_CONTROL, RW); /*  */

  /* Receive Channel-Independent */
  
  REG ('u', P0_MEMORY_DROP, R); /*  */

  /* Regular Receive Channel */
  
  REG ('u', P0_RB, RW);		/*  */
  REG ('u', P0_RPL_CONFIG, RW);	/*  */
  REG ('u', P0_RPL_INDEX, R);	/*  */
  REG ('u', P0_RECV_COUNT, R);	/*  */
  /* REG ('u', P0_RPL_SNAPSHOT, RW); */
  REG ('u', P0_BUFFER_DROP, R);	/*  */

  /* Alternate Receive channel */
  
  REG ('u', P0_A_RB, RW);	/*  */
  REG ('u', P0_A_RPL_CONFIG, RW); /*  */
  REG ('u', P0_A_RPL_INDEX, R);	/*  */
  REG ('u', P0_A_RECV_COUNT, R); /*  */
  /* REG ('u', P0_A_RPL_SNAPSHOT, RW); */
  REG ('u', P0_A_BUFFER_DROP, R); /*  */

  /****************
   * P1
   ****************/

  /* Send Special registers */
  
  REG ('u', P1_SEND, W);	/*  */
  REG ('u', P1_SEND_FREE_COUNT, R); /*  */
  REG ('u', P1_SEND_FREE_LIMIT, RW); /*  */
  REG ('u', P1_SEND_COUNT, RW);	/*  */
  REG ('u', P1_PAUSE_COUNT, RW); /*  */
  REG ('u', P1_SEND_CONTROL, RW); /*  */

  /* Receive Channel-Independent */
  
  REG ('u', P1_MEMORY_DROP, RW);

  /* Regular Receive Channel */
  
  REG ('u', P1_RB, RW);		/*  */
  REG ('u', P1_RPL_CONFIG, RW);	/*  */
  REG ('u', P1_RPL_INDEX, R);	/*  */
  REG ('u', P1_RECV_COUNT, R);	/*  */
  /* REG ('u', P1_RPL_SNAPSHOT, RW); */
  REG ('u', P1_BUFFER_DROP, R);	/*  */

  /* Alternate Receive channel */
  
  REG ('u', P1_A_RB, RW);	/*  */
  REG ('u', P1_A_RPL_CONFIG, RW); /*  */
  REG ('u', P1_A_RPL_INDEX, R);	/*  */
  REG ('u', P1_A_RECV_COUNT, R); /*  */
  /* REG ('u', P1_A_RPL_SNAPSHOT, RW); */
  REG ('u', P1_A_BUFFER_DROP, R); /*  */

  /****************
   * Unlisted
   ****************/

  REG ('u', COPY, W);		/*  */
  REG ('u', COPY_ABORT, W);	/*  */
  REG ('u', CRC32_CONFIG, RW);	/*  */
  REG ('u', CRC32, RW);		/*  */
  REG ('u', CRC32_BYTE, W);	/*  */
  REG ('u', CRC32_HALF, W);	/*  */
  REG ('u', CRC32_WORD, W);	/*  */
  
  /* JTAG */

  REG ('u', JTAG_MASTER, RW);
  
  /* DMA */

  REG ('u', DMA_CONFIG, RW);
  REG ('u', PCI_CLOCK, RW);
  REG ('u', DMA0_COUNT, RW);
  REG ('u', DMA1_COUNT, RW);
  REG ('u', DMA2_COUNT, RW);
  REG ('u', DMA3_COUNT, RW);
  REG ('u', DMA0_POINTER, RW);
  REG ('u', DMA1_POINTER, RW);
  REG ('u', DMA2_POINTER, RW);
  REG ('u', DMA3_POINTER, RW);
  
#if 0
  REG ('u', P0_MODE, RW);	/*  */
  REG ('u', P1_MODE, RW);	/*  */
  REG ('u', P0_TEST, RW);	/*  */
  REG ('u', P1_TEST, RW);	/*  */
  REG ('u', P0_STATE, RW);	/*  */
  REG ('u', P1_STATE, RW);	/*  */
  REG ('u', P0_RECV_CONFIG, RW); /*  */
  REG ('u', P1_RECV_CONFIG, RW); /*  */
  REG ('u', P0_SHORT_DROP, R); /*  */
  REG ('u', P1_SHORT_DROP, R); /*  */
  REG ('u', P0_8B10B_ERROR, R); /*  */
  REG ('u', P1_8B10B_ERROR, R); /*  */
  REG ('u', P0_MYRI_SEND, RW); /*  */
  REG ('u', P1_MYRI_SEND, RW); /*  */
  REG ('u', P0_GEX_CONTROL, RW); /*  */
  REG ('u', P1_GEX_CONTROL, RW); /*  */
  REG ('u', P0_GEX_STATUS, RW); /*  */
  REG ('u', P1_GEX_STATUS, RW); /*  */
  REG ('u', P0_GEX_AN_LP_ABILITY_BP, RW); /*  */
  REG ('u', P1_GEX_AN_LP_ABILITY_BP, RW); /*  */
  REG ('u', P0_GEX_AN_ADVERTISEMENT, RW); /*  */
  REG ('u', P1_GEX_AN_ADVERTISEMENT, RW); /*  */
  REG ('u', P0_GEX_AN_EXPANSION, RW); /*  */
  REG ('u', P1_GEX_AN_EXPANSION, RW); /*  */
  REG ('u', P0_IB_CONFIG, RW);	/*  */
  REG ('u', P1_IB_CONFIG, RW);	/*  */
  REG ('u', P0_IB_LDC, RW);	/*  */
  REG ('u', P1_IB_LDC, RW);	/*  */
  REG ('u', P0_IB_LEDC, RW);	/*  */
  REG ('u', P1_IB_LEDC, RW);	/*  */
  REG ('u', P0_IB_SEC, RW);	/*  */
  REG ('u', P1_IB_SEC, RW);	/*  */
  REG ('u', P0_IB_LEC, RW);	/*  */
  REG ('u', P1_IB_LEC, RW);	/*  */
  REG ('u', P0_IB_REC, RW);	/*  */
  REG ('u', P1_IB_REC, RW);	/*  */
#endif
  
  /* GEX_AN_EXTENDED_STATUS in doc??? */
  /* PCI_CONFIG, PCI_CLK_CNT, DMA*_POINTER, DMA*_COUNT */
  
  /* sort the lists */
  
  r->name = 0;
  qsort (_r, rcnt, sizeof (_r[0]), compare_offsets);
  
  w->name = 0;
  qsort (_w, wcnt, sizeof (_w[0]), compare_offsets);

  /* print the list */

  print_list (_r, "readable");
  print_list (_w, "writable");

  return 0;
}

void
print_list (struct reg_rec *p, const char *type)
{
  int i;
  unsigned int pos;
  
  printf ("struct LANaiX_%s_specials {\n", type);
  
  pos = p[0].offset;
  for (i=0; p[i].name; i++)
    {
      /* Sanity check position, to catch overlapping registers that
	 would break our output. */
      
      if (pos > p[i].offset)
	{
	  fprintf (stderr, "*** offset error ***\n");
	  exit (1);
	}
      
      /* Insert padding */

      if (pos < p[i].offset)
	{
	  printf ("\t\t\t\t\t\tgm_u8_n_t pad%lx[0x%lx];\n",
		  pos - p[0].offset, p[i].offset - pos);
	  pos = p[i].offset;
	}

      /* Print the field */

      printf ("  /* 0x%lx */ gm_%c%lu_n_t %s;\n",
	      p[i].offset, p[i].signedness, p[i].size << 3, p[i].name);
      pos += p[i].size;
    }
  printf ("};\n");
}
